/*
Copyright (c) 2008-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICFileTree.h"
#include <sys/stat.h>

#import "ICDocumentFormat+Constants.h"
#import "NSString+Iceberg.h"

#import "ICIconsUtilities.h"

#import "NSImage+Composition.h"

@implementation ICFileNode

+ (NSImage *) setIconAsDefaultLocation:(NSImage *) inIcon
{
	if (inIcon!=nil)
	{
		NSImage * tIcon;
		NSSize tSize;
	
		tSize=[inIcon size];
		
		tIcon = [[NSImage alloc] initWithSize:tSize];
		
		if (tIcon!=nil)
		{
			NSRect tRect;
			NSRect tInsetRect;
			CGFloat tLineDashPattern[2]={3.0f,3.0f};
			[tIcon lockFocus];
			NSBezierPath * tBezierPath;
			
			[inIcon compositeToPoint:NSZeroPoint operation:NSCompositeCopy];
			
			// Draw the Target pointer
			
			[[NSColor redColor] set];
			
			tRect=NSMakeRect(0.0f,0.0f,tSize.width,tSize.height);
			
			tInsetRect=NSInsetRect(tRect,_CGFloatFloor(tSize.width/4.0f),_CGFloatFloor(tSize.height/4.0f));
			
			tInsetRect.size.width-=1.0f;
			
			tInsetRect.origin.y+=1.0f;
			
			tInsetRect.size.height-=1.0f;
			
			NSFrameRect(tInsetRect);
			
			tBezierPath=[NSBezierPath bezierPath];
			
			if (tBezierPath!=nil)
			{
				[tBezierPath setLineDash:tLineDashPattern count:2 phase:0.0f];
				
				[tBezierPath moveToPoint:NSMakePoint(NSMinX(tRect),_CGFloatFloor(NSMidY(tRect))+0.5f)];
				
				[tBezierPath lineToPoint:NSMakePoint(NSMaxX(tRect),_CGFloatFloor(NSMidY(tRect))+0.5f)];
				
				[tBezierPath stroke];
				
				[tBezierPath removeAllPoints];
				
				[tBezierPath moveToPoint:NSMakePoint(_CGFloatFloor(NSMidX(tRect))-0.5f,NSMinY(tRect)+1.0f)];
				
				[tBezierPath lineToPoint:NSMakePoint(_CGFloatFloor(NSMidX(tRect))-0.5f,NSMaxY(tRect))];
				
				[tBezierPath stroke];
			}
			
			[tIcon unlockFocus];
			
			return [tIcon autorelease];
		}
	}

	return inIcon;
}

+ (NSImage *) cachedIconForBaseNodeAtPath:(NSString *) tPath
{
	static NSMutableDictionary * sCachedBaseNodeIcons=nil;
	
	if (sCachedBaseNodeIcons==nil)
	{
		NSImage * tCachedImage;
		
		sCachedBaseNodeIcons=[[NSMutableDictionary alloc] initWithCapacity:20];
		
		// Hard Drive
		
		tCachedImage=[ICIconsUtilities smallIconForType:NSFileTypeForHFSTypeCode(kGenericHardDiskIcon)];
		
		if (tCachedImage!=nil)
		{
			[sCachedBaseNodeIcons setObject:tCachedImage forKey:@"/"];
			
			// Disabled
			
			[sCachedBaseNodeIcons setObject:[tCachedImage imageWithAlpha:0.5f] forKey:@"/+Disabled"];
		}
		
		// Applications
		
		tCachedImage=[ICIconsUtilities smallIconForType:NSFileTypeForHFSTypeCode(kApplicationsFolderIcon)];
		
		if (tCachedImage!=nil)
		{
			[sCachedBaseNodeIcons setObject:tCachedImage forKey:@"/Applications"];
			
			// Disabled
			
			[sCachedBaseNodeIcons setObject:[tCachedImage imageWithAlpha:0.5f] forKey:@"/Applications+Disabled"];
		}
		
		// Library
		
		tCachedImage=[ICIconsUtilities smallIconForType:NSFileTypeForHFSTypeCode(kSharedLibrariesFolderIcon)];
		
		if (tCachedImage!=nil)
		{
			[sCachedBaseNodeIcons setObject:tCachedImage forKey:@"/Library"];
			
			// Disabled
			
			[sCachedBaseNodeIcons setObject:[tCachedImage imageWithAlpha:0.5f] forKey:@"/Library+Disabled"];
		}
		
		// System
		
		tCachedImage=[ICIconsUtilities smallIconForType:NSFileTypeForHFSTypeCode(kSystemFolderIcon)];
		
		if (tCachedImage!=nil)
		{
			[sCachedBaseNodeIcons setObject:tCachedImage forKey:@"/System"];
			
			// Disabled
			
			[sCachedBaseNodeIcons setObject:[tCachedImage imageWithAlpha:0.5f] forKey:@"/System+Disabled"];
		}
		
		// Users
		
		tCachedImage=[ICIconsUtilities smallIconForImage:[[NSWorkspace sharedWorkspace] iconForFile:@"/Users"]];
		
		if (tCachedImage!=nil)
		{
			[sCachedBaseNodeIcons setObject:tCachedImage forKey:@"/Users"];
			
			// Disabled
			
			[sCachedBaseNodeIcons setObject:[tCachedImage imageWithAlpha:0.5f] forKey:@"/Users+Disabled"];
		}
	}
	
	return [sCachedBaseNodeIcons objectForKey:tPath];
}

+ (NSImage *) cachedIconForPath:(NSString *) inPath
{
    static NSMutableDictionary * sCachedIcons=nil;
    static NSFileManager * sFileManager=nil;
	static NSWorkspace * sWorkspace=nil;
    NSString * tPathExtension;
    NSImage * tIcon;
    NSUInteger tLength;
    static NSImage * sFolderIcon=nil;
    
    if (sCachedIcons==nil)
    {
        NSArray * tArray;
        NSImage * tOtherExtensionIcon;
        
		
        sCachedIcons=[[NSMutableDictionary alloc] initWithCapacity:20];
        
        sFileManager=[NSFileManager defaultManager];
        
        sWorkspace=[NSWorkspace sharedWorkspace];
        
		// Common file extensions
		
		tArray=[NSArray arrayWithObjects:@"nib",@"tiff",@"tif",@"icns",@"rtd",@"rtfd",@"strings",@"plist",@"txt",@"strings",nil];
        
		if (tArray!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[tArray objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSString * tExtension;
				
				while (tExtension=[tEnumerator nextObject])
				{
					NSImage * tExtensionIcon;
					
					tExtensionIcon=[ICIconsUtilities smallIconForType:tExtension];
            
					if (tExtensionIcon!=nil)
					{
						[sCachedIcons setObject:tExtensionIcon forKey:tExtension];
					}
				}
			}
		}
		
        // Other extensions
		
		tOtherExtensionIcon=[ICIconsUtilities smallIconForType:NSFileTypeForHFSTypeCode(kGenericApplicationIcon)];
		
		if (tOtherExtensionIcon!=nil)
		{
			[sCachedIcons setObject:tOtherExtensionIcon forKey:@"app"];
		}
		
		sFolderIcon=[[ICIconsUtilities smallIconForType:NSFileTypeForHFSTypeCode(kGenericFolderIcon)] retain];
		
		if (tOtherExtensionIcon!=nil)
		{
			[sCachedIcons setObject:sFolderIcon forKey:@"lproj"];
			
			[sCachedIcons setObject:sFolderIcon forKey:@"framework"];
		}
		
		// Application
        
        /*tIcon=[NSImage imageNamed:@"package16"];
        [sCachedIcons setObject:tIcon forKey:@"pkg"];
        
        tIcon=[NSImage imageNamed:@"metapackage16"];
        [sCachedIcons setObject:tIcon forKey:@"mpkg"];*/
    }
    
    tIcon=nil;
	
	tPathExtension=[[[inPath lastPathComponent] pathExtension] lowercaseString];
    
    tLength=[tPathExtension length];
    
    if (tLength>0)
    {
        tIcon=[sCachedIcons objectForKey:tPathExtension];
		
		if (tIcon==nil)
		{
			if (tLength>0)
			{
				NSDictionary * tDictionary;
			
				tDictionary=[sFileManager fileAttributesAtPath:inPath traverseLink:NO];
				
				if ([[tDictionary objectForKey:NSFileType] isEqualToString:NSFileTypeDirectory]==YES)
				{
					tIcon=sFolderIcon;
				}
				else
				{
					tIcon=[ICIconsUtilities smallIconForType:tPathExtension];
				}
				
				if (tIcon!=nil)
				{
					[sCachedIcons setObject:tIcon forKey:tPathExtension];
				}
			}
		}
		
		return tIcon;
    }
    else
	{
        NSDictionary * tDictionary;
        
		tDictionary=[sFileManager fileAttributesAtPath:inPath traverseLink:NO];
		
		if ([[tDictionary objectForKey:NSFileType] isEqualToString:NSFileTypeDirectory]==YES)
		{
			tIcon=sFolderIcon;
		}
		else
		{
			tIcon=[sCachedIcons objectForKey:inPath];
			
			if (tIcon==nil)
			{
				tIcon=[sWorkspace iconForFile:inPath];
				
				if (tIcon!=nil)
				{
					NSSize tSize={16.0f,16.0f};
					
					if ([[tDictionary objectForKey:NSFileType] isEqualToString:NSFileTypeSymbolicLink]==YES)
					{
						NSImage * tIconWithArrow;
							
						tIconWithArrow = [[NSImage alloc] initWithSize: NSMakeSize(16.0f,16.0f)];
						
						if (tIconWithArrow!=nil)
						{
							static NSImage * sArrowImage=nil;
						
							if (sArrowImage==nil)
							{
								sArrowImage = [NSImage imageNamed: @"FSIconImage-LinkArrow"];
							
								[sArrowImage setScalesWhenResized: YES];
							
								[sArrowImage setSize: NSMakeSize(16.0f,16.0f)];
							}
						
							[tIcon setScalesWhenResized: YES];
						
							[tIcon setSize:tSize];
						
							[tIconWithArrow lockFocus];
						
							[tIcon compositeToPoint:NSZeroPoint operation:NSCompositeCopy];
						
							[sArrowImage compositeToPoint:NSZeroPoint operation:NSCompositeSourceOver];
						
							[tIconWithArrow unlockFocus];
						
							tIcon=[tIconWithArrow autorelease];
						}
					}
					else
					{
						[tIcon setScalesWhenResized: YES];
						
						[tIcon setSize:tSize];
					}
					
					[sCachedIcons setObject:tIcon forKey:inPath];
				}
			}
		}
    }
    
    return tIcon;
}

+ (id) rootFileNode
{
    ICFileNode * nFileNode;
    
    nFileNode=[[ICFileNode alloc] init];
    
    if (nFileNode!=nil)
    {
        nFileNode->type_=ICDOCUMENT_FILE_ROOT_NODE;
        
        nFileNode->flags_.folder=YES;
        nFileNode->flags_.link=NO;
    }
    
    return [nFileNode autorelease];
}

+ (id) invisibleBaseFileNodeWithName:(NSString *) inName icon:(NSImage *) inIcon user:(int) inUID group:(int) inGID privileges:(int) inPrivileges
{
    ICFileNode * nFileNode;
    
    nFileNode=[ICFileNode alloc];
    
    if (nFileNode!=nil)
    {
        nFileNode->type_=ICDOCUMENT_FILE_BASE_NODE_INVISIBLE;
        
        if (inName!=nil)
        {
            nFileNode->path_=[inName copy];
        }
        
        nFileNode->icon_=inIcon;
        
        nFileNode->uid_=inUID;
        
        nFileNode->gid_=inGID;
        
        nFileNode->privileges_=inPrivileges;
        
        nFileNode->pathType_=ICDOCUMENT_PATH_TYPE_NAME;
        
        nFileNode->flags_.folder=YES;
        
        nFileNode->flags_.link=NO;
    }
    
    return [nFileNode autorelease];
}

+ (id) baseFileNodeWithName:(NSString *) inName icon:(NSImage *) inIcon user:(int) inUID group:(int) inGID privileges:(int) inPrivileges
{
    ICFileNode * nFileNode;
    
    nFileNode=[ICFileNode alloc];
    
    if (nFileNode!=nil)
    {
        nFileNode->type_=ICDOCUMENT_FILE_BASE_NODE;
        
        if (inName!=nil)
        {
            nFileNode->path_=[inName copy];
        }
        
        nFileNode->icon_=inIcon;
        
        nFileNode->uid_=inUID;
        
        nFileNode->gid_=inGID;
        
        nFileNode->privileges_=inPrivileges;
        
        nFileNode->pathType_=ICDOCUMENT_PATH_TYPE_NAME;
        
        nFileNode->flags_.folder=YES;
        
        nFileNode->flags_.link=NO;
    }
    
    return [nFileNode autorelease];
}

+ (id) folderFileNodeWithName:(NSString *) inName icon:(NSImage *) inIcon user:(int) inUID group:(int) inGID privileges:(int) inPrivileges
{
    ICFileNode * nFileNode;
    
    nFileNode=[ICFileNode baseFileNodeWithName:inName
                                          icon:inIcon
                                          user:inUID
                                         group:inGID
                                    privileges:inPrivileges];
    
    if (nFileNode!=nil)
    {
        nFileNode->type_=ICDOCUMENT_FILE_NEWFOLDER_NODE;
    }
    
    return nFileNode;
}

+ (id) fileNodeWithType:(int) inType path:(NSString *) inPath
{
    return [[[ICFileNode alloc] initWithType:inType path:inPath pathType:ICDOCUMENT_PATH_TYPE_GLOBAL] autorelease];
}

+ (id) fileNodeWithType:(int) inType path:(NSString *) inPath pathType:(NSInteger) inPathType
{
    return [[[ICFileNode alloc] initWithType:inType path:inPath pathType:inPathType] autorelease];
}

+ (NSImage *) newSmallIconFromImage:(NSImage *) inImage
{
    NSImage * nImage=nil;
    NSSize tFileIconSize;
        
    tFileIconSize=[inImage size];
    
    nImage=[[NSImage alloc] initWithSize:NSMakeSize(16,16)];
    
    [nImage lockFocus];
    
    [inImage drawInRect:NSMakeRect(0,0,16,16)
                    fromRect:NSMakeRect(0,0,tFileIconSize.width,tFileIconSize.height)
                operation:NSCompositeCopy
                    fraction:1.0f];
    
    [nImage unlockFocus];
    
    return nImage;
}

+ (NSImage *) newIconFromPath:(NSString *) inPath
{
    NSImage * nImage=nil;
    NSImage * tFileIcon;
    
    tFileIcon=[ICFileNode cachedIconForPath:inPath];
    
    if (tFileIcon!=nil)
    {
        NSSize tFileIconSize;
        
        tFileIconSize=[tFileIcon size];
        
        nImage=[[NSImage alloc] initWithSize:NSMakeSize(16,16)];
        
        [nImage lockFocus];
        
        [tFileIcon drawInRect:NSMakeRect(0,0,16,16)
                     fromRect:NSMakeRect(0,0,tFileIconSize.width,tFileIconSize.height)
                    operation:NSCompositeCopy
                     fraction:1.0f];
        
        [nImage unlockFocus];
    }
    
    return nImage;
}

- (id) initWithType:(int) inType path:(NSString *) inPath pathType:(NSInteger) inPathType
{
    self=[super init];
    
    if (self!=NULL)
    {
        type_=inType;
        
        pathType_=inPathType;
        
        [self setPath:inPath];
        
        if (pathType_==ICDOCUMENT_PATH_TYPE_GLOBAL)
        {
            icon_=nil;
        }
        
        if (type_==ICDOCUMENT_FILE_REALFILE_NODE)
        {
            flags_.link=[self isLink];
        }
        
        flags_.folder=NO;
        flags_.existsOnDisk=YES;
    }
    
    return self;
}

- (void) dealloc
{
	[extras_ release];
	
	[super dealloc];
}

#pragma mark -

- (BOOL) link
{
    return flags_.link;
}

- (int) type
{
    return type_;
}

- (NSString *) path
{
    return [[path_ retain] autorelease];
}

- (void) setPath:(NSString *) inPath
{
    if (path_!=inPath)
    {
        [path_ release];
        
        path_=[inPath copy];
    }
}

- (NSString *) fileName
{
    switch(type_)
    {
        case ICDOCUMENT_FILE_BASE_NODE_INVISIBLE:
		case ICDOCUMENT_FILE_BASE_NODE:
        case ICDOCUMENT_FILE_NEWFOLDER_NODE:
            return [[path_ retain] autorelease];
            break;
        case ICDOCUMENT_FILE_REALFILE_NODE:
            return [path_ lastPathComponent];
            break;
    }
    
    return nil;
}

- (NSInteger) pathType
{
    return pathType_;
}

- (void) setPathType:(NSInteger) inPathType
{
    pathType_=inPathType;
}

- (NSImage *) icon
{
    if (icon_==nil && ICDOCUMENT_FILE_REALFILE_NODE==type_)
    {
        icon_=[[ICFileNode cachedIconForPath:path_] retain];
    }
    
    return [[icon_ retain] autorelease];
}

- (int) uid
{
    return uid_;
}

- (void) setUid:(int) inUID
{
    uid_=inUID;
}

- (int) gid
{
    return gid_;
}

- (void) setGid:(int) inGID
{
    gid_=inGID;
}

- (int) privileges
{
    return privileges_;
}

- (void) setPrivileges:(int) inPrivileges
{
    privileges_=inPrivileges;
}

- (NSNumber *) allowDowngradeNumber
{
	return [extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_CAN_DOWNGRADE];
}

- (BOOL) allowDowngrade
{
	NSNumber * tNumber;
	
	tNumber=[extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_CAN_DOWNGRADE];
	
	if (tNumber!=nil)
	{
		return [tNumber boolValue];
	}
	
	return NO;
}

- (void) setAllowDowngrade:(BOOL) inAllowDowngrade
{
	if (extras_==nil)
	{
		extras_=[[NSMutableDictionary alloc] initWithCapacity:3];
	}
	
	[extras_ setObject:[NSNumber numberWithBool:inAllowDowngrade] forKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_CAN_DOWNGRADE];
}

- (NSMutableArray *) locators
{
	return [extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
}

- (void) setLocators:(NSArray *) inLocators
{
	NSArray * tArray;

	if (extras_==nil)
	{
		extras_=[[NSMutableDictionary alloc] initWithCapacity:3];
	}
	
	tArray=[extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
	
	if (tArray!=inLocators)
	{
		if (inLocators==nil)
		{
			[extras_ removeObjectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
		}
		else
		{
			[extras_ setObject:inLocators forKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
		}
	}
}

- (NSMutableDictionary *) preInstallPath
{
	return [extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
}

- (void) setPreInstallPath:(NSDictionary *) inPreInstallPath
{
	NSDictionary * tDictionary;

	if (extras_==nil)
	{
		extras_=[[NSMutableDictionary alloc] initWithCapacity:3];
	}
	
	tDictionary=[extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
	
	if (tDictionary!=inPreInstallPath)
	{
		if (inPreInstallPath==nil)
		{
			[extras_ removeObjectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
		}
		else
		{
			[extras_ setObject:inPreInstallPath forKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
		}
	}
}

- (NSMutableDictionary *) postInstallPath
{
	return [extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
}

- (void) setPostInstallPath:(NSDictionary *) inPostInstallPath
{
	NSDictionary * tDictionary;

	if (extras_==nil)
	{
		extras_=[[NSMutableDictionary alloc] initWithCapacity:3];
	}
	
	tDictionary=[extras_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
	
	if (tDictionary!=inPostInstallPath)
	{
		if (inPostInstallPath==nil)
		{
			[extras_ removeObjectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
		}
		else
		{
			[extras_ setObject:inPostInstallPath forKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
		}
	}
}

#pragma mark -

- (BOOL) isLink
{
    struct stat tStat;
        
    if (lstat([path_ fileSystemRepresentation], &tStat)==0)
    {
        if ((tStat.st_mode & S_IFMT)==S_IFLNK)
        {
            return YES;
        }
    }
    
    return NO;
}

+ (NSString *) privilegesStringRepresentationWithPermission:(int) inPermission mixedPermission:(int) inMixedPermission statType:(char) inType
{
    char ownerExecute,groupExecute,otherExecute;
    char tMixedChar='?';
    
    if (inType==0)
    {
        inType=tMixedChar;
    }
    
    if ((inMixedPermission & S_ISUID)==S_ISUID || (inMixedPermission & S_IXUSR)==S_IXUSR)
    {
        ownerExecute=tMixedChar;
    }
    else
    {
        if ((inPermission & S_ISUID)==S_ISUID)
        {
            ownerExecute=((inPermission & S_IXUSR)==S_IXUSR) ? 's' : 'S';
        }
        else
        {
            ownerExecute=((inPermission & S_IXUSR)==S_IXUSR) ? 'x' : '-';
        }
    }
    
    if ((inMixedPermission & S_ISGID)==S_ISGID || (inMixedPermission & S_IXGRP)==S_IXGRP)
    {
        groupExecute=tMixedChar;
    }
    else
    {
        if ((inPermission & S_ISGID)==S_ISGID)
        {
            groupExecute=((inPermission & S_IXGRP)==S_IXGRP) ? 's' : 'S';
        }
        else
        {
            groupExecute=((inPermission & S_IXGRP)==S_IXGRP) ? 'x' : '-';
        }
    }
    
    if ((inMixedPermission & S_ISTXT)==S_ISTXT || (inMixedPermission & S_IXOTH)==S_IXOTH)
    {
        otherExecute=tMixedChar;
    }
    else
    {
        if ((inPermission & S_ISTXT)==S_ISTXT)
        {
            otherExecute=((inPermission & S_IXOTH)==S_IXOTH) ? 't' : 'T';
        }
        else
        {
            otherExecute=((inPermission & S_IXOTH)==S_IXOTH) ? 'x' : '-';
        }
    }
    
    return [NSString stringWithFormat:@"%c%c%c%c%c%c%c%c%c%c",inType,
                                                            ((inMixedPermission & S_IRUSR)==S_IRUSR) ? tMixedChar : (((inPermission & S_IRUSR)==S_IRUSR) ? 'r' : '-'),
                                                            ((inMixedPermission & S_IWUSR)==S_IWUSR) ? tMixedChar : (((inPermission & S_IWUSR)==S_IWUSR) ? 'w' : '-'),
                                                            ownerExecute,
                                                            ((inMixedPermission & S_IRGRP)==S_IRGRP) ? tMixedChar : (((inPermission & S_IRGRP)==S_IRGRP) ? 'r' : '-'),
                                                            ((inMixedPermission & S_IWGRP)==S_IWGRP) ? tMixedChar : (((inPermission & S_IWGRP)==S_IWGRP) ? 'w' : '-'),
                                                            groupExecute,
                                                            ((inMixedPermission & S_IROTH)==S_IROTH) ? tMixedChar : (((inPermission & S_IROTH)==S_IROTH) ? 'r' : '-'),
                                                            ((inMixedPermission & S_IWOTH)==S_IWOTH) ? tMixedChar : (((inPermission & S_IWOTH)==S_IWOTH) ? 'w' : '-'),
                                                            otherExecute];
}

- (NSString *) privilegesStringRepresentationForLink
{
    char ownerExecute,groupExecute,otherExecute;

    if ((privileges_ & S_ISUID)==S_ISUID)
    {
        ownerExecute=((privileges_ & S_IXUSR)==S_IXUSR) ? 's' : 'S';
    }
    else
    {
        ownerExecute=((privileges_ & S_IXUSR)==S_IXUSR) ? 'x' : '-';
    }
    
    if ((privileges_ & S_ISGID)==S_ISGID)
    {
        groupExecute=((privileges_ & S_IXGRP)==S_IXGRP) ? 's' : 'S';
    }
    else
    {
        groupExecute=((privileges_ & S_IXGRP)==S_IXGRP) ? 'x' : '-';
    }
    
    if ((privileges_ & S_ISTXT)==S_ISTXT)
    {
        otherExecute=((privileges_ & S_IXOTH)==S_IXOTH) ? 't' : 'T';
    }
    else
    {
        otherExecute=((privileges_ & S_IXOTH)==S_IXOTH) ? 'x' : '-';
    }
    
    return [NSString stringWithFormat:@"%c%c%c%c%c%c%c%c%c%c", 'l',
                                                               ((privileges_ & S_IRUSR)==S_IRUSR) ? 'r' : '-',
                                                               ((privileges_ & S_IWUSR)==S_IWUSR) ? 'w' : '-',
                                                               ownerExecute,
                                                               ((privileges_ & S_IRGRP)==S_IRGRP) ? 'r' : '-',
                                                               ((privileges_ & S_IWGRP)==S_IWGRP) ? 'w' : '-',
                                                               groupExecute,
                                                               ((privileges_ & S_IROTH)==S_IROTH) ? 'r' : '-',
                                                               ((privileges_ & S_IWOTH)==S_IWOTH) ? 'w' : '-',
                                                               otherExecute];
}

- (char) statType
{
    char tStatType='d';
    
    if (ICDOCUMENT_FILE_REALFILE_NODE==type_)
    {
        struct stat tStat;
        
        if (lstat([path_ fileSystemRepresentation], &tStat)==0)
        {
            switch((tStat.st_mode & S_IFMT))
            {
                case S_IFDIR:
                    tStatType='d';
                    break;
                case S_IFREG:
                    tStatType='-';
                    break;
                case S_IFLNK:
                    tStatType='l';
                    break;
                case S_IFBLK:
                    tStatType='b';
                    break;
                case S_IFCHR:
                    tStatType='c';
                    break;
                case S_IFSOCK:
                    tStatType='s';
                    break;
                default:
                    tStatType='-';
                    break;
            }
        }
    }
    
    return tStatType;
}

- (NSString *) privilegesStringRepresentation
{
    char tFolder='d';
    char ownerExecute,groupExecute,otherExecute;
    
    if (ICDOCUMENT_FILE_REALFILE_NODE==type_)
    {
        struct stat tStat;
        
        flags_.link=NO;
        
        if (lstat([path_ fileSystemRepresentation], &tStat)==0)
        {
            switch((tStat.st_mode & S_IFMT))
            {
                case S_IFDIR:
                    tFolder='d';
                    break;
                case S_IFREG:
                    tFolder='-';
                    break;
                case S_IFLNK:
                    flags_.link=YES;
                    tFolder='l';
                    break;
                case S_IFBLK:
                    tFolder='b';
                    break;
                case S_IFCHR:
                    tFolder='c';
                    break;
                case S_IFSOCK:
                    tFolder='s';
                    break;
                default:
                    tFolder='-';
                    break;
            }
        }
    }

    if ((privileges_ & S_ISUID)==S_ISUID)
    {
        ownerExecute=((privileges_ & S_IXUSR)==S_IXUSR) ? 's' : 'S';
    }
    else
    {
        ownerExecute=((privileges_ & S_IXUSR)==S_IXUSR) ? 'x' : '-';
    }
    
    if ((privileges_ & S_ISGID)==S_ISGID)
    {
        groupExecute=((privileges_ & S_IXGRP)==S_IXGRP) ? 's' : 'S';
    }
    else
    {
        groupExecute=((privileges_ & S_IXGRP)==S_IXGRP) ? 'x' : '-';
    }
    
    if ((privileges_ & S_ISTXT)==S_ISTXT)
    {
        otherExecute=((privileges_ & S_IXOTH)==S_IXOTH) ? 't' : 'T';
    }
    else
    {
        otherExecute=((privileges_ & S_IXOTH)==S_IXOTH) ? 'x' : '-';
    }
    
    return [NSString stringWithFormat:@"%c%c%c%c%c%c%c%c%c%c", tFolder,
                                                               ((privileges_ & S_IRUSR)==S_IRUSR) ? 'r' : '-',
                                                               ((privileges_ & S_IWUSR)==S_IWUSR) ? 'w' : '-',
                                                               ownerExecute,
                                                               ((privileges_ & S_IRGRP)==S_IRGRP) ? 'r' : '-',
                                                               ((privileges_ & S_IWGRP)==S_IWGRP) ? 'w' : '-',
                                                               groupExecute,
                                                               ((privileges_ & S_IROTH)==S_IROTH) ? 'r' : '-',
                                                               ((privileges_ & S_IWOTH)==S_IWOTH) ? 'w' : '-',
                                                               otherExecute];
}

- (void) setFolder:(BOOL) inFolder
{
    flags_.folder=inFolder;
}

/*- (void) setExpanded:(BOOL) inExpanded
{
    if (type_==ICDOCUMENT_FILE_REALFILE_NODE)
    {
        flags_.expanded=inExpanded;
        flags_.folder=inExpanded;
    }
}*/

- (BOOL) isLeaf
{
    return !flags_.folder;
}

- (BOOL) fileExistenceOnDiskChanged
{
    if (type_==ICDOCUMENT_FILE_REALFILE_NODE)
    {
        struct stat tStat;
		BOOL exists;
        
        exists=(lstat([path_ fileSystemRepresentation], &tStat)==0);
        
        if (exists!=flags_.existsOnDisk)
        {
            flags_.existsOnDisk=exists;
            return YES;
        }
    }
    
    return NO;
}

- (BOOL) existsOnDisk
{
    if (type_==ICDOCUMENT_FILE_REALFILE_NODE)
    {
        return flags_.existsOnDisk;
    }
    
    return YES;
}

@end

@implementation ICFileTree

+ (ICFileTree *) fileNodeWithDictionary:(NSDictionary *) inDictionary projectPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
    ICFileTree * nFileTree=nil;

    if (inDictionary!=nil)
    {
        id tFileNode;
		NSNumber * tNumber;
		NSDictionary * tDictionary;
		NSArray * tArray;
		NSString * tPath;
		NSString * tString;
		int tUID,tGID;
		int tPrivileges;
		int tType;
		int tPathType;
	
        tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_TYPE];
        
        if (tNumber!=nil)
        {
            tType=[tNumber intValue];
        }
        else
        {
            return nil;
        }
        
        // Path
        
        tString=[inDictionary objectForKey:ICDOCUMENT_PATH];
        
        if (tString!=nil)
        {
            tPath=tString;
        }
        else
        {
            tPath=@"";
        }
        
        // Path type
        
        tNumber=[inDictionary objectForKey:ICDOCUMENT_PATH_TYPE];
        
        if (tNumber!=nil)
        {
            tPathType=[tNumber intValue];
        }
        else
        {
            return nil;
        }
        
        // User ID
        
        tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_UID];
        
        if (tNumber!=nil)
        {
            tUID=[tNumber intValue];
        }
        else
        {
            tUID=0;
        }
        
        // GID
        
        tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_GID];
        
        if (tNumber!=nil)
        {
            tGID=[tNumber intValue];
        }
        else
        {
            tGID=80;
        }
        
        // Privileges
        
        tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_PERMISSIONS];
        
        if (tNumber!=nil)
        {
            tPrivileges=[tNumber intValue];
        }
        else
        {
            tPrivileges=0775;
        }
        
        switch (tType)
        {
            case ICDOCUMENT_FILE_BASE_NODE_INVISIBLE:
			
				nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode invisibleBaseFileNodeWithName:tPath
                                                                                       icon:nil
                                                                                       user:tUID
                                                                                      group:tGID
                                                                                 privileges:tPrivileges]
                                                    parent:nil
                                                  children:[NSArray array]];
				
				break;
				
			case ICDOCUMENT_FILE_BASE_NODE:
                
				nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode baseFileNodeWithName:tPath
                                                                                       icon:nil
                                                                                       user:tUID
                                                                                      group:tGID
                                                                                 privileges:tPrivileges]
                                                    parent:nil
                                                  children:[NSArray array]];
                break;
            
			case ICDOCUMENT_FILE_NEWFOLDER_NODE:
				
				nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode folderFileNodeWithName:tPath
                                                                                            icon:nil
                                                                                            user:tUID
                                                                                           group:tGID
                                                                                      privileges:tPrivileges]
                                                    parent:nil
                                                  children:[NSArray array]];
                break;
				
			case ICDOCUMENT_FILE_REALFILE_NODE:
                
                switch (tPathType)
                {
                    case ICDOCUMENT_PATH_TYPE_GLOBAL:
					
                        break;
						
                    case ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT:
                    
                       tPath=[tPath stringByAbsolutingWithPath:inProjectPath];
                        
                        break;
					
					 case ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER:		// A COMPLETER
                    
                        tPath=[tPath stringByAbsolutingWithPath:inReferenceFolderPath];
                        
                        break;
						
                    default:
                        break;
                }
                
                tFileNode=[ICFileNode fileNodeWithType:tType
                                                  path:tPath
                                              pathType:tPathType];
                
                tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_EXPANDED];
        
                if (tNumber!=nil)
                {
                    [tFileNode setFolder:[tNumber boolValue]];
                }
                
                if (tFileNode!=nil)
                {
                    [tFileNode setUid:tUID];
                    [tFileNode setGid:tGID];
                    [tFileNode setPrivileges:tPrivileges];
                    
                    nFileTree=[[ICFileTree alloc] initWithData:tFileNode
                                                        parent:nil
                                                      children:[NSArray array]];
                }
                
                break;
        }
        
        if (nFileTree!=nil)
		{
			tArray=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
        
			if (tArray!=nil)
			{
				[FILENODE_DATA(nFileTree) setLocators:tArray];
			}
			
			tDictionary=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
			
			if (tDictionary!=nil)
			{
				[FILENODE_DATA(nFileTree) setPreInstallPath:tDictionary];
			}
			
			tDictionary=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
			
			if (tDictionary!=nil)
			{
				[FILENODE_DATA(nFileTree) setPostInstallPath:tDictionary];
			}
			
			tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_CAN_DOWNGRADE];
			
			if (tNumber!=nil)
			{
				[FILENODE_DATA(nFileTree) setAllowDowngrade:[tNumber boolValue]];
			}
			
			tArray=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_CHILDREN];
			
			if (tArray!=nil)
			{
				NSUInteger i,tCount;
				ICFileTree * tFileTree;
				
				tCount=[tArray count];
				
				for(i=0;i<tCount;i++)
				{
					NSDictionary * tFileDictionary;
					
					tFileDictionary=[tArray objectAtIndex:i];
					
					tFileTree=[ICFileTree fileNodeWithDictionary:tFileDictionary projectPath:inProjectPath  referenceFolderPath:inReferenceFolderPath];
				
					[nFileTree insertSortedChildReverse:tFileTree];
					
					//[nFileTree insertSortedChild:tFileTree];
				}
			}
		}
	}
    
    
    return [nFileTree autorelease];
}

+ (ICFileTree *) fileTreeWithArray:(NSArray *) inArray projectPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
	ICFileTree * nFileTree=nil;
	
	if (inArray!=nil)
	{
		nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode rootFileNode]
											parent:nil
										  children:[NSArray array]];
		if (nFileTree!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[inArray objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSDictionary * tDictionary;
				
				while (tDictionary=[tEnumerator nextObject])
				{
					ICFileTree * tFileTree;
					
					tFileTree=[ICFileTree fileNodeWithDictionary:tDictionary projectPath:inProjectPath referenceFolderPath:inReferenceFolderPath];

					[nFileTree insertSortedChildReverse:tFileTree];
				}
			}
		}		
	}
	
	return [nFileTree autorelease];
}

+ (ICFileTree *) fileTreeWithDictionary:(NSDictionary *) inDictionary projectPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
    ICFileTree * nFileTree=nil;
	
	if (inDictionary!=nil)
	{
		
    
		nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode rootFileNode]
											parent:nil
										  children:[NSArray array]];
		if (nFileTree!=nil)
		{
			ICFileTree * rootTree;
			
			rootTree=[ICFileTree fileNodeWithDictionary:inDictionary projectPath:inProjectPath referenceFolderPath:inReferenceFolderPath];

			[nFileTree insertChild: rootTree
							atIndex: 0];

		}		
	}
	
	return [nFileTree autorelease];
}

/*+ (ICFileTree *) defaultFileTree
{
    NSString * tPath;
    ICFileTree * nFileTree=nil;
    
    tPath=[[NSBundle mainBundle] pathForResource:@"DefaultTree" ofType:@"plist"];
    
    if (tPath!=nil)
    {
        NSDictionary * tDictionary;
    
        tDictionary=[NSDictionary dictionaryWithContentsOfFile:tPath];
        
        if (tDictionary!=nil)
        {
            
    
            nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode rootFileNode]
                                                parent:nil
                                              children:[NSArray array]];
            if (nFileTree!=nil)
            {
                ICFileTree * rootTree;
                
                rootTree=[ICFileTree fileNodeWithDictionary:tDictionary];
    
                [nFileTree insertChild: rootTree
                               atIndex: 0];
    
            }
        }
    }
        
    return [nFileTree autorelease];
}*/

#pragma mark -

- (ICFileTree *) createMissingBranchesForPath:(NSString *) inPath projectPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
	if (inPath!=nil)
    {
        NSArray * tArray;
        NSUInteger i,tCount;
        ICFileTree * tTree;
		NSString * tRelativePath;
		struct stat tStat;
		
		tTree=self;
        
        tArray=[inPath componentsSeparatedByString:@"/"];
        
		tRelativePath=@"/";
		
        tCount=[tArray count];
        
        for(i=1;i<tCount;i++)
        {
            NSString * tComponent;
            
            tComponent=[tArray objectAtIndex:i];
            
            if ([tComponent length]>0)
            {
                NSArray * tChildren;
                NSUInteger j,tChildrenCount;
                
                tRelativePath=[tRelativePath stringByAppendingPathComponent:tComponent];
				
				tChildren=[tTree children];
                
                tChildrenCount=[tChildren count];
                
                for(j=0;j<tChildrenCount;j++)
                {
                    ICFileTree * tChild;
                    ICFileNode * tNode;
                
                    tChild=[tChildren objectAtIndex:j];
                    
                    tNode=FILENODE_DATA(tChild);
                    
                    if ([[tNode fileName] compare:tComponent]==NSOrderedSame)
                    {
                        tTree=tChild;
						
                        break;
                    }
                }
                
                if (j==tChildrenCount)
                {
                    tRelativePath=[tRelativePath stringByDeletingLastPathComponent];
					
					break;
                }
            }
            else
            {
                break;
            }
        }
		
        
		if (lstat([tRelativePath fileSystemRepresentation], &tStat)==0)
		{
			for(;i<tCount;i++)
			{
				NSString * tComponent;
				
				tComponent=[tArray objectAtIndex:i];
				
				if ([tComponent length]>0)
				{
					ICFileTree * nFileTree;
					
					nFileTree=[[ICFileTree alloc] initWithData:[ICFileNode folderFileNodeWithName:tComponent
																								icon:nil
																								user:tStat.st_uid
																							   group:tStat.st_gid
																						  privileges:tStat.st_mode&UF_SETTABLE]
														parent:tTree
													  children:[NSArray array]];
													  
					if (nFileTree==nil)
					{
						return nil;
					}
					
					[tTree insertSortedChild:nFileTree];
					
					tTree=nFileTree;
					
					[nFileTree release];
				}
					
				tRelativePath=[tRelativePath stringByAppendingPathComponent:tComponent];
			}
		}
		
        return tTree;
    }
    
    return nil;
}

- (NSArray *) arrayWithProjectAtPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
    NSMutableArray * nMutableArray;
    NSArray * tChildren;
    
    tChildren=[self children];
    
    nMutableArray=[NSMutableArray array];
    
    if (tChildren!=nil)
    {
        NSUInteger i,tCount;
        
        tCount=[tChildren count];
        
        for(i=0;i<tCount;i++)
        {
            NSDictionary * tChildDictionary;
            
            tChildDictionary=[[tChildren objectAtIndex:i] dictionaryWithProjectAtPath:inProjectPath referenceFolderPath:inReferenceFolderPath];
            
            [nMutableArray addObject:tChildDictionary];
        }
    }
    
    return nMutableArray;
}

- (NSDictionary *) dictionaryWithProjectAtPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
    NSMutableDictionary * nDictionary;
    NSArray * tChildren;
    NSMutableArray * tChidrenFiles;
    ICFileNode * tFileNode;
    NSNumber * tExpanded=nil;
    NSInteger tPathType;
    NSString * tPath;
    id tObject;
	
    tChildren=[self children];
    
    tChidrenFiles=[NSMutableArray array];
    
    if (tChildren!=nil)
    {
        NSUInteger i,tCount;
        
        tCount=[tChildren count];
        
        for(i=0;i<tCount;i++)
        {
            NSDictionary * tChildDictionary;
            
            tChildDictionary=[[tChildren objectAtIndex:i] dictionaryWithProjectAtPath:inProjectPath referenceFolderPath:inReferenceFolderPath];
            
            [tChidrenFiles addObject:tChildDictionary];
        }
    }
    
    tFileNode=FILENODE_DATA(self);
    
    if ([tFileNode type]==ICDOCUMENT_FILE_REALFILE_NODE)
    {
        if ([tFileNode isLeaf]==NO)
        {
            tExpanded=[NSNumber numberWithBool:YES];
        }
    }
    
    tPathType=[tFileNode pathType];
    tPath=[tFileNode path];
    
    if (inProjectPath!=nil && tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
    {
        tPath=[tPath stringByRelativizingToPath:inProjectPath];
    }
    else  if (inProjectPath!=nil && tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
    {
        tPath=[tPath stringByRelativizingToPath:inReferenceFolderPath];
    }
    else
    {
        tPathType=ICDOCUMENT_PATH_TYPE_GLOBAL;
    }
    
    nDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:tChidrenFiles,ICDOCUMENT_PACKAGE_FILES_FILE_CHILDREN,
                                                           [NSNumber numberWithInt:[tFileNode type]],ICDOCUMENT_PACKAGE_FILES_FILE_TYPE,
                                                           tPath,ICDOCUMENT_PATH,
                                                           [NSNumber numberWithInteger:tPathType],ICDOCUMENT_PATH_TYPE,
                                                           [NSNumber numberWithInt:[tFileNode uid]],ICDOCUMENT_PACKAGE_FILES_FILE_UID,
                                                           [NSNumber numberWithInt:[tFileNode gid]],ICDOCUMENT_PACKAGE_FILES_FILE_GID,
                                                           [NSNumber numberWithInt:[tFileNode privileges]],ICDOCUMENT_PACKAGE_FILES_FILE_PERMISSIONS,
                                                           tExpanded,ICDOCUMENT_PACKAGE_FILES_FILE_EXPANDED,	// Be prepared for tExpanded to be nil
                                                           nil];
    
    tObject=[tFileNode locators];
    
    if (tObject!=nil)
    {
        if ([tObject count]>0)
        {
            [nDictionary setObject:tObject forKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
        }
    }
	
	tObject=[tFileNode preInstallPath];
    
    if (tObject!=nil)
    {
        if ([tObject count]>0)
        {
            [nDictionary setObject:tObject forKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
        }
    }
	
	tObject=[tFileNode postInstallPath];
    
    if (tObject!=nil)
    {
        if ([tObject count]>0)
        {
            [nDictionary setObject:tObject forKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
        }
    }
	
	tObject=[tFileNode allowDowngradeNumber];
    
    if (tObject!=nil)
    {
        [nDictionary setObject:tObject forKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_CAN_DOWNGRADE];
    }
    
    return nDictionary;
}

- (void) recursiveAddBranchChildrenToArray:(NSMutableArray *) inMutableArray
{
	NSArray * tChildrenArray;
	
	tChildrenArray=[self children];
	
	if (tChildrenArray!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[tChildrenArray objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			ICFileTree * tChildFileTree;
			
			while (tChildFileTree=[tEnumerator nextObject])
			{
				ICFileNode * tNodeData;
	
				tNodeData=FILENODE_DATA(self);
				
				if (tNodeData!=nil)
				{
					if ([tNodeData isLeaf]==NO)
					{
						[inMutableArray addObject:tChildFileTree];
						
						[tChildFileTree recursiveAddBranchChildrenToArray:inMutableArray];
					}
				}
			}
		}
	}
}

- (NSMutableArray *) descendantFoldersIncludingSelf:(BOOL) inIncludeSelf
{
	NSMutableArray * tMutableArray=nil;
	ICFileNode * tNodeData;
	
	tNodeData=FILENODE_DATA(self);
	
	if (tNodeData!=nil)
	{
		if ([tNodeData isLeaf]==NO)
		{
			tMutableArray=[NSMutableArray array];
			
			if (tMutableArray!=nil)
			{
				if (inIncludeSelf==YES)
				{
					[tMutableArray addObject:self];
				}
				
				// A COMPLETER
			}
			else
			{
				NSLog(@"[ICFileTree descendantFoldersIncludingSelf:] Memory low");
			}
		}
		
	}
	
	return tMutableArray;
}

- (ICFileTree *) fileTreeAtPath:(NSString *) inPath
{
    if (inPath!=nil)
    {
        NSArray * tArray;
        NSUInteger i,tCount;
        ICFileTree * tTree;
		
		tTree=self;
        
        tArray=[inPath componentsSeparatedByString:@"/"];
        
        tCount=[tArray count];
        
        for(i=1;i<tCount;i++)
        {
            NSString * tComponent;
            
            tComponent=[tArray objectAtIndex:i];
            
            if ([tComponent length]>0)
            {
                NSArray * tChildren;
                NSUInteger j,tChildrenCount;
                
                tChildren=[tTree children];
                
                tChildrenCount=[tChildren count];
                
                for(j=0;j<tChildrenCount;j++)
                {
                    ICFileTree * tChild;
                    ICFileNode * tNode;
                
                    tChild=[tChildren objectAtIndex:j];
                    
                    tNode=FILENODE_DATA(tChild);
                    
                    if ([[tNode fileName] compare:tComponent]==NSOrderedSame)
                    {
                        tTree=tChild;
                        break;
                    }
                    
                }
                
                if (j==tChildrenCount)
                {
                    return nil;
                }
            }
            else
            {
                break;
            }
        }
        
        return tTree;
    }
    
    return nil;
}

- (NSString *) filePath
{
    if ([self nodeParent]!=nil)
	{
		NSString * tString=nil;
		NSMutableString * tMutableString;
		ICFileTree * tFileTree=self;
		
		tMutableString=[NSMutableString stringWithCapacity:1024];
		
		do
		{
			ICFileNode * tFileNode;
			NSString * tFileName;
			
			tFileNode=FILENODE_DATA(tFileTree);
			
			tFileName=[tFileNode fileName];
			
			if (tFileName!=nil)
			{
				if ([tFileName isEqualToString:@"/"]==NO)
				{
					[tMutableString insertString:[tFileNode fileName] atIndex:0];
					[tMutableString insertString:@"/" atIndex:0];
				
					tFileTree=(ICFileTree *) [tFileTree nodeParent];
				}
				else
				{
					if (tFileTree==self)
					{
						[tMutableString insertString:@"/" atIndex:0];
					}
				
					tFileTree=nil;
				}
			}
			else
			{
				tFileTree=nil;
			}
		}
		while (tFileTree!=nil);
		
		if ([tMutableString length]>0)
		{
			tString=[tMutableString copy];
		}
		
		return [tString autorelease];
	}
	
	return @"";
}

- (BOOL) containsTreeWithName:(NSString *) inName
{
    NSArray * tChildren;
    
    tChildren=[self children];
    
    if (tChildren!=nil)
    {
        NSUInteger i,tCount;
        ICFileNode * tNode;
        NSComparisonResult tResult;
        
        tCount=[tChildren count];
        
        for(i=0;i<tCount;i++)
        {
            tNode=FILENODE_DATA([tChildren objectAtIndex:i]);
            
            tResult=[[[tNode path] lastPathComponent] compare:inName options:NSCaseInsensitiveSearch];
            
            if (tResult==NSOrderedSame)
            {
                return YES;
            }
            /*else if (tResult==NSOrderedAscending)	// TO OPTIMIZE (decommenter quand le tri sera fait)
            {
                break;
            }*/
        }
    }
    
    return NO;
}

- (void) insertSortedChildReverse:(ICFileTree *) inChild
{
    NSArray * tChildren;
    NSUInteger tIndex=NSNotFound;
    
    tChildren=[self children];
    
    if (tChildren!=nil)
    {
        NSUInteger i,tCount;
        ICFileNode * tNode;
        NSComparisonResult tResult;
        NSString * tName;
        
        tName=[[FILENODE_DATA(inChild) path] lastPathComponent];
        
        tCount=[tChildren count];
        
		for(i=tCount;i>0;i--)
        {
            tNode=FILENODE_DATA([tChildren objectAtIndex:i-1]);
            
            tResult=[tName compare:[[tNode path] lastPathComponent] options:NSCaseInsensitiveSearch];
            
            if (tResult!=NSOrderedAscending)
            {
                tIndex=i;
                break;
            }
        }
        
        if (tIndex==NSNotFound)
        {
            tIndex=0;
        }
    }
    else
    {
        tIndex=0;
    }
    
    [self insertChild:inChild atIndex:tIndex];
}

- (void) insertSortedChild:(ICFileTree *) inChild
{
    NSArray * tChildren;
    NSUInteger tIndex=NSNotFound;
    
    tChildren=[self children];
    
    if (tChildren!=nil)
    {
        NSUInteger i,tCount;
        ICFileNode * tNode;
        NSComparisonResult tResult;
        NSString * tName;
        
        tName=[[FILENODE_DATA(inChild) path] lastPathComponent];
        
        tCount=[tChildren count];
        
        for(i=0;i<tCount;i++)
        {
            tNode=FILENODE_DATA([tChildren objectAtIndex:i]);
            
            tResult=[tName compare:[[tNode path] lastPathComponent] options:NSCaseInsensitiveSearch];
            
            if (tResult!=NSOrderedDescending)
            {
                tIndex=i;
                break;
            }
        }
        
        if (tIndex==NSNotFound)
        {
            tIndex=tCount;
        }
    }
    else
    {
        tIndex=0;
    }
    
    [self insertChild:inChild atIndex:tIndex];
}

- (void) insertSortedChildren:(NSArray *) inChildren
{
}

- (BOOL) expandAll:(BOOL) inKeepOwnerAndGroup withProjectPath:(NSString *) inProjectPath referenceFolderPath:(NSString *) inReferenceFolderPath
{
	NSUInteger tCount;
	
	tCount=[self numberOfChildren];
        
	if (tCount>0)
	{
		BOOL tExpandedOne=NO;
		NSUInteger i;
		
		for(i=0;i<tCount;i++)
		{
			ICFileTree * tChild;
			
			tChild=(ICFileTree *) [self childAtIndex:i];
			
			if ([tChild expandAll:inKeepOwnerAndGroup withProjectPath:inProjectPath referenceFolderPath:inReferenceFolderPath]==YES)
			{
				tExpandedOne=YES;
			}
		}
		
		return tExpandedOne;
	}
	else
	{
		ICFileNode * tFileNode;
	
		tFileNode=FILENODE_DATA(self);
    
		if ([tFileNode type]==ICDOCUMENT_FILE_REALFILE_NODE)
		{
			NSInteger tPathType;
			NSString * tPath;
			NSDictionary * tDictionary;
			
			tPathType=[tFileNode pathType];
			
			tPath=[tFileNode path];
    
			if (tPath!=nil)
			{
				if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
				{
					// Compute the Absolute Path
			
					tPath=[tPath stringByAbsolutingWithPath:inProjectPath];
				}
				else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
				{
					// Compute the Absolute Path
			
					tPath=[tPath stringByAbsolutingWithPath:inReferenceFolderPath];
				}
			}
			
			tDictionary=[[NSFileManager defaultManager] fileAttributesAtPath:tPath traverseLink:NO];
		
			if ([[tDictionary objectForKey:NSFileType] isEqualToString:NSFileTypeDirectory]==YES)
			{
				[self expand:inKeepOwnerAndGroup];
				
				return YES;
			}
		}
    }
	
	return NO;
}

- (void) expand:(BOOL) inKeepOwnerAndGroup
{
    NSFileManager * tFileManager;
    NSString * tPath;
    ICFileTree * tChild;
    NSArray * directoryContent;
    NSUInteger i,tCount;
    ICFileNode * tTargetNode;
    NSString * tSelfPath;
    
    tTargetNode=FILENODE_DATA(self);
    
    tSelfPath=[tTargetNode path];
    
    tFileManager=[NSFileManager defaultManager];
    
    directoryContent=[tFileManager directoryContentsAtPath:tSelfPath];
    
    tCount=[directoryContent count];
    
    for(i=0;i<tCount;i++)
    {
        id tFileNode;
        
        tPath=[tSelfPath stringByAppendingPathComponent:[directoryContent objectAtIndex:i]];
        
		tFileNode=[ICFileNode fileNodeWithType:ICDOCUMENT_FILE_REALFILE_NODE
                                          path:tPath
                                      pathType:[tTargetNode pathType]];
                                                
        if (tFileNode!=nil)
        {
            NSDictionary * tFileAttributes;
            NSString * tString;
            NSNumber * tNumber;
            
            tFileAttributes=[tFileManager fileAttributesAtPath:tPath traverseLink:NO];
            
            tNumber=[tFileAttributes objectForKey:NSFilePosixPermissions];
                    
            if (tNumber!=nil)
            {
                [tFileNode setPrivileges:[tNumber intValue]];
            }
            else
            {
                // A COMPLETER
            }
            
            if (inKeepOwnerAndGroup==YES)
            {
                tNumber=[tFileAttributes objectForKey:NSFileOwnerAccountID];
                        
                if (tNumber!=nil)
                {
                    [tFileNode setUid:[tNumber intValue]];
                }
                else
                {
                    [tFileNode setUid:[tTargetNode uid]];
                }
                
                tNumber=[tFileAttributes objectForKey:NSFileGroupOwnerAccountID];
                
                if (tNumber!=nil)
                {
                    [tFileNode setGid:[tNumber intValue]];
                }
                else
                {
                    [tFileNode setGid:[tTargetNode gid]];
                }
            }
            else
            {
                [tFileNode setUid:[tTargetNode uid]];
            
                [tFileNode setGid:[tTargetNode gid]];
            }
            
            tString=[tFileAttributes objectForKey:NSFileType];
            
            if (tString!=nil && [tString isEqualToString:NSFileTypeDirectory]==YES)
            {
                tChild=[[ICFileTree alloc] initWithData:tFileNode
                                                 parent:nil
                                               children:[NSArray array]];
                
                [tChild expand:inKeepOwnerAndGroup];
            }
            else
            {
            	tChild=[[ICFileTree alloc] initWithData:tFileNode
                                                 parent:nil
                                               children:nil];

            }
            
            [self insertChild:tChild atIndex:i];
                                
            [tChild release];
        }
    }
    
    [tTargetNode setFolder:YES];
}

- (void) expandOneLevel:(BOOL) inKeepOwnerAndGroup
{
    NSFileManager * tFileManager;
    NSString * tPath;
    ICFileTree * tChild;
    NSArray * directoryContent;
    NSUInteger i,tCount;
    ICFileNode * tTargetNode;
    NSString * tSelfPath;
    
    tTargetNode=FILENODE_DATA(self);
    
    tSelfPath=[tTargetNode path];
    
    tFileManager=[NSFileManager defaultManager];
    
    directoryContent=[tFileManager directoryContentsAtPath:tSelfPath];
    
    tCount=[directoryContent count];
    
    for(i=0;i<tCount;i++)
    {
        id tFileNode;
        
        tPath=[tSelfPath stringByAppendingPathComponent:[directoryContent objectAtIndex:i]];
        
	tFileNode=[ICFileNode fileNodeWithType:ICDOCUMENT_FILE_REALFILE_NODE
                                          path:tPath
                                      pathType:[tTargetNode pathType]];
                                                
        if (tFileNode!=nil)
        {
            NSDictionary * tFileAttributes;
            NSNumber * tNumber;
            
            tFileAttributes=[tFileManager fileAttributesAtPath:tPath traverseLink:NO];
            
            tNumber=[tFileAttributes objectForKey:NSFilePosixPermissions];
                    
            if (tNumber!=nil)
            {
                [tFileNode setPrivileges:[tNumber intValue]];
            }
            
            if (inKeepOwnerAndGroup==YES)
            {
                tNumber=[tFileAttributes objectForKey:NSFileOwnerAccountID];
                        
                if (tNumber!=nil)
                {
                    [tFileNode setUid:[tNumber intValue]];
                }
                else
                {
                    [tFileNode setUid:[tTargetNode uid]];
                }
                
                tNumber=[tFileAttributes objectForKey:NSFileGroupOwnerAccountID];
                
                if (tNumber!=nil)
                {
                    [tFileNode setGid:[tNumber intValue]];
                }
                else
                {
                    [tFileNode setGid:[tTargetNode gid]];
                }
            }
            else
            {
                [tFileNode setUid:[tTargetNode uid]];
            
                [tFileNode setGid:[tTargetNode gid]];
            }
            
            tChild=[[ICFileTree alloc] initWithData:tFileNode
                                             parent:nil
                                           children:nil];
            
            [self insertChild:tChild atIndex:i];
                                
            [tChild release];
        }
    }
    
    [tTargetNode setFolder:YES];
}

- (void) contract
{
    ICFileNode * tTargetNode;
    
    tTargetNode=FILENODE_DATA(self);
    
    [nodeChildren makeObjectsPerformSelector:@selector(setNodeParent:) withObject:nil];
    
    [nodeChildren release];
    
    nodeChildren=[[NSMutableArray array] retain];
    
    [tTargetNode setFolder:NO];
}

- (BOOL) isItemExcludedByBouncer:(ICFileBouncer *) inBouncer
{
	if (inBouncer!=nil)
	{
		ICFileNode * tFileNodeData;
		
		tFileNodeData=FILENODE_DATA(self);
		
		if (tFileNodeData!=nil)
		{
			int tType;
			
			tType=[tFileNodeData type];
			
			if (tType!=ICDOCUMENT_FILE_ROOT_NODE)
			{
				NSString * tFileName;
				
				tFileName=[tFileNodeData fileName];
				
				if (tFileName!=nil)
				{
					BOOL tShouldBeExcluded=NO;
					
					switch(tType)
					{
						case ICDOCUMENT_FILE_BASE_NODE_INVISIBLE:
						case ICDOCUMENT_FILE_BASE_NODE:
						case ICDOCUMENT_FILE_NEWFOLDER_NODE:
						
							tShouldBeExcluded=[inBouncer shouldBounceItemNamed:tFileName ofType:ICFILE_TYPE_FOLDER];
							
							break;
							
						case ICDOCUMENT_FILE_REALFILE_NODE:
						
							{
								NSString * tPath;
								static NSFileManager * sFileManager=nil;
								BOOL isDirectory;
								
								if (sFileManager==nil)
								{
									sFileManager=[NSFileManager defaultManager];
								}
								
								tPath=[tFileNodeData path];
								
								if ([sFileManager fileExistsAtPath:tPath isDirectory:&isDirectory]==YES)
								{
									if (isDirectory==YES)
									{
										tShouldBeExcluded=[inBouncer shouldBounceItemNamed:tFileName ofType:ICFILE_TYPE_FOLDER];
									}
									else
									{
										tShouldBeExcluded=[inBouncer shouldBounceItemNamed:tFileName ofType:ICFILE_TYPE_FILE];
									}
								}
							}
							
							break;
					}
					
					if (tShouldBeExcluded==YES)
					{
						return YES;
					}
					else
					{
						ICFileTree * tParent;
						
						tParent=(ICFileTree *) [self nodeParent];
						
						if (tParent!=nil)
						{
							return [tParent isItemExcludedByBouncer:inBouncer];
						}
					}
				}
			}
		}
	}
	
	return NO;
}

- (BOOL) hasRealChildren
{
    NSUInteger tCount,i;
    
    if ([FILENODE_DATA(self) type]>ICDOCUMENT_FILE_BASE_NODE)
    {
        return YES;
    }
    else
    {
        BOOL tResult;
        
        tCount=[self numberOfChildren];
        
        for(i=0;i<tCount;i++)
        {
            tResult=[((ICFileTree *) [self childAtIndex:i]) hasRealChildren];
            
            if (tResult==YES)
            {
                return YES;
            }
        }
    }
    
    return NO;
}

+ (NSString *) uniqueNameWithParentFileTree:(ICFileTree *) inParentTree
{
    NSUInteger _sIndex=0;
    static NSString * tLocalizedBaseName=nil;
    NSUInteger i,tCount;
    NSString * tString;
    
    if (tLocalizedBaseName==nil)
    {
        tLocalizedBaseName=[NSLocalizedString(@"untitled folder",@"No comment") retain];
    }
    
    tCount=[inParentTree numberOfChildren];
    
    do
    {
        ICFileTree * tFileTree;
        
        if (_sIndex>0)
        {
            tString=[[NSString alloc] initWithFormat:@"%@ %lu",tLocalizedBaseName,_sIndex];
        }
        else
        {
            tString=[[NSString alloc] initWithString:tLocalizedBaseName];
        }
        
        _sIndex++;
        
        for(i=0;i<tCount;i++)
        {
            tFileTree=(ICFileTree *) [inParentTree childAtIndex:i];
            
            if ([tString isEqualToString:[FILENODE_DATA(tFileTree) fileName]]==YES)
            {
                break;
            }
        }
        
        if (i==tCount)
        {
            break;
        }
        
        [tString release];
    }
    while (_sIndex<65535);
    
    return [tString autorelease];
}

@end