/*
Copyright (c) 2007-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICLocalizationUtilities.h"

#import "ICDocumentFormat+Constants.h"

#import "ICLanguageConverter+Native.h"

#import "NSString+Iceberg.h"

@implementation NSDictionary (ICLocalizationUtilities)

- (NSComparisonResult) compareLocalizationLanguage:(NSDictionary *) inDictionary
{
	return [[self objectForKey:ICDOCUMENT_LANGUAGE] caseInsensitiveCompare:[inDictionary objectForKey:ICDOCUMENT_LANGUAGE]];
}

@end

@implementation ICLocalizationUtilities

+ (BOOL) allLanguagesUsedInLocalizations:(NSArray *) inLocalizationsArray
{
	if (inLocalizationsArray!=nil)
	{
		return ([[ICLocalizationUtilities englishLanguages] count]==[inLocalizationsArray count]);
	}

	return NO;
}




+ (NSMutableArray *) englishLanguages
{
	static NSMutableArray * sEnglishLanguageNames=nil;
	
	if (sEnglishLanguageNames==nil)
	{
		sEnglishLanguageNames=[[NSMutableArray alloc] initWithObjects:@"Danish",
																	  @"Dutch",
																	  @"English",
																	  @"Finnish",
																	  @"French",
																	  @"German",
																	  @"Italian",
																	  @"Japanese",
																	  @"Korean",
																	  @"Norwegian",
																	  @"Polish",
																	  @"Brazilian Portuguese",
																	  @"Portuguese",
																	  @"Russian",
																	  @"Spanish",
																	  @"Swedish",
																	  @"Simplified Chinese",
																	  @"Traditional Chinese",
																	  nil];
	
		[sEnglishLanguageNames sortUsingSelector:@selector(caseInsensitiveCompare:)];
	}
	
	return sEnglishLanguageNames;
}

+ (NSMenu *) localizationMenu
{
	static NSMenu * sMenu=nil;
	
	if (sMenu==nil)
	{
		ICLanguageConverter * tLanguageConverter;
		NSMutableArray * tEnglishLanguageNames;
		
		tLanguageConverter=[ICLanguageConverter defaultConverter];
		
		tEnglishLanguageNames=[ICLocalizationUtilities englishLanguages];
		
		if (tEnglishLanguageNames!=nil)
		{
			// Build Menu
			
			sMenu=[[NSMenu allocWithZone:[NSMenu menuZone]] initWithTitle:@""];
			
			if (sMenu!=nil)
			{
				NSEnumerator * tEnumerator;
				
				tEnumerator=[tEnglishLanguageNames objectEnumerator];
				
				if (tEnumerator!=nil)
				{
					NSString * tLanguageName;
					
					while (tLanguageName=[tEnumerator nextObject])
					{
						NSMenuItem * tMenuItem;
						
						tMenuItem=[[NSMenuItem alloc] initWithTitle:tLanguageName action:nil keyEquivalent:@""];
						
						if (tMenuItem!=nil)
						{
							NSImage * tImage;
							NSString * tISOName;
							
							tISOName=[tLanguageConverter ISOFromEnglish:tLanguageName];
							
							if (tISOName!=nil)
							{
								tImage=[NSImage imageNamed:[NSString stringWithFormat:@"flag_%@",tISOName]];
				
								if (tImage!=nil)
								{
									[tMenuItem setImage:tImage];
								}
							}
							
							[sMenu addItem:tMenuItem];
							
							[tMenuItem release];
						}
					}
				}
			}
		}
	}
	
	return [[sMenu copyWithZone:[NSMenu menuZone]] autorelease];;
}

+ (NSArray *) localizationsFromLocalizationsArray:(NSArray *) inLocalizationsArray
{
	NSMutableArray * tLocalizations=nil;
	
	if (inLocalizationsArray!=nil && [inLocalizationsArray count]>0)
	{
		tLocalizations=[NSMutableArray arrayWithCapacity:[inLocalizationsArray count]];
		
		if (tLocalizations!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[inLocalizationsArray objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSDictionary * tItemDictionary;
				
				while (tItemDictionary=[tEnumerator nextObject])
				{
					NSString * tLanguage;
					
					tLanguage=[tItemDictionary objectForKey:ICDOCUMENT_LANGUAGE];
					
					if (tLanguage!=nil)
					{
						[tLocalizations addObject:tLanguage];
					}
					else
					{
						[tLocalizations addObject:@""];
					}
				}
			}
		}
	}
	
	return tLocalizations;
}

+ (id) localizedValueForLanguage:(NSString *) inLanguage inLocalizations:(NSArray *) inLocalizationsArray lookForBestMatch:(BOOL) inLookForBestMatch
{
	id tLocalizedValue=nil;
	
	if (inLanguage!=nil && inLocalizationsArray!=nil)
	{
		NSArray * tAvailableLocalizations;
		
		tAvailableLocalizations=[ICLocalizationUtilities localizationsFromLocalizationsArray:inLocalizationsArray];
		
		if (tAvailableLocalizations!=nil)
		{
			NSUInteger tIndex;
			
			tIndex=[tAvailableLocalizations indexOfObject:inLanguage];
			
			if (tIndex==NSNotFound && inLookForBestMatch==YES)
			{
				// Try with the ISO name
				
				inLanguage=[[ICLanguageConverter defaultConverter] ISOFromEnglish:inLanguage];
				
				tIndex=[tAvailableLocalizations indexOfObject:inLanguage];
			
				if (tIndex==NSNotFound)
				{
					NSArray * tPreferedLocalizations;
				
					tPreferedLocalizations=(NSArray *) CFBundleCopyPreferredLocalizationsFromArray((CFArrayRef) tAvailableLocalizations);
				
					if (tPreferedLocalizations!=nil)
					{
						NSEnumerator * tEnumerator;
						
						tEnumerator=[tPreferedLocalizations objectEnumerator];
						
						if (tEnumerator!=nil)
						{
							NSString * tLanguage;
							
							while (tLanguage=[tEnumerator nextObject])
							{
								tIndex=[tAvailableLocalizations indexOfObject:tLanguage];
								
								if (tIndex!=NSNotFound)
								{
									break;
								}
								else
								{
									tLanguage=[[ICLanguageConverter defaultConverter] ISOFromEnglish:tLanguage];
									
									tIndex=[tAvailableLocalizations indexOfObject:tLanguage];
								
									if (tIndex!=NSNotFound)
									{
										break;
									}
								}
							}
						}
						
						[tPreferedLocalizations release];
					}
				}
            }
			
			if (tIndex!=NSNotFound)
			{
				NSDictionary * tDictionary;
				
				tDictionary=[inLocalizationsArray objectAtIndex:tIndex];
				
				if (tDictionary!=nil)
				{
					tLocalizedValue=[tDictionary objectForKey:ICDOCUMENT_VALUE];
				}
			}
		}
	}
	
	return tLocalizedValue;
}

+ (id) localizedValueForLanguage:(NSString *) inLanguage inLocalizations:(NSArray *) inLocalizationsArray
{
	return [ICLocalizationUtilities localizedValueForLanguage:inLanguage inLocalizations:inLocalizationsArray lookForBestMatch:YES];
}

+ (NSString *) firstNewLanguageInLocalizations:(NSArray *) inLocalizationsArray
{
	NSString * tLanguageName=nil;
	
	if (inLocalizationsArray!=nil)
	{
		NSMutableArray * tEnglishLanguageNames;
		
		tEnglishLanguageNames=[ICLocalizationUtilities englishLanguages];
		
		if (tEnglishLanguageNames!=nil)
		{
			NSArray * tPreferedLocalizations;
			
			if ([inLocalizationsArray count]==0)
			{
				tPreferedLocalizations=(NSArray *) CFBundleCopyPreferredLocalizationsFromArray((CFArrayRef) tEnglishLanguageNames);
				
				[tPreferedLocalizations autorelease];
				
				if (tPreferedLocalizations!=nil && [tPreferedLocalizations count]>0)
				{
					return [[ICLanguageConverter defaultConverter] englishFromISO:[tPreferedLocalizations objectAtIndex:0]];
				}
			}
			else
			{
				tEnglishLanguageNames=[[tEnglishLanguageNames  mutableCopy] autorelease];
		
				if (tEnglishLanguageNames!=nil)
				{
					NSEnumerator * tEnumerator;
					
					tEnumerator=[inLocalizationsArray objectEnumerator];
			
					if (tEnumerator!=nil)
					{
						NSDictionary * tDictionary;
						
						while (tDictionary=[tEnumerator nextObject])
						{
							NSString * tUsedLanguage;
							
							tUsedLanguage=[tDictionary objectForKey:ICDOCUMENT_LANGUAGE];
							
							if (tUsedLanguage!=nil)
							{
								[tEnglishLanguageNames removeObject:tUsedLanguage];
							}
						}
					}
			
					if ([tEnglishLanguageNames count]>0)
					{
						if ([tEnglishLanguageNames count]==1)
						{
							return [tEnglishLanguageNames objectAtIndex:0];
						}
						else
						{
							tPreferedLocalizations=(NSArray *) CFBundleCopyPreferredLocalizationsFromArray((CFArrayRef) tEnglishLanguageNames);
				
							[tPreferedLocalizations autorelease];
							
							if (tPreferedLocalizations!=nil && [tPreferedLocalizations count]>0)
							{
								tLanguageName=[[ICLanguageConverter defaultConverter] englishFromISO:[tPreferedLocalizations objectAtIndex:0]];
								
								if ([tEnglishLanguageNames containsObject:tLanguageName]==NO)	// Workaround for bug in Cocoa
								{
									tLanguageName=[tEnglishLanguageNames objectAtIndex:0];
								}
							}
						}
					}
				}
			}
		}
	}
	
	return tLanguageName;
}

+ (NSArray *) allLanguagesFromLocalizations:(NSArray *) inLocalizationsArray
{
	NSMutableArray * tAllLanguages=nil;
	
	if (inLocalizationsArray!=nil)
	{
		NSEnumerator * tEnumerator;
					
		tEnumerator=[inLocalizationsArray objectEnumerator];

		if (tEnumerator!=nil)
		{
			NSDictionary * tDictionary;
			
			tAllLanguages=[NSMutableArray array];
			
			while (tDictionary=[tEnumerator nextObject])
			{
				NSString * tUsedLanguage;
				
				tUsedLanguage=[tDictionary objectForKey:ICDOCUMENT_LANGUAGE];
				
				[tAllLanguages addObject:[[ICLanguageConverter defaultConverter] englishFromISO:tUsedLanguage]];
			}
		}
	}
	
	return tAllLanguages;
}

+ (BOOL) localizations:(NSArray *) inLocalizationsArray containsLanguage:(NSString *) inLanguage
{
	if (inLocalizationsArray!=nil && inLanguage!=nil)
	{
		NSEnumerator * tEnumerator;
					
		tEnumerator=[inLocalizationsArray objectEnumerator];

		if (tEnumerator!=nil)
		{
			NSDictionary * tDictionary;
			
			while (tDictionary=[tEnumerator nextObject])
			{
				NSString * tUsedLanguage;
				
				tUsedLanguage=[tDictionary objectForKey:ICDOCUMENT_LANGUAGE];
				
				if ([tUsedLanguage isEqualToString:inLanguage]==YES)
				{
					return YES;
				}
			}
		}
	}
	
	return NO;
}

+ (NSUInteger) indexOfLanguage:(NSString *) inLanguage inLocalizations:(NSArray *) inLocalizationsArray
{
	if (inLocalizationsArray!=nil && inLanguage!=nil)
	{
		NSUInteger i,tCount;
		
		tCount=[inLocalizationsArray count];
		
		for(i=0;i<tCount;i++)
		{
			NSDictionary * tDictionary;
			NSString * tUsedLanguage;
			
			tDictionary=[inLocalizationsArray objectAtIndex:i];
			
			tUsedLanguage=[tDictionary objectForKey:ICDOCUMENT_LANGUAGE];
				
			if ([tUsedLanguage isEqualToString:inLanguage]==YES)
			{
				return i;
			}
		}
	}
	
	return NSNotFound;
}

+ (NSMutableDictionary *) validLocalizationsPathFromLocalizationsArray:(NSArray *) inLocalizationsArray  projectPath:(NSString *) inProjectPath referencePath:(NSString *) inReferencePath
{
	NSMutableDictionary * tValidLocalizationsPath=nil;
	
	if (inLocalizationsArray!=nil && inProjectPath!=nil && inReferencePath!=nil)
	{
		tValidLocalizationsPath=[NSMutableDictionary dictionary];
		
		if (tValidLocalizationsPath!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[inLocalizationsArray objectEnumerator];

			if (tEnumerator!=nil)
			{
				NSDictionary * tDictionary;
				NSFileManager * tFileManager;
				
				tFileManager=[NSFileManager defaultManager];
				
				
				while (tDictionary=[tEnumerator nextObject])
				{
					NSString * tLanguage;
					NSDictionary * tPathDictionary;
					
					tLanguage=[[ICLanguageConverter defaultConverter] englishFromISO:[tDictionary objectForKey:ICDOCUMENT_LANGUAGE]];
					
					tPathDictionary=[tDictionary objectForKey:ICDOCUMENT_VALUE];
					
					if (tPathDictionary!=nil)
					{
						NSString * tPath;
						NSInteger tPathType;
						
						tPathType=[[tPathDictionary objectForKey:ICDOCUMENT_PATH_TYPE] integerValue];
							
						tPath=[tPathDictionary objectForKey:ICDOCUMENT_PATH];
						
						if (tPath!=nil && [tPath length]>0)
						{
							if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
							{
								tPath=[tPath stringByAbsolutingWithPath:inProjectPath];
							}
							else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
							{
								tPath=[tPath stringByAbsolutingWithPath:inReferencePath];
							}
								
							if ([tFileManager fileExistsAtPath:tPath]==YES)
							{
								NSString * tNativeLanguage;
								
								tNativeLanguage=[[ICLanguageConverter defaultConverter] nativeForEnglish:tLanguage];
								
								[tValidLocalizationsPath setObject:[NSDictionary dictionaryWithObjectsAndKeys:tPath,ICDOCUMENT_PATH,
																											  tLanguage,ICDOCUMENT_LANGUAGE,
																											  nil]
															forKey:tNativeLanguage];
							}
						}
					}
				}
			}
		}
	}
	
	return tValidLocalizationsPath;
}

+ (NSMutableArray *) localizationsArrayWithValue:(id) inValue forLanguage:(NSString *) inLanguage
{
	NSMutableArray * tLocalizationsArray;
										
	tLocalizationsArray=[NSMutableArray array];

	if (tLocalizationsArray!=nil)
	{
		NSMutableDictionary * tLanguageDictionary;
		
		tLanguageDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:inLanguage,ICDOCUMENT_LANGUAGE,
																				  inValue,ICDOCUMENT_VALUE,
																				  nil];

		if (tLanguageDictionary!=nil)
		{
			[tLocalizationsArray addObject:tLanguageDictionary];
		}
	}
	
	return tLocalizationsArray;
}

+ (NSString *) possibleLanguageForFileAtPath:(NSString *) inPath
{
	if (inPath!=nil)
	{
		NSString * tParentFolder;
		NSString * tExtension;
		
		tParentFolder=[[inPath stringByDeletingLastPathComponent] lastPathComponent];
		
		tExtension=[tParentFolder pathExtension];
		
		if ([tExtension length]==0 || [tExtension caseInsensitiveCompare:@"lproj"]==NSOrderedSame)
		{
			NSString * tLanguage;
			
			tLanguage=[tParentFolder stringByDeletingPathExtension];
			
			if (tLanguage!=nil)
			{
				NSString * tEnglishLanguage;
				ICLanguageConverter * tDefaultConverter;
				
				tDefaultConverter=[ICLanguageConverter defaultConverter];
				
				tEnglishLanguage=[tDefaultConverter englishFromISO:tLanguage];
				
				if ([[tDefaultConverter allEnglishNames] containsObject:tEnglishLanguage]==YES)
				{
					return tEnglishLanguage;
				}
				
				tEnglishLanguage=[tDefaultConverter englishFromISO:[tEnglishLanguage lowercaseString]];
				
				if ([[tDefaultConverter allEnglishNames] containsObject:tEnglishLanguage]==YES)
				{
						return tEnglishLanguage;
				}
					
				tEnglishLanguage=[tDefaultConverter englishFromISO:[tEnglishLanguage capitalizedString]];
				
				if ([[tDefaultConverter allEnglishNames] containsObject:tEnglishLanguage]==YES)
				{
						return tEnglishLanguage;
				}
			}
		}
	}
	
	return nil;
}

@end
