/*
 Copyright (c) 2008-2010, Stephane Sudre
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 
 - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 - Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "ICPackageFilesController.h"

#import "ICPackageFilesHierarchyOutlineView.h"

#include <sys/types.h>

#import "ICPackageFileUtility.h"

#import "ICSmartLocationDetector.h"

// Inspectors

#import "ICPackageFilesInspectorSelectionController.h"


@implementation ICPackageFilesController

+ (ICPackageFilesController *) packageFilesController
{
	ICPackageFilesController * nController=nil;
    
    nController=[ICPackageFilesController alloc];
    
    if (nController!=nil)
    {
        if ([NSBundle loadNibNamed:@"Package_Files" owner:nController]==NO)
        {
            NSBeep();
            
            NSRunCriticalAlertPanel([NSString stringWithFormat:NSLocalizedString(@"Packages can't find the \"%@\".nib resource file.",@"No comment"),@"Package_Files"],NSLocalizedString(@"Packages will stop running to avoid improper behavior.\nYou should check that the application has not been corrupted.",@"No comment"), nil,nil,nil);
            
            [NSApp terminate:nil];
        }
    }
    
    return [nController autorelease];
}

- (void) dealloc
{
	
	
	[super dealloc];
}

#pragma mark -

- (void) awakeFromNib
{
	NSTableColumn * tTableColumn;
    id tPrototypeCell;
	NSFont * tSmallLabelFont;
	
	[super awakeFromNib];
	
	tSmallLabelFont=[NSFont labelFontOfSize:[NSFont smallSystemFontSize]];
	
	// Outline View
	
    tTableColumn = [IBoutlineView_ tableColumnWithIdentifier: @"Owner"];
	
    tPrototypeCell=[tTableColumn dataCell];
	
	if (tPrototypeCell!=nil)
	{
		[tPrototypeCell setFont:tSmallLabelFont];
    }
	
    tTableColumn = [IBoutlineView_ tableColumnWithIdentifier: @"Group"];
    
	tPrototypeCell=[tTableColumn dataCell];
	
	if (tPrototypeCell!=nil)
	{
		[tPrototypeCell setFont:tSmallLabelFont];
    }
    
    
	
	[IBoutlineView_ registerForDraggedTypes:[NSArray arrayWithObjects:NSFilenamesPboardType,
                                                                      ICFilePboardType,
                                                                      ICFileExternalPboardType,
                                                                      nil]];
	
	directoryServicesManager_=[ICDirectoryServicesManager defaultManager];
	
	// Register for Notifications
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(advancedModeDidChange:) 
												 name:ICPREFERENCES_NOTIFICATION_ADVANCED_ADVANCED_MODE_DID_CHANGE
											   object:nil];
}

#pragma mark -

- (void) initializeWithDictionary:(NSMutableDictionary *) inDictionary forDocument:(id) inDocument
{	
	[super initializeWithDictionary:inDictionary forDocument:inDocument];

	dictionary_=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES];
	
	if (dictionary_==nil)
	{
		// Build the default dictionary
		
		dictionary_=[ICPackageFileUtility defaultPackageFileDictionary];
		
		if (dictionary_!=nil)
		{
			[inDictionary setObject:dictionary_ forKey:ICDOCUMENT_PACKAGE_FILES];
		}
	}
	
	if (dictionary_!=nil)
	{
		NSNumber * tNumber;
		
		[cachedPackageUUID_ release];
		
		cachedPackageUUID_=[[inDictionary objectForKey:ICDOCUMENT_PACKAGE_UUID] retain];
		
		[document_ updateFileBouncer];
		
		cachedFileBouncer_=[[document_ fileBouncer] retain];
		
		[dictionary_ retain];
		
		tNumber=[dictionary_ objectForKey:ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE];
		
		if (tNumber!=nil)
		{
			int tPayloadType;
		
			tPayloadType=[tNumber intValue];
			
			if (tPayloadType==ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE_INTERNAL)
			{
				BOOL tState;
				
				tState=[defaults_ boolForKey:ICPREFERENCES_ADVANCED_ADVANCED_MODE_STATE];
				
				[self setHiddenAdvancedOptions:(tState==NO)];
				
				[self performSelector:@selector(delayedHierarchyLoad:) withObject:nil afterDelay:0.1f];
			}
			else if (tPayloadType==ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE_EXTERNAL)
			{
				[IBsplitForkCheckBox_ setHidden:YES];
				
				// A COMPLETER
			}
			else
			{
				NSLog(@"[ initializeWithDictionary:forDocument:] Incorrect Payload Type");
			}
			
			// Register for Notifications
			
			[[NSNotificationCenter defaultCenter] removeObserver:self name:ICFILEHIERARCHYINSPECTOR_NOTIFICATION_FOLDERNAME_SHOULD_CHANGE object:document_];
			
			[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(folderNameShouldChange:)
													 name:ICFILEHIERARCHYINSPECTOR_NOTIFICATION_FOLDERNAME_SHOULD_CHANGE
												   object:document_];
		
			[[NSNotificationCenter defaultCenter] removeObserver:self name:ICFILEHIERARCHYINSPECTOR_NOTIFICATION_ATTRIBUTES_DID_CHANGE object:document_];
			
			[[NSNotificationCenter defaultCenter] addObserver:self
													 selector:@selector(fileAttributesDidChange:)
														 name:ICFILEHIERARCHYINSPECTOR_NOTIFICATION_ATTRIBUTES_DID_CHANGE
													   object:document_];
		}
	}
}

- (void) setHiddenAdvancedOptions:(BOOL) inHidden
{
	NSNumber * tNumber;
	
	tNumber=[dictionary_ objectForKey:ICDOCUMENT_PACKAGE_FILES_SPLIT_FORKS];
				
	if (inHidden==NO || (tNumber!=nil && [tNumber boolValue]==NO))
	{
		if (tNumber==nil || [tNumber boolValue]==YES)
		{
			[IBsplitForkCheckBox_ setState:NSOnState];
		}
		else
		{
			[IBsplitForkCheckBox_ setState:NSOffState];
		}
		
		
		[IBsplitForkCheckBox_ setHidden:NO];
	}
	else
	{
		[IBsplitForkCheckBox_ setHidden:YES];
	}
}

- (void) delayedHierarchyLoad:(id) sender
{
	NSString * tPath;
	
	// Default Install Location
		
	tPath=[dictionary_ objectForKey:ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION];
	
	if (tPath==nil)
	{
		tPath=@"/";
		
		[dictionary_ setObject:tPath forKey:ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION];
	}
	
	[IBdefaultLocationPath_ setStringValue:tPath];
	
	//[IBsetButton_ setEnabled
	
	// File Hierarchy
	
	fileTree_=[ICFileTree fileTreeWithDictionary:[dictionary_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY] projectPath:[document_ folder] referenceFolderPath:[document_ referenceFolderPath]];
	
	if (fileTree_!=nil)
	{
		[fileTree_ retain];
		
		[IBoutlineView_ reloadData];
		
		[self restoreDisclosedStates];
	}
	else
	{
	
		// A COMPLETER
	}
	
	defaultLocation_=[((ICFileTree *) [fileTree_ firstChild]) fileTreeAtPath:tPath];
	
	// A COMPLETER
	
	// Restore Selection
	
	[self restoreSelection];
	
	[self outlineViewSelectionDidChange:nil];
}

- (IBAction) updateFiles:(id) sender
{
	// Save Options
	
	//	Default Install Location
	
	//[dictionary_ setObject:[IBdefaultLocationPath_ stringValue] forKey:ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION];
	
	// Save File Hierarchy if needed
	
	if (hierarchyChanged_==YES)
	{
		ICFileTree * tHierarchyTree_;
		NSDictionary * tDictionary;
		
		tHierarchyTree_=(ICFileTree *) [fileTree_ childAtIndex:0];
		
		if (tHierarchyTree_!=nil)
		{
			tDictionary=[tHierarchyTree_ dictionaryWithProjectAtPath:[document_ folder] referenceFolderPath:[document_ referenceFolderPath]];
			
			if (tDictionary!=nil)
			{
				[dictionary_ setObject:tDictionary forKey:ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY];
			}
		}
		
		hierarchyChanged_=NO;
	}
	
	// Notify Document Change
			
	[self noteDocumentChange];
}

#pragma mark -

- (NSString *) disclosedStateFormatKey
{
	return ICFILE_HIERARCHY_DISCLOSED_STATE;
}

- (NSString *) selectionFormatKey
{
	return ICFILE_HIERARCHY_SELECTION;
}

#pragma mark -

- (id) outlineView:(NSOutlineView *) inOutlineView objectValueForTableColumn:(NSTableColumn *) inTableColumn byItem:(id) inItem
{
    id objectValue = nil;
    ICFileNode * tFileNode;
    
    tFileNode=FILENODE_DATA(inItem);
    
    if([[inTableColumn identifier] isEqualToString: @"File"])
    {
        objectValue = [tFileNode fileName];
	}
    else
    if([[inTableColumn identifier] isEqualToString: @"Owner"])
    {
        if ([tFileNode link]==YES)
        {
            tFileNode=FILENODE_DATA([inItem nodeParent]);
        }
        
       	objectValue=[directoryServicesManager_ userAccountForUID:[tFileNode uid]];
    }
    else
    if([[inTableColumn identifier] isEqualToString: @"Group"])
    {
        if ([tFileNode link]==YES)
        {
            tFileNode=FILENODE_DATA([inItem nodeParent]);
        }
        
        objectValue=[directoryServicesManager_ groupForGID:[tFileNode gid]];
    }
    else
    if([[inTableColumn identifier] isEqualToString: @"Permissions"])
    {
        if ([tFileNode link]==YES)
        {
            objectValue= [FILENODE_DATA([inItem nodeParent]) privilegesStringRepresentationForLink];
        }
        else
        {
            objectValue= [tFileNode privilegesStringRepresentation];
        }
    }
    
    return objectValue;
}

- (void) outlineView:(NSOutlineView *) inOutlineView willDisplayCell:(NSCell *)cell forTableColumn:(NSTableColumn *) inTableColumn item:(id) inItem
{    
    if ([[inTableColumn identifier] isEqualToString: @"File"])
    {
		ICFileNode * tFileNode;
        BOOL tFather=YES;
        NSImage * tIcon=nil;
		NSString * tFileName;
		int tFileType;
		NSColor * tColor;
		NSMutableDictionary * tMutableAttributesDictionary;
		
        tFileNode=FILENODE_DATA(inItem);
        
		tFileType=[tFileNode type];
		
		tFileName=[tFileNode fileName];
		
		tColor=[NSColor blackColor];
		
        switch(tFileType)
        {
            
			case ICDOCUMENT_FILE_BASE_NODE_INVISIBLE:
			case ICDOCUMENT_FILE_BASE_NODE:
			
                tFather=[inItem hasRealChildren];
				
				//tFilePath=[tFileNode path];
				
				if (tFather==YES)
				{
					tIcon=[ICFileNode cachedIconForBaseNodeAtPath:[inItem filePath]];
				}
				else
				{
					tIcon=[ICFileNode cachedIconForBaseNodeAtPath:[[inItem filePath] stringByAppendingString:@"+Disabled"]];
				}
				
            case ICDOCUMENT_FILE_NEWFOLDER_NODE:
				
				if (tIcon==nil)
				{
					if (tFather==YES)
					{
						tIcon=genericFolderIcon_;
					}
					else
					{
						tIcon=genericDisabledFolderIcon_;
					}
                }
				
				if (tIcon!=nil)
				{
					if (defaultLocation_==inItem)
					{
						tIcon=[ICFileNode setIconAsDefaultLocation:tIcon];
					}
					
					[(ImageAndTextCell*)cell setImage:tIcon];
				}
				
				break;
				
            case ICDOCUMENT_FILE_REALFILE_NODE:
                
                [tFileNode fileExistenceOnDiskChanged];
                
                if ([tFileNode existsOnDisk]==NO)
                {
                    tColor=[NSColor redColor];
                }
                
                [(ImageAndTextCell*)cell setImage: [FILENODE_DATA(inItem) icon]];
				
                break;
        }
		
		tMutableAttributesDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:tColor,NSForegroundColorAttributeName,
																					   [cell font],NSFontAttributeName,
																					   nil];
		
		if (cachedHighlightExcludedFiles_==YES && cachedFileBouncer_!=nil)
		{
			if ([inItem isItemExcludedByBouncer:cachedFileBouncer_]==YES)
			{
				[tMutableAttributesDictionary setObject:[NSNumber numberWithInt:NSUnderlineStyleSingle] forKey:NSStrikethroughStyleAttributeName];
			}
		}
		
		[fileNameFormatter_ setCachedAttributes:tMutableAttributesDictionary];
			
		[(ImageAndTextCell*)cell setStringValue:tFileName];
	}
}

#pragma mark -

- (BOOL) outlineView:(NSOutlineView *) inOutlineView writeItems:(NSArray*) inItems toPasteboard:(NSPasteboard*)pboard
{
    NSUInteger i,tCount;
    ICFileTree * tTree;
    NSMutableArray * tExternalDragArray;
	
    // We need to simplify the hierarchy and check no items is a base node
    
    tCount=[inItems count];
    
	[internalDragArray_ release];
	
    internalDragArray_=[[NSMutableArray alloc] initWithCapacity:tCount];
    
    tExternalDragArray=[NSMutableArray arrayWithCapacity:tCount];
    
    for(i=tCount;i>0;i--)
    {
        NSUInteger j;
		int tType;
		
		tTree=[inItems objectAtIndex:i-1];
        
		tType=[FILENODE_DATA(tTree) type];
		
        if (tType==ICDOCUMENT_FILE_BASE_NODE_INVISIBLE ||
			tType==ICDOCUMENT_FILE_BASE_NODE)
        {
            return NO;
        }
        
        for(j=i-1;j>0;j--)
        {
            if ([tTree isDescendantOfNode:[inItems objectAtIndex:j-1]]==YES)
            {
                break;
            }
        }
        
        if (j==0)
        {
            [internalDragArray_ insertObject:tTree atIndex:0];
            
            [tExternalDragArray insertObject:[tTree dictionaryWithProjectAtPath:nil referenceFolderPath:nil] atIndex:0];
        }
    }
    
    [pboard declareTypes:[NSArray arrayWithObjects: ICFilePboardType,ICFileExternalPboardType,nil] owner:self];
    
    [pboard setPropertyList:tExternalDragArray forType:ICFileExternalPboardType];
	
    [pboard setPropertyList:[NSData data] forType:ICFilePboardType]; 
    
    return YES;
}

- (NSDragOperation) outlineView:(NSOutlineView*) inOutlineView validateDrop:(id <NSDraggingInfo>)info proposedItem:(id) inItem proposedChildIndex:(NSInteger)childIndex
{
    NSPasteboard * tPasteBoard;
    NSArray * tArray;
    NSUInteger i,tCount;
    ICFileNode * tFileNode;
    
    tFileNode=FILENODE_DATA(inItem);
    
    tPasteBoard=[info draggingPasteboard];
    
    switch(childIndex)
    {
        case NSOutlineViewDropOnItemIndex:
            break;
        default:
            if ([tFileNode type]==ICDOCUMENT_FILE_ROOT_NODE)
            {
                return NSDragOperationNone;
            }
            
            if ([tPasteBoard availableTypeFromArray:[NSArray arrayWithObject: NSFilenamesPboardType]]!=nil)
            {
                NSString * tPath;
                ICFileTree * tChildTree;
                NSArray * tChildren;
                NSString * tName;
                NSUInteger k,tChildrenCount;
                BOOL find=NO;
                
                tArray=(NSArray *) [tPasteBoard propertyListForType:NSFilenamesPboardType];
                        
                tCount=[tArray count];
                        
                for(i=0;i<tCount;i++)
                {
                    tPath=[tArray objectAtIndex:i];
                            
                    if ([inItem containsTreeWithName:[tPath lastPathComponent]]==YES)
                    {
                        return NSDragOperationNone;
                    }
                }
                
                tPath=[tArray objectAtIndex:0];
                
                tName=[tPath lastPathComponent];
                
                tChildren=[inItem children];
                
                tChildrenCount=[tChildren count];
                
                for(k=0;k<tChildrenCount;k++)
                {
                    tChildTree=[tChildren objectAtIndex:k];
                    
                    if ([tName compare:[FILENODE_DATA(tChildTree) fileName] options:NSCaseInsensitiveSearch]!=NSOrderedDescending)
                    {
                        [inOutlineView setDropItem:inItem dropChildIndex:k];
                        find=YES;
                        
                        break;
                    }
                }
                
                if (find==NO)
                {
                    [inOutlineView setDropItem:inItem dropChildIndex:k];
                }
                
                return NSDragOperationCopy;//[info draggingSourceOperationMask];
            }
            else
            {
                ICFileTree * tFileTree;
                NSString * tName=nil;
                
                if ([info draggingSource]==IBoutlineView_)
                {
					if ([inItem isDescendantOfNodeInArray:internalDragArray_]==YES)
                    {
                        return NSDragOperationNone;
                    }
                    
                    // We need to check the name and eventually switch the drop location
                    
                    tCount=[internalDragArray_ count];
                    
                    for(i=0;i<tCount;i++)
                    {
                        tFileTree=[internalDragArray_ objectAtIndex:i];
                        
                        tName=[FILENODE_DATA(tFileTree) fileName];
                        
                        if ([inItem containsTreeWithName:tName]==YES)
                        {
                            return NSDragOperationNone;
                        }
                    }
                    
                    tFileTree=[internalDragArray_ objectAtIndex:0];
                    
                    tName=[FILENODE_DATA(tFileTree) fileName];
                }
                else
                {
                    // External Drag
                    
                    NSArray * tExternalArray;
                    NSDictionary * tDictionary;
                    
                    tExternalArray=[[info draggingPasteboard] propertyListForType:ICFileExternalPboardType];
                    
                    // We need to check the name and eventually switch the drop location
                    
                    tCount=[tExternalArray count];
                    
                    for(i=0;i<tCount;i++)
                    {
                        tDictionary=[tExternalArray objectAtIndex:i];
                        
                        if (tDictionary!=nil)
                        {
                            tName=[tDictionary objectForKey:ICDOCUMENT_PATH];
                        
                            if (tName==nil)
                            {
                                return NSDragOperationNone;
                            }
                            else
                            {
                                tName=[tName lastPathComponent];
                                
                                if ([inItem containsTreeWithName:tName]==YES)
                                {
                                    return NSDragOperationNone;
                                }
                            }
                        }
                    }
                    
                    tDictionary=[tExternalArray objectAtIndex:0];
                    
                    tName=[[tDictionary objectForKey:ICDOCUMENT_PATH] lastPathComponent];
                }
                    
                // Find the appropriate drop location based on the first name of the dropped objects
                
                {
                    ICFileTree * tChildTree;
                    NSArray * tChildren;
                    NSUInteger k,tChildrenCount;
                    BOOL find=NO;
                
                    tChildren=[inItem children];
                
                    tChildrenCount=[tChildren count];
                    
                    for(k=0;k<tChildrenCount;k++)
                    {
                        tChildTree=[tChildren objectAtIndex:k];
                        
                        if ([tName compare:[FILENODE_DATA(tChildTree) fileName] options:NSCaseInsensitiveSearch]!=NSOrderedDescending)
                        {
                            [inOutlineView setDropItem:inItem dropChildIndex:k];
							
                            find=NO;
                            
                            break;
                        }
                    }
                    
                    if (find==NO)
                    {
                        [inOutlineView setDropItem:inItem dropChildIndex:k];
                    }
                }
                
                return NSDragOperationGeneric;
            }
            break;
    }
    
    return NSDragOperationNone;
}

- (BOOL)outlineView:(NSOutlineView*) inOutlineView acceptDrop:(id <NSDraggingInfo>)info item:(id)targetItem childIndex:(NSInteger)childIndex
{
    ICFileTree * tFileTree;
    NSPasteboard * tPasteBoard;
    NSMutableArray * tNewSelectionArray;
    NSUInteger k,tNewCount;
	NSMutableIndexSet * tMutableIndexSet;
	
    tPasteBoard=[info draggingPasteboard];
    
    tNewSelectionArray=[NSMutableArray array];
    
    if ([tPasteBoard availableTypeFromArray:[NSArray arrayWithObject: NSFilenamesPboardType]]!=nil)
    {
        // Drag & Drop of a real item
        
        NSArray * tArray;
        BOOL tShowCustomizationDialog=YES;
		id tObject;
        
        tArray=(NSArray *) [tPasteBoard propertyListForType:NSFilenamesPboardType];
        
		tObject=[defaults_ objectForKey: ICPREFERENCES_FILES_SHOW_CUSTOMIZATION_DIALOG];
	
		if (tObject!=nil)
		{
			tShowCustomizationDialog=[defaults_ boolForKey: ICPREFERENCES_FILES_SHOW_CUSTOMIZATION_DIALOG];
		}
		
		if (tShowCustomizationDialog==YES)
		{
			[self beginFileSheetForFiles:tArray parents:[NSArray arrayWithObject:targetItem]];
		}
		else
		{
			BOOL tDefaultKeepPermissionMode;
			NSInteger tDefaultReferenceStyle;
			
			tDefaultKeepPermissionMode=[defaults_ boolForKey:ICPREFERENCES_FILES_KEEP_PERMISSION];
			
			tDefaultReferenceStyle=[defaults_ integerForKey:ICPREFERENCES_GENERAL_DEFAULT_REFERENCE_STYLE];
			
			[self _addFiles:tArray toParents:[NSArray arrayWithObject:targetItem] keepOwnerAndGroup:tDefaultKeepPermissionMode referenceStyle:tDefaultReferenceStyle];
		}
        
        return YES;
    }
    else
    {
        if ([info draggingSource]==IBoutlineView_)
        {
            NSMutableDictionary * tDisclosedDictionary;
			NSArray * tCoverageNodesArray;
			
			// Internal Drag, we need to move
            
			tDisclosedDictionary=[document_ documentDefaultForKey:[NSString stringWithFormat:ICFILE_HIERARCHY_DISCLOSED_STATE,cachedPackageUUID_]];
			
            // Find the coverage tree
			
			tCoverageNodesArray=[TreeNode minimumNodeCoverFromNodesInArray:internalDragArray_];
			
			if (tCoverageNodesArray!=nil)
			{
				NSEnumerator * tEnumerator;
				
				tEnumerator=[tCoverageNodesArray objectEnumerator];
				
				if (tEnumerator!=nil)
				{
					NSNumber * tNumber;
					
					while (tFileTree=[tEnumerator nextObject])
					{
						NSMutableArray * tChildFoldersArray;
		
						// Removed item and descendants from disclosed state dictionary (because the key will change)
						
						tChildFoldersArray=[tFileTree descendantFoldersIncludingSelf:YES];
						
						if (tChildFoldersArray!=nil && tDisclosedDictionary!=nil)
						{
							NSUInteger i,tCount;
							
							tCount=[tChildFoldersArray count];
							
							for(i=tCount;i>0;i--)
							{
								ICFileTree * tGroupFileTree;
								NSString * tPath;
								
								tGroupFileTree=[tChildFoldersArray objectAtIndex:(i-1)];
								
								tPath=[tGroupFileTree filePath];
								
								if (tPath!=nil)
								{
									BOOL tRemoved=NO;
									
									tNumber=[tDisclosedDictionary objectForKey:tPath];
									
									if (tNumber==nil)
									{
										tRemoved=YES;
									}
									else
									{
										if ([tNumber boolValue]==NO)
										{
											tRemoved=YES;
										}
										
										[tDisclosedDictionary removeObjectForKey:tPath];
									}
									
									if (tRemoved==YES)
									{
										[tChildFoldersArray removeObject:tGroupFileTree];
									}
								}
							}
						}
						
						[tFileTree retain];
						
						[tFileTree removeFromParent];
						
						[targetItem insertSortedChild:tFileTree];
						
						[tFileTree release];
						
						[tNewSelectionArray addObject:tFileTree];
						
						// Put back the item and descendant folders disclosed state
			
						if (tChildFoldersArray!=nil && tDisclosedDictionary!=nil)
						{
							NSEnumerator * tChildEnumerator;
							
							tChildEnumerator=[tChildFoldersArray objectEnumerator];
							
							if (tChildEnumerator!=nil)
							{
								ICFileTree * tGroupFileTree;
								
								while (tGroupFileTree=[tChildEnumerator nextObject])
								{
									NSString * tPath;
									
									tPath=[tGroupFileTree filePath];
								
									if (tPath!=nil)
									{
										[tDisclosedDictionary setObject:[NSNumber numberWithBool:YES] forKey:tPath];
									}
								}
							}
						}

					}
				}
			}
        }
        else
        {
            // External Drag, we copy
            
            NSArray * tExternalArray;
			NSUInteger i,tCount;
			NSNumber * tPathStyleNumber;
			
			tPathStyleNumber=[NSNumber numberWithLong:[defaults_ integerForKey:ICPREFERENCES_GENERAL_DEFAULT_REFERENCE_STYLE]];
			
            tExternalArray=[tPasteBoard propertyListForType:ICFileExternalPboardType];
            
            tCount=[tExternalArray count];
            
            for(i=0;i<tCount;i++)
            {
                NSMutableDictionary * tDictionary;
                
                tDictionary=[tExternalArray objectAtIndex:i];
                
                [tDictionary setObject:tPathStyleNumber forKey:ICDOCUMENT_PATH_TYPE];
				
				tFileTree=[ICFileTree fileNodeWithDictionary:tDictionary projectPath:[document_ folder] referenceFolderPath:[document_ referenceFolderPath]];
                
                [targetItem insertSortedChild:tFileTree];
                
                [tNewSelectionArray addObject:tFileTree];
            }
        }
    }
    
    [IBoutlineView_ deselectAll:nil];
    
    [IBoutlineView_ reloadData];
    
    tMutableIndexSet=[NSMutableIndexSet indexSet];
	
	if (tMutableIndexSet!=nil)
	{
		tNewCount=[tNewSelectionArray count];
		
		for(k=0;k<tNewCount;k++)
		{
			[tMutableIndexSet addIndex:[IBoutlineView_ rowForItem:[tNewSelectionArray objectAtIndex:k]]];
		}
		
		[IBoutlineView_ selectRowIndexes:tMutableIndexSet byExtendingSelection:NO];
	}
    
    hierarchyChanged_=YES;
	
	[self updateFiles:IBoutlineView_];
    
    return YES;
}

#pragma mark -

- (NSDragOperation) validateFilesDrop:(id <NSDraggingInfo>)info
{
	NSPasteboard * tPasteBoard;
		
	tPasteBoard=[info draggingPasteboard];

	if ([tPasteBoard availableTypeFromArray:[NSArray arrayWithObject:NSFilenamesPboardType]]!=nil)
	{
		NSArray * tArray;
		
		tArray=(NSArray *) [tPasteBoard propertyListForType:NSFilenamesPboardType];
					
		if (tArray!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[tArray objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSString * tPath;
				
				while (tPath=[tEnumerator nextObject])
				{
					NSArray * tPossibleLocations;
					ICFileTree * tTree;
					NSInteger i,tCount;
					
					tPossibleLocations=[ICSmartLocationDetector guessPotentialInstallationLocationsForFileAtPath:tPath];
					
					if (tPossibleLocations==nil)
					{
						return NSDragOperationNone;
					}
					
					tCount=[tPossibleLocations count];
					
					for(i=0;i<tCount;i++)
					{
						NSString * tPossibleParentFolder;
						
						tPossibleParentFolder=[tPossibleLocations objectAtIndex:i];
						
						tTree=[((ICFileTree *) [fileTree_ firstChild]) fileTreeAtPath:tPossibleParentFolder];
						
						if (tTree==nil)
						{
							if ([ICSmartLocationDetector canPathBeCreated:tPossibleParentFolder]==YES)
							{
								break;
							}
						}
						else
						{
							if ([tTree containsTreeWithName:[tPath lastPathComponent]]==YES)
							{
								return NSDragOperationNone;
							}
							
							break;
						}
					}
					
					if (i==tCount)
					{
						return NSDragOperationNone;
					}
				}
			}
		
			return NSDragOperationLink;
		}
	}
	
	return NSDragOperationNone;
}

- (BOOL) acceptFilesDrop:(id <NSDraggingInfo>)info
{
	NSPasteboard * tPasteBoard;
		
	tPasteBoard=[info draggingPasteboard];

	if ([tPasteBoard availableTypeFromArray:[NSArray arrayWithObject:NSFilenamesPboardType]]!=nil)
	{
		NSMutableArray * tMutableArray;
		
		tMutableArray=(NSMutableArray *) [[tPasteBoard propertyListForType:NSFilenamesPboardType] mutableCopy];
					
		if (tMutableArray!=nil)
		{
			NSInteger j,tFilesCount;
			BOOL tShowCustomizationDialog=YES;
			id tObject;
			NSMutableArray * tParentsArray;
				
			tParentsArray=[NSMutableArray array];
				
			tObject=[defaults_ objectForKey: ICPREFERENCES_FILES_SHOW_CUSTOMIZATION_DIALOG];
					
			if (tObject!=nil)
			{
				tShowCustomizationDialog=[defaults_ boolForKey: ICPREFERENCES_FILES_SHOW_CUSTOMIZATION_DIALOG];
			}
			
			tFilesCount=[tMutableArray count];
			
			for(j=tFilesCount;j>0;j--)
			{
				NSString * tPath;
				NSArray * tPossibleLocations;
				ICFileTree * tTree;
				NSInteger i,tCount;
					
				tPath=[tMutableArray objectAtIndex:j-1];
				
				tPossibleLocations=[ICSmartLocationDetector guessPotentialInstallationLocationsForFileAtPath:tPath];
					
				tCount=[tPossibleLocations count];
				
				for(i=0;i<tCount;i++)
				{
					NSString * tPossibleParentFolder;
					
					tPossibleParentFolder=[tPossibleLocations objectAtIndex:i];
					
					tTree=[((ICFileTree *) [fileTree_ firstChild]) fileTreeAtPath:tPossibleParentFolder];
					
					if (tTree==nil)
					{
						if ([ICSmartLocationDetector canPathBeCreated:tPossibleParentFolder]==YES)
						{
							// Create Path
							
							tTree=[(ICFileTree *) [fileTree_ firstChild] createMissingBranchesForPath:tPossibleParentFolder projectPath:[document_ folder] referenceFolderPath:[document_ referenceFolderPath]];
							
							if (tTree!=nil)
							{
								[tParentsArray insertObject:tTree atIndex:0];
							
								break;
							}
						}
					}
					else
					{
						if ([tTree containsTreeWithName:[tPath lastPathComponent]]==NO)
						{
							[tParentsArray insertObject:tTree atIndex:0];
						}
						else
						{
							[tMutableArray removeObjectAtIndex:j-1];
						}
						
						break;
					}
				}
			}
				
			if (tShowCustomizationDialog==YES)
			{
				[self beginFileSheetForFiles:tMutableArray parents:tParentsArray];
			}
			else
			{
				[self _addFiles:tMutableArray 
					  toParents:tParentsArray
			  keepOwnerAndGroup:[defaults_ boolForKey:ICPREFERENCES_FILES_KEEP_PERMISSION] 
				 referenceStyle:[defaults_ integerForKey:ICPREFERENCES_GENERAL_DEFAULT_REFERENCE_STYLE]];
			}
			
			[tMutableArray autorelease];
			
			return YES;
		}
	}
	
	return NO;
}


#pragma mark -


- (BOOL) validateMenuItem:(NSMenuItem *) inMenuItem
{    
	SEL tAction;
	
	tAction=[inMenuItem action];
	
	if (tAction==@selector(setAsDefaultLocation:))
	{
		NSUInteger tSelectedCount;
		
		tSelectedCount=[IBoutlineView_ numberOfSelectedRows];
		 
		if (tSelectedCount==1)
		{
			NSInteger tSelectedRow;
        
			tSelectedRow=[IBoutlineView_ selectedRow];
		
			if (tSelectedRow>=0)
			{
				int tType;
				ICFileTree * tFileTree;
				
				tFileTree=[IBoutlineView_ itemAtRow:tSelectedRow];
				
				tType=[FILENODE_DATA(tFileTree) type];
				
				if (tAction==@selector(setAsDefaultLocation:) &&
					(tType==ICDOCUMENT_FILE_BASE_NODE_INVISIBLE || tType==ICDOCUMENT_FILE_BASE_NODE || tType==ICDOCUMENT_FILE_NEWFOLDER_NODE))
				{
					if (tFileTree!=defaultLocation_)
					{
						return YES;
					}
				}
			}
		}
		
		return NO;
	}
	
	return [super validateMenuItem:inMenuItem];
}

- (IBAction) addFiles:(id)sender
{
    NSOpenPanel * tOpenPanel;
    NSInteger tSelectedRow;
	
    tSelectedRow=[IBoutlineView_ selectedRow];
    
    if (tSelectedRow==-1)
	{
		parentFileTree_=(ICFileTree *) [fileTree_ firstChild];
	}
	else
	{
		parentFileTree_=[IBoutlineView_ itemAtRow:tSelectedRow];
	
		if ([FILENODE_DATA(parentFileTree_) type]==ICDOCUMENT_FILE_REALFILE_NODE)
		{
			parentFileTree_=(ICFileTree *) [parentFileTree_ nodeParent];
		}
    }
	
	[parentFileTree_ retain];
	
    tOpenPanel=[NSOpenPanel openPanel];
    
	if ([[NSUserDefaults standardUserDefaults] boolForKey:ICPREFERENCES_FILES_SHOW_ALL_FILES]==YES)
	{
		[[tOpenPanel _navView] setShowsHiddenFiles:YES];	// @"AppleShowAllFiles" = YES
	}
	
    [tOpenPanel setResolvesAliases:NO];
    [tOpenPanel setCanChooseFiles:YES];
    [tOpenPanel setCanChooseDirectories:YES];
    [tOpenPanel setAllowsMultipleSelection:YES];
    [tOpenPanel setTreatsFilePackagesAsDirectories:YES];
    [tOpenPanel setDelegate:self];
    
    [tOpenPanel setPrompt:NSLocalizedString(@"Add...",@"No comment")];
    
    [tOpenPanel beginSheetForDirectory:nil
                                  file:nil
                                 types:nil
                        modalForWindow:[document_ windowForSheet]
                         modalDelegate:self
                        didEndSelector:@selector(addFilesPanelDidEnd:returnCode:contextInfo:)
                           contextInfo:NULL];
}

- (void) addFilesPanelDidEnd:(NSOpenPanel *)sheet returnCode:(NSInteger)returnCode contextInfo:(void  *)contextInfo
{
	if (returnCode==NSOKButton)
    {
        NSArray * tFileNames;
		
		tFileNames=[sheet filenames];
		
		if (tFileNames!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[tFileNames objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSString * tPath;
				BOOL tShowCustomizationDialog=YES;
				id tObject;
				
				// A AMELIORER
				
				while (tPath=[tEnumerator nextObject])
				{
					if ([parentFileTree_ containsTreeWithName:[tPath lastPathComponent]]==YES)
					{
						NSBeep();
						
						return;
					}
				}
				
				tObject=[defaults_ objectForKey: ICPREFERENCES_FILES_SHOW_CUSTOMIZATION_DIALOG];
	
				if (tObject!=nil)
				{
					tShowCustomizationDialog=[defaults_ boolForKey: ICPREFERENCES_FILES_SHOW_CUSTOMIZATION_DIALOG];
				}
				
				if (tShowCustomizationDialog==YES)
				{
					 NSDictionary * nContextualInformation;
					 
					 nContextualInformation=[[NSDictionary alloc] initWithObjectsAndKeys:tFileNames,IC_ADD_FILES,
																						 [NSArray arrayWithObject:parentFileTree_],IC_ADD_PARENTS,
																						 nil];
					 
					 
					
					[self performSelector:@selector(beginFileSheet:) withObject:nContextualInformation afterDelay:0.f];
				}
				else
				{
					[self _addFiles:tFileNames
						  toParents:[NSArray arrayWithObject:parentFileTree_] 
				  keepOwnerAndGroup:[defaults_ boolForKey:ICPREFERENCES_FILES_KEEP_PERMISSION] 
				     referenceStyle:[defaults_ integerForKey:ICPREFERENCES_GENERAL_DEFAULT_REFERENCE_STYLE]];
				}
			}
		}
    }
	
	[parentFileTree_ release];
	
	parentFileTree_=nil;
}

- (void) beginFileSheetForFiles:(NSArray *) inFiles parents:(id) inParents
{
    NSInteger tDefaultReferenceStyle;
    BOOL tDefaultKeepPermissionMode;
	NSDictionary * nContextualInformation;
	
	tDefaultKeepPermissionMode=[defaults_ boolForKey:ICPREFERENCES_FILES_KEEP_PERMISSION];
	
    tDefaultReferenceStyle=[defaults_ integerForKey:ICPREFERENCES_GENERAL_DEFAULT_REFERENCE_STYLE];
    
    nContextualInformation=[[NSDictionary alloc] initWithObjectsAndKeys:inFiles,IC_ADD_FILES,
																		inParents,IC_ADD_PARENTS,
																		nil];
    
    // Prepare the sheet
    
    [IBownerAndGroup_ setState:(tDefaultKeepPermissionMode==YES) ? NSOnState : NSOffState];
    
    [IBreferenceStyle_ selectItemAtIndex:[IBreferenceStyle_ indexOfItemWithTag:tDefaultReferenceStyle]];
    
    [NSApp beginSheet:IBreferenceWindow_
       modalForWindow:[document_ windowForSheet]
        modalDelegate:self
       didEndSelector:@selector(fileSheetDidEnd:returnCode:contextInfo:)
          contextInfo:nContextualInformation];
}

- (void) fileSheetDidEnd:(NSWindow *)sheet returnCode:(NSInteger)returnCode contextInfo:(void *)contextInfo
{
    NSDictionary * nContextualInformation;
    
    nContextualInformation=(NSDictionary *) contextInfo;
    
    if (returnCode==NSOKButton)
    {
		[self _addFiles:[nContextualInformation objectForKey:IC_ADD_FILES] 
			  toParents:[nContextualInformation objectForKey:IC_ADD_PARENTS] 
	  keepOwnerAndGroup:([IBownerAndGroup_ state]==NSOnState) 
	     referenceStyle:[[IBreferenceStyle_ selectedItem] tag]];
	}
    
    [nContextualInformation release];

    [sheet orderOut:self];
}

- (IBAction) switchSplitForks:(id) sender
{
	BOOL tSplitForks;
	
	tSplitForks=([sender state]==NSOnState);
	
	[dictionary_ setObject:[NSNumber numberWithBool:tSplitForks] forKey:ICDOCUMENT_PACKAGE_FILES_SPLIT_FORKS];
	
	// Notify Document Change
			
	[self noteDocumentChange];
}

- (IBAction) setAsDefaultLocation:(id)sender
{
    NSInteger tRow;
    
    tRow=[IBoutlineView_ selectedRow];

    if (defaultLocation_!=[IBoutlineView_ itemAtRow:tRow])
    {
        defaultLocation_=[IBoutlineView_ itemAtRow:tRow];
    
        [IBoutlineView_ reloadData];
    
    	[IBdefaultLocationPath_ setStringValue:[defaultLocation_ filePath]];
    
        [self updateFiles:IBsetButton_];
		
		[IBsetButton_ setEnabled:NO];
    }
}

#pragma mark -

+ (NSUInteger) _maxDepthOfArray:(NSArray *) inArray
{
	NSUInteger tMaxDepth;
	
	tMaxDepth=0;
	
	if (inArray!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[inArray objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tDictionary;
			
			while (tDictionary=[tEnumerator nextObject])
			{
				NSUInteger tDepth;
				NSArray * tChildren;
				
				tDepth=1;
				
				tChildren=[tDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_CHILDREN];
				
				tDepth+=[ICPackageFilesController _maxDepthOfArray:tChildren];
				
				if (tDepth>tMaxDepth)
				{
					tMaxDepth=tDepth;
				}
			}
		}
	}
	
	return tMaxDepth;
}

- (BOOL) removeInvisibleBaseFoldersFromArray:(NSArray *) inArray depth:(int) inDepth
{
	BOOL tRemoved=NO;
	
	if (inArray!=nil)
	{
		NSUInteger i,tCount;
		NSMutableDictionary * tDisclosedDictionary;
		
		tDisclosedDictionary=[document_ documentDefaultForKey:[NSString stringWithFormat:ICFILE_HIERARCHY_DISCLOSED_STATE,cachedPackageUUID_]];
			
		tCount=[inArray count];
		
		for(i=tCount;i>0;i--)
		{
			ICFileTree * tFileTree;
			
			tFileTree=[inArray objectAtIndex:i-1];
			
			if ([FILENODE_DATA(tFileTree) type]==ICDOCUMENT_FILE_BASE_NODE_INVISIBLE)
			{
				if ([tFileTree hasRealChildren]==YES)
				{
					if (inDepth>1)
					{
						if ([tFileTree numberOfChildren]>1)
						{
							if ([self removeInvisibleBaseFoldersFromArray:[tFileTree children] depth:inDepth-1]==YES)
							{
								tRemoved=YES;
							}
						}
					}
				}
				else
				{
					NSMutableArray * tChildFoldersArray;
							
					// Removed item and descendants from disclosed state dictionary (because the key will change)
		
					tChildFoldersArray=[tFileTree descendantFoldersIncludingSelf:YES];
					
					if (tChildFoldersArray!=nil && tDisclosedDictionary!=nil)
					{
						NSUInteger j,tChildrenCount;
						
						tChildrenCount=[tChildFoldersArray count];
						
						for(j=tChildrenCount;j>0;j--)
						{
							ICFileTree * tGroupFileTree;
							NSString * tPath;
							
							tGroupFileTree=[tChildFoldersArray objectAtIndex:(j-1)];
							
							tPath=[tGroupFileTree filePath];
							
							if (tPath!=nil)
							{
								[tDisclosedDictionary removeObjectForKey:tPath];
							}
						}
					}
					
					// Remove the node
					
					[tFileTree removeFromParent];

					tRemoved=YES;
				}
			}
		}
	}
	
	return tRemoved;
}

- (BOOL) addInvisibleBaseFoldersFromArray:(NSArray *) inArray toFileTree:(ICFileTree *) inFileTree
{
	BOOL tAdded=NO;
	
	if (inArray!=nil && [inArray count]>0 && inFileTree!=nil)
	{
		NSArray * tChildren;
		
		tChildren=[inFileTree children];
		
		if (tChildren!=nil)
		{
			NSEnumerator * tEnumerator;
			NSUInteger tCount;
			
			tCount=[tChildren count];
			
			tEnumerator=[inArray objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSDictionary * tDictionary;
				
				while (tDictionary=[tEnumerator nextObject])
				{
					NSString * tName;
					
					tName=[tDictionary objectForKey:ICDOCUMENT_PATH];
					
					if (tName!=nil)
					{
						NSUInteger i;

						for(i=0;i<tCount;i++)
						{
							ICFileTree * tFileTree;
							NSString * tFileName;
							
							tFileTree=(ICFileTree *) [tChildren objectAtIndex:i];
							
							tFileName=[FILENODE_DATA(tFileTree) fileName];
							
							if (tFileName!=nil)
							{
								NSComparisonResult tResult;
								
								tResult=[tName caseInsensitiveCompare:tFileName];
								
								if (tResult==NSOrderedSame)
								{
									if ([FILENODE_DATA(tFileTree) type]==ICDOCUMENT_FILE_BASE_NODE_INVISIBLE)
									{
										if ([self addInvisibleBaseFoldersFromArray:[tDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_CHILDREN] toFileTree:tFileTree]==YES)
										{
											tAdded=YES;
										}
									}
									
									break;
								}
							}
						}
						
						if (i==tCount)
						{
							ICFileTree * tFileTree;
					
							tAdded=YES;
							
							// Add the sub-file hierarchy
							
							tFileTree=[ICFileTree fileNodeWithDictionary:tDictionary projectPath:[document_ folder] referenceFolderPath:[document_ referenceFolderPath]];

							[inFileTree insertSortedChildReverse:tFileTree];
						}
					}
				}
			}
		}
	}
	
	return tAdded;
}

- (void) switchShowInvisibleFolders
{
	NSNumber * tNumber;
	static NSArray * sInvisibleFoldersArray=nil;
	static NSUInteger sMaxDepth=0;
	
	tNumber=[dictionary_ objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY_SHOW_INVISIBLE];
	
	if (sInvisibleFoldersArray==nil)
	{
		NSString * tPath;
		
		tPath=[[NSBundle mainBundle] pathForResource:@"InvisibleHierarchy" ofType:@"plist"];
		
		if (tPath!=nil)
		{
			sInvisibleFoldersArray=[[NSArray alloc] initWithContentsOfFile:tPath];
			
			if (sInvisibleFoldersArray==nil)
			{
				// A COMPLETER
			}
			else
			{
				sMaxDepth=[ICPackageFilesController _maxDepthOfArray:sInvisibleFoldersArray];
			}
		}
		else
		{
			// A COMPLETER
		}
	}
	
	if (sInvisibleFoldersArray!=nil)
	{
		BOOL tHierarchyChanged=NO;
		NSIndexSet * tIndexSet;
		NSMutableArray * tSelectedNodesMutableArray=nil;
		
		tIndexSet=[IBoutlineView_ selectedRowIndexes];
		
		if (tIndexSet!=nil)
		{
			tSelectedNodesMutableArray=[NSMutableArray array];
			
			if (tSelectedNodesMutableArray!=nil)
			{
				NSUInteger tIndex;
			
				tIndex=[tIndexSet lastIndex];
			
				while (tIndex!=NSNotFound)
				{
					ICFileTree * tFileTree;
					
					tFileTree=[IBoutlineView_ itemAtRow:tIndex];
					
					if (tFileTree!=nil)
					{
						[tSelectedNodesMutableArray addObject:tFileTree];
					}
					
					tIndex=[tIndexSet indexLessThanIndex:tIndex];
				}
			}
		}
		
		if (tNumber==nil || [tNumber boolValue]==NO)
		{
			// Show missing Invisible folders
			
			tHierarchyChanged=[self addInvisibleBaseFoldersFromArray:sInvisibleFoldersArray toFileTree:(ICFileTree *) [fileTree_ firstChild]];
			
			[dictionary_ setObject:[NSNumber numberWithBool:YES] forKey:ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY_SHOW_INVISIBLE];
		}
		else
		{
			// Hide Invisible folders with no children
			
			tHierarchyChanged=[self removeInvisibleBaseFoldersFromArray:[[fileTree_ firstChild] children] depth:sMaxDepth];
			
			[dictionary_ setObject:[NSNumber numberWithBool:NO] forKey:ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY_SHOW_INVISIBLE];
		}
		
		if (tHierarchyChanged==YES)
		{
			[IBoutlineView_ deselectAll:nil];
			
			[IBoutlineView_ reloadData];
			
			// Restore Selection
			
			if (tSelectedNodesMutableArray!=nil && [tSelectedNodesMutableArray count]>0)
			{
				NSEnumerator * tEnumerator;
				
				tEnumerator=[tSelectedNodesMutableArray objectEnumerator];
				
				if (tEnumerator!=nil)
				{
					ICFileTree * tFileTree;
					NSMutableIndexSet * tMutableIndexSet;
					
					tMutableIndexSet=[NSMutableIndexSet indexSet];
					
					while (tFileTree=[tEnumerator nextObject])
					{
						NSInteger tRow;
						
						tRow=[IBoutlineView_ rowForItem:tFileTree];
						
						if (tRow!=-1)
						{
							[tMutableIndexSet addIndex:tRow];
						}
					}
					
					[IBoutlineView_ selectRowIndexes:tMutableIndexSet byExtendingSelection:NO];
				}
			}
			
			hierarchyChanged_=YES;
	
			[self updateFiles:IBoutlineView_];
		}
		
		// Notify Document Change
				
		[self noteDocumentChange];
	}
}

#pragma mark -

- (void) restoreDisclosedStates
{
	NSDictionary * tDictionary;
	
	tDictionary=[document_ documentDefaultForKey:[NSString stringWithFormat:[self disclosedStateFormatKey],cachedPackageUUID_]];
			
	if (tDictionary!=nil)
	{
		NSMutableDictionary * tDictionaryCopy;
		
		tDictionaryCopy=[tDictionary mutableCopy];
		
		if (tDictionaryCopy!=nil)
		{
			[self expandItem:((ICFileTree *) [fileTree_ firstChild]) ifKeyOfDictionary:tDictionaryCopy];
		
			[tDictionaryCopy release];
		}
	}
	else
	{
		if ([self defaultRestoreForItem:fileTree_]==NO)
		{
			ICFileTree * tFileTree;
					
			// Expand '/'
			
			[IBoutlineView_ expandItem:[IBoutlineView_ itemAtRow:0]];
		
			// Expand the /Applications folder
			
			tFileTree=[((ICFileTree *) [fileTree_ firstChild]) fileTreeAtPath:@"/Applications"];
			
			if (tFileTree!=nil)
			{
				[IBoutlineView_ expandItem:tFileTree];
			}
			
			// Expand the /Library folder
			
			tFileTree=[((ICFileTree *) [fileTree_ firstChild]) fileTreeAtPath:@"/Library"];
			
			if (tFileTree!=nil)
			{
				[IBoutlineView_ expandItem:tFileTree];
			}
		}
	}
}

#pragma mark -

- (void) advancedModeDidChange:(NSNotification *) inNotification
{
	BOOL tState;
	
	tState=[defaults_ boolForKey:ICPREFERENCES_ADVANCED_ADVANCED_MODE_STATE];
	
	[self setHiddenAdvancedOptions:(tState==NO)];
}

- (void) outlineViewSelectionDidChange:(NSNotification *) inNotification
{
	NSInteger tSelectedCount;
	NSRect tBounds;
	NSView * tView;
	
	tSelectedCount=[IBoutlineView_ numberOfSelectedRows];
	
	if (tSelectedCount==0)
	{
		if (emptySelectionInspectorController_==nil)
		{
			emptySelectionInspectorController_=[ICFileHierarchyInspectorEmptySelectionController newController];
		}
		
		if (currentInspectorController_!=emptySelectionInspectorController_ || currentInspectorController_==nil)
		{
			tView=[currentInspectorController_ view];
		
			[tView removeFromSuperview];
		
			currentInspectorController_=emptySelectionInspectorController_;
			
			tView=[currentInspectorController_ view];
			
			tBounds=[IBinspectorView_ bounds];
			
			[tView setFrame:tBounds];
			
			[IBinspectorView_ addSubview:tView];
		}
		
		[IBsetButton_ setEnabled:NO];
		
		[IBremoveButton_ setEnabled:NO];
	}
	else
	{
		NSIndexSet * tIndexSet;
		
		[IBremoveButton_ setEnabled:YES];
		
		if (selectionInspectorController_==nil)
		{
			selectionInspectorController_=[ICPackageFilesInspectorSelectionController newController];
		}
		
		if (currentInspectorController_!=selectionInspectorController_ || currentInspectorController_==nil)
		{
			tView=[currentInspectorController_ view];
		
			[tView removeFromSuperview];
		
			currentInspectorController_=selectionInspectorController_;
			
			tView=[currentInspectorController_ view];
			
			tBounds=[IBinspectorView_ bounds];
			
			[tView setFrame:tBounds];
			
			[IBinspectorView_ addSubview:tView];
		}
		
		tIndexSet=[IBoutlineView_ selectedRowIndexes];
		
		if (tIndexSet!=nil)
		{
			NSMutableArray * tSelectedItems;
			
			tSelectedItems=[NSMutableArray array];
			
			if (tSelectedItems!=nil)
			{
				NSUInteger tIndex;
				BOOL tDisableRemoveButton=NO;
				
				tIndex=[tIndexSet firstIndex];
				
				while (tIndex!=NSNotFound)
				{
					id tItem;
					
					tItem=[IBoutlineView_ itemAtRow:tIndex];
					
					if (tItem!=nil)
					{
						if (tDisableRemoveButton==NO)
						{
							int tType;
							
							tType=[FILENODE_DATA(tItem) type];
							
							if (tType==ICDOCUMENT_FILE_BASE_NODE_INVISIBLE ||
								tType==ICDOCUMENT_FILE_BASE_NODE)
							{
								tDisableRemoveButton=YES;
							}
						}
						
						[tSelectedItems addObject:tItem];
						
						tIndex=[tIndexSet indexGreaterThanIndex:tIndex];
					}
				}
				
				if (tDisableRemoveButton==YES)
				{
					[IBremoveButton_ setEnabled:NO];
				}
				
				[IBsetButton_ setEnabled:NO];
				
				if (tSelectedCount==1)
				{
					int tType;
					ICFileTree * tFileTree;
					
					tFileTree=[tSelectedItems lastObject];
					
					tType=[FILENODE_DATA(tFileTree) type];
		
					if (tType==ICDOCUMENT_FILE_BASE_NODE_INVISIBLE || tType==ICDOCUMENT_FILE_BASE_NODE || tType==ICDOCUMENT_FILE_NEWFOLDER_NODE)
					{
						if (tFileTree!=defaultLocation_)
						{
							[IBsetButton_ setEnabled:YES];
						}
					}
					
					[IBaddButton_ setEnabled:YES];
				}
				else
				{
					[IBaddButton_ setEnabled:NO];
				}
				
				[selectionInspectorController_ initializeWithArray:tSelectedItems forDocument:document_];
			}
		}
	}
}

@end
