/*
 Copyright (c) 2008-2010, Stephane Sudre
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 
 - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 - Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "ICQuickBuildWindow.h"

#define STATUS_FRAME_PER_SEC	3.0f

#import "NSBezierPath+RoundedRect.h"

@implementation ICQuickBuildStatusView

- (id) initWithFrame:(NSRect) inFrame
{
	self=[super initWithFrame:inFrame];
	
	if (self!=nil)
	{
		status_=IC_QUICKBUILD_STATUS_BUILDING;
		
		timer_=[NSTimer scheduledTimerWithTimeInterval:1.0f/STATUS_FRAME_PER_SEC
												target:self
											  selector:@selector(updateAnimation:)
											  userInfo:nil
											   repeats:YES];
	
		//[[NSRunLoop currentRunLoop] addTimer:timer_ forMode:NSEventTrackingRunLoopMode];
	}
	
	return self;
}

- (void) dealloc
{
	if (image_!=nil)
	{
		[image_ release];
	}
	
	if (timer_!=nil)
	{
		[timer_ invalidate];
	}

	[super dealloc];
}

#pragma mark -

- (BOOL) isOpaque
{
	return NO;
}

+ (NSImage *) sharedSuccessImage
{
	static NSImage * sSuccessImage=nil;
	
	if (sSuccessImage==nil)
	{
		NSString * tPath;
		
		tPath=[[NSBundle mainBundle] pathForResource:@"QuickBuildSuccess" ofType:@"tif"];
		
		if (tPath!=nil)
		{
			sSuccessImage=[[NSImage alloc] initWithContentsOfFile:tPath];
		}
	}
	
	return sSuccessImage;
}

+ (NSImage *) sharedFailureImage
{
	static NSImage * sFailureImage=nil;
	
	if (sFailureImage==nil)
	{
		NSString * tPath;
		
		tPath=[[NSBundle mainBundle] pathForResource:@"QuickBuildFailure" ofType:@"tif"];
		
		if (tPath!=nil)
		{
			sFailureImage=[[NSImage alloc] initWithContentsOfFile:tPath];
		}
	}
	
	return sFailureImage;
}

- (void) setStatus:(int) inStatus
{
	if (inStatus!=status_)
	{
		[image_ release];
		
		image_=nil;
		
		if (status_==IC_QUICKBUILD_STATUS_BUILDING)
		{
			[timer_ invalidate];
		
			timer_=nil;
		}
		
		status_=inStatus;
		
		switch (status_)
		{
			case IC_QUICKBUILD_STATUS_BUILDING:
		
				timer_=[NSTimer scheduledTimerWithTimeInterval:1.0f/STATUS_FRAME_PER_SEC
														 target:self
													   selector:@selector(updateAnimation:)
														userInfo:nil
														 repeats:YES];
				
				//[[NSRunLoop currentRunLoop] addTimer:timer_ forMode:NSEventTrackingRunLoopMode];
				
				break;
				
			case IC_QUICKBUILD_STATUS_SUCCESS:
			
				image_=[[ICQuickBuildStatusView sharedSuccessImage] retain];
				
				break;
				
			case IC_QUICKBUILD_STATUS_FAILURE:
			
				image_=[[ICQuickBuildStatusView sharedFailureImage] retain];
				
				break;
		}
		
		[self setNeedsDisplay:YES];
	}
}

#pragma mark -

- (void) updateAnimation:(NSTimer *) inAnimation
{
	angle_-=14.0f;
	
	if (angle_<-360.0f)
	{
		angle_+=360.0f;
	}
	
	[self setNeedsDisplay:YES];
}

#define DEG_TO_RAD (2.0f*3.141592654/360.0f)

#define TEETH_COUNT		8

#define TEETH_ARC_LENGTH	35.0f

#define CICLE_RADIUS_PERC	(28.0f/100.0f)

#define TEETH_HEIGHT_PERC	(25.0f/100.0f)

- (void) drawRect:(NSRect) inRect
{
	NSRect tBounds;
	NSSize tImageSize;
	
	
	tBounds=[self bounds];
	
	switch(status_)
	{
		case IC_QUICKBUILD_STATUS_BUILDING:
		{
			NSBezierPath * tBezierPath;
			NSUInteger i;
			CGFloat tHoleLength;
			CGFloat tAngle;
			NSPoint tPoint;
			CGFloat tRadius;
			NSPoint tCenter;
			
			tCenter.x=NSMidX(tBounds);
			tCenter.y=NSMidY(tBounds);
			
			
			tRadius=NSWidth(tBounds)*0.5f;
			
			tHoleLength=(360.0f/TEETH_COUNT)-TEETH_ARC_LENGTH;
			
			tAngle=angle_;
			
			tPoint=NSMakePoint(tCenter.x+(1.0f-TEETH_HEIGHT_PERC)*tRadius*cosf(tAngle*DEG_TO_RAD),tCenter.y+(1.0f-TEETH_HEIGHT_PERC)*tRadius*sinf(tAngle*DEG_TO_RAD));
			
			tBezierPath=[NSBezierPath bezierPath];
			
			[tBezierPath moveToPoint:tPoint];
			
			for(i=0;i<TEETH_COUNT;i++)
			{
				tPoint=NSMakePoint(tCenter.x+tRadius*cosf((tAngle+5.0f)*DEG_TO_RAD),tCenter.y+tRadius*sinf((tAngle+5.0f)*DEG_TO_RAD));
				
				[tBezierPath lineToPoint:tPoint];
				
				[tBezierPath appendBezierPathWithArcWithCenter:tCenter
														radius:tRadius
													startAngle:(tAngle+5.0f)
													  endAngle:tAngle+TEETH_ARC_LENGTH-5.0f
													 clockwise:NO];
													 
				tAngle+=TEETH_ARC_LENGTH;
				
				tPoint=NSMakePoint(tCenter.x+(1.0f-TEETH_HEIGHT_PERC)*tRadius*cosf(tAngle*DEG_TO_RAD),tCenter.y+(1.0f-TEETH_HEIGHT_PERC)*tRadius*sinf(tAngle*DEG_TO_RAD));
				
				[tBezierPath lineToPoint:tPoint];
				
				[tBezierPath appendBezierPathWithArcWithCenter:tCenter
														radius:(1.0f-TEETH_HEIGHT_PERC)*tRadius
													startAngle:tAngle
													  endAngle:tAngle+tHoleLength
													 clockwise:NO];
			
				tAngle+=tHoleLength;
				
			}
			
			[tBezierPath closePath];
			
			[[NSColor colorWithDeviceWhite:0.15f alpha:1.0f] set];
			
			[tBezierPath fill];
			
			tBezierPath=[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(tCenter.x-CICLE_RADIUS_PERC*tRadius,tCenter.y-CICLE_RADIUS_PERC*tRadius,2*CICLE_RADIUS_PERC*tRadius,2*CICLE_RADIUS_PERC*tRadius) ];
			
			[[NSColor colorWithDeviceWhite:0.45f alpha:1.0f] set];
			
			[tBezierPath fill];
			
			// A COMPLETER
			
			break;
		}
			
		case IC_QUICKBUILD_STATUS_SUCCESS:
		case IC_QUICKBUILD_STATUS_FAILURE:
		
			if (image_!=nil)
			{
				tImageSize=[image_ size];
				
				[image_ drawInRect:tBounds fromRect:NSMakeRect(0.0f,0.0f,tImageSize.width,tImageSize.height) operation:NSCompositeSourceOver fraction:1.0f];
			}
			
			break;
	}
}

@end

@implementation ICQuickBuildTextField

- (void) drawRect:(NSRect) inRect
{
	NSRect tBounds;
	
	tBounds=[self bounds];
	
	[self setTextColor:[NSColor colorWithDeviceWhite:0.15f alpha:0.8f]];
	
	[[self cell] drawWithFrame:NSOffsetRect(tBounds,0.0f,3.0f) inView:self];
	
	[self setTextColor:[NSColor colorWithDeviceWhite:1.0f alpha:0.99f]];
	
	[[self cell] drawWithFrame:NSOffsetRect(tBounds,0.0f,2.0f) inView:self];
}

@end

@implementation ICQuickBuildMainView

- (void) drawRect:(NSRect) inRect
{
    NSBezierPath * tBezierPath;
	
	tBezierPath=[NSBezierPath bezierPathWithRoundedRect:[self bounds] radius:10.0f];
	
	if (tBezierPath!=nil)
	{
		[[NSColor colorWithDeviceWhite:0.0f alpha:0.55f] set];
	
		[tBezierPath fill];
		
		[[NSColor colorWithDeviceWhite:0.0f alpha:0.1f] set];
		
		[tBezierPath stroke];
	}
}

@end

@implementation ICQuickBuildWindow

- (id) init
{
	NSScreen * tMainScreen;
	NSRect tVisibleFrame;
	NSRect tWindowFrame;
	
	tMainScreen=[NSScreen mainScreen];
	
	tVisibleFrame=[tMainScreen visibleFrame];
	
	tWindowFrame=NSMakeRect(_CGFloatFloor(NSMidX(tVisibleFrame)-IC_QUICKBUILD_WINDOW_WIDTH*0.5f),
							_CGFloatFloor(NSMidY(tVisibleFrame)-IC_QUICKBUILD_WINDOW_WIDTH*0.5f),
							IC_QUICKBUILD_WINDOW_WIDTH,
							IC_QUICKBUILD_WINDOW_WIDTH);

	self=[super initWithContentRect:tWindowFrame styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:YES];
	
	if (self!=nil)
	{
		ICQuickBuildMainView * tContentView;
		
		[self setIgnoresMouseEvents:YES];
		
		[self setHasShadow:NO];
		
		[self setOpaque:NO];
		
		[self setBackgroundColor:[NSColor clearColor]];
		
		[self setLevel:NSScreenSaverWindowLevel];
		
		tContentView=[[ICQuickBuildMainView alloc] initWithFrame:NSMakeRect(0,0,IC_QUICKBUILD_WINDOW_WIDTH,IC_QUICKBUILD_WINDOW_WIDTH)];
		
		if (tContentView!=nil)
		{
			[self setContentView:tContentView];
			
			[tContentView release];
		}
	}
	
	return self;
}

@end
