/*
Copyright (c) 2004-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICSignatureFactory.h"

#import <Security/Security.h>

#import "ICProjectBuilder+SignatureRequest.h"

#import "ICCertificatesUtilities.h"

#import "ICDocumentFormat+Constants.h"



@implementation ICSignatureFactory

+ (BOOL) performSignatureRequestWithDictionary:(NSDictionary *) inDictionary
{
	BOOL tSuccess=NO;
	
	if (inDictionary!=nil)
	{
		NSString * tMessagePortName;
		NSDictionary * tCertificateDictionary;
		NSData * tDataToSign;
		CFMessagePortRef tRemote=NULL;
		SInt32 tErrorMessageID=IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_FAILURE;
		
		tMessagePortName=[inDictionary objectForKey:IC_BUILDER_REQUEST_MESSAGEPORT_NAME];
		
		tCertificateDictionary=[inDictionary objectForKey:IC_BUILDER_REQUEST_CERTIFICATE];
		
		tDataToSign=[inDictionary objectForKey:IC_BUILDER_REQUEST_DATA];
		
		if (tMessagePortName!=nil && tCertificateDictionary!=nil && [tCertificateDictionary count]>0 && tDataToSign!=nil)
		{
			NSString * tName;
			NSString * tKeychainPath;
			
			tName=[tCertificateDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME];
			
			tKeychainPath=[tCertificateDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH];
			
			if (tName!=nil && tKeychainPath!=nil)
			{
				SecIdentityRef tIdentityRef;
			
				tIdentityRef=[ICCertificatesUtilities identityWithName:tName atPath:tKeychainPath];
				
				if (tIdentityRef!=nil)
				{
					OSStatus tStatus;
					SecKeyRef tPrivateKeyRef;
					
					tStatus=SecIdentityCopyPrivateKey(tIdentityRef,&tPrivateKeyRef);
							
					if (tStatus==noErr)
					{
						const CSSM_ACCESS_CREDENTIALS * tCredentials;
						
						tStatus=SecKeyGetCredentials(tPrivateKeyRef,CSSM_ACL_AUTHORIZATION_SIGN,1,&tCredentials);	// kSecCredentialTypeNoUI
						
						if (tStatus==noErr)
						{
							CSSM_CSP_HANDLE tCSPHandle;
							
							tStatus=SecKeyGetCSPHandle(tPrivateKeyRef,&tCSPHandle);
							
							if (tStatus==noErr)
							{
								const CSSM_KEY * tPtr;
								
								tStatus=SecKeyGetCSSMKey(tPrivateKeyRef,&tPtr);
												
								if (tStatus==noErr)
								{
									CSSM_CC_HANDLE tContextHandle;
									CSSM_RETURN tError;
									
									tError=CSSM_CSP_CreateSignatureContext(tCSPHandle,tPtr->KeyHeader.AlgorithmId,tCredentials,tPtr,&tContextHandle);
									
									if (tError==CSSM_OK)
									{
										CSSM_DATA tData;
										CSSM_DATA tSignatureData;
										
										tData.Length=[tDataToSign length];
										tData.Data=(uint8 *) [tDataToSign bytes];
										
										tSignatureData.Length=0;
										tSignatureData.Data=NULL;
										
										tError=CSSM_SignData(tContextHandle,&tData,1,CSSM_ALGID_SHA1,&tSignatureData);
										
										if (tError==CSSM_OK)
										{
											NSData * tSignedData;
											
											tSignedData=[NSData dataWithBytes:tSignatureData.Data length:tSignatureData.Length];
											
											if (tSignedData!=nil)
											{
												tRemote=CFMessagePortCreateRemote(kCFAllocatorDefault,(CFStringRef) tMessagePortName);
												
												if (tRemote!=NULL)
												{
													SInt32 tResult;
													
													tResult=CFMessagePortSendRequest(tRemote,IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_SUCCESS,(CFDataRef) tSignedData, 20, 10, NULL, NULL);
													
													if (tResult!=kCFMessagePortSuccess)
													{
														switch(tResult)
														{
															case kCFMessagePortSendTimeout:
																
																NSLog(@"Signature Process Error: Message Port Send Time Out");
																
																break;
															
															case kCFMessagePortReceiveTimeout:
																
																NSLog(@"Signature Process Error: Message Port Receive Time Out");
																
																break;
															
															case kCFMessagePortIsInvalid:
															
																NSLog(@"Signature Process Error: Message Port Invalid Port");
															
																break;
																
															case kCFMessagePortTransportError:
																
																NSLog(@"Signature Process Error: Message Port Transport Error");
																
																break;
														}
													}
													else
													{
														tSuccess=YES;
													}
													
													
												}
											}
										}
										else
										{
											switch(tError)
											{
												case CSSMERR_CSP_OPERATION_AUTH_DENIED:
												
													tErrorMessageID=IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_AUTH_DENIED;
													
													break;
											}
										}
										
										CSSM_DeleteContext(tContextHandle);
									}
									
#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5

									/*if (tError!=CSSM_OK)
									{
										NSLog(@"Signature Process Error: %s",CSSMErrorAsString(tError));
									}*/
#endif
								}
							}
						}
						
						// Release Memory
						
						CFRelease(tPrivateKeyRef);
					}
					
					if (tStatus!=noErr)
					{
						NSString * tErrorMessage;
						
						tErrorMessage=(NSString *) SecCopyErrorMessageString(tStatus,NULL);
						
						if (tErrorMessage!=nil)
						{
							NSLog(@"Signature Process Error: %@",tErrorMessage);
						
							[tErrorMessage release];
						}
					}
				}
				else
				{
					tErrorMessageID=IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_CERTIFICATE_NOT_FOUND;
				}
			}
		}
		else
		{
			NSLog(@"[ICSignatureFactory performSignatureRequestWithDictionary: ] Missing Metadata value");
		}
		
		if (tMessagePortName!=nil && tSuccess==NO && tRemote==NULL)
		{
			tRemote=CFMessagePortCreateRemote(kCFAllocatorDefault,(CFStringRef) tMessagePortName);
											
			if (tRemote!=NULL)
			{
				SInt32 tResult;
				
				tResult=CFMessagePortSendRequest(tRemote,tErrorMessageID,NULL, 20, 10, NULL, NULL);
				
				if (tResult!=kCFMessagePortSuccess)
				{
					switch(tResult)
					{
						case kCFMessagePortSendTimeout:
							
							NSLog(@"Signature Process Error: Message Port Send Time Out");
							
							break;
						
						case kCFMessagePortReceiveTimeout:
							
							NSLog(@"Signature Process Error: Message Port Receive Time Out");
							
							break;
						
						case kCFMessagePortIsInvalid:
						
							NSLog(@"Signature Process Error: Message Port Invalid Port");
						
							break;
							
						case kCFMessagePortTransportError:
							
							NSLog(@"Signature Process Error: Message Port Transport Error");
							
							break;
					}
				}
			}
		}
		
		if (tRemote!=NULL)
		{
			// Release Memory
												
			CFRelease(tRemote);
		}
		else
		{
			NSLog(@"[ICSignatureFactory performSignatureRequestWithDictionary: ] Missing Metadata value");
		}
	}
	else
	{
		NSLog(@"[ICSignatureFactory performSignatureRequestWithDictionary: ] Missing Metadata");
	}

	return tSuccess;
}

@end