/*
Copyright (c) 2008-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICFileSystemUtilities.h"

#import "ICBuilderLogManager.h"

#import "ICBuildDistributedNotificationCenter.h"

#include <CoreServices/CoreServices.h>

#include <sys/stat.h>

static OSErr MyFSPathMakeRef(const char * inPath, FSRef * outFSRef );

@implementation ICFileSystemUtilities

+ (BOOL) setFileOwner:(uid_t) inOwner group:(gid_t) inGroup atPath:(NSString *) inPath
{
    const char * tCString;
	NSDictionary * tMetadata=nil;
	
	tCString=[inPath fileSystemRepresentation];
	
	if (tCString!=nil)
	{
		struct stat tStat;
			
		if (lstat(tCString, &tStat)==0)
		{
			if ((tStat.st_mode & S_IFMT)==S_IFLNK)
			{
				return YES;
			}
		}
    
		if (chown(tCString,inOwner,inGroup)==0)
		{
			return YES;
		}

		switch(errno)
		{
			case EPERM:
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
				break;
			
			case ENOENT:
			
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
				break;
				
			case ENAMETOOLONG:
			
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Path(%@) is too long",inPath];
				
				break;
		}
	}
    else
	{
		// Low Memory
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
	}
	
	[[ICBuildDistributedNotificationCenter defaultCenter] postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																		   forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
																		  metadata:tMetadata];
	
	return NO;
}

+ (BOOL) setFilePrivileges:(mode_t) inPrivileges atPath:(NSString *) inPath
{
    const char * tCString;
	NSDictionary * tMetadata=nil;
	
	tCString=[inPath fileSystemRepresentation];
	
	if (tCString!=nil)
	{
		struct stat tStat;
		int tError;
		
		tError=lstat(tCString, &tStat);
		
		// Symbolic Link?
		
		if (tError==0 && (tStat.st_mode & S_IFMT)==S_IFLNK)
		{
			return YES;
		}
		
		if (chmod(tCString,inPrivileges)==0)
		{
			return YES;
		}
		
		switch(errno)
		{
			case EPERM:
			
				if (tError==0 && tStat.st_flags!=0)
				{
					if (chflags(tCString,0)==0)
					{
						if (chmod(tCString,inPrivileges)==0)
						{
							if (chflags(tCString,tStat.st_flags)==0)
							{
								return YES;
							}
						}
					}
				}
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
				break;
			
			case ENOENT:
			
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
				break;
				
			case ENAMETOOLONG:
			
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Path(%@) is too long",inPath];
				
				break;
		}
		
		[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"chmod failed"];
    }
    else
	{
		// Low Memory
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
	}
	
	[[ICBuildDistributedNotificationCenter defaultCenter] postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																		   forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
																		  metadata:tMetadata];
	
	return NO;
}


+ (BOOL) copyPath:(NSString *) fromPath toPath:(NSString *) toDirectoryPath
{
	OSStatus tStatus;
	OSErr tErr;
	FSRef sourceRef, destRef;
	NSDictionary * tMetadata=nil;
	
	tErr = MyFSPathMakeRef( [fromPath UTF8String], &sourceRef );
        
	if( tErr == noErr)	// Get FSRef to destination object
	{
		// We don't have to worry about the symlink problem (2489632) here	
		// cause we would want to copy into the target of the symlink	
		// anyways.  And if its not a symlink, no problems...		
		
		tErr = FSPathMakeRef( (unsigned char *) [toDirectoryPath UTF8String], &destRef, NULL );
		
		if( tErr == noErr )					// make sure the dest is a directory
		{
			tStatus=FSCopyObjectSync(&sourceRef,&destRef,NULL,NULL,kFSFileOperationOverwrite);
	
			//tStatus=FSPathCopyObjectSync([fromPath fileSystemRepresentation],[toPath fileSystemRepresentation],NULL,NULL,kFSFileOperationOverwrite);
	
			if (tStatus==noErr)
			{
				return YES;
			}
			
			switch(tStatus)
			{
				case fnfErr:
				
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"[ICFileSystemUtilities copyPath:toPath:] File not found"];
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_COPIED],IC_BUILDER_FAILURE_REASON,
																	     [NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_SECONDARY_REASON,
																		 fromPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	     toDirectoryPath,IC_BUILDER_INFORMATION_SECONDARY_FILE_PATH,
																	     nil];
					break;
			}
		}
		else
		{
		}
	}
	else
	{
		switch(tErr)
		{
			case fnfErr:
			
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_COPIED],IC_BUILDER_FAILURE_REASON,
																	 [NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_SECONDARY_REASON,
																	 fromPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 toDirectoryPath,IC_BUILDER_INFORMATION_SECONDARY_FILE_PATH,
																	 nil];
				
				break;
		}
	}
	
	if (tMetadata==nil)
	{
		[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_COPIED],IC_BUILDER_FAILURE_REASON,
												   fromPath,IC_BUILDER_INFORMATION_FILE_PATH,
												   toDirectoryPath,IC_BUILDER_INFORMATION_SECONDARY_FILE_PATH,
												   nil];
	}
	
	[[ICBuildDistributedNotificationCenter defaultCenter] postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																			forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
																		   metadata:tMetadata];
	
	return NO;
}

@end

static OSErr MyFSPathMakeRef( const char * inPath, FSRef * outFSRef )
{
	FSRef			tmpFSRef;
	char			tmpPath[ PATH_MAX ],
					*tmpNamePtr;
	OSErr			err;
        char * tCharPtr;
					/* Get local copy of incoming path					*/
	strcpy( tmpPath, (char*)inPath );

					/* Get the name of the object from the given path	*/
					/* Find the last / and change it to a '\0' so		*/
					/* tmpPath is a path to the parent directory of the	*/
					/* object and tmpNamePtr is the name				*/
	tmpNamePtr = strrchr( tmpPath, '/' );
	if( *(tmpNamePtr + 1) == '\0' )
	{				/* in case the last character in the path is a /	*/
		*tmpNamePtr = '\0';
		tmpNamePtr = strrchr( tmpPath, '/' );
	}
	*tmpNamePtr = '\0';
	tmpNamePtr++;
	
        tCharPtr=tmpNamePtr;
        
        while (*tCharPtr!='\0')
        {
            if ((*tCharPtr)==':')
            {
                *tCharPtr='/';
            }
        
            tCharPtr++;
        }
        
					/* Get the FSRef to the parent directory			*/
	err = FSPathMakeRef( (unsigned char*)tmpPath, &tmpFSRef, NULL );
        
	if( err == noErr )
	{				/* Convert the name to a Unicode string and pass it	*/
					/* to FSMakeFSRefUnicode to actually get the FSRef	*/
					/* to the object (symlink)							*/
            UniChar			uniName[255];
            CFStringRef 	tmpStringRef = CFStringCreateWithCString( kCFAllocatorDefault, tmpNamePtr, kCFStringEncodingUTF8 );
            if( tmpStringRef != NULL )
            {
                    err = ( CFStringGetCString( tmpStringRef, (char*)uniName, 255 * sizeof( UniChar ), kCFStringEncodingUnicode ) )				?
                                            FSMakeFSRefUnicode( &tmpFSRef, CFStringGetLength( tmpStringRef ), uniName, kTextEncodingUnknown, &tmpFSRef )	:
                                            1;
                    CFRelease( tmpStringRef );
            }
            else
            {
                err = 1;
            }
        }
	
	if( err == noErr )
        {
            *outFSRef = tmpFSRef;
	}
        
	return err;
}
