/*
Copyright (c) 2008-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICFilesHierarchyUtilities.h"

#include <sys/stat.h>
#include <fts.h>

#import "RegexKitLite.h"

#import "ICProjectBuilder+Constants.h"

@implementation ICFilesHierarchyUtilities

+ (BOOL) isEmptyDirectoryAtPath:(char *) inPath
{
	FTS * ftsp;
    FTSENT * tFile;
    char * tPath[2]={inPath,NULL};
    
    if ((ftsp = fts_open(tPath, FTS_PHYSICAL, NULL)) != NULL)
    {
		while ((tFile = fts_read(ftsp)) != NULL)
		{
			switch (tFile->fts_info)
			{
				case FTS_DNR:
				case FTS_ERR:
				case FTS_NS:
					fts_close(ftsp);
				
					return NO;
					
				case FTS_DOT:
				
					break;
				
				case FTS_DP:
				case FTS_D:
				
					if (!strcmp(inPath,tFile->fts_path))
					{
						break;
					}
				
				
				case FTS_DC:
				case FTS_F:
				case FTS_SL:
				case FTS_SLNONE:
				
					fts_close(ftsp);
					
					return NO;

				default:
					break;
			}
		}
		
		fts_close(ftsp);
    }

	return YES;
}

+ (int) removeDirectoryAtPath:(char *) inPath
{
    FTS * ftsp;
    FTSENT * tFile;
    char * tPath[2]={inPath,NULL};
    
    if ((ftsp = fts_open(tPath, FTS_PHYSICAL, NULL)) == NULL)
    {
        return 1;
    }
	
    while ((tFile = fts_read(ftsp)) != NULL)
    {
        switch (tFile->fts_info)
        {
            case FTS_DNR:
            case FTS_ERR:
            case FTS_NS:
				fts_close(ftsp);
				
				return 1;
			
			case FTS_DP:
				
				if ((tFile->fts_statp->st_flags & (UF_APPEND|UF_IMMUTABLE)) &&
					!(tFile->fts_statp->st_flags & (SF_APPEND|SF_IMMUTABLE)))
				{
					chflags(tFile->fts_accpath,tFile->fts_statp->st_flags &= ~(UF_APPEND|UF_IMMUTABLE));
				}
				
				if (rmdir(tFile->fts_accpath)==-1)
				{
					NSLog(@"delete: dir %d",errno);
					
					return 1;
				}
				
				break;
			
			case FTS_SLNONE:
			case FTS_SL:
			case FTS_F:
				
				if ((tFile->fts_statp->st_flags & (UF_APPEND|UF_IMMUTABLE)) &&
					!(tFile->fts_statp->st_flags & (SF_APPEND|SF_IMMUTABLE)))
				{
					chflags(tFile->fts_accpath,tFile->fts_statp->st_flags &= ~(UF_APPEND|UF_IMMUTABLE));
				}
				
				if (unlink(tFile->fts_accpath)==-1)
				{
					NSLog(@"delete: file %d",errno);
					
					return 1;
				}
				
				break;

            
            default:
				break;
        }
    }
    
    fts_close(ftsp);
    
    if (errno)
    {
        return 1;
    }
    
    return 0;
}


+ (int) changeMode:(mode_t) inMode atPath:(NSString *) inPath
{
    FTS * ftsp;
    FTSENT * tFile;
    char * tPath[2]={(char *) [inPath fileSystemRepresentation],NULL};
    
    if ((ftsp = fts_open(tPath, FTS_PHYSICAL, NULL)) == NULL)
    {
        return 1;
    }

    while ((tFile = fts_read(ftsp)) != NULL)
    {
        switch (tFile->fts_info)
        {
            case FTS_D:
            case FTS_SL:
            case FTS_SLNONE:
                    continue;
            case FTS_DNR:
            case FTS_ERR:
            case FTS_NS:
                    fts_close(ftsp);
                    
                    return 1;
            default:
                    break;
        }
        
        if (inMode == tFile->fts_statp->st_mode)
        {
            continue;
        }
        
        if (chmod(tFile->fts_accpath, inMode) == -1)
        {
            fts_close(ftsp);
            
            return 1;
        }
    }
    
    fts_close(ftsp);
    
    if (errno)
    {
        return 1;
    }
    
    return 0;
}


+ (int) changeOwner:(uid_t) inUID group:(gid_t) inGID atPath:(NSString *) inPath
{
    FTS * ftsp;
    FTSENT * tFile;
    char * tPath[2]={(char *)[inPath fileSystemRepresentation],NULL};
    
    if ((ftsp = fts_open(tPath, FTS_PHYSICAL, NULL)) == NULL)
    {
        return 1;
    }

    while ((tFile = fts_read(ftsp)) != NULL)
    {
        switch (tFile->fts_info)
        {
            case FTS_D:
            case FTS_SL:
            case FTS_SLNONE:
                    continue;
            case FTS_DNR:
            case FTS_ERR:
            case FTS_NS:
                    fts_close(ftsp);
                    
                    return 1;
            default:
                    break;
        }
        
        if (inUID == tFile->fts_statp->st_uid && inGID == tFile->fts_statp->st_gid)
        {
            continue;
        }
        
        if (chown(tFile->fts_accpath, inUID, inGID) == -1)
        {
            fts_close(ftsp);
            
            return 1;
        }
    }
    
    fts_close(ftsp);
    
    if (errno)
    {
        return 1;
    }
    
    return 0;
}

+ (BOOL) getSize:(int64_t *) outSize  andNumberOfFiles:(int64_t *) outNumberOfFiles atPath:(NSString *) inPath
{
	FTS * ftsp;
    FTSENT * tFile;
    char * tPath[2]={(char *) [inPath fileSystemRepresentation],NULL};
    off_t tFolderSize=0;
    int64_t tNumberOfFiles=0;
	
    if ((ftsp = fts_open(tPath, FTS_PHYSICAL, NULL)) == NULL)
    {
        return NO;
    }

    while ((tFile = fts_read(ftsp)) != NULL)
    {
        switch (tFile->fts_info)
        {
            case FTS_SL:
				
				tNumberOfFiles++;		// A VOIR (taille)
				
				continue;
			
			case FTS_D:		// Pre-order directory
            case FTS_SLNONE:
                    continue;
            case FTS_DNR:
            case FTS_ERR:
            case FTS_NS:
                    fts_close(ftsp);
                    
                    return NO;
            default:
                    break;
        }
        
        tFolderSize+=tFile->fts_statp->st_size;
		
		tNumberOfFiles++;
    }
    
    fts_close(ftsp);
    
	*outSize=tFolderSize;
	
	*outNumberOfFiles=tNumberOfFiles;
	
    return YES;
}

+ (BOOL) removeExceptions:(NSDictionary *) inExceptionsDictionary fromFileHierarchyAtPath:(NSString *) inPath;
{
	if (inExceptionsDictionary!=nil && [inExceptionsDictionary count]>0)
	{
		NSArray * tFileStaticExceptions;
		NSArray * tFolderStaticExceptions;
		NSArray * tFileRegExExceptions;
		NSArray * tFolderRegExExceptions;
		FTS * ftsp;
		FTSENT * tFile;
		char * tPath[2]={(char *) [inPath fileSystemRepresentation],NULL};
		NSString * tItemName;
		NSUInteger i,tCount;
		NSError * tError=nil;
		
		tFileStaticExceptions=[inExceptionsDictionary objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_STANDARD_FILES];
		
		tFolderStaticExceptions=[inExceptionsDictionary objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_STANDARD_FOLDERS];
		
		
		tFileRegExExceptions=[inExceptionsDictionary objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_REGEX_FILES];
		
		tFolderRegExExceptions=[inExceptionsDictionary objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_REGEX_FOLDERS];
		
		
		if ((ftsp = fts_open(tPath, FTS_PHYSICAL, NULL)) == NULL)
		{
			return NO;
		}

		while ((tFile = fts_read(ftsp)) != NULL)
		{
			switch (tFile->fts_info)
			{
				case FTS_DNR:
				case FTS_ERR:
				case FTS_NS:
					
					fts_close(ftsp);
						
					return NO;
						
				case FTS_DP:
				
					// Folder
					
					tItemName=[NSString stringWithUTF8String:tFile->fts_name];
					
					if (tFolderStaticExceptions!=nil)
					{
						NSString * tPattern;
						
						tCount=[tFolderStaticExceptions count];
						
						for(i=0;i<tCount;i++)
						{
							tPattern=[tFolderStaticExceptions objectAtIndex:i];
							
							if ([tPattern caseInsensitiveCompare:tItemName]==NSOrderedSame)
							{
								goto cleandir;
							}
						}
					}
					
					if (tFolderRegExExceptions!=nil)
					{
						NSString * tPattern;
						
						tCount=[tFolderRegExExceptions count];
						
						for(i=0;i<tCount;i++)
						{
							tPattern=[tFolderRegExExceptions objectAtIndex:i];
							
							if ([tItemName isMatchedByRegex:tPattern options:RKLNoOptions inRange:((NSRange){0, NSUIntegerMax}) error:&tError]==YES)
							{
								goto cleandir;
							}
							
							if (tError!=nil)
							{
								// A COMPLETER
								
								fts_close(ftsp);
								
								return NO;
							}
						}
					}
					
					continue;
					
				case FTS_F:
				
					// File
					
					tItemName=[NSString stringWithUTF8String:tFile->fts_name];
					
					if (tFileStaticExceptions!=nil)
					{
						NSString * tPattern;
							
						tCount=[tFileStaticExceptions count];
						
						for(i=0;i<tCount;i++)
						{
							tPattern=[tFileStaticExceptions objectAtIndex:i];
							
							if ([tPattern caseInsensitiveCompare:tItemName]==NSOrderedSame)
							{
								goto cleanfile;
							}
						}
					}
					
					if (tFileRegExExceptions!=nil)
					{
						NSString * tPattern;
						
						tCount=[tFileRegExExceptions count];
						
						for(i=0;i<tCount;i++)
						{
							
							
							tPattern=[tFileRegExExceptions objectAtIndex:i];
							
							if ([tItemName isMatchedByRegex:tPattern options:RKLNoOptions inRange:((NSRange){0, NSUIntegerMax}) error:&tError]==YES)
							{
								goto cleanfile;
							}
							
							if (tError!=nil)
							{
								// A COMPLETER
								
								fts_close(ftsp);
								
								return NO;
							}
						}
					}
					
					continue;
				
				default:
					
					continue;
			}
			
		cleandir:
			
			if ([ICFilesHierarchyUtilities removeDirectoryAtPath:tFile->fts_path]!=0)
			{
				fts_close(ftsp);
				
				return NO;
			}
			
			continue;
			
		cleanfile:

			if (unlink(tFile->fts_path)==-1)
			{
				switch(errno)
				{
					case EPERM:
						
						if (chown(tFile->fts_path,0,0)==0)
						{
							unlink(tFile->fts_path);
						}
						
						break;
				}
				
				
			}
		}
		
		fts_close(ftsp);
	}

	return YES;
}

@end
