/*
Copyright (c) 2008-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import <Foundation/Foundation.h>

#include <sys/param.h>
#include <sys/types.h>

#import "ICProjectBuilder+Constants.h"

#import "ICRequirementPluginsManager.h"
#import "ICRequirementPlugin_Converter.h"

#import "ICLocatorPluginsManager.h"
#import "ICLocatorPlugin_Converter.h"

#import <Security/Security.h>

#import "ICBuildDistributedNotificationCenter.h"

#define IC_PACKAGE_BUILD_MODE_PACKAGE_RAW			0

#define IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE	1

#define IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT		2


@interface NSDictionary (RequirementSort)

- (NSComparisonResult) compareRequirementBehavior:(NSDictionary *) inDictionary;

@end

@interface ICProjectBuilder : NSObject
{
	NSFileManager * fileManager_;
	
	ICBuildDistributedNotificationCenter * buildNotificationCenter_;
	
	NSString * projectPath_;
	
	NSString * notificationPath_;
    
	NSString * referenceProjectPath_;
	
	NSString * referenceFolderPath_;
	
	BOOL debug_;
	
	int processID_;
    int userID_;
    gid_t groupID_;
	
	int unknownID_;
    int groupCount_;
    gid_t groups_[NGROUPS + 3];	/* NGROUPS + 1 + Unknwon + groupID_*/
	
	NSDictionary * folderAttributes_;
	mode_t fileAttributes_;
	
	NSString * scratchLocation_;
	
	NSDictionary * documentDictionary_;
	
	NSMutableDictionary * projectDictionary_;
	
	
	NSXMLElement * installerScriptElement_;
	
	
	NSMutableDictionary * options_;

	NSMutableDictionary * buildInformation_;
	
	ICRequirementPluginsManager * sharedRequirementPluginsManager_;
	
	ICLocatorPluginsManager * sharedLocatorPluginsManager_;
	
	
	int buildFormat_;
	
	
	NSUInteger signatureResult_;
	
	SecIdentityRef identity_;
	
	NSData * signedData_;
	
	NSDictionary * certificateDictionary_;
	
	CFMessagePortContext messagePortContext_;
	
	
	BOOL rootIsYouDaddy_;	// Is the build folder located on a volume mounted by root?
}

+ (ICProjectBuilder *) sharedProjectBuilder;

- (void) initializeGroups;

- (NSString *) pathFromPathDictionary:(NSDictionary *) inPathDictionary;

- (BOOL) checkUserPermissionAtPath:(NSString *) inPath;

- (BOOL) copyObjectAtPath:(NSString *) inPath toPath:(NSString *) toPath;

- (BOOL) createDirectoryAtPath:(NSString *) inPath;

- (BOOL) setFileOwnerAtPath:(NSString *) aPath traverseHierarchy:(BOOL) inTraverseHierarchy;


- (BOOL) buildPath:(NSString *) inPath fixedPermissionAfter:(BOOL) inFixedPermissionAfter;


- (void) buildProjectAtPath:(NSString *) inProjectPath forProcessID:(int) inProcessID withUserID:(int) inUserID groupID:(int) inGroupID notificationPath:(NSString *) inNotificationPath splitForksToolName:(NSString *) inSplitForksToolName scratchPath:(NSString *) inScratchPath referenceFolder:(NSString *) inReferenceFolderPath debug:(BOOL) inDebug;

- (BOOL) buildDistributionProjectAtPath:(NSString *) inPath;

- (BOOL) buildPackageProjectAtPath:(NSString *) inPath;


- (BOOL) addBundleFromArray:(NSArray *) inArray toElement:(NSXMLElement *) inElement withPath:(NSString *) inPath packageInfoElement:(NSXMLElement *) inPackageInfoElement downgradableBundles:(NSMutableArray *) inDowngradableArray;

- (BOOL) prepareExceptionRules;

- (BOOL) setDistributionOptions;

- (BOOL) setAdvancedDistributionOptions;

- (BOOL) addLocalizedErrorMessages:(NSArray *) inLocalizationsArray withName:(NSString *) inName errorMessage:(NSString **) outErrorMessage errorDescription:(NSString **) outErrorDescription;

- (BOOL) fillDistributionXMLWithDictionary:(NSDictionary *) inDictionary;

- (BOOL) setExtraAttributes;

- (BOOL) setDistributionResources;

- (BOOL) setExtraResources;

- (BOOL) buildDistributionScriptsAtPath:(NSString *) inPath;

- (BOOL) setTitle:(NSDictionary *) inPresentationDictionary;

- (BOOL) setBackground:(NSDictionary *) inPresentationDictionary;

- (BOOL) setIntroduction:(NSDictionary *) inPresentationDictionary;

- (BOOL) setReadMe:(NSDictionary *) inPresentationDictionary;

- (BOOL) setLicense:(NSDictionary *) inPresentationDictionary;

- (BOOL) setConclusion:(NSDictionary *) inPresentationDictionary;

- (BOOL) setChoiceOutline:(NSDictionary *) inPresentationDictionary;

- (BOOL) addLinesElementsToElement:(NSXMLElement *) inElement withChoicesArray:(NSArray *) inChoicesArray isInstaller:(BOOL) inIsInstaller prefix:(NSString *) inPrefix;

- (BOOL) addPackageRefeferenceElementToChoiceElement:(NSXMLElement *) inElement fromDictionary:(NSDictionary *) inChoiceDictionary;

- (int) hasOneHideAndUnselectRequirementEnabledInArray:(NSArray *) inArray;

- (NSString *) logicStringForDependency:(NSDictionary *) inDictionary;


- (BOOL) addRelocators:(NSArray *) inLocatorsArray forBundle:(NSString *) inBundleIdentifier packageInfoElement:(NSXMLElement *) inPackageInfoElement;


- (BOOL) setRequirementsFunctionForChoiceName:(NSString *) inName withArray:(NSArray *) inArray functionName:(NSString **) outFunctionName;

- (BOOL) createEnabledDependencyFunctionNamed:(NSString *) inFunctionName withDependencyDictionary:(NSDictionary *) inDictionary;

- (BOOL) createSelectedDependencyFunctionNamed:(NSString *) inFunctionName withEnabledState:(int) inEnabledState  withEnabledFunctionName:(NSString *) inEnabledFunctionName withDependencyDictionary:(NSDictionary *) inDictionary;

- (BOOL) setStateAttributesOfChoiceElement:(NSXMLElement *) inElement name:(NSString *) inName withOptionsDictionary:(NSDictionary *) inOptionsDictionary;

- (BOOL) setStateAttributesOfChoiceElement:(NSXMLElement *) inElement name:(NSString *) inName requirementFunctionName:(NSString *) inRequirementFunctionName affectVisible:(BOOL) inAffectVisible withOptionsDictionary:(NSDictionary *) inOptionsDictionary;

- (BOOL) setStateAttributesOfChoiceElement:(NSXMLElement *) inElement name:(NSString *) inName withDictionary:(NSDictionary *) inChoiceDictionary;

- (BOOL) addChoicesElementsToElement:(NSXMLElement *) inElement withChoicesArray:(NSArray *) inChoicesArray isInstaller:(BOOL) inIsInstaller;

- (BOOL) setPackagesReferences:(NSDictionary *) inPresentationDictionary;

- (BOOL) setRequirements;

- (BOOL) setPresentation;

- (BOOL) setJavaScriptScripts;

- (BOOL) setLocalizableStrings;

- (BOOL) setPlugins:(NSString *) inPath;

+ (NSString *) shortFolderNameForEnglishLanguage:(NSString *) inEnglishLanguage;

+ (NSString *) folderNameForEnglishLanguage:(NSString *) inEnglishLanguage;

- (NSString *) getDistributionPathForLanguage:(NSString *) inLanguage;

- (NSString *) getDistributionResourcesPath;

- (NSString *) getDistributionScriptsPath;

- (BOOL) unarchiveXarFileAtPath:(NSString *) inPath toFolder:(NSString *) inFolderPath;

- (BOOL) setCorrectXarPermissionsForFolder:(NSString *) inPath;

- (CFDataRef) processSignatureMessagePortMessage:(SInt32) inMessageID withData:(CFDataRef) inData;

- (void) delayedSignatureRequest:(NSDictionary *) inDictionary;

- (BOOL) createXarFileNamed:(NSString *) inFileName atPath:(NSString *) inPath withContentsOfFolder:(NSString *) inFolderPath;


- (BOOL) buildPackageFromDictionary:(NSMutableDictionary *) inPackageDictionary atPath:(NSString *) inPath mode:(int) inBuildMode;

- (NSString *) copyScriptAtPath:(NSString *) inSourcePath toPath:(NSString *) inDestinationPath;

- (BOOL) buildScriptsAtPath:(NSString *) inPath forScriptsDictionary:(NSDictionary *) inScriptsDictionary ofPackageUUID:(NSString *) inPackageUUID;

- (BOOL) buildPayloadAndBomAtPath:(NSString *) inPath forFilesDictionary:(NSDictionary *) inFilesDictionary ofPackageUUID:(NSString *) inPackageUUID;

- (BOOL) buildFileHierarchyComponent:(NSDictionary *) inDictionary atPath:(NSString *) inPath rootComponent:(BOOL) inRootComponent packagebuildInformation:(NSDictionary *) inPackageBuildInformation;


- (BOOL) buildBundleVersionsDictionaryWithFileHierarchyAtPath:(NSString *) inFileHierarchyPath ofPackageUUID:(NSString *) inPackageUUID;

@end
