/*
Copyright (c) 2004-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import <Foundation/Foundation.h>
#import "ICBuildObserver.h"
#import "ICDispatch+Constants.h"

#import "ICProjectBuilder+Notifications.h"

#include <sys/types.h>
#include <unistd.h>


void usage(void);
int buildToOrder(NSString * inPath,NSString * inScratchPath,NSString * inReferenceFolder,BOOL inDebug);

void usage(void)
{
    (void)fprintf(stderr, "%s\n","usage: packagesbuild [-v] [-d] [-F <reference folder>] [-t <temporary build location>] "
                          " file ...");
    
    exit(1);
}

BOOL wakeupDispatcher(void)
{
	// Send a Mach Port message to start the dipatcher
		
	NSMachBootstrapServer * tBootstrapServer;
	
	tBootstrapServer = [NSMachBootstrapServer sharedInstance];
	
	if (tBootstrapServer!=nil)
	{
		NSMachPort * tMachPort;
		
		tMachPort=(NSMachPort *) [tBootstrapServer portForName:@ICDISPATCH_MACHPORT_NAME];
		
		if (tMachPort!=nil)
		{
			NSPortMessage * tPortMessage;
			
			tPortMessage= [[NSPortMessage alloc] initWithSendPort:tMachPort
													  receivePort:nil
													   components:nil];
			
			if (tPortMessage)
			{
				[tPortMessage setMsgid:0];
				
				[tPortMessage sendBeforeDate:[NSDate distantFuture]];
				
				[tPortMessage release];
				
				return YES;
			}
		}
	}
	
	return NO;
}

int buildToOrder(NSString * inPath,NSString * inScratchPath,NSString * inReferenceFolder,BOOL inDebug)
{
    CFMessagePortRef tRemote;
    
    tRemote=CFMessagePortCreateRemote(NULL,(CFStringRef) ICDISPATCH_MESSAGEPORT_NAME);
	
#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5
	if (tRemote==NULL)
	{
		int tCount1=0;
	
		do
		{
			if (wakeupDispatcher()==YES)
			{
				int tCount2=0;
				struct timespec tSleepDuration={0,100000000};
				
				do
				{
					nanosleep(&tSleepDuration,NULL);
					
					tRemote=CFMessagePortCreateRemote(NULL,(CFStringRef) ICDISPATCH_MESSAGEPORT_NAME);
					
					tCount2++;
				}
				while (tCount2<20 && tRemote==NULL);
			}
			
			tCount1++;
		}
		while (tCount1<2 && tRemote==NULL);
	}
#endif

    if (tRemote!=NULL)
    {
        CFDataRef tDataRef;
        NSMutableDictionary * tMutableDictionary;
        
        tMutableDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:inPath,ICDISPATCH_PROJECT_PATH,
																		     inPath,ICDISPATCH_NOTIFICATION_PATH,
																		     [NSNumber numberWithInt:geteuid()],ICDISPATCH_USER_ID,
																		     [NSNumber numberWithInt:getegid()],ICDISPATCH_GROUP_ID,
																		     [NSNumber numberWithInt:[[NSProcessInfo processInfo] processIdentifier]],ICDISPATCH_PROCESS_ID,
																			 [NSNumber numberWithBool:inDebug],ICDISPATCH_DEBUG_MODE,
																		     nil];
        
        if (tMutableDictionary!=nil)
        {
            if (inReferenceFolder!=nil)
			{
				[tMutableDictionary setObject:inReferenceFolder forKey:ICDISPATCH_REFERENCE_FOLDER];
			}
			
			if (inScratchPath!=nil)
			{
				[tMutableDictionary setObject:inScratchPath forKey:ICDISPATCH_SCRATCH_PATH];
			}
			
			
			tDataRef=CFPropertyListCreateXMLData(kCFAllocatorDefault,(CFPropertyListRef) tMutableDictionary);
        
            if (tDataRef!=NULL)
            {
                if (CFMessagePortSendRequest(tRemote,0,tDataRef, 20, 10, kCFRunLoopDefaultMode, NULL)!=kCFMessagePortSuccess)
                {
                    return -1;
                }
                
                // Release Memory
                
                CFRelease(tDataRef);
            }
        }
        
        // Release memory
        
        CFRelease(tRemote);
    }
    else
    {
        (void)fprintf(stderr, "The packages_dispatcher process is not responding. packagesbuild can't build any project when this process is not running.\n");
        
        return -1;
    }
    
    return 0;
}

#pragma mark -

int main (int argc, const char * argv[])
{
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
    int ch;
    BOOL tVerbose=NO;
	BOOL tDebug=NO;
    char * tCScratchPath=NULL;
    char * tCReferenceFolder=NULL;
	
    // Check the parameters
    
    while ((ch = getopt(argc,(char **) argv,"vd?F:t:")) != -1)
    {
        switch(ch)
        {
             case 'F':
			
                tCReferenceFolder=optarg;
				
                break;
				
            case 'd':
			
                tDebug=YES;
				
                break;
				
			case 't':
			
                tCScratchPath=optarg;
				
                break;
				
            case 'v':
			
                tVerbose=YES;
				
                break;
				
            case '?':
            default:
                usage();
        }
    }
    
    argv+=optind;
    argc-=optind;
    
    if (argc < 1)
    {
        usage();
    }
    else
    {
        ICBuildObserver * nBuildObserver;
        NSString * tProjectPath;
        NSFileManager * tFileManager;
        NSString * tCurrentDirectory;
        NSString * tScratchPath=nil;
        NSString * tReferenceFolder=nil;
		
        tFileManager=[NSFileManager defaultManager];
        
        tCurrentDirectory=[tFileManager currentDirectoryPath];
        
        if (tCScratchPath!=NULL)
        {
            tScratchPath=[[NSString stringWithUTF8String:tCScratchPath] stringByStandardizingPath];
        
            if ([tScratchPath characterAtIndex:0]!='/')
            {
                tScratchPath=[tCurrentDirectory stringByAppendingPathComponent:tScratchPath];
            }
            
            if ([tFileManager fileExistsAtPath:tScratchPath]==NO)
            {
               (void)fprintf(stderr, "Temporary build location \'%s\' does not exist.\n",tCScratchPath);
    
				exit(-1);
            }
        }
		
		 if (tCReferenceFolder!=NULL)
        {
            BOOL isDirectory;
			
			tReferenceFolder=[[NSString stringWithUTF8String:tCReferenceFolder] stringByStandardizingPath];
        
            if ([tReferenceFolder characterAtIndex:0]!='/')
            {
                tReferenceFolder=[tCurrentDirectory stringByAppendingPathComponent:tReferenceFolder];
            }
            
            if ([tFileManager fileExistsAtPath:tReferenceFolder isDirectory:&isDirectory]==NO)
            {
               (void)fprintf(stderr, "Reference folder \'%s\' does not exist.\n",tCReferenceFolder);
    
				exit(-1);
            }
			else if (isDirectory==NO)
			{
				(void)fprintf(stderr, "Reference folder path \'%s\' is not a directory.\n",tCReferenceFolder);
    
				exit(-1);
			}
        }
        
        tProjectPath=[[NSString stringWithUTF8String:argv[0]] stringByStandardizingPath];
        
        if ([tProjectPath characterAtIndex:0]!='/')
        {
            tProjectPath=[tCurrentDirectory stringByAppendingPathComponent:tProjectPath];
        }
        
        if ([tFileManager fileExistsAtPath:tProjectPath]==YES)
        {
            // Install the Notification Observer
    
            nBuildObserver=[ICBuildObserver new];
            
            [nBuildObserver setVerbose:tVerbose];
            
            [[NSDistributedNotificationCenter defaultCenter] addObserver:nBuildObserver
                                                                selector:@selector(processBuildNotification:)
                                                                    name:PACKAGES_BUILDER_NOTIFICATION
                                                                object:nil
                                                    suspensionBehavior:NSNotificationSuspensionBehaviorDeliverImmediately];
            
            // Send the build message
            
            if (buildToOrder(tProjectPath,tScratchPath,tReferenceFolder,tDebug)!=0)
            {
                return 1;
            }
            
            // Run the loop
            
            [[NSRunLoop currentRunLoop] run];
        }
        else
        {
            (void)fprintf(stderr, "No such file or directory (%s)\n",argv[0]);
            
            return 1;
        }
    }
    
    [pool release];
    
    return 0;
}
