/*
Copyright (c) 2008-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICProjectBuilder.h"

#import "ICBuilderLogManager.h"

#include <grp.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <fts.h>

#include <xar/xar.h>
#include <CoreServices/CoreServices.h>

#import "ICDocumentFormat+Constants.h"

#import "ICPresentation+Constants.h"

#import "NSDictionary+Iceberg.h"
#import "NSString+Iceberg.h"

#import "ICLicenseProvider.h"

#import "ICInstallationHierarchyUtilities.h"

#import "ICInstallationTypeUtilities.h"

#import "ICPackageFileUtility.h"

#import "ICPackageUtilities.h"

#import "ICFilesHierarchyUtilities.h"

#import "ICFileSystemUtilities.h"

#import "ICPayloadUtilities.h"

#import "ICProjectRemoverErrorHandler.h"

#import "ICProjectBuilder+Notifications.h"

#import "ICRequirementCheck+Constants.h"

#import "ICBuildStack.h"

#import "ICCertificatesUtilities.h"

#import "ICUUIDFactory.h"

#import "ICProjectBuilder+SignatureRequest.h"

#import "NSString+Random.h"

#define MAX_RESOURCE_NAMING_ATTEMPTS		32768

#define __SET_CORRECT_PERMISSIONS__ 1


NSString * const ICPROJECTBUILDER_AUTHORING_TOOL_NAME=@"Packages";

NSString * const ICPROJECTBUILDER_AUTHORING_TOOL_VERSION=@"1.0";

NSString * const ICPROJECTBUILDER_AUTHORING_TOOL_BUILD=@"2B21";


NSString * const ICPROJECTBUILDER_DITTO_PATH=@"/usr/bin/ditto";

NSString * const ICPROJECTBUILDER_MKBOM_PATH=@"/usr/bin/mkbom";

NSString * const ICPROJECTBUILDER_RESOURCE_FORK_TOOL_SPLITFORKS_PATH=@"/Developer/Tools/SplitForks";

NSString * const ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN=@"goldin";

NSString * const ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN_PATH=@"/usr/local/bin/goldin_64";

NSString * const ICLANGUAGE_DEFAULT_LANGUAGE=@"ICLANGUAGE_DEFAULT_LANGUAGE";


CFDataRef SignatureMessagePortCallBack(CFMessagePortRef local, SInt32 msgid, CFDataRef data, void *info);

CFDataRef SignatureMessagePortCallBack(CFMessagePortRef local, SInt32 msgid, CFDataRef data, void *info)
{
	ICProjectBuilder * tProjectBuilder;
	
	tProjectBuilder=(ICProjectBuilder *) info;
	
	return [tProjectBuilder processSignatureMessagePortMessage:msgid withData:data];
}

@implementation NSDictionary (RequirementSort)

- (NSComparisonResult) compareRequirementBehavior:(NSDictionary *) inDictionary
{
	int tSelfValue,tValue;
	
	tSelfValue=[[self objectForKey:ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR] intValue];
	
	tValue=[[inDictionary objectForKey:ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR] intValue];
	
	if (tSelfValue<tValue)
	{
		return NSOrderedDescending;
	}
	else if (tValue<tSelfValue)
	{
		return NSOrderedAscending;
	}

	return NSOrderedSame;
}

@end

@implementation ICProjectBuilder

+ (ICProjectBuilder *) sharedProjectBuilder
{
	static ICProjectBuilder * sProjectBuilder=nil;
	
	if (sProjectBuilder==nil)
	{
		sProjectBuilder=[ICProjectBuilder new];
	}
	
	return sProjectBuilder;
}

- (id) init
{
	self=[super init];
	
	if (self!=nil)
	{
		fileManager_=[NSFileManager defaultManager];
		
		buildInformation_=[NSMutableDictionary new];
	}
	
	return self;
}

- (void) dealloc
{
	if (identity_!=NULL)
	{
		// Release Memory
									
		CFRelease(identity_);
	}
																	
	[notificationPath_ release];
	
	[folderAttributes_ release];

	[projectDictionary_ release];
	
	[documentDictionary_ release];
	
	[buildInformation_ release];
	
	[super dealloc];
}

#pragma mark -

- (void) initializeGroups
{
    struct passwd * tPassword;
    
    // groupID_
    
    groupCount_=0;
    
    groups_[groupCount_]=groupID_;
    
    groupCount_++;
    
    // Group and user 'unknown'
    
    tPassword = getpwnam("unknown");
    
    if (tPassword!=NULL)
    {
        groups_[groupCount_]=tPassword->pw_gid;
        
        groupCount_++;
        
        unknownID_=tPassword->pw_uid;
    }
    
    // Supplemental Groups
    
    tPassword = getpwuid(userID_);
    
    if (tPassword!=NULL)
    {
        gid_t gid,lastGid=-1;
        int i, tCount;
		gid_t tGroups[NGROUPS + 1];
        
        tCount = NGROUPS + 1;
        
        (void) getgrouplist(tPassword->pw_name, tPassword->pw_gid, tGroups, &tCount);

		for (i=0;i<tCount;i++)
        {
            gid = tGroups[i];
            
            if (lastGid!=gid)
            {
            	groups_[groupCount_]=gid;
            
                groupCount_++;
		
                lastGid = gid;
            }
		}
    }
}

- (NSString *) pathFromPathDictionary:(NSDictionary *) inPathDictionary
{
	NSDictionary * tMetadata=nil;
	
	if (inPathDictionary!=nil)
	{
		NSString * tPath;
		
		tPath=[inPathDictionary objectForKey:ICDOCUMENT_PATH];
		
		if (tPath!=nil)
		{
			NSNumber * tNumber;
			
			tNumber=[inPathDictionary objectForKey:ICDOCUMENT_PATH_TYPE];
			
			if (tNumber!=nil)
			{
				int tPathType;
				
				tPathType=[tNumber intValue];
				
				if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
                {
                    tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
                }
				else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
				{
					tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
				}
				
				return tPath;
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PATH_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
		else
		{
			// Missing Information
				
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return nil;
}

- (BOOL) copyObjectAtPath:(NSString *) inPath toPath:(NSString *) toPath
{
    BOOL tResult;
    
    tResult=[fileManager_ copyPath:inPath toPath:toPath handler:self];
    
    if (tResult==NO)
    {
        [buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_COPIED],IC_BUILDER_FAILURE_REASON,
																									   inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																									   toPath,IC_BUILDER_INFORMATION_SECONDARY_FILE_PATH,
																									   nil]];
    }
    
    return tResult;
}

- (BOOL) createDirectoryAtPath:(NSString *) inPath
{
    BOOL tResult;
    
    tResult=[fileManager_ createDirectoryAtPath:inPath attributes:folderAttributes_];

    if (tResult==NO)
    {
        [buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																									   inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																									   [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																									   nil]];
    }
    /*else
    {
        [permissionsToFixFolderArray_ addObject:inPath];
    }*/
    
    return tResult;
}

- (BOOL) checkUserPermissionAtPath:(NSString *) inPath
{
    struct stat tStat;
    int i;
    
    if (lstat([inPath fileSystemRepresentation], &tStat)!=0)
    {
        [buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:nil];
        
        return NO;
    }

    // Check that the user can read and write
    
    //[buildNotificationCenter_ postNotificationDebugInformation:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"Path: %@ / File Owner ID: %d File Group ID: %d User ID: %d Permissions:%x",inPath,tStat.st_uid,tStat.st_gid,userID_,tStat.st_mode],@"Info",nil]];
    
    if (userID_==tStat.st_uid || unknownID_==tStat.st_uid)
    {
        if ((tStat.st_mode & (S_IRUSR + S_IWUSR)) == (S_IRUSR + S_IWUSR))
        {
            return YES;
        }
    }

    for(i=0;i<groupCount_;i++)
    {
        if (groups_[i]==tStat.st_gid)
        {
            if ((tStat.st_mode & (S_IRGRP + S_IWGRP)) == (S_IRGRP + S_IWGRP))
            {
                return YES;
            }
        }
    }
        
    if ((tStat.st_mode & (S_IROTH + S_IWOTH)) != (S_IROTH + S_IWOTH))
    {
        [buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_WRITE],IC_BUILDER_FAILURE_REASON,
																									   inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																									   nil]];
        
        return NO;
    }
    
    return YES;
}

- (BOOL) setFileOwnerAtPath:(NSString *) inPath traverseHierarchy:(BOOL) inTraverseHierarchy
{
    if (inTraverseHierarchy==NO)
    {
		return [ICFileSystemUtilities setFileOwner:userID_ group:groupID_ atPath:inPath];
    }
	
	if ([ICFilesHierarchyUtilities changeOwner:userID_ group:groupID_ atPath:inPath]!=0)
    {
        NSDictionary * tMetadata=nil;
		
		switch(errno)
        {
            case EPERM:
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
				break;
			
			case ENOENT:
			
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
				break;
				
			case ENAMETOOLONG:
			
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Path(%@) is too long",inPath];
				
				break;
        }
        
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
		
        return NO;
    }
    
    return YES;
}

- (BOOL) buildPath:(NSString *) inPath fixedPermissionAfter:(BOOL) inFixedPermissionAfter
{
    BOOL isDirectory=NO;
    
    if ([fileManager_ fileExistsAtPath:inPath isDirectory:&isDirectory]==NO || isDirectory==NO)
    {
        if ([self buildPath:[inPath stringByDeletingLastPathComponent] fixedPermissionAfter:inFixedPermissionAfter]==YES)
        {
            if ([fileManager_ fileExistsAtPath:inPath]==YES)
            {
                // A File is annoying us
                
                // A COMPLETER
            }
            else
            {
                if (inFixedPermissionAfter==YES)
                {
                    if ([self createDirectoryAtPath:inPath]==NO)
                    {
                        return NO;
                    }
                }
                else
                {
                    if ([fileManager_ createDirectoryAtPath:inPath attributes:folderAttributes_]==NO)
                    {
                        /*[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																	forStep:IC_BUILDER_INFORMATION_STEP_PROJECT
																   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																													   inPath,IC_BUILDER_INFORMATION_FILE_PATH,
																													   [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																													   nil]];*/
						
						return NO;
                    }
                }
                
#ifdef __SET_CORRECT_PERMISSIONS__
                
				if ([self setFileOwnerAtPath:inPath traverseHierarchy:NO]==NO)
                {
                    return NO;
                }
#endif
            }
        }
        else
        {
            return NO;
        }
    }
    else
    {
        // Check the privileges
        
        // A COMPLETER
    }
    
    return YES;
}

#pragma mark -

- (void) buildProjectAtPath:(NSString *) inProjectPath forProcessID:(int) inProcessID withUserID:(int) inUserID groupID:(int) inGroupID notificationPath:(NSString *) inNotificationPath splitForksToolName:(NSString *) inSplitForksToolName scratchPath:(NSString *) inScratchPath  referenceFolder:(NSString *) inReferenceFolderPath debug:(BOOL) inDebug
{
    NSDictionary * tMetadata=nil;
	
	debug_=inDebug;
	
	referenceFolderPath_=inReferenceFolderPath;
	
	notificationPath_=[inNotificationPath copy];
	
	processID_=inProcessID;
	
	buildFormat_=ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT;
	
	buildNotificationCenter_=[ICBuildDistributedNotificationCenter defaultCenter];
	
	[buildNotificationCenter_ attachToProcess:inProcessID andProject:inNotificationPath];
	
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
												forStep:IC_BUILDER_INFORMATION_STEP_PROJECT
											   metadata:[NSDictionary dictionaryWithObjectsAndKeys:inProjectPath,IC_BUILDER_INFORMATION_FILE_PATH,nil]];
									 
	// Read the document and launch the building
    
    documentDictionary_=[[NSMutableDictionary mutableDictionaryWithContentsOfFile:inProjectPath] retain];
	
	if (documentDictionary_!=nil)
	{
		projectDictionary_=[[documentDictionary_ objectForKey:ICDOCUMENT_PROJECT] retain];
		
		if (projectDictionary_!=nil)
		{
			NSNumber * tNumber;
			
			tNumber=[documentDictionary_ objectForKey:ICDOCUMENT_TYPE];
			
			if (tNumber!=nil)
			{
				int tProjectType;
				
				tProjectType=[tNumber intValue];
				
				if (tProjectType!=ICDOCUMENT_TYPE_DISTRIBUTION &&
					tProjectType!=ICDOCUMENT_TYPE_PACKAGE)
				{
					// Incorrect value
						
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_TYPE",tProjectType];
						
					// Missing Information
				
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																										  ICDOCUMENT_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																										  nil];
				}
				else
				{
					int tAttemptCount=0;
					
					projectPath_=[inProjectPath copy];
					
					referenceProjectPath_=[notificationPath_ stringByDeletingLastPathComponent];
					
					rootIsYouDaddy_=YES;
					
					// Scratch location
			
					if (inScratchPath!=nil)
					{
						BOOL isDirectory;
						
						if ([fileManager_ fileExistsAtPath:inScratchPath isDirectory:&isDirectory]==NO || isDirectory==NO)
						{
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Scratch folder (%@) does not exist. Revert to default value.",inScratchPath];
							
							inScratchPath=nil;
						}
					}
					
					if (inScratchPath==nil)
					{
						inScratchPath=[NSString stringWithString:NSTemporaryDirectory()];
					}
					
					// Create a unique temporary folder at the scratchLocation
					
					do
					{
						NSString * tRandomFolderName;
						NSString * tRandomFolderPath;
						BOOL isDirectory;
						
						tRandomFolderName=[NSString randomFolderName];
						
						if (tRandomFolderName==nil)
						{
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
						
							goto buildProjectBail;
						}
						
						tRandomFolderPath=[inScratchPath stringByAppendingPathComponent:tRandomFolderName];
						
						if (tRandomFolderPath==nil)
						{
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
						
							goto buildProjectBail;
						}
						
						if ([fileManager_ fileExistsAtPath:tRandomFolderPath isDirectory:&isDirectory]==NO)
						{
							// Try to create the directory
							
							if ([self createDirectoryAtPath:tRandomFolderPath]==YES)
							{
								scratchLocation_=[tRandomFolderPath copy];
								
								break;
							}
							else
							{
								tMetadata=nil;
							}
						}
						
						
						tAttemptCount++;
					}
					while (tAttemptCount<100);
					
					if (tAttemptCount==100)
					{
						goto buildProjectBail;
					}
					
					options_=[[NSMutableDictionary alloc] initWithCapacity:2];
					
					if (options_==nil)
					{
						// Low Memory
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					}
					else
					{
						NSDictionary * tProjectSettings;
				
						// SplitForks Tool Name
						
						if (inSplitForksToolName==nil)
						{
							inSplitForksToolName=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN;
						}
						
						if (inSplitForksToolName!=nil)
						{
							[options_ setObject:inSplitForksToolName forKey:ICPROJECTBUILDER_BUILD_OPTIONS_SPLITFORK_TOOL];
						}
						
						// Initialize Project info
						
						userID_=inUserID;
						groupID_=inGroupID;
					
						[self initializeGroups];
						
						folderAttributes_=[[NSDictionary alloc] initWithObjectsAndKeys:[NSNumber numberWithInt:/*S_IRWXU+S_IRWXG+S_IRWXO*/S_IRWXU+S_IRGRP+S_IXGRP+S_IROTH+S_IXOTH],NSFilePosixPermissions,
																				   nil];
						
						fileAttributes_=S_IRUSR+S_IXUSR+S_IRGRP+S_IXGRP+S_IROTH+S_IXOTH;
						
						// Prepare the Build folder
						
						tProjectSettings=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];
						
						if (tProjectSettings!=nil)
						{
							NSString * tPath;
							BOOL needsToBuild=YES;
							BOOL isDirectory;
							NSString * tBuildFolderName=@"build";
							
							if (debug_==YES)
							{
								tBuildFolderName=@"build_debug";
							}
							
							
							if (referenceFolderPath_==nil)
							{
								referenceFolderPath_=[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_REFERENCE_FOLDER_PATH];
								
								if (referenceFolderPath_==nil)
								{
									referenceFolderPath_=referenceProjectPath_;
								}
							}
							
							tPath=[self pathFromPathDictionary:[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_BUILD_PATH]];
							
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																		forStep:IC_BUILDER_INFORMATION_STEP_PROJECT_BUILD_FOLDER
																	   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPath,IC_BUILDER_INFORMATION_FILE_PATH,nil]];
							
							if (tPath==nil)
							{
								// We need to create the Build Path by our own
								
								// We use inNotificationPath because it's the real project path
								
								tPath=[[notificationPath_ stringByDeletingLastPathComponent] stringByAppendingPathComponent:tBuildFolderName];
							}
							else
							{
								if (debug_==YES)
								{
									if ([[tPath lastPathComponent] caseInsensitiveCompare:@"build"]==NSOrderedSame)
									{
										tPath=[tPath stringByDeletingLastPathComponent];
									}
									
									tPath=[tPath stringByAppendingPathComponent:tBuildFolderName];
								}
							}
							
							if ([fileManager_ fileExistsAtPath:tPath isDirectory:&isDirectory]==YES)
							{
								struct statfs tStatfs;
								
								if (statfs([tPath fileSystemRepresentation],&tStatfs)==0)
								{
									rootIsYouDaddy_=(tStatfs.f_owner==0);
								}
								
								if (rootIsYouDaddy_==NO)
								{
									seteuid(userID_);
									setegid(groupID_);
								}
								
								if (isDirectory==YES)
								{
									// Check that the user can read and write
									
									if ([self checkUserPermissionAtPath:tPath]==NO)
									{
										// Try to remove the Build file
									
										if ([fileManager_ removeFileAtPath:tPath handler:self]==NO)
										{
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																								 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																								 nil];
										
											goto buildProjectBail;
										}
									}
									
									needsToBuild=NO;
								}
								else
								{
									// Try to remove the Build file
									
									if ([fileManager_ removeFileAtPath:tPath handler:self]==NO)
									{
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																							 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																							 nil];
									
										goto buildProjectBail;
									}
								}
							}
							else
							{
								struct statfs tStatfs;
								
								// The build folder does not exist
								
								if ([fileManager_ fileExistsAtPath:[tPath stringByDeletingLastPathComponent] isDirectory:&isDirectory]==NO)
								{
									// The parent folder does not exist either, we try to fall back to the project folder path
								
									tPath=[[inNotificationPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:tBuildFolderName];
								
									if ([fileManager_ fileExistsAtPath:tPath isDirectory:&isDirectory]==YES)
									{
										if (statfs([tPath fileSystemRepresentation],&tStatfs)==0)
										{
											rootIsYouDaddy_=(tStatfs.f_owner==0);
										}
										
										if (isDirectory==YES)
										{
											// Check that the user can read and write
											
											if ([self checkUserPermissionAtPath:tPath]==NO)
											{
												goto buildProjectBail;
											}
											
											needsToBuild=NO;
										}
										else
										{
											// Try to remove the Build file
											
											if ([fileManager_ removeFileAtPath:tPath handler:self]==NO)
											{
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																									 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																									 nil];
											
												goto buildProjectBail;
											}
										}
									}
									else
									{
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																							 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																							 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																							 nil];
											
										goto buildProjectBail;
									}
								}
								else
								{
									if (statfs([[tPath stringByDeletingLastPathComponent] fileSystemRepresentation],&tStatfs)==0)
									{
										rootIsYouDaddy_=(tStatfs.f_owner==0);
									}
								}
								
								if (rootIsYouDaddy_==NO)
								{
									seteuid(userID_);
									setegid(groupID_);
								}
							}
							
							if (needsToBuild==YES)
							{
								if ([self checkUserPermissionAtPath:[tPath stringByDeletingLastPathComponent]]==NO)
								{
									goto buildProjectBail;
								}
								
								// Create the Build Folder if needed
								
								if ([self createDirectoryAtPath:tPath]==NO)
								{
									goto buildProjectBail;
								}
								
								if (rootIsYouDaddy_==YES)
								{
									if ([self setFileOwnerAtPath:tPath traverseHierarchy:NO]==NO)
									{
										goto buildProjectBail;
									}
								}
							}
							
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																		forStep:IC_BUILDER_INFORMATION_STEP_PROJECT_BUILD_FOLDER
																	   metadata:nil];
							
							// Retrieve Shared Options
							
							// Excluded files
								
							if ([self prepareExceptionRules]==YES)
							{
								BOOL tResult=NO;
								
								// Build the Package or Distribution Script
							
								if (tProjectType==ICDOCUMENT_TYPE_DISTRIBUTION)
								{
									tResult=[self buildDistributionProjectAtPath:tPath];
								}
								else if (tProjectType==ICDOCUMENT_TYPE_PACKAGE)
								{
									tResult=[self buildPackageProjectAtPath:tPath];
								}
								
								// Try to remove the temporary folder
	
								if (scratchLocation_!=nil)
								{
									[fileManager_ removeFileAtPath:scratchLocation_ handler:nil];
									
									scratchLocation_=nil;
								}
								
								if (tResult==YES)
								{
									[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																				forStep:IC_BUILDER_INFORMATION_STEP_PROJECT
																			   metadata:nil];
							
									return;
								}
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_SETTINGS,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					}
				}
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	else
	{
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INCORRECT_TYPE],IC_BUILDER_FAILURE_REASON, nil];
		
		[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Unable to read file at path '%@'",inProjectPath];
	}
	
buildProjectBail:

	// Try to remove the temporary folder
	
	if (scratchLocation_!=nil)
	{
		[fileManager_ removeFileAtPath:scratchLocation_ handler:nil];
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_PROJECT
											   metadata:tMetadata];
}

#pragma mark -

- (BOOL) buildDistributionProjectAtPath:(NSString *) inPath
{
	NSDictionary * tProjectSettings;
	NSDictionary * tMetadata=nil;
	NSNumber * tNumber;
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
												forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION
											   metadata:nil];
	
	// Build the Bundle Skeleton
	
	tProjectSettings=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];
	
	// Get the Format
	
	tNumber=[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT];
	
	if (tNumber!=nil)
	{
		buildFormat_=[tNumber intValue];
		
		if (buildFormat_!=ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT &&
			buildFormat_!=ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
		{
			// Incorrect value
				
			[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT",buildFormat_];
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
		else
		{
			NSString * tString;
			
			// Get the name
		
			tString=[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_NAME];
		
			if (tString!=nil)
			{
				if ([tString length]>0)
				{
					NSMutableDictionary * tLocalizationsDictionary;
					NSMutableArray * tMutableArray;
					NSMutableDictionary * tMutableDictionary;
					NSMutableString * tMutableString;
					NSArray * tArray;
					NSArray * tPackagesToBeBuilt=nil;
					NSMutableDictionary * tProjectPresentationDictionary=nil;
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					
					// Prepare for localizations
										
					tLocalizationsDictionary=[NSMutableDictionary dictionary];
					
					if (tLocalizationsDictionary!=nil)
					{
						[buildInformation_ setObject:tLocalizationsDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
					}
					else
					{
						// Low Memory
						
						goto buildDistributionProjectBail;
					}
					
					tMutableArray=[NSMutableArray array];
					
					if (tMutableArray!=nil)
					{
						[buildInformation_ setObject:tMutableArray forKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_ARRAY];
					}
					else
					{
						// Low Memory
							
						goto buildDistributionProjectBail;
					}
					
					tMutableString=[NSMutableString string];
					
					if (tMutableString!=nil)
					{
						[buildInformation_ setObject:tMutableString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_STRING];
					
						if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
						{
							[tMutableString appendString:@"\tconst __IC_FLAT_DISTRIBUTION__=true;\n"];
						}
						else
						{
							[tMutableString appendString:@"\tconst __IC_FLAT_DISTRIBUTION__=false;\n"];
						}
					}
					else
					{
						// Low Memory
							
						goto buildDistributionProjectBail;
					}
					
					
					tMutableArray=[NSMutableArray array];
					
					if (tMutableArray!=nil)
					{
						[buildInformation_ setObject:tMutableArray forKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_ARRAY];
					}
					else
					{
						// Low Memory
							
						goto buildDistributionProjectBail;
					}
					
					tMutableDictionary=[NSMutableDictionary dictionary];
					
					if (tMutableDictionary!=nil)
					{
						[buildInformation_ setObject:tMutableDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_DICTIONARY];
					}
					else
					{
						// Low Memory
							
						goto buildDistributionProjectBail;
					}
					
					
					tMutableArray=[NSMutableArray array];
					
					if (tMutableArray!=nil)
					{
						[buildInformation_ setObject:tMutableArray forKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_ARRAY];
					}
					else
					{
						// Low Memory
							
						goto buildDistributionProjectBail;
					}
					
					tMutableDictionary=[NSMutableDictionary dictionary];
					
					if (tMutableDictionary!=nil)
					{
						[buildInformation_ setObject:tMutableDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_ATTRIBUTES_DICTIONARY];
					}
					else
					{
						// Low Memory
							
						goto buildDistributionProjectBail;
					}
					
					
					// Build the Packages
									
					tArray=[documentDictionary_ objectForKey:ICDOCUMENT_PACKAGES];
					
					if (tArray!=nil)
					{
						BOOL tContinueProcess=YES;
						NSXMLDocument * tXMLDocument;
						NSString * tContentsPath=nil;
						NSString * tBundlePath=nil;		// Only used for bundle
						
						tProjectPresentationDictionary=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_PRESENTATION];
						
						if (tProjectPresentationDictionary!=nil)
						{
							NSMutableDictionary * tInstallationTypeDictionary;
							
							tInstallationTypeDictionary=[tProjectPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE];
							
							if (tInstallationTypeDictionary==nil)
							{
								tInstallationTypeDictionary=[ICInstallationTypeUtilities defaultInstallationTypeDictionaryWithPackages:tArray];
								
								if (tInstallationTypeDictionary!=nil)
								{
									[tProjectPresentationDictionary setObject:tInstallationTypeDictionary forKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE];
								}
							}
						}
						
						// Find the list of Packages that are required to be be built
						
						tPackagesToBeBuilt=[ICInstallationHierarchyUtilities packagesUUIDEnabledInPresentationDictionary:tProjectPresentationDictionary];
					
					
						if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
						{
							// Bundle
							
							tBundlePath=[inPath stringByAppendingPathComponent:[NSString stringWithFormat:@"%@.mpkg",tString]];
							
							// Remove a previous build if needed
							
							if ([fileManager_ fileExistsAtPath:tBundlePath]==YES)
							{
								if ([fileManager_ removeFileAtPath:tBundlePath handler:self]==NO)
								{
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																						 tBundlePath,IC_BUILDER_INFORMATION_FILE_PATH,
																						 nil];
										
									goto buildDistributionProjectBail;
								}
							}
							
							// Build the Skeleton
							
							if ([self createDirectoryAtPath:tBundlePath]==YES)	// .mpkg/
							{
								tContentsPath=[tBundlePath stringByAppendingPathComponent:@"Contents"];
								
								if ([self createDirectoryAtPath:tContentsPath]==YES)	// Contents/
								{
									NSString * tPackagesPath;
									
									[buildInformation_ setObject:tContentsPath forKey:ICPROJECTBUILDER_BUILD_INFORMATION_CONTENTS_PATH];
									
									tPackagesPath=[tContentsPath stringByAppendingPathComponent:@"Packages"];
									
									if ([self createDirectoryAtPath:tPackagesPath]==YES)	// Packages/
									{
										// Build the list of Packages that are required to be be built
										
										if (tPackagesToBeBuilt!=nil && [tPackagesToBeBuilt count]>0)
										{
											NSEnumerator * tEnumerator;
											
											[buildNotificationCenter_ postNotificationBuildProgressionInformation:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:[tPackagesToBeBuilt count]],IC_BUILDER_INFORMATION_BUILD_PROGRESSION_PACKAGES_NUMBER,
																												   nil]];
											
											tEnumerator=[tArray objectEnumerator];
											
											if (tEnumerator!=nil)
											{
												NSMutableDictionary * tPackageDictionary;
												
												while (tPackageDictionary=[tEnumerator nextObject])
												{
													NSString * tUUID;
													
													tUUID=[tPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_UUID];
													
													if (tUUID!=nil)
													{
														if ([tPackagesToBeBuilt containsObject:tUUID]==YES)
														{
															if ([self buildPackageFromDictionary:tPackageDictionary atPath:tBundlePath mode:IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE]==NO)
															{
																goto buildDistributionProjectBail;
															}
														}
													}
													else
													{
														// Missing Information
					
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																											 ICDOCUMENT_PACKAGE_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																											 nil];
														
														goto buildDistributionProjectBail;
													}
												}
											}
											else
											{
												// Low Memory
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
												
												goto buildDistributionProjectBail;
											}
										}
									}
									else
									{
										tMetadata=nil;
									}
								}
								else
								{
									tMetadata=nil;
								}
							}
							else
							{
								tMetadata=nil;
							}
						}
						else if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
						{
							// Flat
							
							tBundlePath=[inPath stringByAppendingPathComponent:[NSString stringWithFormat:@"%@.pkg",tString]];
							
							tContentsPath=[scratchLocation_ stringByAppendingPathComponent:@"_root_"];
							
							if ([fileManager_ fileExistsAtPath:tContentsPath]==NO)
							{
								// Try to create the directory
								
								if ([self createDirectoryAtPath:tContentsPath]==NO)
								{
									goto buildDistributionProjectBail;
								}
							}
							
							[buildInformation_ setObject:tContentsPath forKey:ICPROJECTBUILDER_BUILD_INFORMATION_CONTENTS_PATH];
							
							// Build the list of Packages that are required to be be built
							
							if (tPackagesToBeBuilt!=nil && [tPackagesToBeBuilt count]>0)
							{
								NSEnumerator * tEnumerator;
								
								[buildNotificationCenter_ postNotificationBuildProgressionInformation:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:[tPackagesToBeBuilt count]],IC_BUILDER_INFORMATION_BUILD_PROGRESSION_PACKAGES_NUMBER,
																									   nil]];
								
								tEnumerator=[tArray objectEnumerator];
								
								if (tEnumerator!=nil)
								{
									NSMutableDictionary * tPackageDictionary;
									
									while (tPackageDictionary=[tEnumerator nextObject])
									{
										NSString * tUUID;
										
										tUUID=[tPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_UUID];
										
										if (tUUID!=nil)
										{
											if ([tPackagesToBeBuilt containsObject:tUUID]==YES)
											{
												if ([self buildPackageFromDictionary:tPackageDictionary atPath:tBundlePath mode:IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT]==NO)
												{
													goto buildDistributionProjectBail;
												}
											}
										}
										else
										{
											// Missing Information
		
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PACKAGE_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
											
											goto buildDistributionProjectBail;
										}
									}
								}
								else
								{
									// Low Memory
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
									
									goto buildDistributionProjectBail;
								}
							}
						}
						
						// Build the Distribution Script
										
						[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																	forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_SCRIPT
																   metadata:nil];
						
						// Preset the metadata
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
						
						installerScriptElement_=(NSXMLElement *) [NSXMLNode elementWithName:@"installer-script"];
						
						if (installerScriptElement_!=nil)
						{
							id tAttribute;
							
							// authoringTool
							
							tAttribute=[NSXMLNode attributeWithName:@"authoringTool" stringValue:ICPROJECTBUILDER_AUTHORING_TOOL_NAME];
							
							if (tAttribute!=nil)
							{
								[installerScriptElement_ addAttribute:tAttribute];
								
								// authoringToolVersion
								
								tAttribute=[NSXMLNode attributeWithName:@"authoringToolVersion" stringValue:ICPROJECTBUILDER_AUTHORING_TOOL_VERSION];
							
								if (tAttribute!=nil)
								{
									[installerScriptElement_ addAttribute:tAttribute];
									
									// authoringToolBuild
									
									tAttribute=[NSXMLNode attributeWithName:@"authoringToolBuild" stringValue:ICPROJECTBUILDER_AUTHORING_TOOL_BUILD];
							
									if (tAttribute!=nil)
									{
										id tComment;
										
										tContinueProcess=NO;
										
										[installerScriptElement_ addAttribute:tAttribute];
										
										// Options
							
										if ([self setDistributionOptions]==YES)
										{
											// Requirements
										
											if ([self setRequirements]==YES)
											{
												// Presentation
										
												if ([self setPresentation]==YES)
												{
													// Choices/Requirements
										
													tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n        |    JavaScript Scripts    |\n        +==========================+"];
				
													if (tComment!=nil)
													{
														[installerScriptElement_ addChild:tComment];
													}
													
													// JavaScript Scripts
										
													if ([self setJavaScriptScripts]==YES)
													{
														// Create Localizable.strings files
										
														if ([self setLocalizableStrings]==YES)
														{
															// Advanced Options
							
															if ([self setAdvancedDistributionOptions]==YES)
															{
																// Extra Attributes
										
																if ([self setExtraAttributes]==YES)
																{
																	// Add the Resources
									
																	if ([self setDistributionResources]==YES)
																	{
																		// Extra Resources
										
																		if ([self setExtraResources]==YES)
																		{
																			if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
																			{
																				// Build the Scripts distribution archive
																				
																				if ([self buildDistributionScriptsAtPath:tContentsPath]==YES)
																				{
																					// Plug-ins
										
																					if ([self setPlugins:tContentsPath]==YES)
																					{
																						tContinueProcess=YES;
																					}
																				}
																			}
																			else
																			{
																				// Plug-ins
																				
																				if ([self setPlugins:tContentsPath]==YES)
																				{
																					tContinueProcess=YES;
																				}
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										
										if (tContinueProcess==YES)
										{
											// XML Document

											tXMLDocument=[[[NSXMLDocument alloc] initWithRootElement:installerScriptElement_] autorelease];
											
											if (tXMLDocument!=nil)
											{
												NSData * tData;
												NSError * tError;
												
												[tXMLDocument setCharacterEncoding:@"UTF-8"];
												
												tData=[tXMLDocument XMLDataWithOptions:NSXMLDocumentTidyHTML|NSXMLNodePrettyPrint|NSXMLNodeCompactEmptyElement];
												
												if (tData!=nil)
												{
													NSString * tDistributionScriptFileName=@"";
													
													if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
													{
														tDistributionScriptFileName=@"distribution.dist";
													}
													else if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
													{
														tDistributionScriptFileName=@"Distribution";
													}
													
													if ([tData writeToFile:[tContentsPath stringByAppendingPathComponent:tDistributionScriptFileName] options:0 error:&tError]==YES)
													{
														NSMutableDictionary * tProjectMetaData;
														
														// Set the permissions to the correct user
														
														if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
														{
															if ([ICFilesHierarchyUtilities changeOwner:userID_ group:groupID_ atPath:tBundlePath]!=0)
															{
																[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"An error occurred while changing owner and group attributes of %@",tBundlePath];
																
																// A COMPLETER
															}
														}
														
														[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																									forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_SCRIPT
																								   metadata:nil];
														
														
														if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
														{
															if ([self setCorrectXarPermissionsForFolder:tContentsPath]==YES)
															{
																if (identity_==NULL)
																{
																	if (certificateDictionary_==nil)
																	{
																		certificateDictionary_=[[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS] objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE];
																	}
																	
																	if (certificateDictionary_!=nil && [certificateDictionary_ count]>0)
																	{
																		NSString * tName;
																		NSString * tKeychainPath;
																		
																		tName=[certificateDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME];
																		
																		if (tName!=nil)
																		{
																			// A COMPLETER
																		}
																		else
																		{
																			// Missing Information
																			
																			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																																 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																																 nil];
																	
																			goto buildDistributionProjectBail;
																		}
																		
																		tKeychainPath=[certificateDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH];
																		
																		if (tKeychainPath!=nil)
																		{
																			// A COMPLETER
																		}
																		else
																		{
																			// Missing Information
																			
																			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																																 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																																 nil];
																	
																			goto buildDistributionProjectBail;
																		}
																		
																		identity_=[ICCertificatesUtilities identityWithName:tName atPath:tKeychainPath];
																	
																		if (identity_==NULL)
																		{
																			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_SIGNATURE_GENERIC_ERROR],IC_BUILDER_FAILURE_REASON,
																																 nil];
																			
																			goto buildDistributionProjectBail;
																		}
																		
																		CFRetain(identity_);
																	}
																}
																
																// Build the xar archive
															
																if ([self createXarFileNamed:[NSString stringWithFormat:@"%@.pkg",tString] atPath:inPath withContentsOfFolder:tContentsPath]==YES)
																{
																	// Remove the Package Working folder
																	
																	if ([fileManager_ removeFileAtPath:tContentsPath handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==NO)
																	{	
																		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																															 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																															 tContentsPath,IC_BUILDER_FAILURE_TAG_NAME,
																															 nil];
																	
																		goto buildDistributionProjectBail;
																	}
																}
																else
																{
																	tMetadata=nil;
																
																	goto buildDistributionProjectBail;
																}
															}
															else
															{
																tMetadata=nil;
																
																goto buildDistributionProjectBail;
															}
														}
														
														[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																									forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION
																								   metadata:nil];
														   
														tProjectMetaData=[[[ICBuildStack sharedBuildStack] metadataForStep:IC_BUILDER_INFORMATION_STEP_PROJECT] mutableCopy];
														
														if (tProjectMetaData==nil)
														{
															tProjectMetaData=[[NSMutableDictionary alloc] initWithCapacity:1];
														}
														
														if (tProjectMetaData!=nil)
														{
															[tProjectMetaData setObject:tBundlePath forKey:IC_BUILDER_INFORMATION_FILE_PATH];
														
															[[ICBuildStack sharedBuildStack] setMetadata:tProjectMetaData forStep:IC_BUILDER_INFORMATION_STEP_PROJECT];
															
															[tProjectMetaData release];
														}
														
														return YES;
													}
													else
													{
														// Unable to write file
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																											 [tContentsPath stringByAppendingPathComponent:@"distribution.dist"],IC_BUILDER_INFORMATION_FILE_PATH,
																											 //,IC_BUILDER_INFORMATION_SYSTEM_ERROR_DESCRIPTION,
																											 nil];
													}
												}
												else
												{
													// Low Memory
													
													//tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
												}
											}
											else
											{
												// Low Memory
										
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create XML Document"];
											}
										}
										else
										{
											tMetadata=nil;
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'authoringToolBuild' attribute"];
									}
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'authoringToolVersion' attribute"];
								}
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'authoringTool' attribute"];
							}
						}
						else
						{
							// Low Memory
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <installer-script> element"];
						}
						
buildDistributionprojectScriptBail:
						
						if (tContinueProcess==YES)
						{
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																		forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_SCRIPT
																	   metadata:tMetadata];
						}

					}
					else
					{
						// Missing Information
	
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PACKAGES,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				}
				else
				{
					// Empty String Value
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_PROJECT_SETTINGS_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
				
			}
			else
			{
				// Missing Information
					
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PROJECT_SETTINGS_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
	}
	else
	{
		// Missing Information
				
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
															 ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT,IC_BUILDER_FAILURE_TAG_NAME,
															 nil];
	}
	
	
buildDistributionProjectBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION
											   metadata:tMetadata];
	
	return NO;
}

#pragma mark -

- (BOOL) buildPackageProjectAtPath:(NSString *) inPath
{
	NSDictionary * tProjectSettings;
	NSString * tString;
	NSDictionary * tMetadata=nil;
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
												forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE
											   metadata:nil];
	
	tProjectSettings=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];
	
	// Get the name
		
	tString=[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_NAME];
		
	if (tString!=nil)
	{
		if ([tString length]>0)
		{
			NSString * tPath;
			
			tPath=[inPath stringByAppendingPathComponent:[NSString stringWithFormat:@"%@.pkg",tString]];
			
			// Package Build Process: Start
			
			if ([self buildPackageFromDictionary:projectDictionary_ atPath:tPath mode:IC_PACKAGE_BUILD_MODE_PACKAGE_RAW]==YES)
			{
				NSMutableDictionary * tProjectMetaData;
				
				// Set the permissions to the correct user
				
				if ([ICFilesHierarchyUtilities changeOwner:userID_ group:groupID_ atPath:tPath]!=0)
				{
					// A COMPLETER
				}
				
				tProjectMetaData=[[[ICBuildStack sharedBuildStack] metadataForStep:IC_BUILDER_INFORMATION_STEP_PROJECT] mutableCopy];
																		
				if (tProjectMetaData==nil)
				{
					tProjectMetaData=[[NSMutableDictionary alloc] initWithCapacity:1];
				}
				
				if (tProjectMetaData!=nil)
				{
					[tProjectMetaData setObject:tPath forKey:IC_BUILDER_INFORMATION_FILE_PATH];
				
					[[ICBuildStack sharedBuildStack] setMetadata:tProjectMetaData forStep:IC_BUILDER_INFORMATION_STEP_PROJECT];
					
					[tProjectMetaData release];
				
				}
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
															forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE
														   metadata:nil];
	
				return YES;
			}
		}
		else
		{
			// String is Empty
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_SETTINGS_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	else
	{
		// Missing Information
				
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
															 ICDOCUMENT_PROJECT_SETTINGS_NAME,IC_BUILDER_FAILURE_TAG_NAME,
															 nil];
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE
											   metadata:tMetadata];
	
	return NO;
}

#pragma mark -

- (BOOL) setDistributionOptions
{
	NSXMLElement * tOptionsElement;
	NSDictionary * tMetadata;
	
	tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
	
	tOptionsElement=(NSXMLElement *) [NSXMLNode elementWithName:@"options"];
	
	if (tOptionsElement!=nil)
	{
		NSDictionary * tDictionary;
		NSNumber * tNumber;
		id tAttribute;
		
		// rootVolumeOnly
		
		tDictionary=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_REQUIREMENTS];
		
		if (tDictionary!=nil)
		{
			tNumber=[tDictionary objectForKey:ICDOCUMENT_PROJECT_REQUIREMENTS_ROOT_VOLUME_ONLY];
			
			if (tNumber!=nil && [tNumber boolValue]==YES)
			{
				tAttribute=[NSXMLNode attributeWithName:@"rootVolumeOnly" stringValue:@"true"];
										
				if (tAttribute!=nil)
				{
					[tOptionsElement addAttribute:tAttribute];
				}
				else
				{
					// Low Memory
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'rootVolumeOnly' attribute"];
					
					goto setDistributionOptionsBail;
				}
			}
		}
		
		// customize
		
		tDictionary=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_PRESENTATION];
		
		if (tDictionary!=nil)
		{
			NSDictionary * tInstallationTypeDictionary;
			int tInstallationTypeMode=ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_MODE_ALLOW;
			
			tInstallationTypeDictionary=[tDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE];
			
			if (tInstallationTypeDictionary!=nil)
			{
				tNumber=[tInstallationTypeDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_MODE];
				
				if (tNumber!=nil)
				{
					tInstallationTypeMode=[tNumber intValue];
					
					if (tInstallationTypeMode<ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_MODE_ALLOW || tInstallationTypeMode>ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_MODE_ALWAYS)
					{
						// Incorrect Value
					
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_SCALING",tInstallationTypeMode];
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_MODE,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
						
						goto setDistributionOptionsBail;
					}
				}
			}
			
			if (tInstallationTypeMode==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_MODE_NEVER)
			{
				tAttribute=[NSXMLNode attributeWithName:@"customize" stringValue:@"never"];
										
				if (tAttribute!=nil)
				{
					[tOptionsElement addAttribute:tAttribute];
				}
				else
				{
					// Low Memory
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'customize' attribute"];
					
					goto setDistributionOptionsBail;
				}
			}
			else if (tInstallationTypeMode==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_MODE_ALWAYS)
			{
				tAttribute=[NSXMLNode attributeWithName:@"customize" stringValue:@"always"];
										
				if (tAttribute!=nil)
				{
					[tOptionsElement addAttribute:tAttribute];
				}
				else
				{
					// Low Memory
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'customize' attribute"];
					
					goto setDistributionOptionsBail;
				}
			}
		}
		
		// allow-external-scripts
		
		// A COMPLETER
		
		[installerScriptElement_ addChild:tOptionsElement];
		
		return YES;
	}
	else
	{
		// Low Memory
	
		[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <options> element"];
	}
	

setDistributionOptionsBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) fillDistributionXMLWithDictionary:(NSDictionary *) inDictionary
{
	NSDictionary * tMetadata=nil;
	
	if (inDictionary!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		tEnumerator=[inDictionary keyEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSString * tKey;
			
			while (tKey=[tEnumerator nextObject])
			{
				id tValue;
				id tAttribute=nil;
				
				tValue=[inDictionary objectForKey:tKey];
				
				if (tValue!=nil)
				{
					NSArray * tKeyPath;
					
					tKeyPath=[tKey componentsSeparatedByString:@":"];
					
					if (tKeyPath!=nil && [tKeyPath count]==2)
					{
						NSArray * tElementsPathArray;
						NSString * tAttributeName;
						
						tAttributeName=[tKeyPath objectAtIndex:1];
						
						tElementsPathArray=[[tKeyPath objectAtIndex:0] componentsSeparatedByString:@"."];
						
						if (tElementsPathArray!=nil)
						{
							NSUInteger tCount;
							
							tCount=[tElementsPathArray count];
							
							if (tCount>0)
							{
								NSXMLElement * tParentElement=nil;
								NSXMLElement * tElement;
								NSUInteger i;
								NSString * tString=nil;
								
								tElement=installerScriptElement_;
							
								for(i=1;i<tCount;i++)
								{
									NSArray * tArray;
									NSString * tElementName;
									
									tParentElement=tElement;
									
									tElementName=[tElementsPathArray objectAtIndex:i];
									
									tArray=[tParentElement elementsForName:tElementName];
									
									if (tArray==nil || [tArray count]==0)
									{
										// Create the element
										
										tElement=[NSXMLNode elementWithName:tElementName];
										
										if (tElement!=nil)
										{
											[tParentElement addChild:tElement];
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create <%@> element",tElementName];
											
											goto fillDistributionXMLWithDictionaryBail;
										}
									}
									else
									{
										tElement=[tArray objectAtIndex:0];
									}
								}
								
								if ([tValue isKindOfClass:[NSNumber class]]==YES)
								{
									tString=([tValue boolValue]==YES) ? @"true" : @"false";
								}
								else if ([tValue isKindOfClass:[NSString class]]==YES)
								{
									tString=tValue;
								}
								else if ([tValue isKindOfClass:[NSArray class]]==YES)
								{
									tString=[tValue componentsJoinedByString:@","];
								}
								
								if (tString!=nil)
								{
									tAttribute=[tElement attributeForName:tAttributeName];
									
									if (tAttribute==nil)
									{
										[tElement removeAttributeForName:tAttributeName];
									}
									
									tAttribute=[NSXMLNode attributeWithName:tAttributeName stringValue:tString];
									
									if (tAttribute!=nil)
									{
										[tElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
			
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create '%@' attribute",tAttributeName];
									
										goto fillDistributionXMLWithDictionaryBail;
									}
								}
								else
								{
									// A COMPLETER
								}
								
								continue;
							}
						}
					}
					
					// Incorrect Value
						
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,nil];
								
					goto fillDistributionXMLWithDictionaryBail;
				}
			}
			
			return YES;
		}
	}
	
fillDistributionXMLWithDictionaryBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) setAdvancedDistributionOptions
{
	NSDictionary * tProjectSettings;
	NSDictionary * tAdvancedOptionsDictionary;
	id tAttribute;
	BOOL tMinSpecVersionSet=NO;
	
	tProjectSettings=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];
	
	tAdvancedOptionsDictionary=[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_ADVANCED_OPTIONS];
	
	if (tAdvancedOptionsDictionary!=nil)
	{
		if ([tAdvancedOptionsDictionary objectForKey:@"installer-script:minSpecVersion"]!=nil)
		{
			tMinSpecVersionSet=YES;
		}
		
		if ([self fillDistributionXMLWithDictionary:tAdvancedOptionsDictionary]==NO)
		{
			return NO;
		}
	}
	
	if (tMinSpecVersionSet==NO)
	{
		// At least set the minSpecVersion
		
		tAttribute=[NSXMLNode attributeWithName:@"minSpecVersion" stringValue:@"1.0"];
										
		if (tAttribute!=nil)
		{
			[installerScriptElement_ addAttribute:tAttribute];
		}
		else
		{
			// Low Memory
			
			[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'minSpecVersion' attribute"];
			
			[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
														forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
													   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil]];
			
			return NO;
		}
	}
	
	return YES;
}

- (BOOL) setExtraAttributes
{
	NSDictionary * tExtraAttributesDictionary;
		
	tExtraAttributesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_ATTRIBUTES_DICTIONARY];
				
	return [self fillDistributionXMLWithDictionary:tExtraAttributesDictionary];
}

- (BOOL) setDistributionResources
{
	NSDictionary * tProjectRequirements;
	
	tProjectRequirements=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_REQUIREMENTS];
	
	if (tProjectRequirements!=nil)
	{
		NSMutableArray * tHierarchyArray;
		NSDictionary * tMetadata=nil;
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_RESOURCES
												   metadata:nil];
		
		tHierarchyArray=[tProjectRequirements objectForKey:ICDOCUMENT_PROJECT_REQUIREMENTS_FILE_HIERARCHY];
		
		if (tHierarchyArray!=nil)
		{
			NSString * tResourcesPath;
			
			tResourcesPath=[self getDistributionScriptsPath];
			
			if (tResourcesPath!=nil)
			{
				NSNumber * tNumber;
				NSUInteger i,tCount;
				NSDictionary * tChildDictionary;
				
				// Copy the file hierarchy
				
				tCount=[tHierarchyArray count];
				
				for(i=0;i<tCount;i++)
				{
					tChildDictionary=[tHierarchyArray objectAtIndex:i];
					
					if ([self buildFileHierarchyComponent:tChildDictionary atPath:tResourcesPath rootComponent:NO packagebuildInformation:nil]==NO)
					{
						return NO;
					}
				}
				
				// Remove the files defined as exceptions
				
				tNumber=[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_PAYLOAD_ONLY];
				
				if (tNumber==nil || [tNumber boolValue]==NO)
				{
					if ([ICFilesHierarchyUtilities removeExceptions:[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS] fromFileHierarchyAtPath:tResourcesPath]==NO)
					{
						return NO;
					}
				}
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_RESOURCES
														   metadata:nil];
												   
				return YES;
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_REQUIREMENTS_FILE_HIERARCHY,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
		
		return NO;
	}
	
	return YES;
}

- (BOOL) setExtraResources
{
	NSArray * tExtraResourcesArray;
	NSEnumerator * tEnumerator;
	NSDictionary * tMetadata=nil;
	
	tExtraResourcesArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_ARRAY];
	
	tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
	
	tEnumerator=[tExtraResourcesArray objectEnumerator];
	
	if (tEnumerator!=nil)
	{
		
		NSString * tDestinationDirectory;
		
		tDestinationDirectory=[self getDistributionScriptsPath];
				
		if (tDestinationDirectory!=nil)
		{
			NSMutableDictionary * tMergedExtraResources;
			
			tMergedExtraResources=[NSMutableDictionary dictionary];
			
			if (tMergedExtraResources!=nil)
			{
				NSDictionary * tSourceDictionary;
				
				// Make sure we only add an extra resource once.
				
				while (tSourceDictionary=[tEnumerator nextObject])
				{
					NSString * tPath;
					NSNumber * tNumber;
						
					tNumber=[tSourceDictionary objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_PERMISSIONS];
					
					if (tNumber!=nil)
					{
						tPath=[self pathFromPathDictionary:[tSourceDictionary objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_PATH]];
					}
					else
					{
						tPath=[self pathFromPathDictionary:tSourceDictionary];
					}
					
					if (tPath!=nil)
					{
						if (tNumber!=nil)
						{
							[tMergedExtraResources setObject:tNumber forKey:tPath];
						}
						else
						{
							[tMergedExtraResources setObject:[NSNumber numberWithInt:-1] forKey:tPath];
						}
					}
					else
					{
						tMetadata=nil;
						
						// A COMPLETER
						
						break;
					}
				}
				
				tEnumerator=[tMergedExtraResources keyEnumerator];
	
				if (tEnumerator!=nil)
				{
					NSString * tSourcePath;
					
					while (tSourcePath=[tEnumerator nextObject])
					{
						if ([fileManager_ fileExistsAtPath:tSourcePath]==YES)
						{
							if ([fileManager_ fileExistsAtPath:[tDestinationDirectory stringByAppendingPathComponent:[tSourcePath lastPathComponent]]]==NO)
							{
								// Copy the file
							
								if ([ICFileSystemUtilities copyPath:tSourcePath toPath:tDestinationDirectory]==YES)
								{
									NSNumber * tSourcePermissions;
									int tPrivileges=-1;
									
									tSourcePermissions=[tMergedExtraResources objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_PERMISSIONS];
									
									if (tSourcePermissions!=nil)
									{
										tPrivileges=[tSourcePermissions intValue];
									}
									
									if (tPrivileges==-1)
									{
										tPrivileges=0775;
									}
									
									// Set the Permissions
									
									if ([ICFileSystemUtilities setFilePrivileges:tPrivileges atPath:[tDestinationDirectory stringByAppendingPathComponent:[tSourcePath lastPathComponent]]]==NO)
									{
										return NO;
									}
									
									continue;
								}
							}
							else
							{
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_COPIED],IC_BUILDER_FAILURE_REASON,
																													  tSourcePath,IC_BUILDER_INFORMATION_FILE_PATH,
																													  [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_OVERWRITE,
																													  nil];
							}
						}
						else
						{
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																												  tSourcePath,IC_BUILDER_INFORMATION_FILE_PATH,
																												  nil];
						}
						
						break;
					}
						
					if (tSourceDictionary==nil)
					{
						return YES;
					}
				}
			}
		}
		else
		{
			// A COMPLETER
			
			tMetadata=nil;
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
	
	return NO;
}

#pragma mark -

- (BOOL) addLocalizedErrorMessages:(NSArray *) inLocalizationsArray withName:(NSString *) inName errorMessage:(NSString **) outErrorMessage errorDescription:(NSString **) outErrorDescription
{
	NSDictionary * tMetadata=nil;
	
	if (inLocalizationsArray!=nil && inName!=nil && outErrorMessage!=nil)
	{
		NSEnumerator * tLocalizationsEnumerator;

		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		(*outErrorMessage)=nil;
	
		if (outErrorDescription!=nil)
		{
			(*outErrorDescription)=nil;
		}
		
		tLocalizationsEnumerator=[inLocalizationsArray objectEnumerator];
			
		if (tLocalizationsEnumerator!=nil)
		{
			NSMutableDictionary * tLocalizationsDictionary;
			NSDictionary * tLanguageDictionary;
			BOOL atLeastOneErrorMessage=NO;
			BOOL atLeastOneErrorDescription=NO;
			
			tLocalizationsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
			
			while (tLanguageDictionary=[tLocalizationsEnumerator nextObject])
			{
				NSString * tLanguage;
				
				tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
				
				if (tLanguage!=nil)
				{
					NSMutableDictionary * tLanguageLocalizationsDictionary;
					NSString * tLocalization;
					
					tLanguageLocalizationsDictionary=[tLocalizationsDictionary objectForKey:tLanguage];
					
					if (tLanguageLocalizationsDictionary==nil)
					{
						tLanguageLocalizationsDictionary=[NSMutableDictionary dictionary];
						
						if (tLanguageLocalizationsDictionary!=nil)
						{
							[tLocalizationsDictionary setObject:tLanguageLocalizationsDictionary forKey:tLanguage];
						}
						else
						{
							// Low Memory
							
							goto addLocalizedErrorMessagesBail;
						}
					}
					
					tLocalization=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
				
					if (tLocalization!=nil)
					{
						if ((*outErrorMessage)==nil)
						{
							(*outErrorMessage)=[NSString stringWithFormat:@"REQUIREMENT_FAILED_MESSAGE_%@",[inName uppercaseString]];
						}
							
						if ([tLocalization length]==0)
						{
							tLocalization=@" ";
						}
						
						[tLanguageLocalizationsDictionary setObject:tLocalization forKey:*outErrorMessage];
						
						atLeastOneErrorMessage=YES;
					}
					else
					{
						// Missing Information
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_VALUE,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
						
						goto addLocalizedErrorMessagesBail;
					}
					
					if (outErrorDescription!=nil)
					{
						tLocalization=[tLanguageDictionary objectForKey:ICDOCUMENT_SECONDARY_VALUE];
					
						if (tLocalization!=nil)
						{
							if ((*outErrorDescription)==nil)
							{
								(*outErrorDescription)=[NSString stringWithFormat:@"REQUIREMENT_FAILED_DESCRIPTION_%@",[inName uppercaseString]];
							}
							
							if ([tLocalization length]==0)
							{
								tLocalization=@" ";
							}
							
							[tLanguageLocalizationsDictionary setObject:tLocalization forKey:*outErrorDescription];
							
							atLeastOneErrorDescription=YES;
						}
					}
				}
				else
				{
					// Missing Information
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
					
					goto addLocalizedErrorMessagesBail;
				}
			}
			
			if (atLeastOneErrorDescription==YES || atLeastOneErrorMessage==YES)
			{
				NSMutableDictionary * tDefaultLanguageDictionary;
																		
				// Set the default value

				tDefaultLanguageDictionary=[tLocalizationsDictionary objectForKey:ICLANGUAGE_DEFAULT_LANGUAGE];
						
				if (tDefaultLanguageDictionary==nil)
				{
					tDefaultLanguageDictionary=[NSMutableDictionary dictionary];
					
					if (tDefaultLanguageDictionary!=nil)
					{
						[tLocalizationsDictionary setObject:tDefaultLanguageDictionary forKey:ICLANGUAGE_DEFAULT_LANGUAGE];
					}
					else
					{
						// Low Memory
						
						goto addLocalizedErrorMessagesBail;
					}
				}
				
				if (atLeastOneErrorMessage==YES)
				{
					[tDefaultLanguageDictionary setObject:@" " forKey:(*outErrorMessage)];
				}
										
				if (atLeastOneErrorDescription==YES)
				{
					[tDefaultLanguageDictionary setObject:@" " forKey:(*outErrorDescription)];
				}
			}
			
			return YES;
		}
	}
	
addLocalizedErrorMessagesBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
												   
	return NO;
}

- (BOOL) setRequirements
{
	NSDictionary * tProjectRequirements;
	NSDictionary * tProjectSettings;
	NSNumber * tNumber;
	BOOL tLeopardInstallationCheckRequired=NO;
	
	tProjectSettings=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];
	
	tProjectRequirements=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_REQUIREMENTS];
	
	if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
	{
		tLeopardInstallationCheckRequired=YES;
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLATION_VOLUME_REQUIREMENTS
												   metadata:nil];		
	}
	
	if (tProjectRequirements==nil && tLeopardInstallationCheckRequired==YES)
	{
		tProjectRequirements=[NSDictionary dictionary];
		
		if (tProjectRequirements==nil)
		{
			[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
														forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
													   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil]];
			
			return NO;
		}
	}
	
	if (tProjectRequirements!=nil)
	{
		NSMutableArray * tRequirementsArray;
		NSDictionary * tMetadata=nil;
		
		tRequirementsArray=[tProjectRequirements objectForKey:ICDOCUMENT_PROJECT_REQUIREMENTS_LIST];
		
		if (tRequirementsArray==nil && tLeopardInstallationCheckRequired==YES)
		{
			tRequirementsArray=[NSMutableArray array];
			
			if (tRequirementsArray==nil)
			{
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
															forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
														   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil]];
				
				return NO;
			}
		}
		
		if (tRequirementsArray!=nil)
		{
			if (tLeopardInstallationCheckRequired==YES)
			{
				// We need to add a OS requirement
				
				NSMutableDictionary * tMutableOSRequirementDictionary;
				
				tMutableOSRequirementDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInteger:1],@"IC_REQUIREMENT_OS_DISK_TYPE",
																								  [NSNumber numberWithInteger:0],@"IC_REQUIREMENT_OS_DISTRIBUTION_TYPE",
																								  [NSNumber numberWithInteger:100500],@"IC_REQUIREMENT_OS_MINIMUM_VERSION",
																								  nil];
				
				if (tMutableOSRequirementDictionary!=nil)
				{
					NSMutableDictionary * tMutableRequirementDictionary;
				
					tMutableRequirementDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInteger:ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_FAIL],ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,
																									tMutableOSRequirementDictionary,ICDOCUMENT_REQUIREMENT_DICTIONARY,
																									[NSNumber numberWithInteger:IC_REQUIREMENT_CHECK_TYPE_INSTALLATION],IC_REQUIREMENT_CHECK_TYPE,
																									[NSString stringWithString:@"fr.whitebox.Packages.requirement.os"],ICDOCUMENT_REQUIREMENT_IDENTIFIER,
																									[NSArray array],ICDOCUMENT_REQUIREMENT_MESSAGE,
																									[NSNumber numberWithBool:YES],ICDOCUMENT_REQUIREMENT_STATE,
																									nil];
				
					if (tMutableRequirementDictionary!=nil)
					{
						[tRequirementsArray insertObject:tMutableRequirementDictionary atIndex:0];
					}
				}
			}
			
			if ([tRequirementsArray count]>0)
			{
				NSEnumerator * tEnumerator;
			
				if (tLeopardInstallationCheckRequired==NO)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLATION_VOLUME_REQUIREMENTS
															   metadata:nil];
				}
				
				[tRequirementsArray sortUsingSelector:@selector(compareRequirementBehavior:)];
			
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				
				tEnumerator=[tRequirementsArray objectEnumerator];
			
				if (tEnumerator!=nil )
				{
					NSDictionary * tDictionary;
					NSMutableString * tVolumeCheckFunctionCode;
					NSMutableString * tInstallationCheckFunctionCode; 
					int tInstallationCheckDepth=-1,tVolumeCheckDepth=-1;
					NSString * tInstallationCheckTabulationDepth=@"\t",* tVolumeCheckTabulationDepth=@"\t";
					int tInstallationCheckIndex=0,tVolumeCheckIndex=0;
					int tRequirementIndex=0;
					NSString * tVolumeCheckPreTest=@"", * tInstallationCheckPreTest=@"";
					
					tVolumeCheckFunctionCode=[NSMutableString stringWithString:@"\tfunction volume_check()\n\
\t{\n\
\t\tvar tResult;\n\n"];

					tInstallationCheckFunctionCode=[NSMutableString stringWithString:@"\tfunction installation_check()\n\
\t{\n\
\t\tvar tResult;\n\n"];
				
					if (tVolumeCheckFunctionCode!=nil && tInstallationCheckFunctionCode!=nil)
					{
						NSMutableArray * tJavascriptFunctionsArray;
						NSMutableDictionary * tJavascriptFunctionsDictionary;
						NSMutableString * tJavascriptConstantsString;
						NSMutableArray * tJavascriptConstantsArray;
						NSMutableArray * tExtraResourcesArray;
						NSMutableDictionary * tExtraAttributesDictionary;
						int i;
						
						tExtraAttributesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_ATTRIBUTES_DICTIONARY];
						
						tExtraResourcesArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_ARRAY];
						
						tJavascriptConstantsString=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_STRING];
						
						tJavascriptConstantsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_ARRAY];
						
						tJavascriptFunctionsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_ARRAY];
									
						tJavascriptFunctionsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_DICTIONARY];
						
						if (sharedRequirementPluginsManager_==nil)
						{
							sharedRequirementPluginsManager_=[ICRequirementPluginsManager defaultManager];
						}
					
						if (sharedRequirementPluginsManager_!=nil)
						{
							BOOL tContinue=YES;
							
							while (tContinue==YES && (tDictionary=[tEnumerator nextObject]))
							{
								NSNumber * tStateNumber;
								
								tStateNumber=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_STATE];
								
								tContinue=NO;
								
								if (tStateNumber!=nil)
								{
									if ([tStateNumber boolValue]==YES)
									{
										NSString * tIdentifier;
										
										tIdentifier=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_IDENTIFIER];
										
										if (tIdentifier!=nil)
										{
											id tRequirementConverter;
											
											// Find the appropriate Requirement Converter plugin
											
											tRequirementConverter=[sharedRequirementPluginsManager_ createPluginConverterControllerForIdentifier:tIdentifier documentDictionary:documentDictionary_];
											
											if (tRequirementConverter!=nil)
											{
												NSDictionary * tRequirementSettingsDictionary;
												
												tRequirementSettingsDictionary=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_DICTIONARY];
												
												if (tRequirementSettingsDictionary!=nil)
												{
													NSString * tInvocationCode;
													NSDictionary * tErrorDictionary;
													int tRequirementType;
													int tIndex;
													int tDepth;
													NSString * tPretest;
													NSMutableString * tFunctionCode;
													NSString * tTabulationDepth;
													
													// Find if it's an installation or volumeCheck requirements
															
													tRequirementType=[tRequirementConverter requirementTypeWithParameters:tRequirementSettingsDictionary];
															
													if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_UNDEFINED)
													{
														tNumber=[tDictionary objectForKey:IC_REQUIREMENT_CHECK_TYPE];
														
														if (tNumber!=nil)
														{
															tRequirementType=[tNumber intValue];
														}
													}
													
													if (tRequirementType!=IC_REQUIREMENT_CHECK_TYPE_INSTALLATION && tRequirementType!=IC_REQUIREMENT_CHECK_TYPE_TARGET)
													{
														// Incorrect Value
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																											 IC_REQUIREMENT_CHECK_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																											 nil];
														
														break;
													}
													
													if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_INSTALLATION)
													{
														tIndex=tInstallationCheckIndex;
														
														tFunctionCode=tInstallationCheckFunctionCode;
														
														tTabulationDepth=tInstallationCheckTabulationDepth;
														
														tDepth=tInstallationCheckDepth;
														
														tPretest=tInstallationCheckPreTest;
													}
													else if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_TARGET)
													{
														tIndex=tVolumeCheckIndex;
														
														tFunctionCode=tVolumeCheckFunctionCode;
														
														tTabulationDepth=tVolumeCheckTabulationDepth;
														
														tDepth=tVolumeCheckDepth;
														
														tPretest=tVolumeCheckPreTest;
													}
													
													tInvocationCode=[tRequirementConverter invocationWithParameters:tRequirementSettingsDictionary index:tIndex error:&tErrorDictionary];
													
													if (tInvocationCode!=nil)
													{
														tNumber=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR];
														
														if (tNumber!=nil)
														{
															int tFailureBehavior;
															NSString * tCodePart;
															NSDictionary * tSharedFunctionDictionary;
															NSString * tVariables;
															NSMutableArray * tSharedConstants;
															NSDictionary * tAttributesDictionary;
															NSArray * tExtraResources;
															NSString * tErrorMessage,* tErrorDescription;
															NSArray * tLocalizations;
															
															tFailureBehavior=[tNumber intValue];
															
															// Add the shared functions
																
															tSharedFunctionDictionary=[tRequirementConverter sharedFunctionsImplementation];
															
															if (tSharedFunctionDictionary!=nil && [tSharedFunctionDictionary count]>0)
															{
																NSArray * tFunctionNamesKeys;
																
																tFunctionNamesKeys=[tSharedFunctionDictionary allKeys];
																
																if (tFunctionNamesKeys!=nil && [tFunctionNamesKeys count]>0)
																{
																	NSEnumerator * tFunctionEnumerator;
																	
																	tFunctionEnumerator=[tFunctionNamesKeys objectEnumerator];
																	
																	if (tFunctionEnumerator!=nil)
																	{
																		NSString * tSharedRequirementFunctionName;
																		
																		while (tSharedRequirementFunctionName=[tFunctionEnumerator nextObject])
																		{
																			if ([tJavascriptFunctionsDictionary objectForKey:tSharedRequirementFunctionName]==nil)
																			{
																				NSString * tImplementationCode;
																				
																				tImplementationCode=[tSharedFunctionDictionary objectForKey:tSharedRequirementFunctionName];
																				
																				if (tImplementationCode!=nil && [tImplementationCode length]>0)
																				{
																					[tJavascriptFunctionsArray addObject:tImplementationCode];
																					
																					[tJavascriptFunctionsDictionary setObject:tImplementationCode forKey:tSharedRequirementFunctionName];
																				}
																			}
																		}
																	}
																}
															}
																
															// Add the shared constants
															
															tSharedConstants=[[tRequirementConverter sharedConstantsNames] mutableCopy];
															
															if (tSharedConstants!=nil)
															{
																[tSharedConstants removeObjectsInArray:tJavascriptConstantsArray];
															
																if ([tSharedConstants count]>0)
																{
																	NSString * tString;
																	
																	tString=[tRequirementConverter constantsForNames:tSharedConstants];
																	
																	if (tString!=nil)
																	{
																		[tJavascriptConstantsArray addObjectsFromArray:tSharedConstants];
																		
																		[tJavascriptConstantsString appendFormat:@"%@\n",tString];
																	}
																	else
																	{
																		tMetadata=nil;
																		
																		break;
																	}
																}
																
																[tSharedConstants release];
															}
															
															
															
															tDepth++;
															
															tTabulationDepth=[tTabulationDepth stringByAppendingString:@"\t"];
															
															// Add the vars if needed
															
															tVariables=[tRequirementConverter variablesWithIndex:tIndex tabulationDepth:tTabulationDepth parameters:tRequirementSettingsDictionary error:&tErrorDictionary];
															
															if (tVariables==nil)
															{
																tMetadata=nil;
																
																break;
															}
															
															// Get the list of attributes to set
															
															tAttributesDictionary=[tRequirementConverter requiredOptionsValuesWithParameters:tRequirementSettingsDictionary];

															if (tAttributesDictionary!=nil && [tAttributesDictionary count]>0)
															{
																[tExtraAttributesDictionary addEntriesFromDictionary:tAttributesDictionary];
															}
															
															// Get the list of resources to add
															
															tExtraResources=[tRequirementConverter requiredAdditionalResourcesWithParameters:tRequirementSettingsDictionary];
															
															if (tExtraResources!=nil && [tExtraResources count]>0)
															{
																[tExtraResourcesArray addObjectsFromArray:tExtraResources];
															}
															
															tIndex++;
															
															tLocalizations=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_MESSAGE];
																
															if (tLocalizations!=nil)
															{
																NSString * tLocalizationKeyPrefix;
																BOOL noErrorMessage=NO;
																
																if ([tLocalizations count]==0)
																{
																	noErrorMessage=YES;
																}
																else
																{
																	tRequirementIndex++;
																}
																
																if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_INSTALLATION)
																{
																	if (noErrorMessage==NO)
																	{
																		tLocalizationKeyPrefix=[NSString stringWithFormat:@"INSTALLATION_CHECK_%d",tRequirementIndex];
																		
																		if ([self addLocalizedErrorMessages:tLocalizations withName:tLocalizationKeyPrefix errorMessage:&tErrorMessage errorDescription:&tErrorDescription]==NO)
																		{
																			return NO;
																		}
																		
																		noErrorMessage=(tErrorMessage==nil);
																	}
																	
																	tCodePart=[NSString stringWithFormat:@"%@%@%@tResult=%@;\n\n\
%@if (tResult==false)\n\
%@{\n",tPretest,tVariables,tTabulationDepth,tInvocationCode,tTabulationDepth,tTabulationDepth];

																	if (noErrorMessage==NO)
																	{
																		tCodePart=[tCodePart stringByAppendingFormat:@"%@\tmy.result.title = system.localizedString(\'%@\');\n\
%@\tmy.result.message = system.localizedString(\'%@\');\n",tTabulationDepth,tErrorMessage,tTabulationDepth,tErrorDescription];
																	}
																	else
																	{
																		tCodePart=[tCodePart stringByAppendingFormat:@"%@\tmy.result.title = system.localizedStandardStringWithFormat(\'InstallationCheckError\', system.localizedString(\'DISTRIBUTION_TITLE\'));\n\
%@\tmy.result.message = ' ';\n",tTabulationDepth,tTabulationDepth];
																	}
																	
																	if (tFailureBehavior==ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_FAIL)
																	{
																		tCodePart=[tCodePart stringByAppendingFormat:@"%@\tmy.result.type = 'Fatal';\n",tTabulationDepth];
																		
																	}
																	else if (tFailureBehavior==ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_WARN)
																	{
																		tCodePart=[tCodePart stringByAppendingFormat:@"%@\tmy.result.type = 'Warn';\n",tTabulationDepth];
																	}
																	else
																	{
																		// Incorrect Value
																		
																		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																															 ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,IC_BUILDER_FAILURE_TAG_NAME,
																															 nil];
																		
																		break;
																	}
																}
																else if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_TARGET)
																{
																	if (noErrorMessage==NO)
																	{
																		tLocalizationKeyPrefix=[NSString stringWithFormat:@"VOLUME_CHECK_%d",tRequirementIndex];
																		
																		if ([self addLocalizedErrorMessages:tLocalizations withName:tLocalizationKeyPrefix errorMessage:&tErrorMessage errorDescription:nil]==NO)
																		{
																			return NO;
																		}
																		
																		noErrorMessage=(tErrorMessage==nil);
																	}
																	
																	tCodePart=[NSString stringWithFormat:@"%@%@%@tResult=%@;\n\n\
%@if (tResult==false)\n\
%@{\n",tPretest,tVariables,tTabulationDepth,tInvocationCode,tTabulationDepth,tTabulationDepth];

																	if (noErrorMessage==NO)
																	{
																		tCodePart=[tCodePart stringByAppendingFormat:@"%@\tmy.result.message = system.localizedString(\'%@\');\n",tTabulationDepth,tErrorMessage];
																	}
																	
																	if (tFailureBehavior==ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_FAIL)
																	{
																		tCodePart=[tCodePart stringByAppendingFormat:@"%@\tmy.result.type = 'Fatal';\n",tTabulationDepth];
																		
																	}
																	else
																	{
																		// Incorrect Value
																		
																		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																															 ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,IC_BUILDER_FAILURE_TAG_NAME,
																															 nil];
																		
																		break;
																	}
																}
																
																tCodePart=[tCodePart stringByAppendingFormat:@"%@}\n",tTabulationDepth];
															
																tPretest=[NSString stringWithFormat:@"\n%@if (tResult==true)\n\%@{\n",tTabulationDepth,tTabulationDepth];
																
																[tFunctionCode appendString:tCodePart];
																
																if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_INSTALLATION)
																{
																	tInstallationCheckIndex=tIndex;
																	
																	tInstallationCheckTabulationDepth=tTabulationDepth;
																	
																	tInstallationCheckDepth=tDepth;
																	
																	tInstallationCheckPreTest=tPretest;
																}
																else if (tRequirementType==IC_REQUIREMENT_CHECK_TYPE_TARGET)
																{
																	tVolumeCheckIndex=tIndex;
																
																	tVolumeCheckTabulationDepth=tTabulationDepth;
																	
																	tVolumeCheckDepth=tDepth;
																	
																	tVolumeCheckPreTest=tPretest;
																}
																
																tContinue=YES;
															}
															else
															{
																// Missing Information
																
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																													 ICDOCUMENT_REQUIREMENT_MESSAGE,IC_BUILDER_FAILURE_TAG_NAME,
																													 nil];
															}
														}
														else
														{
															// Missing Information
															
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
														}
													}
													else
													{
														// Code not generated
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_REQUIREMENT_MISSING_CODE],IC_BUILDER_FAILURE_REASON,
																											 tIdentifier,IC_BUILDER_FAILURE_TAG_NAME,
																											 tErrorDictionary,IC_BUILDER_FAILURE_SECONDARY_REASON,
																											 nil];
													}
												}
												else
												{
													// Missing Information
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																										 ICDOCUMENT_REQUIREMENT_DICTIONARY,IC_BUILDER_FAILURE_TAG_NAME,
																										 nil];
												}
												
												[tRequirementConverter release];
											}
											else
											{
												// Converter not found
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_REQUIREMENT_MISSING_CONVERTER],IC_BUILDER_FAILURE_REASON,
																								     tIdentifier,IC_BUILDER_FAILURE_TAG_NAME,
																								     nil];
											}
										}
										else
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_REQUIREMENT_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
									else
									{
										tContinue=YES;
									}
								}
								else
								{
									// Missing Information
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_REQUIREMENT_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
								}
							}
						
							if (tContinue==YES)
							{
								if (tInstallationCheckDepth!=-1)
								{
									NSXMLElement * tElement;
									
									for(i=0;i<tInstallationCheckDepth;i++)
									{
										NSString * tClosingLine;
										
										tInstallationCheckTabulationDepth=[tInstallationCheckTabulationDepth substringToIndex:[tInstallationCheckTabulationDepth length]-1];
										
										tClosingLine=[NSString stringWithFormat:@"%@}\n",tInstallationCheckTabulationDepth];
										
										[tInstallationCheckFunctionCode appendString:tClosingLine];
									}
									
									[tInstallationCheckFunctionCode appendString:@"\n\t\treturn tResult;\n\t}"];
									
									[tJavascriptFunctionsArray addObject:tInstallationCheckFunctionCode];
									
									[tJavascriptFunctionsDictionary setObject:tInstallationCheckFunctionCode forKey:@"installation_check"];
									
									tElement=(NSXMLElement *) [NSXMLNode elementWithName:@"installation-check"];
									
									if (tElement!=nil)
									{
										id tAttribute;
									
										tAttribute=[NSXMLNode attributeWithName:@"script" stringValue:@"installation_check()"];
															
										if (tAttribute!=nil)
										{
											[tElement addAttribute:tAttribute];
											
											[installerScriptElement_ addChild:tElement];
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'script' attribute"];
										
											tContinue=NO;
										}
									}
									else
									{
										// Low Memory
											
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'installation-check' element"];
									
										tContinue=NO;
									}
								}
							
								if (tContinue==YES)
								{
									if (tVolumeCheckDepth!=-1)
									{
										NSXMLElement * tElement;
										
										for(i=0;i<tVolumeCheckDepth;i++)
										{
											NSString * tClosingLine;
											
											tVolumeCheckTabulationDepth=[tVolumeCheckTabulationDepth substringToIndex:[tVolumeCheckTabulationDepth length]-1];
											
											tClosingLine=[NSString stringWithFormat:@"%@}\n",tVolumeCheckTabulationDepth];
											
											[tVolumeCheckFunctionCode appendString:tClosingLine];
										}
										
										[tVolumeCheckFunctionCode appendString:@"\n\t\treturn tResult;\n\t}"];
										
										[tJavascriptFunctionsArray addObject:tVolumeCheckFunctionCode];
										
										[tJavascriptFunctionsDictionary setObject:tVolumeCheckFunctionCode forKey:@"volume_check"];
										
										tElement=(NSXMLElement *) [NSXMLNode elementWithName:@"volume-check"];
										
										if (tElement!=nil)
										{
											id tAttribute;
										
											tAttribute=[NSXMLNode attributeWithName:@"script" stringValue:@"volume_check()"];
																
											if (tAttribute!=nil)
											{
												[tElement addAttribute:tAttribute];
												
												[installerScriptElement_ addChild:tElement];
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'script' attribute"];
											
												tContinue=NO;
											}
										}
										else
										{
											// Low Memory
												
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_WARNING format:@"Low Memory: Unable to create 'volume-check' element"];
										
											tContinue=NO;
										}
									}
								
									if (tContinue==YES)
									{
										[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																					forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLATION_VOLUME_REQUIREMENTS
																				   metadata:nil];
										
										return YES;
									}
								}
							}
						}
					}
				}
			}
			else
			{
				return YES;
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_REQUIREMENTS_LIST,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
		return NO;
	}
	
	return YES;
}

#pragma mark -

- (BOOL) setTitle:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tProjectSettings;
	NSString * tProjectName;
	NSDictionary * tMetadata;
	
	// Get the project name
	
	tProjectSettings=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];

	tProjectName=[tProjectSettings objectForKey:ICDOCUMENT_PROJECT_SETTINGS_NAME];
	
	if (tProjectName!=nil)
	{
		NSDictionary * tTitleDictionary;
		NSArray * tLocalizations=nil;
		NSXMLNode * tNode=nil;
		
		tTitleDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_TITLE];

		if (tTitleDictionary!=nil)
		{
			tLocalizations=[tTitleDictionary objectForKey:ICDOCUMENT_LOCALIZATIONS];
		}
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		// title 
		
		if (tLocalizations==nil || [tLocalizations count]==0)
		{
			// Use the name of the project
			
			tNode=[NSXMLNode textWithStringValue:tProjectName];
		}
		else
		{
			// Make the title localizable
			
			NSString * tKey;
			
			tKey=@"DISTRIBUTION_TITLE";
			
			tNode=[NSXMLNode textWithStringValue:tKey];
			
			if (tNode!=nil)
			{
				NSEnumerator * tEnumerator;
				
				tEnumerator=[tLocalizations objectEnumerator];
				
				if (tEnumerator!=nil)
				{
					NSMutableDictionary * tLocalizationsDictionary=nil;
					NSDictionary * tLanguageDictionary;
					BOOL tAtLeastOneLocalizedTitle=NO;
					
					tLocalizationsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
					
					while (tLanguageDictionary=[tEnumerator nextObject])
					{
						NSString * tLanguage;
						
						tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
						
						if (tLanguage!=nil)
						{
							NSMutableDictionary * tLanguageLocalizationsDictionary;
							NSString * tLocalization;
							
							tLanguageLocalizationsDictionary=[tLocalizationsDictionary objectForKey:tLanguage];
							
							if (tLanguageLocalizationsDictionary==nil)
							{
								tLanguageLocalizationsDictionary=[NSMutableDictionary dictionary];
								
								if (tLanguageLocalizationsDictionary!=nil)
								{
									[tLocalizationsDictionary setObject:tLanguageLocalizationsDictionary forKey:tLanguage];
								}
								else
								{
									// Low Memory
									
									goto setTitleBail;
								}
							}
							
							tLocalization=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
						
							if (tLocalization!=nil)
							{
								if ([tLocalization length]>0)
								{
									[tLanguageLocalizationsDictionary setObject:tLocalization forKey:tKey];
									
									tAtLeastOneLocalizedTitle=YES;
								}
							}
							else
							{
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_VALUE,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							
								goto setTitleBail;
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							goto setTitleBail;
						}
					}
					
					if (tAtLeastOneLocalizedTitle==NO)
					{
						tNode=[NSXMLNode textWithStringValue:tProjectName];
					}
					else
					{
						// Set the default value
						
						NSMutableDictionary * tDefaultLanguageDictionary;
						
						tDefaultLanguageDictionary=[tLocalizationsDictionary objectForKey:ICLANGUAGE_DEFAULT_LANGUAGE];
								
						if (tDefaultLanguageDictionary==nil)
						{
							tDefaultLanguageDictionary=[NSMutableDictionary dictionary];
							
							if (tDefaultLanguageDictionary!=nil)
							{
								[tLocalizationsDictionary setObject:tDefaultLanguageDictionary forKey:ICLANGUAGE_DEFAULT_LANGUAGE];
							}
							else
							{
								// Low Memory
								
								tNode=nil;
							}
						}
								
						[tDefaultLanguageDictionary setObject:tProjectName forKey:tKey];
					}
				}
				else
				{
					// Low Memory
					
					tNode=nil;
				}
			}
		}
		
		if (tNode!=nil)
		{
			NSXMLElement * tTitleElement;
			
			tTitleElement=(NSXMLElement *) [NSXMLNode elementWithName:@"title"];
			
			if (tTitleElement!=nil)
			{
				[tTitleElement addChild:tNode];
			
				[installerScriptElement_ addChild:tTitleElement];
				
				return YES;
			}
			else
			{
				// Low Memory
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <title> element"];
			}
		}
	}
	else
	{
		// Missing Information
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
															 ICDOCUMENT_PROJECT_SETTINGS_NAME,IC_BUILDER_FAILURE_TAG_NAME,
															 nil];
	}

setTitleBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
														forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
													   metadata:tMetadata];

	return NO;
}

- (BOOL) setBackground:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tBackgroundDictionary;
	
	tBackgroundDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND];
	
	if (tBackgroundDictionary!=nil)
	{
		NSNumber * tNumber;
		NSDictionary * tMetadata=nil;
		
		tNumber=[tBackgroundDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_CUSTOM];
		
		if (tNumber!=nil)
		{
			if ([tNumber boolValue]==YES)
			{
				NSDictionary * tPathDictionary;
				
				tPathDictionary=[tBackgroundDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_PATH];
				
				if (tPathDictionary!=nil)
				{
					if ([tPathDictionary count]>0)
					{
						NSString * tAbsolutePath;
						
						[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																	forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_BACKGROUND_IMAGE
																   metadata:tMetadata];
						
						tAbsolutePath=[self pathFromPathDictionary:tPathDictionary];
						
						if (tAbsolutePath!=nil)
						{
							if ([fileManager_ fileExistsAtPath:tAbsolutePath]==YES)
							{
								NSXMLElement * tBackgroundElement;
								NSString * tDestinationPath;
								NSString * tResourcesPath;
								NSString * tFileName;
								NSString * tBaseName;
								int tIndex;
								
								// Copy the Image to the appropriate location
								
								tResourcesPath=[self getDistributionResourcesPath];
									
								if (tResourcesPath!=nil)
								{
									tFileName=tBaseName=@"background";
									
									tDestinationPath=[tResourcesPath stringByAppendingPathComponent:tFileName];
									
									tIndex=0;
									
									while ([fileManager_ fileExistsAtPath:tDestinationPath]==YES && tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										tFileName=[NSString stringWithFormat:@"%@_%d",tBaseName,tIndex];
										
										tDestinationPath=[tResourcesPath stringByAppendingPathComponent:tFileName];
									
										tIndex++;
									}
									
									if (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										if ([self copyObjectAtPath:tAbsolutePath toPath:tDestinationPath]==NO)
										{
											return NO;
										}
										
										if ([ICFileSystemUtilities setFilePrivileges:fileAttributes_ atPath:tDestinationPath]==NO)
										{
											return NO;
										}
									
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
										
										tBackgroundElement=(NSXMLElement *) [NSXMLNode elementWithName:@"background"];
									
										if (tBackgroundElement!=nil)
										{
											id tAttribute;
														
											// file [Required]
											
											tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFileName];
														
											if (tAttribute!=nil)
											{
												[tBackgroundElement addAttribute:tAttribute];
											
												// scaling [optional]
											
												tNumber=[tBackgroundDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_SCALING];
							
												if (tNumber!=nil)
												{
													int tScaling=-1;
													
													tScaling=[tNumber intValue];
													
													if (tScaling==ICSCALING_PROPORTIONALLY)
													{
														tAttribute=[NSXMLNode attributeWithName:@"scaling" stringValue:@"proportional"];
													}
													else if (tScaling==ICSCALING_TO_FIT)
													{
														tAttribute=[NSXMLNode attributeWithName:@"scaling" stringValue:@"tofit"];
													}
													else if (tScaling==ICSCALING_NONE)
													{
														tAttribute=[NSXMLNode attributeWithName:@"scaling" stringValue:@"none"];
													}
													else 
													{
														tScaling=-1;
														
														// Incorrect Value
														
														[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_SCALING",tScaling];
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_SCALING,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
													}
													
													if (tScaling!=-1)
													{
														if (tAttribute!=nil)
														{
															[tBackgroundElement addAttribute:tAttribute];
															
															// alignment [optional]
													
															tNumber=[tBackgroundDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_ALIGNMENT];
									
															if (tNumber!=nil)
															{
																int tAlignment;
																NSString * tAlignmentString=@"";
																
																tAlignment=[tNumber intValue];
																
																switch(tAlignment)
																{
																	case ICALIGNMENT_CENTER:
																	
																		tAlignmentString=@"center";
																		
																		break;
																	
																	case ICALIGNMENT_TOP:
																
																		tAlignmentString=@"top";
																		
																		break;
																		
																	case ICALIGNMENT_TOP_LEFT:
																
																		tAlignmentString=@"topleft";
																		
																		break;
																		
																	case ICALIGNMENT_TOP_RIGHT:
																
																		tAlignmentString=@"topright";
																		
																		break;
																		
																	case ICALIGNMENT_LEFT:
																
																		tAlignmentString=@"left";
																		
																		break;
																		
																	case ICALIGNMENT_BOTTOM:
																
																		tAlignmentString=@"bottom";
																		
																		break;
																		
																	case ICALIGNMENT_BOTTOM_LEFT:
																
																		tAlignmentString=@"bottomleft";
																		
																		break;
																		
																	case ICALIGNMENT_BOTTOM_RIGHT:
																
																		tAlignmentString=@"bottomright";
																		
																		break;
																	
																	case ICALIGNMENT_RIGHT:
																	
																		tAlignmentString=@"right";
																		
																		break;
																	
																	default:
																	
																		// Incorrect Value
																		
																		[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_ALIGNMENT",tAlignment];
																		
																		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																															 ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_ALIGNMENT,IC_BUILDER_FAILURE_TAG_NAME,
																															 nil];
																		
																		break;
																}
																
																if ([tAlignmentString length]>0)
																{
																	tAttribute=[NSXMLNode attributeWithName:@"alignment" stringValue:tAlignmentString];
																	
																	if (tAttribute!=nil)
																	{
																		[tBackgroundElement addAttribute:tAttribute];
																		
																		[installerScriptElement_ addChild:tBackgroundElement];
																	
																		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																													forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_BACKGROUND_IMAGE
																												   metadata:tMetadata];
																		
																		return YES;
																	}
																	else
																	{
																		// Low Memory
																		
																		[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'alignment' attribute"];
																	}
																}
															}
															else
															{
																// // Missing Information
																
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																													 ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_ALIGNMENT,IC_BUILDER_FAILURE_TAG_NAME,
																													 nil];
															}
														}
														else
														{
															// Low Memory
															
															[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'scaling' attribute"];
														}
													}
												}
												else
												{
													// // Missing Information
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																										 ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_SCALING,IC_BUILDER_FAILURE_TAG_NAME,
																										 nil];
												}
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
											}
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <background> element"];
										}
									}
								}
							}
							else
							{
								// File does not exist
								
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																					 tAbsolutePath,IC_BUILDER_INFORMATION_FILE_PATH,
																					 nil];
							}
						}
					}
					else
					{
						return YES;
					}
				}
				else
				{
					// Missing Information
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				return YES;
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_PRESENTATION_BACKGROUND_CUSTOM,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
		
		return NO;
	}
	
	return YES;
}

- (BOOL) setIntroduction:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tWelcomeDictionary;
	NSDictionary * tMetadata=nil;
	
	tWelcomeDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INTRODUCTION];
	
	if (tWelcomeDictionary!=nil)
	{
		NSArray * tLocalizationsArray;
		BOOL tUsesCustomIntroduction=NO;
		NSString * tFileName;
		
		tLocalizationsArray=[tWelcomeDictionary objectForKey:ICDOCUMENT_LOCALIZATIONS];
		
		if (tLocalizationsArray!=nil)
		{
			BOOL tContinue=YES;
			
			if ([tLocalizationsArray count]>0)
			{
				NSEnumerator * tEnumerator;
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_WELCOME_MESSAGE
														   metadata:nil];
				
				tEnumerator=[tLocalizationsArray objectEnumerator];
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				
				if (tEnumerator!=nil)
				{
					NSDictionary * tLanguageDictionary;
					NSMutableDictionary * tLocalizedPathDictionary;
					
					tLocalizedPathDictionary=[NSMutableDictionary dictionaryWithCapacity:[tLocalizationsArray count]];
					
					if (tLocalizedPathDictionary!=nil)
					{
						while (tContinue==YES && (tLanguageDictionary=[tEnumerator nextObject]))
						{
							NSDictionary * tPathDictionary;
							
							tPathDictionary=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
							
							if (tPathDictionary!=nil)
							{
								NSString * tPath;
								
								tPath=[tPathDictionary objectForKey:ICDOCUMENT_PATH];
								
								if (tPath!=nil && [tPath length]>0)
								{
									int tPathType;
									
									tContinue=NO;
									
									tPathType=[[tPathDictionary objectForKey:ICDOCUMENT_PATH_TYPE] intValue];
									
									if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
									}
									else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
									}
									
									// Check that the file exists
									
									if ([fileManager_ fileExistsAtPath:tPath]==YES)
									{
										NSString * tLanguage;
										
										tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
										
										if (tLanguage!=nil)
										{
											if ([tLanguage length]>0)
											{
												// Copy the file to the appropriate location and rename it to DISTRIBUTION_WELCOME
												
												NSString * tLanguagePath;
												
												tLanguagePath=[self getDistributionPathForLanguage:tLanguage];
												
												if (tLanguagePath!=nil)
												{
													[tLocalizedPathDictionary setObject:tPath forKey:tLanguagePath];
													
													tUsesCustomIntroduction=YES;
													
													tContinue=YES;
												}
												else
												{
													tMetadata=nil;
												}
											}
											else
											{
												// Incorrect Value
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
										}
										else
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
									else
									{
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"File not found: '%@'",tPath];
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																							 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																							 nil];
									}
								}
							}
						}
						
						if (tContinue==YES)
						{
							if (tUsesCustomIntroduction==YES)
							{
								NSArray * tAllLocalizedPaths;
								
								tAllLocalizedPaths=[tLocalizedPathDictionary allKeys];
								
								tContinue=NO;
								
								if (tAllLocalizedPaths!=nil && [tAllLocalizedPaths count]>0)
								{
									NSString * tBaseName;
									NSUInteger tIndex;
									NSUInteger tCount;
									NSString * tExtension;
									
									tExtension=[[tLocalizedPathDictionary objectForKey:[tAllLocalizedPaths objectAtIndex:0]] pathExtension];
									
									tIndex=0;
									
									tBaseName=@"welcome";
									
									if (tExtension!=nil && [tExtension length]>0)
									{
										tFileName=[tBaseName stringByAppendingPathExtension:tExtension];
									}
									else
									{
										tFileName=tBaseName;
									}
									
									tCount=[tAllLocalizedPaths count];
									
									while (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										NSUInteger i;
										NSString * tLocalizedPath;
										NSString * tDestinationPath;
											
										for(i=0;i<tCount;i++)
										{
											tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
											
											tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
											
											if ([fileManager_ fileExistsAtPath:tDestinationPath]==YES)
											{
												break;
											}
										}
										
										if (i==tCount)
										{
											for(i=0;i<tCount;i++)
											{
												tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
												
												tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
												
												if ([self copyObjectAtPath:[tLocalizedPathDictionary objectForKey:tLocalizedPath] toPath:tDestinationPath]==NO)
												{
													return NO;
												}
													
												if ([ICFileSystemUtilities setFilePrivileges:fileAttributes_ atPath:tDestinationPath]==NO)
												{
													return NO;
												}
											}
											
											break;
										}
										else
										{
											tIndex++;
											
											if (tExtension!=nil && [tExtension length]>0)
											{
												tFileName=[[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex] stringByAppendingPathExtension:tExtension];
											}
											else
											{
												tFileName=[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex];
											}
										}
									}
									
									if (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										tContinue=YES;
									}
									else
									{
										tMetadata=nil;
									}
								}
							
						
								if (tContinue==YES)
								{
									NSXMLElement * tWelcomeElement;
									
									tContinue=NO;
									
									// welcome
									
									tWelcomeElement=(NSXMLElement *) [NSXMLNode elementWithName:@"welcome"];
													
									if (tWelcomeElement!=nil)
									{
										id tAttribute;
													
										// file
										
										tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFileName];
													
										if (tAttribute!=nil)
										{
											[tWelcomeElement addAttribute:tAttribute];
											
											[installerScriptElement_ addChild:tWelcomeElement];
											
											tContinue=YES;
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
											
										}
									}
									else
									{
										// Low Memory
									
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <welcome> element"];
									}
								}
							}
						}
					}
					else
					{
						// Low Memory
						
						tContinue=NO;
					}
				}
				else
				{
					// Low Memory
					
					tContinue=NO;
				}
				
				if (tContinue==NO)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
															   metadata:tMetadata];
	
					return NO;
				}
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_WELCOME_MESSAGE
														   metadata:nil];
			}
		}
	}

	return YES;
}

- (BOOL) setReadMe:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tReadmeDictionary;
	NSDictionary * tMetadata=nil;
										
	tReadmeDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_README];
	
	if (tReadmeDictionary!=nil)
	{
		NSArray * tLocalizationsArray;
		BOOL tUsesCustomReadMe=NO;
		NSString * tFileName;
	
		tLocalizationsArray=[tReadmeDictionary objectForKey:ICDOCUMENT_LOCALIZATIONS];
		
		if (tLocalizationsArray!=nil)
		{
			BOOL tContinue=YES;
			
			if ([tLocalizationsArray count]>0)
			{
				NSEnumerator * tEnumerator;
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_README_MESSAGE
														   metadata:nil];
				
				tEnumerator=[tLocalizationsArray objectEnumerator];
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				
				if (tEnumerator!=nil)
				{
					NSDictionary * tLanguageDictionary;
					NSMutableDictionary * tLocalizedPathDictionary;
					
					tLocalizedPathDictionary=[NSMutableDictionary dictionaryWithCapacity:[tLocalizationsArray count]];
					
					if (tLocalizedPathDictionary!=nil)
					{
						while (tContinue==YES && (tLanguageDictionary=[tEnumerator nextObject]))
						{
							NSDictionary * tPathDictionary;
							
							tPathDictionary=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
							
							if (tPathDictionary!=nil)
							{
								NSString * tPath;
								
								tPath=[tPathDictionary objectForKey:ICDOCUMENT_PATH];
								
								if (tPath!=nil && [tPath length]>0)
								{
									int tPathType;
									
									tContinue=NO;
									
									tPathType=[[tPathDictionary objectForKey:ICDOCUMENT_PATH_TYPE] intValue];
									
									if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
									}
									else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
									}
									
									// Check that the file exists
									
									if ([fileManager_ fileExistsAtPath:tPath]==YES)
									{
										NSString * tLanguage;
										
										tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
										
										if (tLanguage!=nil)
										{
											if ([tLanguage length]>0)
											{
												// Copy the file to the appropriate location and rename it to DISTRIBUTION_WELCOME
												
												NSString * tLanguagePath;
												
												tLanguagePath=[self getDistributionPathForLanguage:tLanguage];
												
												if (tLanguagePath!=nil)
												{
													[tLocalizedPathDictionary setObject:tPath forKey:tLanguagePath];
													
													tUsesCustomReadMe=YES;
													
													tContinue=YES;
												}
												else
												{
													tMetadata=nil;
												}
											}
											else
											{
												// Incorrect Value
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
										}
										else
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
									else
									{
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"File not found: '%@'",tPath];
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																							 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																							 nil];
									}
								}
							}
						}
						
						if (tContinue==YES)
						{
							if (tUsesCustomReadMe==YES)
							{
								NSArray * tAllLocalizedPaths;
								
								tAllLocalizedPaths=[tLocalizedPathDictionary allKeys];
								
								tContinue=NO;
								
								if (tAllLocalizedPaths!=nil && [tAllLocalizedPaths count]>0)
								{
									NSString * tBaseName;
									NSUInteger tIndex;
									NSUInteger tCount;
									NSString * tExtension;
									
									tExtension=[[tLocalizedPathDictionary objectForKey:[tAllLocalizedPaths objectAtIndex:0]] pathExtension];
									
									tIndex=0;
									
									tBaseName=@"readme";
									
									if (tExtension!=nil && [tExtension length]>0)
									{
										tFileName=[tBaseName stringByAppendingPathExtension:tExtension];
									}
									else
									{
										tFileName=tBaseName;
									}
									
									tCount=[tAllLocalizedPaths count];
									
									while (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										NSUInteger i;
										NSString * tLocalizedPath;
										NSString * tDestinationPath;
											
										for(i=0;i<tCount;i++)
										{
											tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
											
											tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
											
											if ([fileManager_ fileExistsAtPath:tDestinationPath]==YES)
											{
												break;
											}
										}
										
										if (i==tCount)
										{
											for(i=0;i<tCount;i++)
											{
												tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
												
												tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
												
												if ([self copyObjectAtPath:[tLocalizedPathDictionary objectForKey:tLocalizedPath] toPath:tDestinationPath]==NO)
												{
													return NO;
												}
													
												if ([ICFileSystemUtilities setFilePrivileges:fileAttributes_ atPath:tDestinationPath]==NO)
												{
													return NO;
												}

											}
											
											break;
										}
										else
										{
											tIndex++;
											
											if (tExtension!=nil && [tExtension length]>0)
											{
												tFileName=[[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex] stringByAppendingPathExtension:tExtension];
											}
											else
											{
												tFileName=[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex];
											}
										}
									}
									
									if (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										tContinue=YES;
									}
									else
									{
										tMetadata=nil;
									}
								}
							
						
								if (tContinue==YES)
								{
									NSXMLElement * tReadMeElement;
									
									tContinue=NO;
									
									// readme
									
									tReadMeElement=(NSXMLElement *) [NSXMLNode elementWithName:@"readme"];
													
									if (tReadMeElement!=nil)
									{
										id tAttribute;
													
										// file
										
										tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFileName];
													
										if (tAttribute!=nil)
										{
											[tReadMeElement addAttribute:tAttribute];
											
											[installerScriptElement_ addChild:tReadMeElement];
											
											tContinue=YES;
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
											
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <readme> element"];
										
									}
								}
							}
						}
					}
					else
					{
						// Low Memory
						
						tContinue=NO;
					}
				}
				else
				{
					// Low Memory
					
					tContinue=NO;
				}
				
				if (tContinue==NO)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
																forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
															   metadata:tMetadata];
	
					return NO;
				}
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_README_MESSAGE
														   metadata:nil];
			}
		}
	}

	return YES;
}

- (BOOL) setLicense:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tLicenseDictionary;
	NSDictionary * tMetadata=nil;
										
	tLicenseDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_LICENSE];
	
	if (tLicenseDictionary!=nil)
	{
		BOOL tUsesCustomLicense=NO;
		NSString * tFileName;
		NSNumber * tNumber;
		int tLicenseMode=ICLICENSE_MODE_CUSTOM;
		NSArray * tLocalizationsArray=nil;
		NSString * tSlaString=nil;
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_LICENSE_DOCUMENT
												   metadata:nil];
		
		tNumber=[tLicenseDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_LICENSE_MODE];
		
		if (tNumber!=nil)
		{
			tLicenseMode=[tNumber intValue];
			
			if (tLicenseMode==ICLICENSE_MODE_TEMPLATE)
			{
				NSString * tLicenseTemplateName;
				
				tLicenseTemplateName=[tLicenseDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_LICENSE_TEMPLATE_NAME];
				
				if (tLicenseTemplateName!=nil)
				{
					tSlaString=[[ICLicenseProvider defaultProvider] slaForLicenseWithName:tLicenseTemplateName];
					
					tLocalizationsArray=[[ICLicenseProvider defaultProvider] localizationsForLicenseWithName:tLicenseTemplateName];
					
					if (tLocalizationsArray==nil || [tLocalizationsArray count]==0)
					{
						tLocalizationsArray=nil;
						
						// Missing Templates
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_LICENSE_MISSING_TEMPLATE],IC_BUILDER_FAILURE_REASON,
																			 tLicenseTemplateName,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				}
			}
			else if (tLicenseMode==ICLICENSE_MODE_CUSTOM)
			{
				tLocalizationsArray=[tLicenseDictionary objectForKey:ICDOCUMENT_LOCALIZATIONS];
				
				if (tLocalizationsArray==nil)
				{
					// Missing Information
			
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_LOCALIZATIONS,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				// Incorrect Value
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_LICENSE_MODE",tLicenseMode];
									
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PROJECT_PRESENTATION_LICENSE_MODE,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		
			if (tLocalizationsArray!=nil)
			{
				BOOL tContinue=YES;
				
				if ([tLocalizationsArray count]>0)
				{
					NSEnumerator * tEnumerator;
						
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					
					tEnumerator=[tLocalizationsArray objectEnumerator];
						
					if (tEnumerator!=nil)
					{
						NSDictionary * tLanguageDictionary;
						NSMutableDictionary * tLocalizedPathDictionary;
						
						tLocalizedPathDictionary=[NSMutableDictionary dictionaryWithCapacity:[tLocalizationsArray count]];
						
						if (tLocalizedPathDictionary!=nil)
						{
							while (tContinue==YES && (tLanguageDictionary=[tEnumerator nextObject]))
							{
								NSDictionary * tPathDictionary;
								
								tPathDictionary=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
								
								if (tPathDictionary!=nil)
								{
									NSString * tPath;
									
									tPath=[tPathDictionary objectForKey:ICDOCUMENT_PATH];
									
									if (tPath!=nil && [tPath length]>0)
									{
										int tPathType;
										
										tContinue=NO;
										
										tPathType=[[tPathDictionary objectForKey:ICDOCUMENT_PATH_TYPE] intValue];
										
										if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
										{
											tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
										}
										else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
										{
											tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
										}
										
										// Check that the file exists
										
										if ([fileManager_ fileExistsAtPath:tPath]==YES)
										{
											NSString * tLanguage;
											
											tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
											
											if (tLanguage!=nil)
											{
												if ([tLanguage length]>0)
												{
													// Copy the file to the appropriate location and rename it to DISTRIBUTION_WELCOME
													
													NSString * tLanguagePath;
													
													tLanguagePath=[self getDistributionPathForLanguage:tLanguage];
													
													if (tLanguagePath!=nil)
													{
														[tLocalizedPathDictionary setObject:tPath forKey:tLanguagePath];
														
														tUsesCustomLicense=YES;
														
														tContinue=YES;
													}
													else
													{
														tMetadata=nil;
													}
												}
												else
												{
													// Incorrect Value
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																										 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																										 nil];
												}
											}
											else
											{
												// Missing Information
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
										}
										else
										{
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"File not found: '%@'",tPath];
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																								 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																								 nil];
										}
									}
								}
							}
							
							if (tContinue==YES)
							{
								if (tUsesCustomLicense==YES)
								{
									NSArray * tAllLocalizedPaths;
									
									tAllLocalizedPaths=[tLocalizedPathDictionary allKeys];
									
									tContinue=NO;
									
									if (tAllLocalizedPaths!=nil && [tAllLocalizedPaths count]>0)
									{
										NSString * tBaseName;
										NSUInteger tIndex;
										NSUInteger tCount;
										NSString * tExtension;
									
										tExtension=[[tLocalizedPathDictionary objectForKey:[tAllLocalizedPaths objectAtIndex:0]] pathExtension];
									
										tIndex=0;
										
										tContinue=YES;
										
										tBaseName=@"license";
										
										if (tExtension!=nil && [tExtension length]>0)
										{
											tFileName=[tBaseName stringByAppendingPathExtension:tExtension];
										}
										else
										{
											tFileName=tBaseName;
										}
										
										tCount=[tAllLocalizedPaths count];
										
										while (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
										{
											NSUInteger i;
											NSString * tLocalizedPath;
											NSString * tDestinationPath;
												
											for(i=0;i<tCount;i++)
											{
												tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
												
												tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
												
												if ([fileManager_ fileExistsAtPath:tDestinationPath]==YES)
												{
													break;
												}
											}
											
											if (i==tCount)
											{
												if (tLicenseMode==ICLICENSE_MODE_TEMPLATE)
												{
													NSDictionary * tKeywordsDictionary;
													
													tKeywordsDictionary=[tLicenseDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_LICENSE_TEMPLATE_KEYWORDS];
													
													for(i=0;i<tCount;i++)
													{
														NSMutableString * tMutableString;
														
														NSStringEncoding tEncoding;
														NSError * tError;
														
														tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
														
														tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
														
														tMutableString=[[NSMutableString alloc] initWithContentsOfFile:[tLocalizedPathDictionary objectForKey:tLocalizedPath]
																										  usedEncoding:&tEncoding
																												 error:&tError];
														if (tMutableString==nil)
														{
															int tEncodingIndex;
															NSStringEncoding tPotentialEncodings[4]={
																								NSMacOSRomanStringEncoding,
																								NSUTF8StringEncoding,
																								NSISOLatin1StringEncoding,
																								NSASCIIStringEncoding
																								};
																								
															for(tEncodingIndex=0;tEncodingIndex<4;tEncodingIndex++)
															{
																tEncoding=tPotentialEncodings[tEncodingIndex];
																
																tMutableString=[[NSMutableString alloc] initWithContentsOfFile:[tLocalizedPathDictionary objectForKey:tLocalizedPath]
																													  encoding:tEncoding
																														 error:&tError];
																if (tMutableString!=nil)
																{
																	break;
																}
															}
														}
														
														if (tMutableString!=nil)
														{
															if (tKeywordsDictionary!=nil)
															{
																[ICLicenseProvider replaceKeywords:tKeywordsDictionary inString:tMutableString];
															}

															if ([tMutableString writeToFile:tDestinationPath atomically:NO encoding:tEncoding error:&tError]==NO)
															{
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																													 tDestinationPath,IC_BUILDER_INFORMATION_FILE_PATH,
																													 [tError localizedDescription],IC_BUILDER_INFORMATION_SYSTEM_ERROR_DESCRIPTION,
																													 nil];
																
																tContinue=NO;
																
																break;
															}
															
															[tMutableString release];
															
															if ([ICFileSystemUtilities setFilePrivileges:fileAttributes_ atPath:tDestinationPath]==NO)
															{
																return NO;
															}
														}
														else
														{
															// Low Memory?
															
															tContinue=NO;
																
															break;
														}
													}
												}
												else if (tLicenseMode==ICLICENSE_MODE_CUSTOM)
												{
													for(i=0;i<tCount;i++)
													{
														tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
														
														tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
														
														if ([self copyObjectAtPath:[tLocalizedPathDictionary objectForKey:tLocalizedPath] toPath:tDestinationPath]==NO)
														{
															return NO;
														}
															
														if ([ICFileSystemUtilities setFilePrivileges:fileAttributes_ atPath:tDestinationPath]==NO)
														{
															return NO;
														}
													}
												}
													
												break;
											}
											else
											{
												tIndex++;
												
												if (tExtension!=nil && [tExtension length]>0)
												{
													tFileName=[[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex] stringByAppendingPathExtension:tExtension];
												}
												else
												{
													tFileName=[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex];
												}
											}
										}
										
										if (tContinue==YES)
										{
											if (tIndex>=MAX_RESOURCE_NAMING_ATTEMPTS)
											{
												tMetadata=nil;
											
												tContinue=NO;
											}
										}
									}
								
							
									if (tContinue==YES)
									{
										NSXMLElement * tLicenseElement;
										
										tContinue=NO;
										
										// license
										
										tLicenseElement=(NSXMLElement *) [NSXMLNode elementWithName:@"license"];
														
										if (tLicenseElement!=nil)
										{
											id tAttribute;
														
											// file
											
											tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFileName];
														
											if (tAttribute!=nil)
											{
												[tLicenseElement addAttribute:tAttribute];
												
												[installerScriptElement_ addChild:tLicenseElement];
												
												tContinue=YES;
												
												// sla
											
												if (tSlaString!=nil)
												{
													tAttribute=[NSXMLNode attributeWithName:@"sla" stringValue:tSlaString];
															
													if (tAttribute!=nil)
													{
														[tLicenseElement addAttribute:tAttribute];
													}
													else
													{
														// Low Memory
														
														[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'sla' attribute"];
														
														tContinue=NO;
													}
												}
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
											}
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <license> element"];
										}
									}
								}
							}
						}
						else
						{
							// Low Memory
							
							tContinue=NO;
						}
					}
					else
					{
						// Low Memory
						
						tContinue=NO;
					}
				}
				
				if (tContinue==YES)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_LICENSE_DOCUMENT
															   metadata:nil];
													   
					return YES;
				}
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_PRESENTATION_LICENSE_MODE,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	else
	{
		return YES;
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) setConclusion:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tSummaryDictionary;
	NSDictionary * tMetadata=nil;
	
	tSummaryDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_SUMMARY];
	
	if (tSummaryDictionary!=nil)
	{
		NSArray * tLocalizationsArray;
		BOOL tUsesCustomSummary=NO;
		NSString * tFileName;

		tLocalizationsArray=[tSummaryDictionary objectForKey:ICDOCUMENT_LOCALIZATIONS];
		
		if (tLocalizationsArray!=nil)
		{
			BOOL tContinue=YES;
			
			if ([tLocalizationsArray count]>0)
			{
				NSEnumerator * tEnumerator;
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_CONCLUSION_MESSAGE
														   metadata:nil];
												   
				tEnumerator=[tLocalizationsArray objectEnumerator];
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				
				if (tEnumerator!=nil)
				{
					NSDictionary * tLanguageDictionary;
					NSMutableDictionary * tLocalizedPathDictionary;
					
					tLocalizedPathDictionary=[NSMutableDictionary dictionaryWithCapacity:[tLocalizationsArray count]];
					
					if (tLocalizedPathDictionary!=nil)
					{
						while (tContinue==YES && (tLanguageDictionary=[tEnumerator nextObject]))
						{
							NSDictionary * tPathDictionary;
							
							tPathDictionary=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
							
							if (tPathDictionary!=nil)
							{
								NSString * tPath;
								
								tPath=[tPathDictionary objectForKey:ICDOCUMENT_PATH];
								
								if (tPath!=nil && [tPath length]>0)
								{
									int tPathType;
									
									tContinue=NO;
									
									tPathType=[[tPathDictionary objectForKey:ICDOCUMENT_PATH_TYPE] intValue];
									
									if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
									}
									else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
									}
									
									// Check that the file exists
									
									if ([fileManager_ fileExistsAtPath:tPath]==YES)
									{
										NSString * tLanguage;
										
										tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
										
										if (tLanguage!=nil)
										{
											if ([tLanguage length]>0)
											{
												// Copy the file to the appropriate location and rename it to DISTRIBUTION_WELCOME
												
												NSString * tLanguagePath;
												
												tLanguagePath=[self getDistributionPathForLanguage:tLanguage];
												
												if (tLanguagePath!=nil)
												{
													[tLocalizedPathDictionary setObject:tPath forKey:tLanguagePath];
													
													tUsesCustomSummary=YES;
													
													tContinue=YES;
												}
												else
												{
													tMetadata=nil;
												}
											}
											else
											{
												// Incorrect Value
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
										}
										else
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
									else
									{
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"File not found: '%@'",tPath];
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																							 tPath,IC_BUILDER_INFORMATION_FILE_PATH,
																							 nil];
									}
								}
							}
						}
						
						if (tContinue==YES)
						{
							if (tUsesCustomSummary==YES)
							{
								NSArray * tAllLocalizedPaths;
								
								tAllLocalizedPaths=[tLocalizedPathDictionary allKeys];
								
								tContinue=NO;
								
								if (tAllLocalizedPaths!=nil && [tAllLocalizedPaths count]>0)
								{
									NSString * tBaseName;
									NSUInteger tIndex;
									NSUInteger tCount;
									NSString * tExtension;
									
									tExtension=[[tLocalizedPathDictionary objectForKey:[tAllLocalizedPaths objectAtIndex:0]] pathExtension];
									
									tIndex=0;
									
									tBaseName=@"conclusion";
									
									if (tExtension!=nil && [tExtension length]>0)
									{
										tFileName=[tBaseName stringByAppendingPathExtension:tExtension];
									}
									else
									{
										tFileName=tBaseName;
									}

									
									tCount=[tAllLocalizedPaths count];
									
									while (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										NSUInteger i;
										NSString * tLocalizedPath;
										NSString * tDestinationPath;
											
										for(i=0;i<tCount;i++)
										{
											tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
											
											tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
											
											if ([fileManager_ fileExistsAtPath:tDestinationPath]==YES)
											{
												break;
											}
										}
										
										if (i==tCount)
										{
											for(i=0;i<tCount;i++)
											{
												tLocalizedPath=[tAllLocalizedPaths objectAtIndex:i];
												
												tDestinationPath=[tLocalizedPath stringByAppendingPathComponent:tFileName];
												
												if ([self copyObjectAtPath:[tLocalizedPathDictionary objectForKey:tLocalizedPath] toPath:tDestinationPath]==NO)
												{
													return NO;
												}
													
												if ([ICFileSystemUtilities setFilePrivileges:fileAttributes_ atPath:tDestinationPath]==NO)
												{
													return NO;
												}

											}
											
											break;
										}
										else
										{
											tIndex++;
											
											if (tExtension!=nil && [tExtension length]>0)
											{
												tFileName=[[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex] stringByAppendingPathExtension:tExtension];
											}
											else
											{
												tFileName=[NSString stringWithFormat:@"%@_%lu",tBaseName,tIndex];
											}
										}
									}
									
									if (tIndex<MAX_RESOURCE_NAMING_ATTEMPTS)
									{
										tContinue=YES;
									}
									else
									{
										tMetadata=nil;
									}
								}
								
								if (tContinue==YES)
								{
									NSXMLElement * tConclusionElement;
		
									tContinue=NO;
									
									// conclusion
									
									tConclusionElement=(NSXMLElement *) [NSXMLNode elementWithName:@"conclusion"];
													
									if (tConclusionElement!=nil)
									{
										id tAttribute;
													
										// conclusion
										
										tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFileName];
													
										if (tAttribute!=nil)
										{
											[tConclusionElement addAttribute:tAttribute];
											
											[installerScriptElement_ addChild:tConclusionElement];
											
											tContinue=YES;
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
											
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <readme> element"];
										
									}
								}
							}
						}
					}
					else
					{
						// Low Memory
						
						tContinue=NO;
					}
				}
				else
				{
					// Low Memory
					
					tContinue=NO;
				}
				
				if (tContinue==NO)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
				
					return NO;
				}							   
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_CONCLUSION_MESSAGE
														   metadata:nil];
			}
		}
	}
	
	return YES;
}

- (BOOL) setChoiceOutline:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tMetadata=nil;
	
	if (inPresentationDictionary!=nil)
	{
		NSDictionary * tInstallationTypeDictionary;
		NSDictionary * tHierarchiesListDictionary;
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_CHOICES_HIERARCHIES
												   metadata:nil];
		
		tInstallationTypeDictionary=[inPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE];
		
		tHierarchiesListDictionary=[tInstallationTypeDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHIES];
		
		if (tHierarchiesListDictionary!=nil)
		{
			NSEnumerator * tKeyEnumerator;
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			
			tKeyEnumerator=[tHierarchiesListDictionary keyEnumerator];
			
			if (tKeyEnumerator!=nil)
			{
				NSString * tKey;
				NSMutableDictionary * tChoicesNamesDictionary;
					
				tChoicesNamesDictionary=[NSMutableDictionary dictionary];
			
				if (tChoicesNamesDictionary!=nil)
				{
					[buildInformation_ setObject:tChoicesNamesDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES];
					
					while (tKey=[tKeyEnumerator nextObject])
					{
						NSDictionary * tHierarchyDictionary;
						
						tHierarchyDictionary=[tHierarchiesListDictionary objectForKey:tKey];
						
						if (tHierarchyDictionary!=nil)
						{
							NSArray * tHierarchyArray;
							
							tHierarchyArray=[tHierarchyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHIES_LIST];
							
							if (tHierarchyArray!=nil)
							{
								NSXMLElement * tChoicesOutlineElement;
			
								// choices-outline
								
								tChoicesOutlineElement=(NSXMLElement *) [NSXMLNode elementWithName:@"choices-outline"];
												
								if (tChoicesOutlineElement!=nil)
								{
									BOOL isInstaller=YES;
									id tComment;
									NSString * tPrefix;
									
									// ui
									
									if ([tKey isEqualToString:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHY_INSTALLER]==NO)
									{
										id tAttribute=nil;
										
										if ([tKey isEqualToString:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHY_SOFTWAREUPDATE]==YES)
										{
											tAttribute=[NSXMLNode attributeWithName:@"ui" stringValue:@"SoftwareUpdate"];
											
											tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n        |      Software Update     |\n        +==========================+"];
											
											tPrefix=@"software_update_choice_";
										}
										else if ([tKey isEqualToString:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHY_INVISIBLE]==YES)
										{
											tAttribute=[NSXMLNode attributeWithName:@"ui" stringValue:@"Invisible"];
											
											tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n        |         Invisible        |\n        +==========================+"];
											
											tPrefix=@"invisible_choice_";
										}
										else
										{
											// Incorrect Value
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHIES,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
											
											break;
										}
									
										isInstaller=NO;
									
										if (tAttribute!=nil)
										{
											[tChoicesOutlineElement addAttribute:tAttribute];
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'ui' attribute"];
											
											break;
										}
									}
									else
									{
										tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n        |         Installer        |\n        +==========================+"];
										
										tPrefix=@"installer_choice_";
									}
									
									if (tComment!=nil)
									{
										[installerScriptElement_ addChild:tComment];
									}
									
									if ([self addLinesElementsToElement:tChoicesOutlineElement withChoicesArray:tHierarchyArray isInstaller:isInstaller prefix:tPrefix]==YES)
									{
										[installerScriptElement_ addChild:tChoicesOutlineElement];
									
										if ([self addChoicesElementsToElement:installerScriptElement_ withChoicesArray:tHierarchyArray isInstaller:isInstaller]==YES)
										{
											continue;
										}
									}
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <choices-outline> element"];
								}
								
								break;
							}
							else
							{
								// Missing Information
				
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHIES_LIST,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							}
							
							break;
						}
					}
					
					if (tKey==nil)
					{
						[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																	forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_CHOICES_HIERARCHIES
																   metadata:nil];
											   
						return YES;
					}
				}
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_HIERARCHIES,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) addPackageRefeferenceElementToChoiceElement:(NSXMLElement *) inElement fromDictionary:(NSDictionary *) inChoiceDictionary
{
	NSDictionary * tMetadata=nil;
	
	if (inElement!=nil && inChoiceDictionary!=nil)
	{
		NSString * tPackageUUID;
									
		tPackageUUID=[inChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_PACKAGE_UUID];

		if (tPackageUUID==nil)
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_PACKAGE_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
		else
		{
			NSXMLElement * tPackageRefElement;
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			
			// Add the pkg-ref
		
			tPackageRefElement=(NSXMLElement *) [NSXMLNode elementWithName:@"pkg-ref"];

			if (tPackageRefElement!=nil)
			{
				id tAttribute;
				NSMutableDictionary * tPackagesDictionary;
				NSString * tPackageReferenceID=nil;
				NSDictionary * tPackageBuildInformation;
				
				tPackagesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES];
				
				tPackageBuildInformation=[tPackagesDictionary objectForKey:tPackageUUID];
					
				if (tPackageBuildInformation!=nil)
				{
					tPackageReferenceID=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID];
				}
				
				if (tPackageReferenceID!=nil)
				{
					tAttribute=[NSXMLNode attributeWithName:@"id" stringValue:tPackageReferenceID];
				
					if (tAttribute!=nil)
					{
						[tPackageRefElement addAttribute:tAttribute];
						
						[inElement addChild:tPackageRefElement];
						
						return YES;
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'id' attribute"];
					}
				}
				else
				{
					// Missing Build Data
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																		 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				// Low Memory
			
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <pkg-ref> element"];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
	
	return NO;
}

- (int) hasOneHideAndUnselectRequirementEnabledInArray:(NSArray *) inArray
{
	if (inArray!=nil)
	{
		NSEnumerator * tEnumerator;
		NSDictionary * tMetadata=nil;
		
		tEnumerator=[inArray objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tRequirementDictionary;
				
			while (tRequirementDictionary=[tEnumerator nextObject])
			{
				NSNumber * tNumber;
				
				tNumber=[tRequirementDictionary objectForKey:ICDOCUMENT_REQUIREMENT_STATE];
				
				if (tNumber!=nil)
				{
					if ([tNumber boolValue]==YES)
					{
						tNumber=[tRequirementDictionary objectForKey:ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR];
						
						if (tNumber!=nil)
						{
							if ([tNumber intValue]==ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_CHOICE_UNSELECT_AND_HIDE)
							{
								return YES;
							}
						}
						else
						{
							// Missing Information
					
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							break;
						}
					}
				}
				else
				{
					// Missing Information
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_REQUIREMENT_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
					
					break;
				}
			}
			
			if (tMetadata==nil)
			{
				return NO;
			}
		}
		else
		{
			// Low Memory
		
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		}
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
											   
		return -1;
	}
												   
	return NO;
}

- (BOOL) setRequirementsFunctionForChoiceName:(NSString *) inName withArray:(NSArray *) inArray functionName:(NSString **) outFunctionName
{
	NSDictionary * tMetadata=nil;
	
	if (inName!=nil && inArray !=nil && outFunctionName!=nil)
	{
		NSMutableArray * tJavascriptFunctionsArray;
		NSMutableDictionary * tJavascriptFunctionsDictionary;
		NSMutableString * tJavascriptConstantsString;
		NSMutableArray * tJavascriptConstantsArray;
		NSMutableArray * tExtraResourcesArray;
		NSMutableDictionary * tExtraAttributesDictionary;
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		tExtraAttributesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_ATTRIBUTES_DICTIONARY];
		
		tExtraResourcesArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_EXTRA_RESOURCES_ARRAY];
		
		tJavascriptConstantsString=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_STRING];
		
		tJavascriptConstantsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_ARRAY];
		
		tJavascriptFunctionsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_ARRAY];
					
		tJavascriptFunctionsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_DICTIONARY];
		
		(*outFunctionName)=[inName stringByAppendingString:@"_requirement"];
		
		if ((*outFunctionName)!=nil)
		{
			NSMutableString * tFunctionCode; 
			
			tFunctionCode=[NSMutableString stringWithFormat:@"\tfunction %@(inCheckVisibilityOnly,inShowFailedToolTip)\n\
\t{\n\
\t\tvar tResult;\n\n",*outFunctionName];
			
			if (tFunctionCode!=nil)
			{
				BOOL tContinue=YES;
				
				if (sharedRequirementPluginsManager_==nil)
				{
					sharedRequirementPluginsManager_=[ICRequirementPluginsManager defaultManager];
					
					if (sharedRequirementPluginsManager_==nil)
					{
						// A COMPLETER
						
						tContinue=NO;
					}
				}
				
				if (tContinue==YES)
				{
					NSEnumerator * tEnumerator;
				
					tEnumerator=[inArray objectEnumerator];
				
					if (tEnumerator!=nil)
					{
						NSDictionary * tDictionary;
						int i,tDepth=-1;
						NSString * tTabulationDepth=@"\t";
						NSString * tPretest=@"";
						int tIndex=0;
						int tErrorIndex=1;
						
						while (tContinue==YES && (tDictionary=[tEnumerator nextObject]))
						{
							NSNumber * tNumber;
							
							tNumber=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_STATE];
							
							tContinue=NO;
							
							if (tNumber!=nil)
							{
								if ([tNumber boolValue]==YES)
								{
									NSString * tIdentifier;
									
									tIdentifier=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_IDENTIFIER];
									
									if (tIdentifier!=nil)
									{
										id tRequirementConverter;
										
										// Find the appropriate Requirement Converter plugin
										
										tRequirementConverter=[sharedRequirementPluginsManager_ createPluginConverterControllerForIdentifier:tIdentifier documentDictionary:documentDictionary_];
										
										if (tRequirementConverter!=nil)
										{
											NSDictionary * tRequirementSettingsDictionary;
											
											tRequirementSettingsDictionary=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_DICTIONARY];
											
											if (tRequirementSettingsDictionary!=nil)
											{
												NSString * tInvocationCode;
												NSDictionary * tErrorDictionary;
												
												tInvocationCode=[tRequirementConverter invocationWithParameters:tRequirementSettingsDictionary index:tIndex error:&tErrorDictionary];
												
												if (tInvocationCode!=nil)
												{
													tNumber=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR];
													
													if (tNumber!=nil)
													{
														int tFailureBehavior;
														NSString * tCodePart;
														NSDictionary * tSharedFunctionDictionary;
														NSString * tVariables;
														NSMutableArray * tSharedConstants;
														NSDictionary * tAttributesDictionary;
														NSArray * tExtraResources;
														
														// Add the shared functions
															
														tSharedFunctionDictionary=[tRequirementConverter sharedFunctionsImplementation];
														
														if (tSharedFunctionDictionary!=nil && [tSharedFunctionDictionary count]>0)
														{
															NSArray * tFunctionNamesKeys;
															
															tFunctionNamesKeys=[tSharedFunctionDictionary allKeys];
															
															if (tFunctionNamesKeys!=nil && [tFunctionNamesKeys count]>0)
															{
																NSEnumerator * tFunctionEnumerator;
																
																tFunctionEnumerator=[tFunctionNamesKeys objectEnumerator];
																
																if (tFunctionEnumerator!=nil)
																{
																	NSString * tSharedRequirementFunctionName;
																	
																	while (tSharedRequirementFunctionName=[tFunctionEnumerator nextObject])
																	{
																		if ([tJavascriptFunctionsDictionary objectForKey:tSharedRequirementFunctionName]==nil)
																		{
																			NSString * tImplementationCode;
																			
																			tImplementationCode=[tSharedFunctionDictionary objectForKey:tSharedRequirementFunctionName];
																			
																			if (tImplementationCode!=nil && [tImplementationCode length]>0)
																			{
																				[tJavascriptFunctionsArray addObject:tImplementationCode];
																				
																				[tJavascriptFunctionsDictionary setObject:tImplementationCode forKey:tSharedRequirementFunctionName];
																			}
																		}
																	}
																}
															}
														}
															
														// Add the shared constants
														
														tSharedConstants=[[tRequirementConverter sharedConstantsNames] mutableCopy];
														
														if (tSharedConstants!=nil)
														{
															[tSharedConstants removeObjectsInArray:tJavascriptConstantsArray];
														
															if ([tSharedConstants count]>0)
															{
																NSString * tString;
																
																tString=[tRequirementConverter constantsForNames:tSharedConstants];
																
																if (tString!=nil)
																{
																	[tJavascriptConstantsArray addObjectsFromArray:tSharedConstants];
																	
																	[tJavascriptConstantsString appendFormat:@"%@\n",tString];
																}
																else
																{
																	// A COMPLETER
																}
															}
															
															[tSharedConstants release];
														}
														
														tFailureBehavior=[tNumber intValue];
														
														tDepth++;
														
														tTabulationDepth=[tTabulationDepth stringByAppendingString:@"\t"];
														
														// Add the vars if needed
														
														tVariables=[tRequirementConverter variablesWithIndex:tIndex tabulationDepth:tTabulationDepth parameters:tRequirementSettingsDictionary error:&tErrorDictionary];
														
														if (tVariables!=nil)
														{
															// Get the list of resources to add
														
															tAttributesDictionary=[tRequirementConverter requiredOptionsValuesWithParameters:tRequirementSettingsDictionary];

															if (tAttributesDictionary!=nil && [tAttributesDictionary count]>0)
															{
																[tExtraAttributesDictionary addEntriesFromDictionary:tAttributesDictionary];
															}
															
															// Get the list of attributes to set
															
															tExtraResources=[tRequirementConverter requiredAdditionalResourcesWithParameters:tRequirementSettingsDictionary];
															
															if (tExtraResources!=nil && [tExtraResources count]>0)
															{
																[tExtraResourcesArray addObjectsFromArray:tExtraResources];
															}
															
															tIndex++;
															
															if (tFailureBehavior==ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_CHOICE_UNSELECT_AND_HIDE)
															{
																tCodePart=[NSString stringWithFormat:@"%@%@%@tResult=%@;\n\n",tPretest,tVariables,tTabulationDepth,tInvocationCode];

																tPretest=[NSString stringWithFormat:@"%@if (tResult==true)\n\%@{\n",tTabulationDepth,tTabulationDepth];
																
																[tFunctionCode appendString:tCodePart];
																
																tContinue=YES;
															}
															else if (tFailureBehavior==ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR_CHOICE_UNSELECT_AND_DISABLE)
															{
																NSArray * tLocalizations;
																
																tLocalizations=[tDictionary objectForKey:ICDOCUMENT_REQUIREMENT_MESSAGE];
																
																if (tLocalizations!=nil)
																{
																	BOOL noErrorMessage=NO;
																	NSString * tErrorMessage;
																
																	tContinue=YES;
																	
																	if ([tLocalizations count]==0)
																	{
																		noErrorMessage=YES;
																	}
																	else
																	{
																		NSEnumerator * tLocalizationsEnumerator;
					
																		tLocalizationsEnumerator=[tLocalizations objectEnumerator];
																		
																		if (tLocalizationsEnumerator!=nil)
																		{
																			NSMutableDictionary * tLocalizationsDictionary;
																			NSDictionary * tLanguageDictionary;
																			BOOL atLeastOneErrorMessage=NO;
																			
																			tLocalizationsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
																			
																			tErrorMessage=[NSString stringWithFormat:@"%@_REQUIREMENT_FAILED_%d",[inName uppercaseString],tErrorIndex];
																			
																			while (tLanguageDictionary=[tLocalizationsEnumerator nextObject])
																			{
																				NSString * tLanguage;
																				
																				tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
																				
																				if (tLanguage!=nil)
																				{
																					NSMutableDictionary * tLanguageLocalizationsDictionary;
																					NSString * tLocalization;
																					
																					tLanguageLocalizationsDictionary=[tLocalizationsDictionary objectForKey:tLanguage];
																					
																					if (tLanguageLocalizationsDictionary==nil)
																					{
																						tLanguageLocalizationsDictionary=[NSMutableDictionary dictionary];
																						
																						if (tLanguageLocalizationsDictionary!=nil)
																						{
																							[tLocalizationsDictionary setObject:tLanguageLocalizationsDictionary forKey:tLanguage];
																						}
																						else
																						{
																							// Low Memory
																							
																							tContinue=NO;
																							
																							break;
																						}
																					}
																					
																					tLocalization=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
																				
																					if (tLocalization!=nil)
																					{
																						atLeastOneErrorMessage=YES;
																							
																						[tLanguageLocalizationsDictionary setObject:tLocalization forKey:tErrorMessage];
																						
																						tErrorIndex++;
																					}
																					else
																					{
																						// Missing Information
																						
																						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																																			 ICDOCUMENT_VALUE,IC_BUILDER_FAILURE_TAG_NAME,
																																			 nil];
																						
																						tContinue=NO;
																							
																						break;
																					}
																				}
																				else
																				{
																					// Missing Information
																					
																					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																																		 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																																		 nil];
																					
																					tContinue=NO;
																							
																					break;
																				}
																			}
																			
																			if (tContinue==YES && atLeastOneErrorMessage==NO)
																			{
																				NSMutableDictionary * tDefaultLanguageDictionary;
																				
																				noErrorMessage=NO;
																			
																				// Set the default value
										
																				tDefaultLanguageDictionary=[tLocalizationsDictionary objectForKey:ICLANGUAGE_DEFAULT_LANGUAGE];
																						
																				if (tDefaultLanguageDictionary==nil)
																				{
																					tDefaultLanguageDictionary=[NSMutableDictionary dictionary];
																					
																					if (tDefaultLanguageDictionary!=nil)
																					{
																						[tLocalizationsDictionary setObject:tDefaultLanguageDictionary forKey:ICLANGUAGE_DEFAULT_LANGUAGE];
																					}
																					else
																					{
																						// Low Memory
																						
																						tContinue=NO;
																					}
																				}
																						
																				[tDefaultLanguageDictionary setObject:@"" forKey:tErrorMessage];
																			}
																		}
																		else
																		{
																			// Low Memory
																			
																			tContinue=NO;
																		}

																	}
																	
																	if (tContinue==YES)
																	{
																		if (noErrorMessage==YES)
																		{
																			tCodePart=[NSString stringWithFormat:@"%@%@%@tResult=%@;\n\n\
%@if (tResult==false)\n\
%@{\n\
%@\tif (inCheckVisibilityOnly==true)\n\
%@\t{\n\
%@\t\ttResult=true;\n\
%@\t}\n\
%@}\n",tPretest,tVariables,tTabulationDepth,tInvocationCode,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth];
																		}
																		else
																		{
																			if (tErrorMessage!=nil)
																			{
																				tCodePart=[NSString stringWithFormat:@"%@%@%@tResult=%@;\n\n\
%@if (tResult==false)\n\
%@{\n\
%@\tif (inCheckVisibilityOnly==true)\n\
%@\t{\n\
%@\t\ttResult=true;\n\
%@\t}\n\
%@\telse\n\
%@\t{\n\
%@\t\tif (inShowFailedToolTip==true)\n\
%@\t\t{\n\
%@\t\t\tchoices.%@.tooltip=system.localizedString(\'%@\');\n\
%@\t\t}\n\
%@\t}\n\
%@}\n",tPretest,tVariables,tTabulationDepth,tInvocationCode,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,tTabulationDepth,inName,tErrorMessage,tTabulationDepth,tTabulationDepth,tTabulationDepth];
																			}
																			else
																			{
																				tMetadata=nil;
																			
																				tContinue=NO;
																			}
																		}
																	}
																	
																	if (tContinue==YES)
																	{
																		tPretest=[NSString stringWithFormat:@"\n%@if (tResult==true)\n\%@{\n",tTabulationDepth,tTabulationDepth];
																	
																		[tFunctionCode appendString:tCodePart];
																	}
																}
																else
																{
																	// Missing Information
																	
																	tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																														 ICDOCUMENT_REQUIREMENT_MESSAGE,IC_BUILDER_FAILURE_TAG_NAME,
																														 nil];
																}
															}
															else
															{
																// Incorrect Value
																
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																													 ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,IC_BUILDER_FAILURE_TAG_NAME,
																													 nil];
															}
														}
														else
														{
															// A COMPLETER
													
															tMetadata=nil;
														}
													}
													else
													{
														// Missing Information
										
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																											 ICDOCUMENT_REQUIREMENT_FAILURE_BEHAVIOR,IC_BUILDER_FAILURE_TAG_NAME,
																											 nil];
													}
												}
												else
												{
													// A COMPLETER
													
													tMetadata=nil;
												}
											}
											else
											{
												// Missing Information
										
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_REQUIREMENT_DICTIONARY,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
											
											[tRequirementConverter release];
										}
										else
										{
											// Converter not found
											
											tMetadata=nil;
											
											// A COMPLETER
										}
									}
									else
									{
										// Missing Information
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_REQUIREMENT_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
									}
								}
								else
								{
									tContinue=YES;
								}
							}
							else
							{
								// Missing Information
								
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_REQUIREMENT_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							}
						}
						
						if (tDictionary==nil)
						{
							for(i=0;i<tDepth;i++)
							{
								NSString * tClosingLine;
								
								tTabulationDepth=[tTabulationDepth substringToIndex:[tTabulationDepth length]-1];
								
								tClosingLine=[NSString stringWithFormat:@"%@}\n",tTabulationDepth];
								
								[tFunctionCode appendString:tClosingLine];
								
								
							}
							
							[tFunctionCode appendString:@"\n\t\treturn tResult;\n\t}"];
							
							[tJavascriptFunctionsArray addObject:tFunctionCode];
							
							[tJavascriptFunctionsDictionary setObject:tFunctionCode forKey:*outFunctionName];
							
							return YES;
						}
					}
				}
			}
		}
	}
	
setRequirementsFunctionForChoiceNameBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (NSString *) logicStringForDependency:(NSDictionary *) inDictionary
{
	NSString * tString=nil;
	NSDictionary * tMetadata=nil;
	
	if (inDictionary!=nil)
	{
		NSDictionary * tDictionary;
		NSNumber * tNumber;
		
		// Is it a branch or leaf?
		
		tDictionary=[inDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_TOP];
		
		if (tDictionary==nil)
		{
			// Leaf
			
			NSString * tChoiceUUID;
			
			tChoiceUUID=[inDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_UUID];
			
			if (tChoiceUUID!=nil)
			{
				NSDictionary * tChoicesNamesDictionary;
						
				tChoicesNamesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES];
			
				if (tChoicesNamesDictionary!=nil)
				{
					NSString * tChoiceReadableID;
					
					tChoiceReadableID=[tChoicesNamesDictionary objectForKey:tChoiceUUID];
					
					if (tChoiceReadableID!=nil)
					{
						tNumber=[inDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_COMPARATOR];
						
						if (tNumber!=nil)
						{
							int tComparator;
							
							tComparator=[tNumber intValue];
							
							tNumber=[inDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_STATE_OBJECT];
						
							if (tNumber!=nil)
							{
								int tStateObject;
								
								tStateObject=[tNumber intValue];
								
								if (tComparator==ICCOMPARATOR_IS)
								{
									tString=[NSString stringWithFormat:@"choices.%@.",tChoiceReadableID];
								}
								else if (tComparator==ICCOMPARATOR_IS_NOT)
								{
									tString=[NSString stringWithFormat:@"!choices.%@.",tChoiceReadableID];
								}
								else
								{
									// Incorrect Value
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_COMPARATOR",tComparator];
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_COMPARATOR,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
								}
								
								if (tString!=nil)
								{
									if (tStateObject==ICSTATE_ENABLED)
									{
										tString=[tString stringByAppendingString:@"enabled"];
									}
									else if (tStateObject==ICSTATE_SELECTED)
									{
										tString=[tString stringByAppendingString:@"selected"];
									}
									else
									{
										// Incorrect Value
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_STATE_OBJECT",tStateObject];
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_STATE_OBJECT,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										tString=nil;
									}
								}
							}
							else
							{
								// Missing Information
								
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_STATE_OBJECT,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_COMPARATOR,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					}
					else
					{
						// Missing Information
				
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 tChoiceUUID,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				}
				else
				{
					// Missing Build Data
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES"];
				
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																		 ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_LEAF_ITEM_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
		else
		{
			// Branch
			
			NSString * tTopString;
			
			tTopString=[self logicStringForDependency:tDictionary];
			
			if (tTopString!=nil)
			{
				tDictionary=[inDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_BOTTOM];
				
				if (tDictionary!=nil)
				{
					NSString * tBottomString;
			
					tBottomString=[self logicStringForDependency:tDictionary];
					
					if (tBottomString!=nil)
					{
						tNumber=[inDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_OPERATOR];
						
						if (tNumber!=nil)
						{
							int tOperator;
							
							tOperator=[tNumber intValue];
							
							if (tOperator==ICOPERATOR_AND)
							{
								tString=[NSString stringWithFormat:@"(%@ && %@)",tTopString,tBottomString];
							}
							else if (tOperator==ICOPERATOR_OR)
							{
								tString=[NSString stringWithFormat:@"(%@ || %@)",tTopString,tBottomString];
							}
							else
							{
								// Incorrect Value
							
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_OPERATOR",tOperator];
								
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_OPERATOR,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_OPERATOR,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					}
				}
				else
				{
					// Missing Information
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_CHOICE_TREE_BRANCH_BOTTOM,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
		}
	}
	
	if (tMetadata!=nil)
	{
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
	}
	
	return tString;
}

- (BOOL) createEnabledDependencyFunctionNamed:(NSString *) inFunctionName withDependencyDictionary:(NSDictionary *) inDictionary
{
	NSMutableDictionary * tMetadata=nil;
	
	if (inFunctionName!=nil && inDictionary!=nil)
	{
		NSString * tLogicString;
		
		tLogicString=[self logicStringForDependency:inDictionary];
		
		if (tLogicString!=nil)
		{
			NSString * tFunctionCode;
			
			tFunctionCode=[NSString stringWithFormat:@"\tfunction %@()\n\t{\n\t\treturn %@;\t\n\t}",inFunctionName,tLogicString];
			
			if (tFunctionCode!=nil)
			{
				NSMutableArray * tJavascriptFunctionsArray;
				NSMutableDictionary * tJavascriptFunctionsDictionary;
				
				tJavascriptFunctionsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_ARRAY];
				
				if (tJavascriptFunctionsArray!=nil)
				{
					[tJavascriptFunctionsArray addObject:tFunctionCode];
				}
				
				tJavascriptFunctionsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_DICTIONARY];
				
				if (tJavascriptFunctionsDictionary!=nil)
				{
					[tJavascriptFunctionsDictionary setObject:tFunctionCode forKey:inFunctionName];
					
					return YES;
				}
			}
			else
			{
				// Low Memory
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) createSelectedDependencyFunctionNamed:(NSString *) inFunctionName withEnabledState:(int) inEnabledState withEnabledFunctionName:(NSString *) inEnabledFunctionName withDependencyDictionary:(NSDictionary *) inDictionary
{
	NSMutableDictionary * tMetadata=nil;
	
	if (inFunctionName!=nil && inDictionary!=nil && inEnabledFunctionName!=nil)
	{
		NSString * tLogicString;
		
		tLogicString=[self logicStringForDependency:inDictionary];
		
		if (tLogicString!=nil)
		{
			NSString * tFunctionCode;
			
			if (inEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_ALWAYS)
			{
				tFunctionCode=[NSString stringWithFormat:@"\tfunction %@(isStart)\n\t{\n\t\tvar tSelected;\n\n\t\ttSelected=%@;\n\n\t\tif (isStart==true)\n\t\t{\n\t\t\treturn tSelected;\n\t\t}\n\n\t\treturn (tSelected && my.choice.selected);\n\t}",inFunctionName,tLogicString];
			}
			else if (inEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_NEVER)
			{
				tFunctionCode=[NSString stringWithFormat:@"\tfunction %@()\n\t{\n\t\treturn %@;\n\t}",inFunctionName,tLogicString];
			}
			else
			{
				tFunctionCode=[NSString stringWithFormat:@"\tfunction %@(isStart)\n\t{\n\t\tvar tSelected;\n\n\t\ttSelected=%@;\n\n\t\tif (%@()==false || isStart==true)\n\t\t{\n\t\t\treturn tSelected;\n\t\t}\n\n\t\treturn (tSelected && my.choice.selected);\n\t}",inFunctionName,tLogicString,inEnabledFunctionName];
			}
						
			if (tFunctionCode!=nil)
			{
				NSMutableArray * tJavascriptFunctionsArray;
				NSMutableDictionary * tJavascriptFunctionsDictionary;
				
				tJavascriptFunctionsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_ARRAY];
				
				if (tJavascriptFunctionsArray!=nil)
				{
					[tJavascriptFunctionsArray addObject:tFunctionCode];
				}
				
				tJavascriptFunctionsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_DICTIONARY];
				
				if (tJavascriptFunctionsDictionary!=nil)
				{
					[tJavascriptFunctionsDictionary setObject:tFunctionCode forKey:inFunctionName];
					
					return YES;
				}
			}
			else
			{
				// Low Memory
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) setStateAttributesOfChoiceElement:(NSXMLElement *) inElement name:(NSString *) inName requirementFunctionName:(NSString *) inRequirementFunctionName affectVisible:(BOOL) inAffectVisible withOptionsDictionary:(NSDictionary *) inOptionsDictionary
{
	NSDictionary * tMetadata=nil;
	
	// Requirements

	if (inElement!=nil && inOptionsDictionary!=nil && inRequirementFunctionName!=nil)
	{
		NSNumber * tNumber;
		id tAttribute;
		NSDictionary * tDependencyDictionary;
		BOOL tContinue=YES;
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		// start_visible
		
		tNumber=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_HIDDEN];
		
		if (tNumber!=nil && [tNumber boolValue]==YES)
		{
			tAttribute=[NSXMLNode attributeWithName:@"start_visible" stringValue:@"false"];

			if (tAttribute!=nil)
			{
				[inElement addAttribute:tAttribute];
			}
			else
			{
				// Low Memory
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_visible' attribute"];
				
				tContinue=NO;
			}
		}
		else
		{
			if (inAffectVisible==YES)
			{
				tAttribute=[NSXMLNode attributeWithName:@"start_visible" stringValue:[NSString stringWithFormat:@"%@(true,false)",inRequirementFunctionName]];
			
				if (tAttribute!=nil)
				{
					[inElement addAttribute:tAttribute];
				}
				else
				{
					// Low Memory
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_visible' attribute"];
					
					tContinue=NO;
				}
			}
		}
		
		if (tContinue==YES)
		{
		
			tNumber=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE];
		
			if (tNumber!=nil)
			{
				int tState;
				
				tState=[tNumber intValue];
				
				tContinue=NO;
				
				switch(tState)
				{
					// Choice or Merged Choices
					
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_REQUIRED:
						
						// start_enabled
						
						tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:@"false"];

						if (tAttribute!=nil)
						{
							[inElement addAttribute:tAttribute];
							
							// start_selected
						
							tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:[NSString stringWithFormat:@"%@(false,true)",inRequirementFunctionName]];

							if (tAttribute!=nil)
							{
								[inElement addAttribute:tAttribute];
								
								tContinue=YES;
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
							}
						}
						else
						{
							// Low Memory
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
						}
						
						break;
						
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_SELECTED:
					
						// start_enabled
						
						tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@(false,true)",inRequirementFunctionName]];

						if (tAttribute!=nil)
						{
							[inElement addAttribute:tAttribute];
							
							// start_selected
						
							tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:[NSString stringWithFormat:@"%@(false,false)",inRequirementFunctionName]];

							if (tAttribute!=nil)
							{
								[inElement addAttribute:tAttribute];
								
								tContinue=YES;
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
							}
						}
						else
						{
							// Low Memory
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
						}
						
						break;
						
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_UNSELECTED:
					
						// start_enabled
						
						tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@(false,true)",inRequirementFunctionName]];

						if (tAttribute!=nil)
						{
							[inElement addAttribute:tAttribute];
							
							// start_selected
						
							tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:@"false"];

							if (tAttribute!=nil)
							{
								[inElement addAttribute:tAttribute];
								
								tContinue=YES;
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
							}
						}
						else
						{
							// Low Memory
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
						}
						
						break;
						
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENT:
					
						tDependencyDictionary=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY];
							
						if (tDependencyDictionary!=nil)
						{
							int tEnabledState=ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_ALWAYS;
							NSString * tSelectedFunctionName;
							
							// enabled
							
							tNumber=[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_MODE];
							
							if (tNumber!=nil)
							{
								tEnabledState=[tNumber intValue];
							}
							
							if (tEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_ALWAYS)
							{
								tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@(false,false)",inRequirementFunctionName]];
								
								if (tAttribute!=nil)
								{
									[inElement addAttribute:tAttribute];
									
									tContinue=YES;
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
								}
							}
							else if (tEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_NEVER)
							{
								tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:@"false"];

								if (tAttribute!=nil)
								{
									[inElement addAttribute:tAttribute];
									
									tContinue=YES;
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
								}
							}
							else  if (tEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_DEPENDENT)
							{
								NSString * tEnabledFunctionName;
							
								tEnabledFunctionName=[inName stringByAppendingString:@"_enabled"];
								
								if (tEnabledFunctionName!=nil)
								{
									if ([self createEnabledDependencyFunctionNamed:tEnabledFunctionName
												   withDependencyDictionary:[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_DICTIONARY]]==YES)
									{
										tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@() && %@(false,false)",tEnabledFunctionName,inRequirementFunctionName]];

										if (tAttribute!=nil)
										{
											[inElement addAttribute:tAttribute];
											
											tAttribute=[NSXMLNode attributeWithName:@"enabled" stringValue:[NSString stringWithFormat:@"%@() && %@(false,false)",tEnabledFunctionName,inRequirementFunctionName]];

											if (tAttribute!=nil)
											{
												[inElement addAttribute:tAttribute];
												
												tContinue=YES;
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'enabled' attribute"];
											}
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
										}
									}
									else
									{
										tMetadata=nil;
									}
								}
							}
							else 
							{
								// Incorrect Value
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_MODE",tEnabledState];
								
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_MODE,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							}
							
							if (tContinue==YES)
							{
								tContinue=NO;
								
								// selected
							
								tSelectedFunctionName=[inName stringByAppendingString:@"_selected"];
							
								if (tSelectedFunctionName!=nil)
								{
									if ([self createSelectedDependencyFunctionNamed:tSelectedFunctionName
																   withEnabledState:tEnabledState
															withEnabledFunctionName:[inName stringByAppendingString:@"_enabled"]
														   withDependencyDictionary:[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_SELECTED_DICTIONARY]]==YES)
									{
										tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:[NSString stringWithFormat:@"%@(true) && %@(false,true)",tSelectedFunctionName,inRequirementFunctionName]];

										if (tAttribute!=nil)
										{
											[inElement addAttribute:tAttribute];
											
											tAttribute=[NSXMLNode attributeWithName:@"selected" stringValue:[NSString stringWithFormat:@"%@(false) && %@(false,true)",tSelectedFunctionName,inRequirementFunctionName]];

											if (tAttribute!=nil)
											{
												[inElement addAttribute:tAttribute];
												
												tContinue=YES;
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'selected' attribute"];
											}
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
										}
									}
									else
									{
										tMetadata=nil;
									}
								}
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					
						break;
					
					// Groups
					
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_ENABLED_ALWAYS:
					
						// Nothing to do
						
						tContinue=YES;
						
						break;
						
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_ENABLED_NEVER:
						
						// start_enabled
						
						tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:@"false"];

						if (tAttribute!=nil)
						{
							[inElement addAttribute:tAttribute];
							
							tContinue=YES;
						}
						else
						{
							// Low Memory
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
						}
						
						break;
						
					case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_ENABLED_DEPENDENT:
					
						tDependencyDictionary=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY];
							
						if (tDependencyDictionary!=nil)
						{
							NSString * tEnabledFunctionName;
							
							tEnabledFunctionName=[inName stringByAppendingString:@"_enabled"];
							
							if (tEnabledFunctionName!=nil)
							{
								if ([self createEnabledDependencyFunctionNamed:tEnabledFunctionName
											   withDependencyDictionary:[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_DICTIONARY]]==YES)
								{
									tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@()",tEnabledFunctionName]];

									if (tAttribute!=nil)
									{
										[inElement addAttribute:tAttribute];
										
										tAttribute=[NSXMLNode attributeWithName:@"enabled" stringValue:[NSString stringWithFormat:@"%@()",tEnabledFunctionName]];

										if (tAttribute!=nil)
										{
											[inElement addAttribute:tAttribute];
											
											tContinue=YES;
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'enabled' attribute"];
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
									}
								}
								else
								{
									tMetadata=nil;
								}
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					
						break;
						
					default:
					
						// Incorrect Value
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (@d) for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE",tState];
								
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
						break;
				}
			
				if (tContinue==YES)
				{
					return YES;
				}
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) setStateAttributesOfChoiceElement:(NSXMLElement *) inElement name:(NSString *) inName withOptionsDictionary:(NSDictionary *) inOptionsDictionary
{
	NSDictionary * tMetadata=nil;
	
	// No requirements

	if (inElement!=nil && inOptionsDictionary!=nil)
	{
		NSNumber * tNumber;
		id tAttribute;
		NSDictionary * tDependencyDictionary;
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		// start_visible
		
		tNumber=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_HIDDEN];
		
		if (tNumber!=nil && [tNumber boolValue]==YES)
		{
			tAttribute=[NSXMLNode attributeWithName:@"start_visible" stringValue:@"false"];

			if (tAttribute!=nil)
			{
				[inElement addAttribute:tAttribute];
			}
			else
			{
				// Low Memory
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_visible' attribute"];
				
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
															forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
													       metadata:tMetadata];
											   
				return NO;
			}
		}
		
		tNumber=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE];
		
		if (tNumber!=nil)
		{
			int tState;
			
			tState=[tNumber intValue];
			
			switch(tState)
			{
				// Choice or Merged Choices
				
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_REQUIRED:
					
					// start_enabled
					
					tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:@"false"];

					if (tAttribute!=nil)
					{
						[inElement addAttribute:tAttribute];
						
						// start_selected
					
						tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:@"true"];

						if (tAttribute!=nil)
						{
							[inElement addAttribute:tAttribute];
							
							return YES;
						}
						else
						{
							// Low Memory
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
						}
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
					}
					
					break;
					
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_SELECTED:
				
					// Nothing to do
					
					return YES;
					
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_UNSELECTED:
				
					// start_selected
					
					tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:@"false"];

					if (tAttribute!=nil)
					{
						[inElement addAttribute:tAttribute];
						
						return YES;
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
					}
					
					break;
					
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENT:
				
					tDependencyDictionary=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY];
						
					if (tDependencyDictionary!=nil)
					{
						int tEnabledState=ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_ALWAYS;
						NSString * tSelectedFunctionName;
						BOOL tContinue=YES;
						
						// enabled
						
						tNumber=[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_MODE];
						
						if (tNumber!=nil)
						{
							tEnabledState=[tNumber intValue];
						}
						
						if (tEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_ALWAYS)
						{
							// Nothing to do
						}
						else if (tEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_NEVER)
						{
							tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:@"false"];

							if (tAttribute!=nil)
							{
								[inElement addAttribute:tAttribute];
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
								
								tContinue=NO;
							}
						}
						else  if (tEnabledState==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_DEPENDENT)
						{
							NSString * tEnabledFunctionName;
						
							tEnabledFunctionName=[inName stringByAppendingString:@"_enabled"];
							
							tContinue=NO;
							
							if (tEnabledFunctionName!=nil)
							{
								if ([self createEnabledDependencyFunctionNamed:tEnabledFunctionName
											   withDependencyDictionary:[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_DICTIONARY]]==YES)
								{
									tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@()",tEnabledFunctionName]];

									if (tAttribute!=nil)
									{
										[inElement addAttribute:tAttribute];
										
										tAttribute=[NSXMLNode attributeWithName:@"enabled" stringValue:[NSString stringWithFormat:@"%@()",tEnabledFunctionName]];

										if (tAttribute!=nil)
										{
											[inElement addAttribute:tAttribute];
											
											tContinue=YES;
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'enabled' attribute"];
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
									}
								}
								else
								{
									tMetadata=nil;
								}
							}
						}
						else 
						{
							// Incorrect Value
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (@d) for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_MODE",tEnabledState];
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_STATE_MODE,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							tContinue=NO;
						}
						
						if (tContinue==YES)
						{
							// selected
							
							tSelectedFunctionName=[inName stringByAppendingString:@"_selected"];
							
							if (tSelectedFunctionName!=nil)
							{
								if ([self createSelectedDependencyFunctionNamed:tSelectedFunctionName
															   withEnabledState:tEnabledState
														withEnabledFunctionName:[inName stringByAppendingString:@"_enabled"]
													   withDependencyDictionary:[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_SELECTED_DICTIONARY]]==YES)
								{
									tAttribute=[NSXMLNode attributeWithName:@"start_selected" stringValue:[NSString stringWithFormat:@"%@(true)",tSelectedFunctionName]];

									if (tAttribute!=nil)
									{
										[inElement addAttribute:tAttribute];
										
										tAttribute=[NSXMLNode attributeWithName:@"selected" stringValue:[NSString stringWithFormat:@"%@(false)",tSelectedFunctionName]];

										if (tAttribute!=nil)
										{
											[inElement addAttribute:tAttribute];
											
											return YES;
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'selected' attribute"];
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_selected' attribute"];
									}
								}
								else
								{
									tMetadata=nil;
								}
							}
						}
					}
					else
					{
						// Missing Information
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				
					break;
				
				// Groups
				
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_ENABLED_ALWAYS:
				
					// Nothing to do
					
					return YES;
					
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_ENABLED_NEVER:
					
					// start_enabled
					
					tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:@"false"];

					if (tAttribute!=nil)
					{
						[inElement addAttribute:tAttribute];
						
						return YES;
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
					}
					
					break;
					
				case ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_ENABLED_DEPENDENT:
				
					tDependencyDictionary=[inOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY];
						
					if (tDependencyDictionary!=nil)
					{
						NSString * tEnabledFunctionName;
						
						tEnabledFunctionName=[inName stringByAppendingString:@"_enabled"];
						
						if (tEnabledFunctionName!=nil)
						{
							if ([self createEnabledDependencyFunctionNamed:tEnabledFunctionName
												  withDependencyDictionary:[tDependencyDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY_ENABLED_DICTIONARY]]==YES)
							{
								tAttribute=[NSXMLNode attributeWithName:@"start_enabled" stringValue:[NSString stringWithFormat:@"%@()",tEnabledFunctionName]];

								if (tAttribute!=nil)
								{
									[inElement addAttribute:tAttribute];
									
									tAttribute=[NSXMLNode attributeWithName:@"enabled" stringValue:[NSString stringWithFormat:@"%@()",tEnabledFunctionName]];

									if (tAttribute!=nil)
									{
										[inElement addAttribute:tAttribute];
										
										return YES;
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'enabled' attribute"];
									}
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'start_enabled' attribute"];
								}
							}
							else
							{
								tMetadata=nil;
							}
						}
					}
					else
					{
						// Missing Information
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE_DEPENDENCY,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				
					break;
					
				default:
				
					// Incorrect value
				
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (@d)for key ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE",tState];
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																										  ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																										  nil];
				
					break;
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
											   
	return NO;
}

- (BOOL) setStateAttributesOfChoiceElement:(NSXMLElement *) inElement name:(NSString *) inName withDictionary:(NSDictionary *) inChoiceDictionary
{
	NSDictionary * tMetadata=nil;
	
	if (inElement!=nil && inChoiceDictionary!=nil)
	{
		NSDictionary * tOptionsDictionary;
		NSArray * tRequirementsArray;
		NSString * tRequirementFunctionName=nil;
		int tAtLeastOneUnselectAndHideRequirementEnabled=NO;
		BOOL tContinue=YES;
		
		tRequirementsArray=[inChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_REQUIREMENTS];
		
		if (tRequirementsArray!=nil && [tRequirementsArray count]>0)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[tRequirementsArray objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				NSDictionary * tRequirementDictionary;
				
				while (tRequirementDictionary=[tEnumerator nextObject])
				{
					NSNumber * tStateNumber;
					
					tStateNumber=[tRequirementDictionary objectForKey:ICDOCUMENT_REQUIREMENT_STATE];
					
					if (tStateNumber!=nil)
					{
						if ([tStateNumber boolValue]==YES)
						{
							tContinue=NO;
							
							if ([self setRequirementsFunctionForChoiceName:inName withArray:tRequirementsArray functionName:&tRequirementFunctionName]==YES)
							{
								tAtLeastOneUnselectAndHideRequirementEnabled=[self hasOneHideAndUnselectRequirementEnabledInArray:tRequirementsArray];
							
								if (tAtLeastOneUnselectAndHideRequirementEnabled!=-1)
								{
									tContinue=YES;
								}
							}
							
							break;
						}
					}
					else
					{
						// Missing Information
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_REQUIREMENT_STATE,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
						
						tContinue=NO;
					}
				}
			}
			else
			{
				// Low Memory
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				
				tContinue=NO;
			}
		}
		
		if (tContinue==YES)
		{
			tOptionsDictionary=[inChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS];
		
			if (tOptionsDictionary!=nil)
			{
				if (tRequirementFunctionName==nil)
				{
					// No Requirements
									
					if ([self setStateAttributesOfChoiceElement:inElement name:inName withOptionsDictionary:tOptionsDictionary]==YES)
					{
						return YES;
					}
				}
				else
				{
					// Requirements
					
					if ([self setStateAttributesOfChoiceElement:inElement name:inName requirementFunctionName:tRequirementFunctionName affectVisible:tAtLeastOneUnselectAndHideRequirementEnabled withOptionsDictionary:tOptionsDictionary]==YES)
					{
						return YES;
					}
				}
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) addChoicesElementsToElement:(NSXMLElement *) inElement withChoicesArray:(NSArray *) inChoicesArray isInstaller:(BOOL) inIsInstaller
{
	NSDictionary * tMetadata=nil;
	
	if (inElement!=nil && inChoicesArray!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[inChoicesArray objectEnumerator];
	
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tChoiceDictionary;
			NSMutableDictionary * tChoicesNamesDictionary;
						
			tChoicesNamesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES];
			
			if (tChoicesNamesDictionary!=nil)
			{
				BOOL tContinue=YES;
				
				while (tContinue==YES && (tChoiceDictionary=[tEnumerator nextObject]))
				{
					NSXMLElement * tChoiceElement;
					
					tChoiceElement=(NSXMLElement *) [NSXMLNode elementWithName:@"choice"];
					
					if (tChoiceElement!=nil)
					{
						NSString * tChoiceUUID;
						
						// id
						
						tChoiceUUID=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_UUID];
						
						if (tChoiceUUID!=nil)
						{
							NSString * tChoiceReadableID;
							
							tChoiceReadableID=[tChoicesNamesDictionary objectForKey:tChoiceUUID];
							
							if (tChoiceReadableID!=nil)
							{
								id tAttribute;
							
								tAttribute=[NSXMLNode attributeWithName:@"id" stringValue:tChoiceReadableID];
							
								if (tAttribute!=nil)
								{
									NSNumber * tTypeNumber;
									NSDictionary * tOptionsDictionary;
									NSArray * tLocalizations;
									NSString * tString=nil;
									NSString * tDefaultTitle=nil;
									NSString * tDefaultDescription=nil;
									
									[tChoiceElement addAttribute:tAttribute];
									
									tTypeNumber=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_TYPE];
									
									tOptionsDictionary=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS];
										
									if (tOptionsDictionary!=nil)
									{
										if (tTypeNumber!=nil && [tTypeNumber intValue]==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_TYPE_GROUP)
										{
											// Group choice
											
											NSArray * tChildren;
													
											tChildren=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_GROUP_CHILDREN];
												
											if (tChildren!=nil && [tChildren count]>0)
											{
												NSNumber * tHideChildrenNumber;
												
												tHideChildrenNumber=[tOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_HIDE_CHILDREN];
									
												if (tHideChildrenNumber==nil || [tHideChildrenNumber boolValue]==NO)
												{
													if ([self addChoicesElementsToElement:inElement withChoicesArray:tChildren isInstaller:inIsInstaller]==NO)
													{
														tContinue=NO;
													}
												}
												else
												{
													NSEnumerator * tChildrenEnumerator;
												
													// Add the pkg-refs
													
													tChildrenEnumerator=[tChildren objectEnumerator];
													
													if (tChildrenEnumerator!=nil)
													{
														NSDictionary * tChildDictionary;
														
														while (tChildDictionary=[tChildrenEnumerator nextObject])
														{
															if ([self addPackageRefeferenceElementToChoiceElement:tChoiceElement fromDictionary:tChildDictionary]==NO)
															{
																tContinue=NO;
							
																break;
															}
														}
													}
													else
													{
														// Low Memory
														
														tContinue=NO;
													}
												}
											}
											else
											{
												// If it's not a hierarchy for Installer.app, we skip empty folder
												
												if (inIsInstaller==NO)
												{
													continue;
												}
											}
										}
										else
										{
											// Package choice
											
											if ([self addPackageRefeferenceElementToChoiceElement:tChoiceElement fromDictionary:tChoiceDictionary]==NO)
											{
												tContinue=NO;
											}
										}
										
										if (tContinue==NO)
										{
											break;
										}
										
										if ([self setStateAttributesOfChoiceElement:tChoiceElement name:tChoiceReadableID withDictionary:tChoiceDictionary]==NO)
										{
											tContinue=NO;
							
											break;
										}
									
										// title 
										
										// Get the default title
									
										if (tTypeNumber==nil || [tTypeNumber intValue]==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_TYPE_PACKAGE)
										{
											// Use the name of the package (if it's a package)
											
											NSString * tPackageUUID;
											
											tPackageUUID=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_PACKAGE_UUID];
											
											if (tPackageUUID!=nil)
											{
												NSMutableDictionary * tPackagesDictionary;
												NSDictionary * tPackageBuildInformation;
												
												tPackagesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES];
												
												tPackageBuildInformation=[tPackagesDictionary objectForKey:tPackageUUID];
												
												if (tPackageBuildInformation!=nil)
												{
													tDefaultTitle=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NAME];
												}
												else
												{
													// Missing Build Data
													
													[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: %@",tPackageUUID];
						
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																										 tPackageUUID,IC_BUILDER_FAILURE_TAG_NAME,
																										 nil];
													
													tContinue=NO;
												}
											}
											else
											{
												// Missing Information
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_PACKAGE_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
												
												tContinue=NO;
											}
										}
										else
										{
											NSArray * tComponents;
											
											tDefaultTitle=@"Choice";
											
											// Try to get something better
											
											tComponents=[tChoiceReadableID componentsSeparatedByString:@"_"];
											
											if (tComponents!=nil)
											{
												NSUInteger tCount;
												
												tCount=[tComponents count];
												
												if (tCount>2)
												{
													NSArray * tIndexesArray;
													
													tIndexesArray=[tComponents subarrayWithRange:NSMakeRange(2,tCount-2)];
													
													tDefaultTitle=[NSString stringWithFormat:@"%@ %@",[tComponents objectAtIndex:1],[tIndexesArray componentsJoinedByString:@"."]];
												}
											}
										}
									
										if (tContinue==NO)
										{
											break;
										}
										
										tLocalizations=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_TITLE];
										
										if (tLocalizations==nil || [tLocalizations count]==0)
										{
											tString=tDefaultTitle;
										}
										else
										{
											// Make the title localizable
												
											NSEnumerator * tLocalizationsEnumerator;
											
											tString=[NSString stringWithFormat:@"%@_TITLE",[tChoiceReadableID uppercaseString]];
											
											tLocalizationsEnumerator=[tLocalizations objectEnumerator];
												
											if (tLocalizationsEnumerator!=nil)
											{
												NSMutableDictionary * tLocalizationsDictionary;
												NSDictionary * tLanguageDictionary;
												BOOL tAtLeastOneLocalizedTitle=NO;
												
												tLocalizationsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
												
												while (tLanguageDictionary=[tLocalizationsEnumerator nextObject])
												{
													NSString * tLanguage;
													
													tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
													
													if (tLanguage!=nil)
													{
														NSMutableDictionary * tLanguageLocalizationsDictionary;
														NSString * tLocalization;
														
														tLanguageLocalizationsDictionary=[tLocalizationsDictionary objectForKey:tLanguage];
														
														if (tLanguageLocalizationsDictionary==nil)
														{
															tLanguageLocalizationsDictionary=[NSMutableDictionary dictionary];
															
															if (tLanguageLocalizationsDictionary!=nil)
															{
																[tLocalizationsDictionary setObject:tLanguageLocalizationsDictionary forKey:tLanguage];
															}
															else
															{
																// Low Memory
																
																tContinue=NO;
																
																break;
															}
														}
														
														tLocalization=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
													
														if (tLocalization!=nil)
														{
															if ([tLocalization length]>0)
															{
																[tLanguageLocalizationsDictionary setObject:tLocalization forKey:tString];
															
																tAtLeastOneLocalizedTitle=YES;
															}
														}
														else
														{
															// Missing Information
															
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_VALUE,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
															
															tContinue=NO;
																
															break;
														}
													}
													else
													{
														// Missing Information
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																											 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																											 nil];
														
														tContinue=NO;
														
														break;
													}
												}
												
												if (tContinue==NO)
												{
													break;
												}
												
												if (tAtLeastOneLocalizedTitle==YES)
												{
													// Set the default value
									
													NSMutableDictionary * tDefaultLanguageDictionary;
													
													tDefaultLanguageDictionary=[tLocalizationsDictionary objectForKey:ICLANGUAGE_DEFAULT_LANGUAGE];
															
													if (tDefaultLanguageDictionary==nil)
													{
														tDefaultLanguageDictionary=[NSMutableDictionary dictionary];
														
														if (tDefaultLanguageDictionary!=nil)
														{
															[tLocalizationsDictionary setObject:tDefaultLanguageDictionary forKey:ICLANGUAGE_DEFAULT_LANGUAGE];
														}
														else
														{
															// Low Memory
															
															tContinue=NO;
							
															break;
														}
													}
															
													[tDefaultLanguageDictionary setObject:tDefaultTitle forKey:tString];
												}
												else
												{
													tString=tDefaultTitle;
												}
											}
											else
											{
												// Low Memory
												
												tContinue=NO;
							
												break;
											}
										}
									
										if (tString!=nil)
										{
											tAttribute=[NSXMLNode attributeWithName:@"title" stringValue:tString];
								
											if (tAttribute!=nil)
											{
												[tChoiceElement addAttribute:tAttribute];
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'title' attribute"];
												
												tContinue=NO;
							
												break;
											}
										}
										else
										{
											// A COMPLETER
													
											tContinue=NO;
							
											break;
										}
										
										
										// description 
										
										tDefaultDescription=@"";
										
										tLocalizations=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_DESCRIPTION];
										
										if (tLocalizations==nil || [tLocalizations count]==0)
										{
											tString=tDefaultDescription;
										}
										else
										{
											// Make the description localizable
												
											NSEnumerator * tLocalizationsEnumerator;
											
											tString=[NSString stringWithFormat:@"%@_DESCRIPTION",[tChoiceReadableID uppercaseString]];
											
											tLocalizationsEnumerator=[tLocalizations objectEnumerator];
												
											if (tLocalizationsEnumerator!=nil)
											{
												NSMutableDictionary * tLocalizationsDictionary;
												NSDictionary * tLanguageDictionary;
												BOOL tAtLeastOneLocalizedDescription=NO;
												
												tLocalizationsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
												
												while (tLanguageDictionary=[tLocalizationsEnumerator nextObject])
												{
													NSString * tLanguage;
													
													tLanguage=[tLanguageDictionary objectForKey:ICDOCUMENT_LANGUAGE];
													
													if (tLanguage!=nil)
													{
														NSMutableDictionary * tLanguageLocalizationsDictionary;
														NSString * tLocalization;
														
														tLanguageLocalizationsDictionary=[tLocalizationsDictionary objectForKey:tLanguage];
														
														if (tLanguageLocalizationsDictionary==nil)
														{
															tLanguageLocalizationsDictionary=[NSMutableDictionary dictionary];
															
															if (tLanguageLocalizationsDictionary!=nil)
															{
																[tLocalizationsDictionary setObject:tLanguageLocalizationsDictionary forKey:tLanguage];
															}
															else
															{
																// Low Memory
																
																tContinue=NO;
							
																break;
															}
														}
														
														tLocalization=[tLanguageDictionary objectForKey:ICDOCUMENT_VALUE];
													
														if (tLocalization!=nil)
														{
															if ([tLocalization length]>0)
															{
																[tLanguageLocalizationsDictionary setObject:tLocalization forKey:tString];
																
																tAtLeastOneLocalizedDescription=YES;
															}
														}
														else
														{
															// Missing Information
														
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_VALUE,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
															
															tContinue=NO;
							
															break;
														}
													}
													else
													{
														// Missing Information
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																											 ICDOCUMENT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																											 nil];
														
														tContinue=NO;
							
														break;
													}
												}
												
												if (tContinue==NO)
												{
													break;
												}
												
												if (tAtLeastOneLocalizedDescription==YES)
												{
													// Set the default value
									
													NSMutableDictionary * tDefaultLanguageDictionary;
													
													tDefaultLanguageDictionary=[tLocalizationsDictionary objectForKey:ICLANGUAGE_DEFAULT_LANGUAGE];
															
													if (tDefaultLanguageDictionary==nil)
													{
														tDefaultLanguageDictionary=[NSMutableDictionary dictionary];
														
														if (tDefaultLanguageDictionary!=nil)
														{
															[tLocalizationsDictionary setObject:tDefaultLanguageDictionary forKey:ICLANGUAGE_DEFAULT_LANGUAGE];
														}
														else
														{
															// Low Memory
															
															tContinue=NO;
							
															break;
														}
													}
															
													[tDefaultLanguageDictionary setObject:tDefaultDescription forKey:tString];
												}
												else
												{
													tString=tDefaultDescription;
												}
											}
											else
											{
												// Low Memory
												
												tContinue=NO;
							
												break;
											}
										}
									
										if (tString!=nil)
										{
											tAttribute=[NSXMLNode attributeWithName:@"description" stringValue:tString];
								
											if (tAttribute!=nil)
											{
												[tChoiceElement addAttribute:tAttribute];
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'description' attribute"];
												
												tContinue=NO;
							
												break;
											}
										}
										
										[inElement addChild:tChoiceElement];
										
										continue;
									}
									else
									{
										// Missing Information
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
									}
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'id' attribute"];
								}
							}
							else
							{
								// A COMPLETER
								
								tMetadata=nil;
							}
						}
						else
						{
							// Missing Build Data
				
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES"];
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <choice> element"];
						
					}
					
					tContinue=NO;
				}
				
				if (tContinue==YES)
				{
					return YES;
				}
			}
			else
			{
				// Missing Build Data
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES"];
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																	 ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) addLinesElementsToElement:(NSXMLElement *) inElement withChoicesArray:(NSArray *) inChoicesArray isInstaller:(BOOL) inIsInstaller prefix:(NSString *) inPrefix
{
	NSDictionary * tMetadata=nil;
	
	if (inElement!=nil && inChoicesArray!=nil && inPrefix!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[inChoicesArray objectEnumerator];
	
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tChoiceDictionary;
			unsigned long tIndex=1;
			NSMutableDictionary * tChoicesNamesDictionary;
						
			tChoicesNamesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES];
			
			if (tChoicesNamesDictionary!=nil)
			{
				BOOL tContinue=YES;
				
				while (tContinue==YES && (tChoiceDictionary=[tEnumerator nextObject]))
				{
					NSXMLElement * tLineElement;
					
					tLineElement=(NSXMLElement *) [NSXMLNode elementWithName:@"line"];
					
					if (tLineElement!=nil)
					{
						NSString * tChoiceUUID;
						
						// choice
						
						tChoiceUUID=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_UUID];
						
						if (tChoiceUUID==nil)
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_UUID,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							tContinue=NO;
						
							break;
						}
						else
						{
							NSString * tChoiceReadableID;
							
							tChoiceReadableID=[NSString stringWithFormat:@"%@%lu",inPrefix,tIndex];
							
							if (tChoiceReadableID!=nil)
							{
								id tAttribute;
							
								tIndex++;
								
								[tChoicesNamesDictionary setObject:tChoiceReadableID forKey:tChoiceUUID];
								
								
								tAttribute=[NSXMLNode attributeWithName:@"choice" stringValue:tChoiceReadableID];
								
								if (tAttribute!=nil)
								{
									NSNumber * tNumber;
									
									[tLineElement addAttribute:tAttribute];
									
									tNumber=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_TYPE];
									
									if (tNumber!=nil && [tNumber intValue]==ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_TYPE_GROUP)
									{
										NSDictionary * tOptionsDictionary;
										
										tOptionsDictionary=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS];
										
										if (tOptionsDictionary==nil)
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
											
											tContinue=NO;
						
											break;
										}
										else
										{
											NSArray * tChildren;
												
											tChildren=[tChoiceDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_GROUP_CHILDREN];
												
											if (tChildren!=nil && [tChildren count]>0)
											{
												tNumber=[tOptionsDictionary objectForKey:ICDOCUMENT_PROJECT_PRESENTATION_INSTALLATION_TYPE_LIST_ITEM_OPTIONS_HIDE_CHILDREN];
									
												if (tNumber==nil || [tNumber boolValue]==NO)
												{
													if ([self addLinesElementsToElement:tLineElement withChoicesArray:tChildren isInstaller:inIsInstaller prefix:[tChoiceReadableID stringByAppendingString:@"_"]]==NO)
													{
														tContinue=NO;
						
														break;
													}
												}
											}
											else
											{
												// If it's not a hierarchy for Installer.app, we skip empty folder
												
												if (inIsInstaller==NO)
												{
													continue;
												}
											}
										}
									}
									
									[inElement addChild:tLineElement];
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'choice' attribute"];
									
									tContinue=NO;
						
									break;
								}
							}
							else
							{
								// Low Memory
								
								tContinue=NO;
						
								break;
							}
						}
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <line> element"];
						
						tContinue=NO;
						
						break;
					}
				}
			
				if (tContinue==YES)
				{
					return YES;
				}
			}
			else
			{
				// Missing Build Data
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES"];
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																	 ICPROJECTBUILDER_BUILD_INFORMATION_CHOICES_NAMES,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
											   
	return NO;
}

- (BOOL) setPackagesReferences:(NSDictionary *) inPresentationDictionary
{
	NSDictionary * tMetadata=nil;
	
	if (inPresentationDictionary!=nil)
	{
		NSMutableDictionary * tPackagesDictionary;
		NSEnumerator * tEnumerator;
		
		tPackagesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES];
				
		tEnumerator=[tPackagesDictionary objectEnumerator];
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tPackageBuildInformation;
			BOOL tContinue=YES;
			
			while (tContinue==YES && (tPackageBuildInformation=[tEnumerator nextObject]))
			{
				NSNumber * tNumber;
				
				tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCATION_TYPE];
				
				if (tNumber!=nil)
				{
					NSXMLElement * tPkgRefElement;
					int tLocationType;
					
					tLocationType=[tNumber intValue];
					
					tPkgRefElement=(NSXMLElement *) [NSXMLNode elementWithName:@"pkg-ref"];
					
					if (tPkgRefElement!=nil)
					{
						NSString * tString;
						id tAttribute;
						
						// id
						
						tString=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID];
						
						if (tString!=nil)
						{
							tAttribute=[NSXMLNode attributeWithName:@"id" stringValue:tString];
							
							if (tAttribute!=nil)
							{
								[tPkgRefElement addAttribute:tAttribute];
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'id' attribute"];
								
								tContinue=NO;
					
								break;
							}
						}
						else
						{
							// Missing Build Data
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID"];
						
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																				 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
					
							tContinue=NO;
					
							break;
						}
						
						// version
						
						tString=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_VERSION];
						
						if (tString!=nil)
						{
							tAttribute=[NSXMLNode attributeWithName:@"version" stringValue:tString];
							
							if (tAttribute!=nil)
							{
								[tPkgRefElement addAttribute:tAttribute];
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'version' attribute"];
								
								tContinue=NO;
					
								break;
							}
						}
						else
						{
							// Missing Build Data
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_VERSION"];
						
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																				 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
					
							tContinue=NO;
					
							break;
						}
						
						// path
						
						tString=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
						
						if (tString!=nil)
						{
							NSXMLNode * tNode;
							
							tNode=[NSXMLNode textWithStringValue:tString];
							
							if (tNode!=nil)
							{
								[tPkgRefElement addChild:tNode];
							}
							else
							{
								// Low Memory
								
								tContinue=NO;
					
								break;
							}
						}
						else
						{
							// Missing Build Data
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH"];
						
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																				 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
					
							tContinue=NO;
					
							break;
						}
						
						// auth
						
						tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_AUTHENTICATION];
						
						if (tNumber!=nil && [tNumber boolValue]==YES)
						{
							tAttribute=[NSXMLNode attributeWithName:@"auth" stringValue:@"Root"];
						
							if (tAttribute!=nil)
							{
								[tPkgRefElement addAttribute:tAttribute];
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'auth' attribute"];
								
								tContinue=NO;
				
								break;
							}
						}
						
						// Optional data
						
						tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE];
						
						if (tNumber!=nil)
						{
							tAttribute=[NSXMLNode attributeWithName:@"installKBytes" stringValue:[tNumber stringValue]];
									
							if (tAttribute!=nil)
							{
								[tPkgRefElement addAttribute:tAttribute];
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'installKBytes' attribute"];
								
								tContinue=NO;
				
								break;
							}
						}
						
						/*if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL ||
							tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)*/
						{
														
							// onConclusion
						
							tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_CONCLUSION_ACTION];
							
							if (tNumber!=nil)
							{
								tString=nil;
								
								switch([tNumber intValue])
								{
									case ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_REQUIRE_RESTART:
									
										tString=@"RequireRestart";
									
										break;
									
									case ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_REQUIRE_SHUTDOWN:
									
										tString=@"RequireShutdown";
									
										break;
								
									case ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_REQUIRE_LOGOUT:
									
										tString=@"RequireLogout";
									
										break;
								}
								
								if (tString!=nil)
								{
									tAttribute=[NSXMLNode attributeWithName:@"onConclusion" stringValue:tString];
							
									if (tAttribute!=nil)
									{
										[tPkgRefElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'onConclusion' attribute"];
										
										tContinue=NO;
					
										break;
									}
								}
							}
							else
							{
								// Missing Build Data
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_CONCLUSION_ACTION"];
						
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																		 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_CONCLUSION_ACTION,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
					
								tContinue=NO;
					
								break;
							}
						}
							
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <pkg-ref> element"];
						
						tContinue=NO;
					
						break;
					}
					
					[installerScriptElement_ addChild:tPkgRefElement];
				}
				else
				{
					// Missing Build Data
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_LOCATION_TYPE"];
						
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																		 ICPROJECTBUILDER_BUILD_INFORMATION_LOCATION_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
					
					tContinue=NO;
					
					break;
				}
			}
			
			if (tContinue==YES)
			{
				return YES;
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) setPresentation
{
	NSDictionary * tProjectPresentation;
	id tComment;
	
	tProjectPresentation=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_PRESENTATION];
		
	tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n        |       Presentation       |\n        +==========================+"];
	
	if (tComment!=nil)
	{
		[installerScriptElement_ addChild:tComment];
	}
	
	// Title
	
	if ([self setTitle:tProjectPresentation]==YES)
	{
		// Background
	
		if ([self setBackground:tProjectPresentation]==YES)
		{
			// Introduction
	
			if ([self setIntroduction:tProjectPresentation]==YES)
			{
				// ReadMe
	
				if ([self setReadMe:tProjectPresentation]==YES)
				{
					// License
	
					if ([self setLicense:tProjectPresentation]==YES)
					{
						// Conclusion
	
						if ([self setConclusion:tProjectPresentation]==YES)
						{
							// Package Hierarchy
	
							if ([self setChoiceOutline:tProjectPresentation]==YES)
							{
								// Packages References
	
								tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n        |    Package References    |\n        +==========================+"];
								
								if (tComment!=nil)
								{
									[installerScriptElement_ addChild:tComment];
								}
								
								if ([self setPackagesReferences:tProjectPresentation]==YES)
								{
									return YES;
								}
							}
						}
					}
				}
			}
		}
	}
	
	return NO;
}

- (BOOL) setJavaScriptScripts
{
	NSMutableArray * tJavaScriptFunctionsArray;
	
	tJavaScriptFunctionsArray=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_FUNCTIONS_ARRAY];
	
	if ([tJavaScriptFunctionsArray count]>0)
	{
		NSEnumerator * tEnumerator;
		NSDictionary * tMetadata=nil;
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_JAVASCRIPT
												   metadata:nil];
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		tEnumerator=[tJavaScriptFunctionsArray objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSXMLElement * tScriptElement;
			
			tScriptElement=(NSXMLElement *) [NSXMLNode elementWithName:@"script"];
						
			if (tScriptElement!=nil)
			{
				NSMutableString * tMutableString;
				
				tMutableString=[NSMutableString stringWithString:@"\n\n"];
				
				if (tMutableString!=nil)
				{
					id tNode;
					NSString * tFunctionCode;
					NSMutableString * tJavascriptConstantsString;
					
					// Constants
				
					tJavascriptConstantsString=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_JAVASCRIPT_CONSTANTS_STRING];
					
					if ([tJavascriptConstantsString length]>0)
					{
						[tMutableString appendFormat:@"%@",tJavascriptConstantsString];
					}
					
					// Functions
					
					while (tFunctionCode=[tEnumerator nextObject])
					{
						[tMutableString appendFormat:@"%@\n\n",tFunctionCode];
					}
				
					
					// To re-align the </script> tag
					
					[tMutableString appendString:@"    "];
					
					tNode=[NSXMLNode textWithStringValue:tMutableString];
					
					if (tNode!=nil)
					{
						[tScriptElement addChild:tNode];
						
						[installerScriptElement_ addChild:tScriptElement];
						
						[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																	forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_JAVASCRIPT
																   metadata:nil];
												   
						return YES;
					}
				}
			}
			else
			{
				// Low Memory
				
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create <scripts> element"];
			}
		}
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
											   
		return NO;
	}

	return YES;
}

- (BOOL) setLocalizableStrings
{
	NSDictionary * tLocalizationsDictionary;
	NSEnumerator * tLanguageEnumerator;
	NSDictionary * tMetadata=nil;
	
	tLocalizationsDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCALIZATIONS];
	
	tLanguageEnumerator=[tLocalizationsDictionary keyEnumerator];
	
	if (tLanguageEnumerator!=nil)
	{
		NSMutableDictionary * tDefaultLanguageDictionary;
								
		// Get the default values and the list of keys
		
		tDefaultLanguageDictionary=[tLocalizationsDictionary objectForKey:ICLANGUAGE_DEFAULT_LANGUAGE];
		
		if (tDefaultLanguageDictionary==nil)
		{
			// Missing Build Data
			
			[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICLANGUAGE_DEFAULT_LANGUAGE"];
						
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																 ICLANGUAGE_DEFAULT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
		else
		{
			NSString * tLanguage;
			BOOL tContinue=YES;
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			
			while (tContinue==YES && (tLanguage=[tLanguageEnumerator nextObject]))
			{
				if ([tLanguage isEqualToString:ICLANGUAGE_DEFAULT_LANGUAGE]==NO)
				{
					NSString * tLanguagePath;
												
					tLanguagePath=[self getDistributionPathForLanguage:tLanguage];
					
					if (tLanguagePath!=nil)
					{
						NSDictionary * tLocalizedStringsDictionary;
						
						tLocalizedStringsDictionary=[tLocalizationsDictionary objectForKey:tLanguage];
						
						if (tLocalizedStringsDictionary!=nil)
						{
							NSMutableString * tMutableString;
						
							tMutableString=[NSMutableString string];
							
							if (tMutableString!=nil)
							{
								NSEnumerator * tKeyEnumerator;
			
								tKeyEnumerator=[tDefaultLanguageDictionary keyEnumerator];
										
								if (tKeyEnumerator!=nil)
								{
									NSString * tKey;
									NSError * tError;
								
									while (tContinue==YES && (tKey=[tKeyEnumerator nextObject]))
									{
										NSString * tLocalizedString;
										
										tLocalizedString=[tLocalizedStringsDictionary objectForKey:tKey];
										
										if (tLocalizedString==nil)
										{
											tLocalizedString=[tDefaultLanguageDictionary objectForKey:tKey];
										}
										
										if (tLocalizedString==nil)
										{
											// Missing Build Data
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: %@",tKey];
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																								 ICLANGUAGE_DEFAULT_LANGUAGE,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
											
											tContinue=NO;
											
											break;
										}
										else
										{
											NSMutableString * tMutableLocalizedString;
										
											tMutableLocalizedString=[tLocalizedString mutableCopy];
										
											if (tMutableLocalizedString!=nil)
											{
												[tMutableLocalizedString replaceOccurrencesOfString:@"\\" withString:@"\\\\" options:0 range:NSMakeRange(0,[tMutableLocalizedString length])];
												
												[tMutableLocalizedString replaceOccurrencesOfString:@"\"" withString:@"\\\"" options:0 range:NSMakeRange(0,[tMutableLocalizedString length])];
												
												[tMutableLocalizedString replaceOccurrencesOfString:@"\n" withString:@"\\n" options:0 range:NSMakeRange(0,[tMutableLocalizedString length])];
												
												[tMutableLocalizedString replaceOccurrencesOfString:@"\r" withString:@"\\r" options:0 range:NSMakeRange(0,[tMutableLocalizedString length])];
												
												[tMutableLocalizedString replaceOccurrencesOfString:@"\t" withString:@"\\t" options:0 range:NSMakeRange(0,[tMutableLocalizedString length])];
												
												[tMutableString appendString:[NSString stringWithFormat:@"\"%@\" = \"%@\";\n\n",tKey,tMutableLocalizedString]];
												
												[tMutableLocalizedString release];
											}
											else
											{
												// Low Memory
												
												tContinue=NO;
											
												break;
											}
										}
									}
									
									if (tContinue==YES)
									{
										if ([tMutableString writeToFile:[tLanguagePath stringByAppendingPathComponent:@"Localizable.strings"] atomically:NO encoding:NSUnicodeStringEncoding error:&tError]==NO)
										{
											// A COMPLETER
									
											[tError release];
									
											tContinue=NO;
											
											break;
										}
									}
								}
								else
								{
									// Low Memory
									
									tContinue=NO;
											
									break;
								}
							}
							else
							{
								// Low Memory
								
								tContinue=NO;
											
								break;
							}
						}
						else
						{
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,nil];
							
							tContinue=NO;
											
							break;
						}
					}
					else
					{
						tMetadata=nil;
						
						tContinue=NO;
											
						break;
					}
				}
			}
			
			if (tContinue==YES)
			{
				return YES;
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) setPlugins:(NSString *) inPath
{
	NSDictionary * tMetadata=nil;
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
												forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLER_PLUGINS
											   metadata:nil];
	
	if (inPath!=nil)
	{
		NSDictionary * tProjectPresentationDictionary;
		
		tProjectPresentationDictionary=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_PRESENTATION];
		
		if (tProjectPresentationDictionary!=nil)
		{
			NSArray * tInstallationSteps;
			
			tInstallationSteps=[tProjectPresentationDictionary objectForKey:ICDOCUMENT_PROJECT_INSTALLATION_STEPS];
		
			if (tInstallationSteps!=nil)
			{
				NSEnumerator * tEnumerator;
				NSMutableArray * tSectionOrdersMutableArray;
				
				tEnumerator=[tInstallationSteps objectEnumerator];
				
				tSectionOrdersMutableArray=[NSMutableArray array];
				
				if (tEnumerator!=nil && tSectionOrdersMutableArray!=nil)
				{
					NSDictionary * tStepDictionary;
					BOOL hasCustomPlugin=NO;
					BOOL tDirectoryCreated=NO;
					
					while (tStepDictionary=[tEnumerator nextObject])
					{
						NSString * tClassName;
						
						tClassName=[tStepDictionary objectForKey:ICPRESENTATION_CHAPTER_VIEW_CONTROLLER_CLASS];
						
						if ([tClassName isEqualToString:@"ICPresentationViewIntroductionController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"Introduction"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewReadMeController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"ReadMe"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewLicenseController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"License"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewDestinationSelectController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"Target"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewInstallationTypeController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"PackageSelection"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewInstallationController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"Install"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewSummaryController"]==YES)
						{
							[tSectionOrdersMutableArray addObject:@"FinishUp"];
						}
						else if ([tClassName isEqualToString:@"ICPresentationViewInstallerPluginController"]==YES)
						{
							NSDictionary * tPathDictionary;
							
							hasCustomPlugin=YES;
							
							tPathDictionary=[tStepDictionary objectForKey:ICPRESENTATION_CHAPTER_VIEW_CONTROLLER_INSTALLERPLUGIN_PATH];
				
							if (tPathDictionary!=nil)
							{
								NSString * tPath;
								
								tPath=[self pathFromPathDictionary:tPathDictionary];
								
								if (tPath!=nil)
								{
									BOOL isDirectory;
									
									if ([fileManager_ fileExistsAtPath:tPath isDirectory:&isDirectory]==YES && isDirectory==YES)
									{
										NSString * tBundleName;
										NSString * tDestinationPath;
										
										tDestinationPath=[inPath stringByAppendingPathComponent:@"Plugins"];
										
										tBundleName=[tPath lastPathComponent];
										
										// Build the Plugins folder if needed
										
										if (tDirectoryCreated==NO)
										{
											if ([self createDirectoryAtPath:tDestinationPath]==NO)
											{
												tMetadata=nil;
												
												goto setPluginsBail;
											}
											else
											{
												tDirectoryCreated=YES;
											}
										}
										
										// Copy the plugin to the Plugins folder
										
										if ([ICFileSystemUtilities copyPath:tPath toPath:tDestinationPath]==NO)
										{
											goto setPluginsBail;
										}
										
										[tSectionOrdersMutableArray addObject:tBundleName];
									}
									else
									{
										// A COMPLETER
										
										goto setPluginsBail;
									}
								}
								else
								{
									goto setPluginsBail;
								}
							}
							else
							{
								// Missing Information
				
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICPRESENTATION_CHAPTER_VIEW_CONTROLLER_INSTALLERPLUGIN_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							
								goto setPluginsBail;
							}
						}
					}
					
					if (hasCustomPlugin==YES)
					{
						NSDictionary * tDictionary;
						
						tDictionary=[NSDictionary dictionaryWithObjectsAndKeys:tSectionOrdersMutableArray,@"SectionOrder",nil];
						
						if (tDictionary!=nil)
						{
							NSString * tPlistPath;
							
							tPlistPath=[inPath stringByAppendingPathComponent:@"Plugins/InstallerSections.plist"];
						
							if ([tDictionary writeToFile:tPlistPath atomically:NO]==YES)
							{
								//[ICFileSystemUtilities setFilePrivileges:S_IRUSR+S_IWUSR+S_IRGRP+S_IROTH atPath:tPlistPath];		// A COMPLETER
								
								[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																			forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLER_PLUGINS
																		   metadata:nil];
								
								return YES;
							}
							else
							{
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																					 tPlistPath,IC_BUILDER_INFORMATION_FILE_PATH,
																					 nil];
							}
						}
						else
						{
							// Low Memory
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
						}
					}
					else
					{
						[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																	forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLER_PLUGINS
																   metadata:nil];
						
						return YES;
					}
				}
				else
				{
					// Low Memory
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				}
			}
			else
			{
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
															forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLER_PLUGINS
														   metadata:nil];
				
				return YES;
			}
		}
	}

setPluginsBail:
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_INSTALLER_PLUGINS
											   metadata:tMetadata];
	
	return NO;
}

+ (NSString *) shortFolderNameForEnglishLanguage:(NSString *) inEnglishLanguage
{
	if (inEnglishLanguage!=nil)
	{
		static NSDictionary * sFoldersNamesDictionary=nil;
	
		if (sFoldersNamesDictionary==nil)
		{
			sFoldersNamesDictionary=[NSDictionary dictionaryWithObjectsAndKeys:@"da.lproj",@"Danish",
																			   @"du.lproj",@"Dutch",
																			   @"en.lproj",@"English",
																			   @"fi.lproj",@"Finnish",
																			   @"fr.lproj",@"French",
																			   @"de.lproj",@"German",
																			   @"it.lproj",@"Italian",
																			   @"ja.lproj",@"Japanese",
																			   @"ko.lproj",@"Korean",
																			   @"no.lproj",@"Norwegian",
																			   @"pl.lproj",@"Polish",
																			   @"pt.lproj",@"Brazilian Portuguese",
																			   @"pt_PT.lproj",@"Portuguese",
																			   @"ru.lproj",@"Russian",
																			   @"es.lproj",@"Spanish",
																			   @"sv.lproj",@"Swedish",
																			   @"zh_CN.lproj",@"Simplified Chinese",
																			   @"zh_TW.lproj",@"Traditional Chinese",
																			   nil];
																			   
																			   
																			   
		}
	
		return [sFoldersNamesDictionary objectForKey:inEnglishLanguage];
	}
	
	return nil;
}

+ (NSString *) folderNameForEnglishLanguage:(NSString *) inEnglishLanguage
{
	if (inEnglishLanguage!=nil)
	{
		static NSDictionary * sFoldersNamesDictionary=nil;
	
		if (sFoldersNamesDictionary==nil)
		{
			sFoldersNamesDictionary=[NSDictionary dictionaryWithObjectsAndKeys:@"Danish.lproj",@"Danish",
																			   @"Dutch.lproj",@"Dutch",
																			   @"English.lproj",@"English",
																			   @"Finnish.lproj",@"Finnish",
																			   @"French.lproj",@"French",
																			   @"German.lproj",@"German",
																			   @"Italian.lproj",@"Italian",
																			   @"Japanese.lproj",@"Japanese",
																			   @"Korean.lproj",@"Korean",
																			   @"Norwegian.lproj",@"Norwegian",
																			   @"Polish.lproj",@"Polish",
																			   @"pt.lproj",@"Brazilian Portuguese",
																			   @"pt_PT.lproj",@"Portuguese",
																			   @"Russian.lproj",@"Russian",
																			   @"Spanish.lproj",@"Spanish",
																			   @"Swedish.lproj",@"Swedish",
																			   @"zh_CN.lproj",@"Simplified Chinese",
																			   @"zh_TW.lproj",@"Traditional Chinese",
																			   nil];
																			   
																			   
																			   
		}
	
		return [sFoldersNamesDictionary objectForKey:inEnglishLanguage];
	}
	
	return nil;
}

- (NSString *) getDistributionPathForLanguage:(NSString *) inLanguage
{
	NSString * tLanguagePath;
	NSString * tLanguagePathKey;
	
	tLanguagePathKey=[NSString stringWithFormat:@"_%@",inLanguage];
	
	tLanguagePath=[buildInformation_ objectForKey:tLanguagePathKey];
	
	if (tLanguagePath==nil)
	{
		NSString * tResourcesPath;
		
		tResourcesPath=[self getDistributionResourcesPath];
		
		if (tResourcesPath!=nil)
		{
			NSString * tFolderName;
			
			if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
			{
				tFolderName=[ICProjectBuilder shortFolderNameForEnglishLanguage:inLanguage];
			}
			else if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
			{
				tFolderName=[ICProjectBuilder folderNameForEnglishLanguage:inLanguage];
			}
			else
			{
				tFolderName=nil;
			}
			
			if (tFolderName!=nil)
			{
				tLanguagePath=[tResourcesPath stringByAppendingPathComponent:tFolderName];
				
				if (tLanguagePath!=nil)
				{
					if ([self createDirectoryAtPath:tLanguagePath]==YES)
					{
						[buildInformation_ setObject:tLanguagePath forKey:tLanguagePathKey];
					}
					else
					{
						return nil;
					}
				}
			}
		}
	}
	
	return tLanguagePath;
}

- (NSString *) getDistributionScriptsPath
{
	if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_BUNDLE)
	{
		return [self getDistributionResourcesPath];
	}
	else if (buildFormat_==ICDOCUMENT_PROJECT_SETTINGS_BUILD_FORMAT_FLAT)
	{
		NSString * tDistributionScriptsPath;
		
		tDistributionScriptsPath=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_SCRIPTS_PATH];
	
		if (tDistributionScriptsPath==nil)
		{
			tDistributionScriptsPath=[scratchLocation_ stringByAppendingPathComponent:@"_dist_scripts_"];
								
			if ([fileManager_ fileExistsAtPath:tDistributionScriptsPath]==NO)
			{
				// Try to create the directory
				
				if ([self createDirectoryAtPath:tDistributionScriptsPath]==YES)
				{
					[buildInformation_ setObject:tDistributionScriptsPath forKey:ICPROJECTBUILDER_BUILD_INFORMATION_SCRIPTS_PATH];
				}
				else
				{
					tDistributionScriptsPath=nil; 
				}
			}
			else
			{
				return nil;
			}
		}
		
		return tDistributionScriptsPath;
	}
	
	return nil;
}

- (NSString *) getDistributionResourcesPath
{
	NSString * tResourcesPath;
	
	tResourcesPath=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_RESOURCES_PATH];
	
	if (tResourcesPath==nil)
	{
		NSString * tContentsPath;
			
		tContentsPath=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_CONTENTS_PATH];
		
		if (tContentsPath!=nil)
		{
			tResourcesPath=[tContentsPath stringByAppendingPathComponent:@"Resources"];
			
			if (tResourcesPath!=nil)
			{
				if ([self createDirectoryAtPath:tResourcesPath]==YES)
				{
					[buildInformation_ setObject:tResourcesPath forKey:ICPROJECTBUILDER_BUILD_INFORMATION_RESOURCES_PATH];
				}
				else
				{
					return nil;
				}
			}
		}
	}
	
	return tResourcesPath;
}

#pragma mark -

- (BOOL) prepareExceptionRules
{
	NSDictionary * tMetadata=nil;
	NSDictionary * tProjectSettingsDictionary;
	
	tProjectSettingsDictionary=[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS];
	
	if (tProjectSettingsDictionary!=nil)
	{
		NSMutableArray * tMutableArray;
		NSNumber * tNumber;
		
		tNumber=[tProjectSettingsDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES_PAYLOAD_ONLY];
		
		if (tNumber!=nil)
		{
			[options_ setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_PAYLOAD_ONLY];
		}
		
		tMutableArray=[tProjectSettingsDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES];
		
		if (tMutableArray!=nil)
		{
			NSMutableDictionary * tMutableDictionary;
			
			tMutableDictionary=[NSMutableDictionary dictionary];
			
			if (tMutableDictionary!=nil)
			{
				NSEnumerator * tEnumerator;
				NSMutableArray * tFileStaticExceptions;
				NSMutableArray * tFolderStaticExceptions;
				NSMutableArray * tFileRegExExceptions;
				NSMutableArray * tFolderRegExExceptions;
				
				tFileStaticExceptions=[NSMutableArray array];
				
				tFolderStaticExceptions=[NSMutableArray array];
				
				tFileRegExExceptions=[NSMutableArray array];
				
				tFolderRegExExceptions=[NSMutableArray array];
				
				tEnumerator=[tMutableArray objectEnumerator];
				
				if (tEnumerator!=nil)
				{
					NSDictionary * tExceptionDictionary;
					
					while (tExceptionDictionary=[tEnumerator nextObject])
					{
						
						
						tNumber=[tExceptionDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES_PATTERNS_STATE];
						
						if (tNumber!=nil && [tNumber boolValue]==YES)
						{
							NSArray * tPatternArray;
							
							tPatternArray=[tExceptionDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES_PATTERNS_ARRAY];
							
							if (tPatternArray!=nil)
							{
								NSEnumerator * tPatternEnumerator;
								
								tPatternEnumerator=[tPatternArray objectEnumerator];
								
								if (tPatternEnumerator!=nil)
								{
									NSDictionary * tPatternDictionary;
									
									while (tPatternDictionary=[tPatternEnumerator nextObject])
									{
										NSString * tPattern;
											
										tPattern=[tPatternDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES_PATTERN_ITEM_STRING];
										
										if (tPattern!=nil && [tPattern length]>1)
										{
											unsigned int tType;
											
											tNumber=[tPatternDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES_PATTERN_ITEM_FILE_TYPE];
											
											tType=[tNumber unsignedIntValue];
											
											tNumber=[tPatternDictionary objectForKey:ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES_PATTERN_ITEM_REGULAR_EXPRESSION];
											
											if ([tNumber boolValue]==NO)
											{
												if (tType==ICFILE_TYPE_FILE)
												{
													[tFileStaticExceptions addObject:tPattern];
												}
												else if (tType==ICFILE_TYPE_FOLDER)
												{
													[tFolderStaticExceptions addObject:tPattern];
												}
												else if (tType==ICFILE_TYPE_FILE_OR_FOLDER)
												{
													[tFileStaticExceptions addObject:tPattern];
													
													[tFolderStaticExceptions addObject:tPattern];
												}
											}
											else
											{
												if (tType==ICFILE_TYPE_FILE)
												{
													[tFileRegExExceptions addObject:tPattern];
												}
												else if (tType==ICFILE_TYPE_FOLDER)
												{
													[tFolderRegExExceptions addObject:tPattern];
												}
												else if (tType==ICFILE_TYPE_FILE_OR_FOLDER)
												{
													[tFileRegExExceptions addObject:tPattern];
													
													[tFolderRegExExceptions addObject:tPattern];
												}
											}
										}
									}
								}
							}
						}
					}
					
					if ([tFileStaticExceptions count]>0)
					{
						[tMutableDictionary setObject:tFileStaticExceptions forKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_STANDARD_FILES];
					}
					
					if ([tFolderStaticExceptions count]>0)
					{
						[tMutableDictionary setObject:tFolderStaticExceptions forKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_STANDARD_FOLDERS];
					}
					
					if ([tFileRegExExceptions count]>0)
					{
						[tMutableDictionary setObject:tFileRegExExceptions forKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_REGEX_FILES];
					}
					
					if ([tFolderRegExExceptions count]>0)
					{
						[tMutableDictionary setObject:tFolderRegExExceptions forKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_REGEX_FOLDERS];
					}
					
					if ([tMutableDictionary count]>0)
					{
						[options_ setObject:tMutableDictionary forKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS];
					}
					
					return YES;
				}
				else
				{
					// Low Memory
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				}
			}
			else
			{
				// Low Memory
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PROJECT_SETTINGS_EXCLUDED_FILES,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
	else
	{
		// Missing Information
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
															 ICDOCUMENT_PROJECT_SETTINGS,IC_BUILDER_FAILURE_TAG_NAME,
															 nil];
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

#pragma mark -

- (BOOL) setCorrectXarPermissionsForFolder:(NSString *) inPath
{
	NSDictionary * tMetadata=nil;
	
	if (inPath!=nil)
	{
		FTS * ftsp;
		FTSENT * tFile;
		 char * tPath[2]={(char *) [inPath fileSystemRepresentation],NULL};
		
		if ((ftsp = fts_open(tPath, FTS_PHYSICAL, 0)) == NULL)
		{
			return NO;
		}

		while ((tFile = fts_read(ftsp)) != NULL)
		{
			mode_t tMode;
			
			switch (tFile->fts_info)
			{
				case FTS_D:
					tMode=S_IRWXU+S_IRGRP+S_IXGRP+S_IROTH+S_IXOTH;	// 0755
					
					break;
				case FTS_DP:
				case FTS_SL:
				case FTS_SLNONE:
						continue;
				case FTS_DNR:
				case FTS_ERR:
				case FTS_NS:
						fts_close(ftsp);
						
						return NO;
				default:				// At least 0644
				
					tMode=0;
					
					if ((tFile->fts_statp->st_mode & 0700) < (S_IRUSR+S_IWUSR))
					{
						tMode=S_IRUSR+S_IWUSR;
					}
					else
					{
						tMode=(tFile->fts_statp->st_mode & 0700);
					}
					
					if ((tFile->fts_statp->st_mode & 0070) < S_IRGRP)
					{
						tMode+=S_IRGRP;
					}
					else
					{
						tMode+=(tFile->fts_statp->st_mode & 0070);
					}
					
					if ((tFile->fts_statp->st_mode & 0007) < S_IROTH)
					{
						tMode+=S_IROTH;
					}
					else
					{
						tMode+=(tFile->fts_statp->st_mode & 0007);
					}
					
					break;
			}
			
			// Set Permissions
			
			if (tMode != tFile->fts_statp->st_mode)
			{
				if (chmod(tFile->fts_accpath, tMode) == -1)
				{
					switch(errno)
					{
						case EPERM:
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																				 [NSString stringWithUTF8String:tFile->fts_accpath],IC_BUILDER_INFORMATION_FILE_PATH,
																				 nil];
							break;
						
						case ENOENT:
						
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																				 [NSString stringWithUTF8String:tFile->fts_accpath],IC_BUILDER_INFORMATION_FILE_PATH,
																				 nil];
							break;
							
						case ENAMETOOLONG:
						
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Path(%@) is too long",inPath];
							
							break;
					}
				}
			}
			
			// Set Owner:Group
			
#ifdef __SET_CORRECT_PERMISSIONS__
			
			if (chown(tFile->fts_accpath, 0, 0) == -1)
			{
				switch(errno)
				{
					case EPERM:
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																			 [NSString stringWithUTF8String:tFile->fts_accpath],IC_BUILDER_INFORMATION_FILE_PATH,
																			 nil];
						break;
					
					case ENOENT:
					
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																			 [NSString stringWithUTF8String:tFile->fts_accpath],IC_BUILDER_INFORMATION_FILE_PATH,
																			 nil];
						break;
						
					case ENAMETOOLONG:
					
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Path(%@) is too long",inPath];
						
						break;
				}
				
				break;
			}
#endif			
		}
		
		fts_close(ftsp);
		
		if (errno==0)
		{
			return YES;
		}
    }
	
    [buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;	
}

#define SIGNATURE_REQUEST_TIME_OUT		10*60.0f // 10 minutes

- (void) _signatureTimedOut:(id) sender
{
	signatureResult_=IC_BUILDER_FAILURE_REASON_SIGNATURE_TIME_OUT;
	
	CFRunLoopStop(CFRunLoopGetCurrent());
}

- (CFDataRef) processSignatureMessagePortMessage:(SInt32) inMessageID withData:(CFDataRef) inData
{
	[NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(_signatureTimedOut:) object:nil];
	
	switch(inMessageID)
	{
		case IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_SUCCESS:
			
			signedData_=[(NSData *) inData copy];
			
			break;
			
		case IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_AUTH_DENIED:
		
			signatureResult_=IC_BUILDER_FAILURE_REASON_SIGNATURE_AUTH_DENIED;
			
			break;
		
		case IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA_CERTIFICATE_NOT_FOUND:
		
			signatureResult_=IC_BUILDER_FAILURE_REASON_SIGNATURE_CERTIFICATE_NOT_FOUND;
		
			break;
		
		default:
			
			signatureResult_=IC_BUILDER_FAILURE_REASON_SIGNATURE_GENERIC_ERROR;
			
			break;
	}
	
	CFRunLoopStop(CFRunLoopGetCurrent());
	
	return NULL;
}


- (void) delayedSignatureRequest:(NSDictionary *) inDictionary
{
	[buildNotificationCenter_ postNotificationRequest:IC_BUILDER_NOTIFICATION_REQUEST_SIGN_DATA
											 metadata:inDictionary];
}

static int32_t signatureCallback(xar_signature_t inSignature, void *context, uint8_t * inData, uint32_t inLength, uint8_t **signed_data, uint32_t *signed_len)
{
	int32_t tResult=-1;
	NSString * tUniqueMessageID;
	
	tUniqueMessageID=[ICUUIDFactory string];
	
	if (tUniqueMessageID!=nil)
	{
		CFMessagePortRef tLocalMessagePort;
		Boolean shouldFreeInfo=FALSE;
		ICProjectBuilder * tProjectBuilderInstance;
	
		tProjectBuilderInstance=(ICProjectBuilder *) context;
		
		tProjectBuilderInstance->messagePortContext_.version=0;
		tProjectBuilderInstance->messagePortContext_.info=(void *) tProjectBuilderInstance;
		tProjectBuilderInstance->messagePortContext_.retain=NULL;
		tProjectBuilderInstance->messagePortContext_.release=NULL;
		tProjectBuilderInstance->messagePortContext_.copyDescription=NULL;
		
		tLocalMessagePort =CFMessagePortCreateLocal(kCFAllocatorDefault, (CFStringRef) tUniqueMessageID,SignatureMessagePortCallBack, &tProjectBuilderInstance->messagePortContext_, &shouldFreeInfo);

		if (tLocalMessagePort!=NULL)
		{
			CFRunLoopSourceRef tRunLoopSource;
			
			tRunLoopSource =CFMessagePortCreateRunLoopSource(kCFAllocatorDefault, tLocalMessagePort,0);
			
			if (tRunLoopSource!=NULL)
			{
				NSData * tDataToSign=nil;
				
				CFRunLoopAddSource(CFRunLoopGetCurrent(), tRunLoopSource,kCFRunLoopDefaultMode);
				
				// Add a timeout
		
				[tProjectBuilderInstance performSelector:@selector(_signatureTimedOut:) withObject:nil afterDelay:SIGNATURE_REQUEST_TIME_OUT];
				
				// Post Distributed Notification
				
				tDataToSign=[NSData dataWithBytes:inData length:inLength];
				
				if (tDataToSign!=nil)
				{
					[tProjectBuilderInstance performSelector:@selector(delayedSignatureRequest:)
												  withObject:[NSDictionary dictionaryWithObjectsAndKeys:tDataToSign,IC_BUILDER_REQUEST_DATA,
																										tUniqueMessageID,IC_BUILDER_REQUEST_MESSAGEPORT_NAME,
																										tProjectBuilderInstance->certificateDictionary_,IC_BUILDER_REQUEST_CERTIFICATE,
																										nil]
												  afterDelay:0.01f];
				
					// Run the loop
							
					CFRunLoopRun();
					
					CFRunLoopRemoveSource(CFRunLoopGetCurrent(), tRunLoopSource,kCFRunLoopDefaultMode);
					
					// Release Memory
					
					CFRelease(tRunLoopSource);
					
					CFRelease(tLocalMessagePort);
					
					if (tProjectBuilderInstance->signedData_!=nil)
					{
						*signed_len=[tProjectBuilderInstance->signedData_ length];
						
						*signed_data = (uint8_t *) malloc(*signed_len);

						memcpy(*signed_data,[tProjectBuilderInstance->signedData_ bytes], *signed_len);
						
						[tProjectBuilderInstance->signedData_ release];
						
						tProjectBuilderInstance->signedData_=nil;
						
						tResult=0;
					}
				}
			}
		}
	}
	
	return tResult;
}

- (BOOL) unarchiveXarFileAtPath:(NSString *) inPath toFolder:(NSString *) inFolderPath
{
	NSDictionary * tMetadata=nil;
	
	if (inPath!=nil && inFolderPath!=nil)
	{
		// Check if the file exists
		
		if ([fileManager_ fileExistsAtPath:inPath]==YES)
		{
			xar_t tXarArchive;
				
			tXarArchive=xar_open([inPath fileSystemRepresentation],READ);
				
			if (tXarArchive!=NULL)
			{
				int32_t tError;
				xar_iter_t tIterator;
				
				tIterator=xar_iter_new();
				
				if (tIterator!=NULL)
				{
					xar_file_t tFile;
					
					for (tFile=xar_file_first(tXarArchive,tIterator);tFile!=NULL;tFile=xar_file_next(tIterator))
					{
						NSString * tFileRelativePath;
							
						tFileRelativePath = [NSString stringWithUTF8String:xar_get_path(tFile)];
					
						tError=xar_extract_tofile(tXarArchive,tFile,[[inFolderPath stringByAppendingPathComponent:tFileRelativePath] fileSystemRepresentation]);
						
						if (tError!=0)
						{
							xar_close(tXarArchive);
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																				 [inFolderPath stringByAppendingPathComponent:tFileRelativePath],IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							goto unarchiveXarFileAtPathBail;
						}
					}
						
					tError=xar_close(tXarArchive);
					
					if (tError==0)
					{
						return YES;
					}
				}
				else
				{
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
				}
			}
			else
			{
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_OPENED],IC_BUILDER_FAILURE_REASON,
																	 inPath,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
		else
		{
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																 inPath,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
																	 
		}
	}
	else
	{
	}

unarchiveXarFileAtPathBail:
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (BOOL) createXarFileNamed:(NSString *) inFileName atPath:(NSString *) inPath withContentsOfFolder:(NSString *) inFolderPath
{
	NSDictionary * tMetadata=nil;
	
	signatureResult_=0;
	
	if (inFileName!=nil && inFolderPath!=nil && inPath!=nil)
	{
		if ([inFileName length]>0)
		{
			NSString * tFilePath;
			
			tFilePath=[inPath stringByAppendingPathComponent:inFileName];
			
			if (tFilePath!=nil)
			{
				xar_t nXarFile;
				
				// Check if the file already exists
				
				if ([fileManager_ fileExistsAtPath:tFilePath]==YES)
				{
					[fileManager_ removeFileAtPath:tFilePath handler:nil];
				}
				
				nXarFile=xar_open([tFilePath fileSystemRepresentation],WRITE);
				
				if (nXarFile!=NULL)
				{
					NSDirectoryEnumerator * tDirectoryEnumerator;
					
					if ([fileManager_ changeCurrentDirectoryPath:inFolderPath]==NO)
					{
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to change current directory to \'%@\'",inFolderPath];
							
						xar_close(nXarFile);
						
						return NO;
					}
					
					tDirectoryEnumerator=[fileManager_ enumeratorAtPath:inFolderPath];
					
					if (tDirectoryEnumerator!=nil)
					{
						NSString * tItemPath;
						
						if (identity_!=NULL)
						{
							SecKeyRef tPrivateKeyRef;
							const CSSM_KEY * tPtr;
							SecCertificateRef tCertificateRef;
							OSStatus tStatus;
							char * tSignatureType=NULL;
							CSSM_DATA_PTR tDataPtr;
							int32_t tSignatureSize;
							
							tStatus=SecIdentityCopyPrivateKey(identity_,&tPrivateKeyRef);
							
							if (tStatus!=noErr)
							{
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to copy identity private key"];
							
								xar_close(nXarFile);
								
								return NO;
							}
							
							tStatus=SecKeyGetCSSMKey(tPrivateKeyRef,&tPtr);
							
							if (tStatus!=noErr)
							{
								// Release Memory
							
								CFRelease(tPrivateKeyRef);
							
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to get CSSM key"];
								
								xar_close(nXarFile);
								
								return NO;
							}
							
							// Size
							
							tSignatureSize=(tPtr->KeyHeader.LogicalKeySizeInBits)>>3;
							
							// Algorithm
							
							switch(tPtr->KeyHeader.AlgorithmId)
							{
								case CSSM_ALGID_RSA:
								
									tSignatureType="RSA";
								
									break;
									
								case CSSM_ALGID_SHA1WithRSA:
								
									tSignatureType="Sha1WithRSA";
									
									break;
							}
							
							// Release Memory
							
							CFRelease(tPrivateKeyRef);
							
							
							if (tSignatureType==NULL)
							{
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unsupported algorithm"];
								
								xar_close(nXarFile);
								
								return NO;
							}
							
							tStatus=SecIdentityCopyCertificate(identity_,&tCertificateRef);
							
							if (tStatus!=noErr)
							{
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to copy certificate"];
								
								xar_close(nXarFile);
								
								return NO;
							}
							
							tDataPtr=(CSSM_DATA_PTR) malloc(sizeof(CSSM_DATA));
							
							if (tDataPtr!=NULL)
							{
								xar_signature_t tSignature;
								
								tStatus=SecCertificateGetData(tCertificateRef,tDataPtr);
								
								if (tStatus!=noErr)
								{
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to get certificate data"];
								
									xar_close(nXarFile);
								
									// Release Memory
							
									CFRelease(tCertificateRef);
							
									free(tDataPtr);
									
									return NO;
								}
								
								tSignature=xar_signature_new(nXarFile,tSignatureType="RSA", tSignatureSize , signatureCallback, self);
			
								if (tSignature!=NULL)
								{
									xar_signature_add_x509certificate(tSignature,tDataPtr->Data,tDataPtr->Length);
								}
								else
								{
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"xar_signature_new: low memory"];
								
									xar_close(nXarFile);
								
									// Release Memory
							
									CFRelease(tCertificateRef);
							
									free(tDataPtr);
									
									return NO;
								}
								
								free(tDataPtr);
							}
							else
							{
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to allocate memory"];
								
								xar_close(nXarFile);
							
								return NO;
							}
							
							// Release Memory
							
							CFRelease(tCertificateRef);
						}
						
						while (tItemPath=[tDirectoryEnumerator nextObject])
						{
							NSString * tLastPathComponent;
							
							tLastPathComponent=[tItemPath lastPathComponent];
							
							if (tLastPathComponent==nil)
							{
								tLastPathComponent=tItemPath;
							}
							
							if ([tLastPathComponent hasPrefix:@"."]==NO)
							{
								if (([tLastPathComponent isEqualToString:@"Payload"]==YES || [tLastPathComponent isEqualToString:@"Scripts"]==YES) && 
									 [tItemPath rangeOfString:@"Resources/"].location==NSNotFound)
								{
									xar_opt_set(nXarFile,XAR_OPT_COMPRESSION,XAR_OPT_VAL_NONE);
								}
								else
								{
									xar_opt_set(nXarFile,XAR_OPT_COMPRESSION,XAR_OPT_VAL_GZIP);
								}
								
								if (xar_add(nXarFile,[tItemPath fileSystemRepresentation])==NULL)
								{
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to add file \'%@\' to xar archive",tItemPath];
									
									xar_close(nXarFile);
									
									return NO;
								}
							}
							
						}
					}
					
					// Close file
					
					if (xar_close(nXarFile)==0)
					{
						return YES;
					}
					else
					{
						tMetadata=nil;
						
						if (signatureResult_!=0)
						{
							switch(signatureResult_)
							{
								case IC_BUILDER_FAILURE_REASON_SIGNATURE_GENERIC_ERROR:
								
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Signature Error: Generic Error"];
								
									break;
								
								case IC_BUILDER_FAILURE_REASON_SIGNATURE_TIME_OUT:
								
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Signature Error: Timed Out"];
								
									break;
									
								case IC_BUILDER_FAILURE_REASON_SIGNATURE_AUTH_DENIED:
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Signature Error: Authorization Denied"];
									
									break;
									
								case IC_BUILDER_FAILURE_REASON_SIGNATURE_CERTIFICATE_NOT_FOUND:
								
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Signature Error: Certificate not found"];
									
									break;
							}
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:signatureResult_],IC_BUILDER_FAILURE_REASON,
																				 nil];
						}
					}
				}
				else
				{
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Unable to create xar file at path \'%@\'.",tFilePath];
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																		 tFilePath,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				// Low Memory
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			}
		}
		else
		{
			// Incorrect value (Empty String)
					
			[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"File name is empty"];
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,nil];
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

#pragma mark -

- (BOOL) buildPackageFromDictionary:(NSMutableDictionary *) inPackageDictionary atPath:(NSString *) inPath mode:(int) inBuildMode
{
	NSDictionary * tSettingsDictionary;
	NSString * tPackageFolder=nil;
	NSDictionary * tMetadata=nil;
	
	tSettingsDictionary=[inPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS];
	
	if (tSettingsDictionary!=nil)
	{
		NSNumber * tNumber;
		int tPackageType=ICDOCUMENT_PACKAGE_TYPE_PROJECT;
		NSString * tPackageUUID=nil;
		NSMutableDictionary * tPackageBuildInformation=nil;
		NSString * tString;
		NSString * tPackageName=nil;
	
		if (inBuildMode==IC_PACKAGE_BUILD_MODE_PACKAGE_RAW)
		{
			tPackageUUID=@"StandAlone Package";
			
			tPackageType=ICDOCUMENT_PACKAGE_TYPE_PROJECT;
		}
		else
		{
			tPackageUUID=[inPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_UUID];
			
			tNumber=[inPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_TYPE];
		
			if (tNumber!=nil)
			{
				tPackageType=[tNumber intValue];
			}
		}

		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		// Build Information

		tPackageBuildInformation=[NSMutableDictionary dictionary];
			
		if (tPackageBuildInformation!=nil)
		{
			NSMutableDictionary * tPackagesDictionary;
			
			tPackagesDictionary=[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES];
			
			if (tPackagesDictionary==nil)
			{
				tPackagesDictionary=[NSMutableDictionary dictionary];
				
				if (tPackagesDictionary!=nil)
				{
					[buildInformation_ setObject:tPackagesDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES];
				}
				else
				{
					// Low Memory
					
					goto buildPackageBail;
				}
			}
			
			[tPackagesDictionary setObject:tPackageBuildInformation forKey:tPackageUUID];
		
			// Package Name
					
			if (inBuildMode==IC_PACKAGE_BUILD_MODE_PACKAGE_RAW)
			{
				tPackageName=[[inPath lastPathComponent] stringByDeletingPathExtension];
				
				inPath=[inPath stringByDeletingLastPathComponent];
			}
			else
			{
				tPackageName=[[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_NAME] stringByDeletingPathExtension];
				
				if (tPackageName!=nil)
				{
					[tPackageBuildInformation setObject:tPackageName forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NAME];
				}
			}
		
			if (tPackageType==ICDOCUMENT_PACKAGE_TYPE_REFERENCED)
			{
				[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
															forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_REFERENCE
														   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																											   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																											   nil]];
				
				tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_LOCATION];
				
				if (tNumber!=nil)
				{
					int tLocationType;
					NSString * tLocationPath;
					
					[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCATION_TYPE];
					
					tLocationType=[tNumber intValue];
					
					tLocationPath=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH];
					
					if (tLocationPath!=nil)
					{
						NSString * tURLPrefix=nil;
						
						if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
						{
							tURLPrefix=@"http://";
						}
						else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
						{
							tURLPrefix=@"x-disc://";
						}
						else
						{
							// Incorrect value
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PACKAGE_SETTINGS_LOCATION",tLocationType];
												
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
																												  
							goto buildPackageBail;
						}
						
						if ([tLocationPath hasPrefix:tURLPrefix]==YES)
						{
							tLocationPath=[tLocationPath substringFromIndex:[tURLPrefix length]];
						}
						
						if ([tLocationPath hasPrefix:@"/"]==YES)
						{
							tLocationPath=[tLocationPath substringFromIndex:1];
						}
							
						if ([tLocationPath length]>0)
						{
							tLocationPath=[tLocationPath stringByAppendingPathComponent:tPackageName];
							
							if (tLocationPath!=nil)
							{
								[tPackageBuildInformation setObject:[NSString stringWithFormat:@"%@%@.pkg",tURLPrefix,[tLocationPath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
							}
							else
							{
								// Low Memory
							
								goto buildPackageBail;
							}
						}
						else
						{
							// Incorrect Value (Empty String)
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH string can not be empty."];
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							goto buildPackageBail;
						}
					}
					else
					{
						// Missing Information
							
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
																			 
						goto buildPackageBail;
					}
					
					// Get the Package ID
					
					tString=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER];
							
					if (tString!=nil)
					{
						if ([tString length]>0)
						{
							[tPackageBuildInformation setObject:tString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID];
							
							// Get the Version Number
					
							tString=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_VERSION];
									
							if (tString!=nil)
							{
								if ([tString length]>0)
								{
									[tPackageBuildInformation setObject:tString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_VERSION];
									
									// Get the Conclusion Action
					
									tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION];
											
									if (tNumber!=nil)
									{
										[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_CONCLUSION_ACTION];
										
										// Get the Authentication
					
										tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_AUTHENTICATION];
												
										if (tNumber!=nil)
										{
											[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_AUTHENTICATION];
											
											// Payload Size (May not be there)
											
											tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_PAYLOAD_SIZE];
													
											if (tNumber!=nil)
											{
												[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE];
											}
											
											
											[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																						forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_REFERENCE
																					   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																																		   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																																		   nil]];
											
											return YES;
										}
										else
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PACKAGE_SETTINGS_AUTHENTICATION,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
									else
									{
										// Missing Information
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
									}
								}
								else
								{
									// Incorrect Value (Empty String)
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_VERSION string can not be empty."];
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_SETTINGS_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
								}
							}
							else
							{
								// Missing Information
							
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_PACKAGE_SETTINGS_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
							}
						}
						else
						{
							// Incorrect Value
							
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER string can not be empty."];
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					}
					else
					{
						// Missing Information
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				}
				else
				{
					// Missing Information
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				int tLocationType=ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_EMBEDDED;
				NSString * tLocationPath;
				NSString * tPackageFinalDestination=nil;
				id tFinalPackageName=nil;
				
				if (inBuildMode!=IC_PACKAGE_BUILD_MODE_PACKAGE_RAW)
				{
					tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_LOCATION];
				
					if (tNumber!=nil)
					{
						tLocationType=[tNumber intValue];
					}
				
					[tPackageBuildInformation setObject:[NSNumber numberWithInt:tLocationType] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_LOCATION_TYPE];
				}
				
				if (inBuildMode!=IC_PACKAGE_BUILD_MODE_PACKAGE_RAW && tPackageType==ICDOCUMENT_PACKAGE_TYPE_IMPORTED)
				{
					NSDictionary * tPathDictionary;
					
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_IMPORT
															   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																												  tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																												  nil]];
					
					tPathDictionary=[inPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_REFERENCE_PATH];
					
					if (tPathDictionary!=nil)
					{
						NSString * tImportedPackagePath;
						
						tImportedPackagePath=[self pathFromPathDictionary:tPathDictionary]; 
						
						if (tImportedPackagePath!=nil)
						{
							// Get the Package Name
							
							tFinalPackageName=tPackageName=[[[tImportedPackagePath lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:@"pkg"];
							
							if (tPackageName!=nil)
							{
								NSMutableDictionary * tPackageSettingsDictionary;
				
								
								
								tPackageSettingsDictionary=[ICPackageUtilities packageSettingsDictionaryWithContentsOfFile:tImportedPackagePath];
								
								if (tPackageSettingsDictionary==nil)
								{
									// Does the file exist on disk
									
									if ([fileManager_ fileExistsAtPath:tImportedPackagePath]==NO)
									{
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_REFERENCE_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																							 tImportedPackagePath,IC_BUILDER_INFORMATION_FILE_PATH,
																							 nil];
									}
								
								
									goto buildPackageBail;
								}
								
								// We need to copy the package except if it's to be embedded in a flat distribution
							
								if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_EMBEDDED)
								{
									if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE)
									{
										NSString * tRelativePath;
											
										tRelativePath=[[NSString stringWithString:@"./Contents/Packages"] stringByAppendingPathComponent:tPackageName];
										
										if (tRelativePath!=nil)
										{
											[tPackageBuildInformation setObject:[NSString stringWithFormat:@"file:%@",[tRelativePath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
										}
										else
										{
											// Low Memory
											
											goto buildPackageBail;
										}
									
										tPackageFinalDestination=[[inPath stringByAppendingPathComponent:tRelativePath] stringByStandardizingPath];
									}
									else if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT)
									{
										tFinalPackageName=[NSMutableString stringWithString:tPackageName];
										
										if (tFinalPackageName!=nil)
										{
											//CFStringTrimWhitespace((CFMutableStringRef) tFinalPackageName); 
											
											[tFinalPackageName replaceOccurrencesOfString:@" " withString:@"_" options:0 range:NSMakeRange(0,[tFinalPackageName length])];
										}
										else
										{
											tFinalPackageName=tPackageName;
										}
										
										[tPackageBuildInformation setObject:[NSString stringWithFormat:@"#%@",tFinalPackageName] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
									}
								}
								else
								{
									tLocationPath=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH];
									
									if (tLocationPath==nil)
									{
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
									}
									else
									{
										if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_CUSTOM_PATH)
										{
											if ([tLocationPath hasPrefix:@"file:"]==YES)
											{
												tLocationPath=[tLocationPath substringFromIndex:5];
											}
											
											if ([tLocationPath length]>0)
											{
												NSString * tRelativePath;
												
												tRelativePath=[tLocationPath stringByAppendingPathComponent:tPackageName];

												if (tRelativePath!=nil)
												{
													NSString * tFilePath=nil;
													
													if ([tRelativePath hasPrefix:@"/"]==YES)
													{
														tFilePath=[NSString stringWithFormat:@"file:%@",tRelativePath];
													}
													else
													{
														if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE)
														{
															tFilePath=[[NSString stringWithFormat:@"file:../%@",tRelativePath] stringByStandardizingPath];
														}
														else
														{
															tFilePath=[[NSString stringWithFormat:@"file:./%@",tRelativePath] stringByStandardizingPath];
														}
														
													}
													
													[tPackageBuildInformation setObject:[tFilePath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
												}
												else
												{
													// Low Memory
													
													goto buildPackageBail;
												}
												
												if ([tLocationPath isEqualToString:@"./"]==YES ||
													[tLocationPath isEqualToString:@"../"]==YES)
												{
													tPackageFinalDestination=[[[inPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:tRelativePath] stringByStandardizingPath];
												}
												else
												{
													tPackageFinalDestination=[[[inPath stringByAppendingPathComponent:@"../packages"] stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
												}
												
											}
											else
											{
												// Incorrect Value (Empty String)
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH string can not be empty."];
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
												
												goto buildPackageBail;
											}
										}
										else
										{
											NSString * tURLPrefix=nil;
						
											if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
											{
												tURLPrefix=@"http://";
											}
											else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
											{
												tURLPrefix=@"x-disc://";
											}
											else
											{
												// Incorrect value
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PACKAGE_SETTINGS_LOCATION",tLocationType];
																	
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
																																	  
												goto buildPackageBail;
											}
											
											if ([tLocationPath hasPrefix:tURLPrefix]==YES)
											{
												tLocationPath=[tLocationPath substringFromIndex:[tURLPrefix length]];
											}
											
											if ([tLocationPath hasPrefix:@"/"]==YES)
											{
												tLocationPath=[tLocationPath substringFromIndex:1];
											}
												
											if ([tLocationPath length]>0)
											{
												tLocationPath=[tLocationPath stringByAppendingPathComponent:tPackageName];
												
												if (tLocationPath!=nil)
												{
													[tPackageBuildInformation setObject:[NSString stringWithFormat:@"%@%@",tURLPrefix,[tLocationPath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
												}
												else
												{
													// Low Memory
												
													goto buildPackageBail;
												}
												
												if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
												{
													tPackageFinalDestination=[[[inPath stringByAppendingPathComponent:@"../http"] stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
												}
												else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
												{
													tPackageFinalDestination=[[[inPath stringByAppendingPathComponent:@"../x-disc"] stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
												}
											}
											else
											{
												// Incorrect Value (Empty String)
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH string can not be empty."];
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
												
												goto buildPackageBail;
											}
										
											// When it's not local we need these pieces of informartion
											
											
											
											// Get the Archive Size
											
											// A COMPLETER
										}
									}
								}
								
								// Get the Conclusion Action
							
								tNumber=[tPackageSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION];
										
								if (tNumber!=nil)
								{
									[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_CONCLUSION_ACTION];
								}
								else
								{
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION missing for import."];
									
									goto buildPackageBail;
								}
								
								// Get the Authentication
								
								tNumber=[tPackageSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_AUTHENTICATION];
										
								if (tNumber!=nil)
								{
									[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_AUTHENTICATION];
								}
								else
								{
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_AUTHENTICATION missing for import."];
									
									goto buildPackageBail;
								}
								
								// Get the Install Size
								
								tNumber=[tPackageSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_PAYLOAD_SIZE];
										
								if (tNumber!=nil)
								{
									[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE];
								}
							
								// Get the Package ID
						
								tString=[tPackageSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER];
										
								if (tString!=nil)
								{
									if ([tString length]>0)
									{
										[tPackageBuildInformation setObject:tString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID];
										
										// Get the Version Number
									
										tString=[tPackageSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_VERSION];
												
										if (tString!=nil)
										{
											if ([tString length]>0)
											{
												[tPackageBuildInformation setObject:tString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_VERSION];
												
												// Create the intermediate folder if needed
										
												if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_EMBEDDED && inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT)
												{
													// Create the Package folder
													 
													tPackageFolder=[[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_CONTENTS_PATH] stringByAppendingPathComponent:tFinalPackageName];
													
													if ([self createDirectoryAtPath:tPackageFolder]==NO)
													{
														return NO;
													}
													
													// Extract all the files in the xar archive into the folder
													
													if ([self unarchiveXarFileAtPath:tImportedPackagePath toFolder:tPackageFolder]==YES)
													{
														[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																									forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_IMPORT
																								   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																																					   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																																					   nil]];
																
														return YES;
													}
													else
													{
														return NO;
													}
												}
												else
												{
													if ([self buildPath:[tPackageFinalDestination stringByDeletingLastPathComponent] fixedPermissionAfter:YES]==YES)
													{
														if (identity_==NULL)
														{
															certificateDictionary_=[[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS] objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE];
														
															if (certificateDictionary_!=nil && [certificateDictionary_ count]>0)
															{
																NSString * tName;
																NSString * tKeychainPath;
																
																tName=[certificateDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME];
																
																if (tName!=nil)
																{
																	// A COMPLETER
																}
																else
																{
																	// Missing Information
																	
																	tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																														 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																														 nil];
															
																	goto buildPackageBail;
																}
																
																tKeychainPath=[certificateDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH];
																
																if (tKeychainPath!=nil)
																{
																	// A COMPLETER
																}
																else
																{
																	// Missing Information
																	
																	tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																														 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																														 nil];
															
																	goto buildPackageBail;
																}
																
																identity_=[ICCertificatesUtilities identityWithName:tName atPath:tKeychainPath];
															
																if (identity_==NULL)
																{
																	tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_SIGNATURE_GENERIC_ERROR],IC_BUILDER_FAILURE_REASON,
																														 nil];
															
																	goto buildPackageBail;
																}
																
																CFRetain(identity_);
															}
														}
														
														if (identity_==NULL)
														{
															// We don't need to sign the package
															
															// Copy the package
											
															if ([self copyObjectAtPath:tImportedPackagePath toPath:tPackageFinalDestination]==YES)
															{
																[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																											forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_IMPORT
																										   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																																							   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																																							   nil]];
																
																return YES;
															}
															else
															{
																tMetadata=nil;
															}
														}
														else
														{
															// We need to sign the package
															
															tPackageFolder=[scratchLocation_ stringByAppendingPathComponent:tPackageName];
							
															if (tPackageFolder!=nil)
															{
																if ([self createDirectoryAtPath:tPackageFolder]==NO)
																{
																	goto buildPackageBail;
																}
															}
															else
															{
																goto buildPackageBail;
															}
															
															// Extract the package to a folder
															
															if ([self unarchiveXarFileAtPath:tImportedPackagePath toFolder:tPackageFolder]==YES)
															{
																// Recreate the xar archive
															
																if ([self createXarFileNamed:tPackageName atPath:[tPackageFinalDestination stringByDeletingLastPathComponent] withContentsOfFolder:tPackageFolder]==YES)
																{
																	// Compute the SHA256 hash
																	
																	if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
																	{
																		// A COMPLETER
																	}
																	/*else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
																	{
																	}*/
																	
																	// Remove the Package Working folder
																	
																	if ([fileManager_ removeFileAtPath:tPackageFolder handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==YES)
																	{
																		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																													forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_IMPORT
																												   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																																									   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																																									   nil]];
																		
																		return YES;
																	}
																	else
																	{	
																		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																															 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																															 tPackageFolder,IC_BUILDER_FAILURE_TAG_NAME,
																															 nil];
																	}
																}
																else
																{
																	tMetadata=nil;
																}
															}
															else
															{
																return NO;
															}
														}
													}
													else
													{
														tMetadata=nil;	// A COMPLETER
													}
												}
											}
											else
											{
												// Incorrect Value (Empty String)
									
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PACKAGE_SETTINGS_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
										}
										else
										{
											// Missing Information
									
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PACKAGE_SETTINGS_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
									else
									{
										// Incorrect Value (Empty String)
									
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
									}
								}
								else
								{
									// Missing Information
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
								}
							}
							else
							{
								// Low Memory
							
								goto buildPackageBail;
							}
						}
						else
						{
							// A COMPLETER
						
							tMetadata=nil;
						
							goto buildPackageBail;
						}
					}
					else
					{
						// Missing Information
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICDOCUMENT_PACKAGE_REFERENCE_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
				}
				else if (tPackageType==ICDOCUMENT_PACKAGE_TYPE_PROJECT)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_CREATE
															   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																												  tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																												  nil]];
					
					if (tPackageName==nil)
					{
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																			 nil];
					}
					else
					{
						NSDictionary * tFilesDictionary;
						NSDictionary * tScriptsDictionary;
						int tPayloadType=0;	// A COMPLETER
					
						tFinalPackageName=tPackageName;
						
						if ([tPackageName length]>0)
						{
							// Create the Package Working Directory
							
							if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT &&
								tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_EMBEDDED)
							{
								tFinalPackageName=[NSMutableString stringWithString:tPackageName];
										
								if (tFinalPackageName!=nil)
								{
									//CFStringTrimWhitespace((CFMutableStringRef) tFinalPackageName); 
									
									[tFinalPackageName replaceOccurrencesOfString:@" " withString:@"_" options:0 range:NSMakeRange(0,[tFinalPackageName length])];
								}
								else
								{
									tFinalPackageName=tPackageName;
								}

								
								tPackageFolder=[[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_CONTENTS_PATH] stringByAppendingPathComponent:[NSString stringWithFormat:@"%@.pkg",tFinalPackageName]];
							}
							else
							{
								tPackageFolder=[scratchLocation_ stringByAppendingPathComponent:tPackageName];
							}
							
							if (tPackageFolder!=nil)
							{
								if ([self createDirectoryAtPath:tPackageFolder]==NO)
								{
									goto buildPackageBail;
								}
							}
							else
							{
								goto buildPackageBail;
							}
						}
						else
						{
							// Incorrect value (Empty String)
					
							[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_NAME string can not be empty."];
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PACKAGE_SETTINGS_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
							
							goto buildPackageBail;
						}
							
							
						// Build the Payload
						
						tFilesDictionary=[inPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES];
						
						if (tFilesDictionary==nil)
						{
							tFilesDictionary=[ICPackageFileUtility defaultEmptyPackageFileDictionary];
							
							if (tFilesDictionary!=nil)
							{
								[inPackageDictionary setObject:tFilesDictionary forKey:ICDOCUMENT_PACKAGE_FILES];
							}
						}
						
						if (tFilesDictionary!=nil)
						{
							tNumber=[tFilesDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE];
							
							if (tNumber!=nil)
							{
								NSMutableDictionary * tMutableDictionary;
								NSMutableArray * tMutableArray;
								
								tPayloadType=[tNumber intValue];
								
								tMutableArray=[NSMutableArray array];
								
								if (tMutableArray!=nil)
								{
									[tPackageBuildInformation setObject:tMutableArray forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_DOWNGRADABLE];
								}
								else
								{
									// Low Memory
									
									goto buildPackageBail;
								}
								
								tMutableDictionary=[NSMutableDictionary dictionary];
								
								if (tMutableDictionary!=nil)
								{
									[tPackageBuildInformation setObject:tMutableDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_LOCATORS];
								}
								else
								{
									// Low Memory
									
									goto buildPackageBail;
								}
								
								
								tMutableDictionary=[NSMutableDictionary dictionary];
								
								if (tMutableDictionary!=nil)
								{
									[tPackageBuildInformation setObject:tMutableDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS];
								}
								else
								{
									// Low Memory
									
									goto buildPackageBail;
								}
								
								if (tPayloadType==ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE_INTERNAL)
								{
									if ([self buildPayloadAndBomAtPath:tPackageFolder forFilesDictionary:tFilesDictionary ofPackageUUID:tPackageUUID]==NO)
									{
										tMetadata=nil;
										
										goto buildPackageBail;
									}
								}
								else if (tPayloadType==ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE_EXTERNAL)
								{
									// A COMPLETER
								}
								else
								{
									// Incorrect Value
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE",tPayloadType];
								
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									goto buildPackageBail;
								}
							}
							else
							{
								// Missing Information
							
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																					 ICDOCUMENT_PACKAGE_FILES_PAYLOAD_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
													 
								goto buildPackageBail;
							}
						}
						
						// Prepare the Scripts if needed
						
						tScriptsDictionary=[inPackageDictionary objectForKey:ICDOCUMENT_PACKAGE_SCRIPTS];
						
						if (tScriptsDictionary!=nil || [[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS] count]>0)
						{
							if ([self buildScriptsAtPath:tPackageFolder forScriptsDictionary:tScriptsDictionary ofPackageUUID:tPackageUUID]==NO)
							{
								tMetadata=nil;
								
								goto buildPackageBail;
							}
						}
						
						// Settings
						
						if (tSettingsDictionary!=nil)
						{
							NSXMLElement * tPackageInfoElement;

							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																		forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_INFO
																	   metadata:nil];
							
							tPackageInfoElement=(NSXMLElement *) [NSXMLNode elementWithName:@"pkg-info"];
							
							if (tPackageInfoElement!=nil)
							{
								
								id tAttribute;
								NSXMLElement * tPayloadElement;
								NSXMLElement * tScriptsElement=nil;
								//NSXMLDocument * tXMLDocument;
								BOOL tBoolean;
								NSError * tError;
								NSArray * tArray;
								NSDictionary * tDictionary;
								NSDictionary * tBundleScriptDictionary;
								NSDictionary * tFinalScriptDictionary;
								
								// Create the XML Document
								
								// format-version
								
								tAttribute=[NSXMLNode attributeWithName:@"format-version" stringValue:@"2"];
								
								if (tAttribute!=nil)
								{
									[tPackageInfoElement addAttribute:tAttribute];
								}
								else
								{
									// Low Memory
								
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'format-version' attribute"];
														 
									goto buildPackageBail;
								}
								
								// identifier
								
								tString=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER];
								
								if (tString!=nil)
								{
									if ([tString length]>0)
									{
										[tPackageBuildInformation setObject:tString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ID];
										
										tAttribute=[NSXMLNode attributeWithName:@"identifier" stringValue:tString];
										
										if (tAttribute!=nil)
										{
											[tPackageInfoElement addAttribute:tAttribute];
										}
										else
										{
											// Low Memory
										
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'identifier' attribute"];
											
											goto buildPackageBail;
										}
									}
									else
									{
										// Incorrect Value (Empty String)
									
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER string can not be empty."];
								
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										goto buildPackageBail;
									}
								}
								else
								{
									// Missing Information
								
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_SETTINGS_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									goto buildPackageBail;
								}
								
								// version
								
								tString=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_VERSION];
								
								if (tString!=nil)
								{
									if ([tString length]>0)
									{
										[tPackageBuildInformation setObject:tString forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_VERSION];
										
										tAttribute=[NSXMLNode attributeWithName:@"version" stringValue:tString];
										
										if (tAttribute!=nil)
										{
											[tPackageInfoElement addAttribute:tAttribute];
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'version' attribute"];
											
											goto buildPackageBail;
										}
									}
									else
									{
										// Incorrect Value
									
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_VERSION string can not be empty."];
								
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_SETTINGS_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										goto buildPackageBail;
									}
								}
								else
								{
									// Missing Information
								
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_SETTINGS_VERSION,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									goto buildPackageBail;
								}
								
								// relocatable
								
								tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_RELOCATABLE];
								
								if (tNumber!=nil)
								{
									tBoolean=[tNumber boolValue];
									
									tAttribute=[NSXMLNode attributeWithName:@"relocatable" stringValue:(tBoolean==YES) ? @"true" : @"false"];
								
									if (tAttribute!=nil)
									{
										[tPackageInfoElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'relocatable' attribute"];
										
										goto buildPackageBail;
									}
								}
								
								// overwrite-permissions
								
								tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_OVERWRITE_PERMISSIONS];
								
								if (tNumber!=nil)
								{
									tBoolean=[tNumber boolValue];
									
									tAttribute=[NSXMLNode attributeWithName:@"overwrite-permissions" stringValue:(tBoolean==YES) ? @"true" : @"false"];
								
									if (tAttribute!=nil)
									{
										[tPackageInfoElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'overwrite-permissions' attribute"];
										
										goto buildPackageBail;
									}
								}
								
								// followSymLinks
								
								tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_FOLLOW_SYMBOLIC_LINKS];
								
								if (tNumber!=nil)
								{
									tBoolean=[tNumber boolValue];
									
									tAttribute=[NSXMLNode attributeWithName:@"followSymLinks" stringValue:(tBoolean==YES) ? @"true" : @"false"];
								
									if (tAttribute!=nil)
									{
										[tPackageInfoElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'followSymLinks' attribute"];
										
										goto buildPackageBail;
									}
								}
								
								// useHFSPlusCompression
								
								tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_USE_HFSPLUS_COMPRESSION];
								
								if (tNumber!=nil)
								{
									tBoolean=[tNumber boolValue];
									
									if (tBoolean==YES)
									{
										tAttribute=[NSXMLNode attributeWithName:@"useHFSPlusCompression" stringValue:@"true"];
								
										if (tAttribute!=nil)
										{
											[tPackageInfoElement addAttribute:tAttribute];
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'useHFSPlusCompression' attribute"];
											
											goto buildPackageBail;
										}
									}
								}
								
								// install-location
								
								if (tPayloadType==1 || tFilesDictionary==nil)	// A COMPLETER
								{
									tString=@"/";
								}
								else
								{
									tString=[tFilesDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION];
									
									if (tString!=nil)
									{
										if ([tString hasPrefix:@"/"]==NO)
										{
											// Incorrect Value
								
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
											
											goto buildPackageBail;
										}
									}
									else
									{
										// Missing Information
								
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																							 ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										goto buildPackageBail;
									}
								}
								
								if (tString!=nil)
								{
									tAttribute=[NSXMLNode attributeWithName:@"install-location" stringValue:tString];
								
									if (tAttribute!=nil)
									{
										[tPackageInfoElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
									
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'install-location' attribute"];
										
										goto buildPackageBail;
									}
								}
								
								// auth
								
								tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_AUTHENTICATION];
								
								if (tNumber!=nil)
								{
									[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_AUTHENTICATION];
									
									tString=([tNumber boolValue]==YES) ? @"root" : @"none";
									
									tAttribute=[NSXMLNode attributeWithName:@"auth" stringValue:tString];
								
									if (tAttribute!=nil)
									{
										[tPackageInfoElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'auth' attribute"];
										
										goto buildPackageBail;
									}
								}
								else
								{
									// Missing Information
								
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_SETTINGS_AUTHENTICATION,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									goto buildPackageBail;
								}
								
								// post-install-action
								
								tNumber=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION];
								
								if (tNumber!=nil)
								{
									int tConclusionAction;
									
									[tPackageBuildInformation setObject:tNumber forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_CONCLUSION_ACTION];
									
									tConclusionAction=[tNumber intValue];
									
									if (tConclusionAction!=ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_NONE)
									{
										switch(tConclusionAction)
										{
											case ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_REQUIRE_RESTART:
											
												tString=@"restart";
											
												break;
											
											case ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_REQUIRE_SHUTDOWN:
											
												tString=@"shutdown";
											
												break;
										
											case ICDOCUMENT_PACKAGES_SETTINGS_CONCLUSION_ACTION_REQUIRE_LOGOUT:
											
												tString=@"logout";
											
												break;
										
											default:
												
												// Incorrect Value
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION",tConclusionAction];
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
												
												goto buildPackageBail;
												
												break;
										}
										
										tAttribute=[NSXMLNode attributeWithName:@"postinstall-action" stringValue:tString];
									
										if (tAttribute!=nil)
										{
											[tPackageInfoElement addAttribute:tAttribute];
										}
										else
										{
											// Low Memory
											
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'post-install-action' attribute"];
											
											goto buildPackageBail;
										}
									}
								}
								else
								{
									// Missing Information
								
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_SETTINGS_CONCLUSION_ACTION,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									goto buildPackageBail;
								}
								
								// payload
								
								tPayloadElement=(NSXMLElement *) [NSXMLNode elementWithName:@"payload"];
				
								if (tPayloadElement!=nil)
								{
									
									// installKBytes
									
									tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE];
									
									if (tNumber==nil)
									{
										// Missing Build Data
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE"];
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																							 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										goto buildPackageBail;
									}
									
									tAttribute=[NSXMLNode attributeWithName:@"installKBytes" stringValue:[tNumber stringValue]];
									
									if (tAttribute!=nil)
									{
										[tPayloadElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'installKBytes' attribute"];
										
										goto buildPackageBail;
										
									}
									
									// numberOfFiles
									
									tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NUMBER_OF_FILES];
									
									if (tNumber==nil)
									{
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NUMBER_OF_FILES"];
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																							 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NUMBER_OF_FILES,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										goto buildPackageBail;
									}
									
									tAttribute=[NSXMLNode attributeWithName:@"numberOfFiles" stringValue:[tNumber stringValue]];
									
									if (tAttribute!=nil)
									{
										[tPayloadElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'numberOfFiles' attribute"];
										
										goto buildPackageBail;
									}
								
									[tPackageInfoElement addChild:tPayloadElement];
								}
								else
								{
									// Low Memory
									
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <payload> element"];
									
									goto buildPackageBail;
								}
								
								// scripts
								
								tNumber=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_HAS_SCRIPTS];
								
								if (tNumber!=nil && [tNumber boolValue]==YES)
								{
									tScriptsElement=(NSXMLElement *) [NSXMLNode elementWithName:@"scripts"];
									
									if (tScriptsElement!=nil)
									{
										NSXMLElement * tScriptsFileElement;
										
										// preinstall
										
										tString=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_PREINSTALL_SCRIPT_PATH];
										
										if (tString!=nil)
										{
											tScriptsFileElement=(NSXMLElement *) [NSXMLNode elementWithName:@"preinstall"];
											
											if (tScriptsFileElement!=nil)
											{
												tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tString];
									
												if (tAttribute!=nil)
												{
													[tScriptsFileElement addAttribute:tAttribute];
													
													[tScriptsElement addChild:tScriptsFileElement];
												}
												else
												{
													// Low Memory
													
													[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
																
													goto buildPackageBail;
												}
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <preinstall> element"];
												
												goto buildPackageBail;
											}
										}
										
										// postinstall
										
										tString=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_POSTINSTALL_SCRIPT_PATH];
										
										if (tString!=nil)
										{
											tScriptsFileElement=(NSXMLElement *) [NSXMLNode elementWithName:@"postinstall"];
											
											if (tScriptsFileElement!=nil)
											{
												tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tString];
									
												if (tAttribute!=nil)
												{
													[tScriptsFileElement addAttribute:tAttribute];
													
													[tScriptsElement addChild:tScriptsFileElement];
												}
												else
												{
													// Low Memory
													
													[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
																
													goto buildPackageBail;
												}
											}
											else
											{
												// Low Memory
												
												[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <postinstall> element"];
												
												goto buildPackageBail;
											}
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <scripts> element"];
										
										goto buildPackageBail;
									}
								}
								
								tBundleScriptDictionary=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS];
								
								tFinalScriptDictionary=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_FINAL];
								
								if (tBundleScriptDictionary!=nil && tFinalScriptDictionary!=nil)
								{
									if (tScriptsElement==nil)
									{
										tScriptsElement=(NSXMLElement *) [NSXMLNode elementWithName:@"scripts"];
									}
									
									if (tScriptsElement!=nil)
									{
										NSEnumerator * tEnumerator;
								
										tEnumerator=[tBundleScriptDictionary keyEnumerator];
										
										if (tEnumerator!=nil)
										{
											NSString * tBundleIdentifier;
											
											while (tBundleIdentifier=[tEnumerator nextObject])
											{
												NSDictionary * tPathDictionary;
												NSString * tPath;
												NSXMLElement * tScriptsFileElement;
												NSString * tFinalScriptPath;
												
												tPathDictionary=[tBundleScriptDictionary objectForKey:tBundleIdentifier];
												
												// Pre-installation
												
												tPath=[tPathDictionary objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_PREINSTALL];
												
												if (tPath!=nil)
												{
													tFinalScriptPath=[tFinalScriptDictionary objectForKey:tPath];
													
													if (tFinalScriptPath!=nil)
													{
														tScriptsFileElement=(NSXMLElement *) [NSXMLNode elementWithName:@"preinstall"];
														
														if (tScriptsFileElement!=nil)
														{
															tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFinalScriptPath];
												
															if (tAttribute!=nil)
															{
																[tScriptsFileElement addAttribute:tAttribute];
															}
															else
															{
																// Low Memory
																
																[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
																
																goto buildPackageBail;
															}
															
															tAttribute=[NSXMLNode attributeWithName:@"component-id" stringValue:tBundleIdentifier];
												
															if (tAttribute!=nil)
															{
																[tScriptsFileElement addAttribute:tAttribute];
															}
															else
															{
																// Low Memory
																
																[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'component-id' attribute"];
																
																goto buildPackageBail;
															}
															
															[tScriptsElement addChild:tScriptsFileElement];
														}
														else
														{
															// Low Memory
															
															[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <preinstall> element"];
															
															goto buildPackageBail;
														}
													}
												}
												
												// Post-installation
												
												tPath=[tPathDictionary objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_POSTINSTALL];
												
												if (tPath!=nil)
												{
													tFinalScriptPath=[tFinalScriptDictionary objectForKey:tPath];
													
													if (tFinalScriptPath!=nil)
													{
														tScriptsFileElement=(NSXMLElement *) [NSXMLNode elementWithName:@"postinstall"];
														
														if (tScriptsFileElement!=nil)
														{
															tAttribute=[NSXMLNode attributeWithName:@"file" stringValue:tFinalScriptPath];
												
															if (tAttribute!=nil)
															{
																[tScriptsFileElement addAttribute:tAttribute];
															}
															else
															{
																// Low Memory
																
																[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'file' attribute"];
																
																goto buildPackageBail;
															}
															
															tAttribute=[NSXMLNode attributeWithName:@"component-id" stringValue:tBundleIdentifier];
												
															if (tAttribute!=nil)
															{
																[tScriptsFileElement addAttribute:tAttribute];
															}
															else
															{
																// Low Memory
																
																[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'component-id' attribute"];
																
																goto buildPackageBail;
															}
															
															[tScriptsElement addChild:tScriptsFileElement];
														}
														else
														{
															// Low Memory
															
															[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <postinstall> element"];
															
															goto buildPackageBail;
														}
													}
													else
													{
														// Missing Value
														
														tMetadata=nil;
														
														// A COMPLETER
														
														goto buildPackageBail;
													}
												}
											}
										}
										else
										{
											// Low Memory
											
											goto buildPackageBail;
										}
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <scripts> element"];
										
										goto buildPackageBail;
									}
								}
								
								if (tScriptsElement!=nil && [tScriptsElement childCount]>0)
								{
									[tPackageInfoElement addChild:tScriptsElement];
								}
								
								// bundle-version
								
								tArray=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_VERSIONS];
								
								if (tArray!=nil)
								{
									NSXMLElement * tBundleVersionElement;
									
									tBundleVersionElement=(NSXMLElement *) [NSXMLNode elementWithName:@"bundle-version"];
									
									if (tBundleVersionElement!=nil)
									{
										if ([self addBundleFromArray:tArray toElement:tBundleVersionElement withPath:@"" packageInfoElement:tPackageInfoElement downgradableBundles:[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_DOWNGRADABLE]]==NO)
										{
											tMetadata=nil;
											
											goto buildPackageBail;
										}
										
										if ([tBundleVersionElement childCount]>0)
										{
											[tPackageInfoElement addChild:tBundleVersionElement];
										}
									}
									else
									{
										// Low Memory
									
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <bundle-version> element"];
										
										goto buildPackageBail;
									}
								}
								
								// relocate
								
								if (debug_==NO)
								{
									tDictionary=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_LOCATORS];
									
									if (tDictionary!=nil && [tDictionary count]>0)
									{
										NSEnumerator * tBundleEnumerator;
										NSXMLNode * tComment;
										
										tComment=[NSXMLNode commentWithStringValue:@"+==========================+\n    |         Locators         |\n    +==========================+"];
									
										if (tComment!=nil)
										{
											[tPackageInfoElement addChild:tComment];
										}
										
										tBundleEnumerator=[tDictionary keyEnumerator];
										
										if (tBundleEnumerator!=nil)
										{
											NSString * tBundleIdentifier;
											
											while (tBundleIdentifier=[tBundleEnumerator nextObject])
											{
												NSArray * tLocatorsArray;
												
												tLocatorsArray=[tDictionary objectForKey:tBundleIdentifier];
												
												if ([self addRelocators:tLocatorsArray forBundle:tBundleIdentifier packageInfoElement:tPackageInfoElement]==NO)
												{
													tMetadata=nil;
												
													goto buildPackageBail;
												}
											}
										}
										else
										{
											// Low Memory
										
											goto buildPackageBail;
										}
									}
								}
								
								// Save the XML
								
								tString=[tPackageInfoElement XMLStringWithOptions:NSXMLNodePrettyPrint|NSXMLNodeCompactEmptyElement];
								
								if (tString!=nil)
								{
									if ([tString writeToFile:[tPackageFolder stringByAppendingPathComponent:@"PackageInfo"] atomically:NO encoding:NSUTF8StringEncoding error:&tError]==YES)
									{
										[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																					forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_INFO
																				   metadata:nil];
										
										
										tPackageName=[tPackageName stringByAppendingPathExtension:@"pkg"];
									
										// Create the package at the appropriate location
								
										if (inBuildMode!=IC_PACKAGE_BUILD_MODE_PACKAGE_RAW)
										{
											if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_EMBEDDED)
											{
												if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE)		// Bundle Distribution
												{
													NSString * tRelativePath;
													
													tRelativePath=[[NSString stringWithString:@"./Contents/Packages"] stringByAppendingPathComponent:tPackageName];
													
													if (tRelativePath!=nil)
													{
														[tPackageBuildInformation setObject:[NSString stringWithFormat:@"file:%@",[tRelativePath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
													}
													else
													{
														// Low Memory
														
														goto buildPackageBail;
													}
													
													tPackageFinalDestination=[[inPath stringByAppendingPathComponent:tRelativePath] stringByStandardizingPath];
												}
												else															// Flat Distribution
												{
													[tPackageBuildInformation setObject:[NSString stringWithFormat:@"#%@.pkg",tFinalPackageName] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
												
													[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																								forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_CREATE
																							   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																																					   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																																					   nil]];
														
													return YES;
												}
											}
											else
											{
												tLocationPath=[tSettingsDictionary objectForKey:ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH];
												
												if (tLocationPath!=nil)
												{
													if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_CUSTOM_PATH)
													{
														if ([tLocationPath hasPrefix:@"file:"]==YES)
														{
															tLocationPath=[tLocationPath substringFromIndex:5];
														}
														
														if ([tLocationPath length]>0)
														{
															NSString * tRelativePath;
															
															tRelativePath=[tLocationPath stringByAppendingPathComponent:tPackageName];
															
															if (tRelativePath!=nil)
															{
																NSString * tFilePath=nil;
																
																if ([tRelativePath hasPrefix:@"/"]==YES)
																{
																	tFilePath=[NSString stringWithFormat:@"file:%@",tRelativePath];
																}
																else
																{
																	if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE)
																	{
																		tFilePath=[[NSString stringWithFormat:@"file:../%@",tRelativePath] stringByStandardizingPath];
																	}
																	else
																	{
																		tFilePath=[[NSString stringWithFormat:@"file:./%@",tRelativePath] stringByStandardizingPath];
																	}
																}
																
																[tPackageBuildInformation setObject:[tFilePath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
															}
															else
															{
																// Low Memory
																
																goto buildPackageBail;
															}
															
															if ([tLocationPath isEqualToString:@"./"]==YES ||
																[tLocationPath isEqualToString:@"../"]==YES)
															{
																tPackageFinalDestination=[[[inPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:tRelativePath] stringByStandardizingPath];
															}
															else
															{
																tPackageFinalDestination=[[[inPath stringByAppendingPathComponent:@"../packages"] stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
															}
														}
														else
														{
															// Incorrect Value (Empty String)
															
															[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH string can not be empty."];
									
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
															
															goto buildPackageBail;
														}
													}
													else
													{
														NSString * tURLPrefix=nil;
						
														if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
														{
															tURLPrefix=@"http://";
														}
														else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
														{
															tURLPrefix=@"x-disc://";
														}
														else
														{
															// Incorrect value
															
															[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PACKAGE_SETTINGS_LOCATION",tLocationType];
																				
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
																																				  
															goto buildPackageBail;
														}
														
														if ([tLocationPath hasPrefix:tURLPrefix]==YES)
														{
															tLocationPath=[tLocationPath substringFromIndex:[tURLPrefix length]];
														}
														
														if ([tLocationPath hasPrefix:@"/"]==YES)
														{
															tLocationPath=[tLocationPath substringFromIndex:1];
														}
															
														if ([tLocationPath length]>0)
														{
															tLocationPath=[tLocationPath stringByAppendingPathComponent:tPackageName];
															
															if (tLocationPath!=nil)
															{
																[tPackageBuildInformation setObject:[NSString stringWithFormat:@"%@%@",tURLPrefix,[tLocationPath stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_REFERENCE_PATH];
															}
															else
															{
																// Low Memory
															
																goto buildPackageBail;
															}
															
															if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
															{
																tPackageFinalDestination=[[[inPath stringByAppendingPathComponent:@"../http"] stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
															}
															else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
															{
																tPackageFinalDestination=[[[inPath stringByAppendingPathComponent:@"../x-disc"] stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
															}
														}
														else
														{
															// Incorrect Value (Empty String)
															
															[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH string can not be empty."];
															
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_STRING_EMPTY],IC_BUILDER_FAILURE_REASON,
																												 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																												 nil];
															
															goto buildPackageBail;
														}
													}
												}
												else
												{
													// Missing Information
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																										 ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																										 nil];
													
													goto buildPackageBail;
												}
											}
										}
										else
										{
											tPackageFinalDestination=[[inPath stringByAppendingPathComponent:tPackageName] stringByStandardizingPath];
										}
										
										// Create the intermediate folder if needed
										
										if ([self buildPath:[tPackageFinalDestination stringByDeletingLastPathComponent] fixedPermissionAfter:NO]==YES)
										{
											// Set the appropriate permissions
											
											if ([self setCorrectXarPermissionsForFolder:tPackageFolder]==YES)
											{
												if (inBuildMode==IC_PACKAGE_BUILD_MODE_PACKAGE_RAW ||
													inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT)
												{
													if (identity_==NULL)
													{
														certificateDictionary_=[[projectDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS] objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE];
													
														if (certificateDictionary_!=nil && [certificateDictionary_ count]>0)
														{
															NSString * tName;
															NSString * tKeychainPath;
															
															tName=[certificateDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME];
															
															if (tName!=nil)
															{
																// A COMPLETER
															}
															else
															{
																// Missing Information
																
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																													 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																													 nil];
														
																goto buildPackageBail;
															}
															
															tKeychainPath=[certificateDictionary_ objectForKey:ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH];
															
															if (tKeychainPath!=nil)
															{
																// A COMPLETER
															}
															else
															{
																// Missing Information
																
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																													 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_KEYCHAIN_PATH,IC_BUILDER_FAILURE_TAG_NAME,
																													 nil];
														
																goto buildPackageBail;
															}
															
															identity_=[ICCertificatesUtilities identityWithName:tName atPath:tKeychainPath];
														
															if (identity_==NULL)
															{
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																													 ICDOCUMENT_PROJECT_SETTINGS_CERTIFICATE_NAME,IC_BUILDER_FAILURE_TAG_NAME,
																													 nil];
																
																goto buildPackageBail;
															}
															
															CFRetain(identity_);
														}
													}
												}
												
												// Build the xar archive
											
												if ([self createXarFileNamed:tPackageName atPath:[tPackageFinalDestination stringByDeletingLastPathComponent] withContentsOfFolder:tPackageFolder]==YES)
												{
													if (inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_BUNDLE || 
														inBuildMode==IC_PACKAGE_BUILD_MODE_DISTRIBUTION_FLAT)
													{
														// Compute the SHA256 hash
														
														if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_HTTP_URL)
														{
															// A COMPLETER
														}
														/*else if (tLocationType==ICDOCUMENT_PACKAGE_SETTINGS_LOCATION_REMOVABLE_MEDIA)
														{
														}*/
													}
													
													// Remove the Package Working folder
													
													if ([fileManager_ removeFileAtPath:tPackageFolder handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==YES)
													{
														[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																									forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_CREATE
																								   metadata:[NSDictionary dictionaryWithObjectsAndKeys:tPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,
																																					   tPackageName,IC_BUILDER_INFORMATION_PACKAGE_NAME,
																																					   nil]];
														
														return YES;
													}
													else
													{	
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																											 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																											 tPackageFolder,IC_BUILDER_FAILURE_TAG_NAME,
																											 nil];
													}
												}
												else
												{
													tMetadata=nil;
												}
											}
											else
											{
												tMetadata=nil;
											}
										}
										else
										{
											tMetadata=nil;	// A COMPLETER
										}
									}
									else
									{
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																							 [tPackageFolder stringByAppendingPathComponent:@"PackageInfo"],IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
									}
								}
							}
							else
							{
								// Low Memory
								
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <pkg-info> element"];
							}
						}
					}
				}
				else
				{
					// Incorrect value
				
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Incorrect Value (%d) for key ICDOCUMENT_PACKAGE_TYPE",tPackageType];
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_PACKAGE_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
		}
	}
	else
	{
		// Missing Information
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
															 ICDOCUMENT_PACKAGE_SETTINGS,IC_BUILDER_FAILURE_TAG_NAME,
															 nil];
	}
	
buildPackageBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];

	if (tPackageFolder!=nil)
	{
		// Remove the Package Working folder
	
		if ([fileManager_ removeFileAtPath:tPackageFolder handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==NO)
		{
			[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
														forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
													   metadata:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																										   [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																										   tPackageFolder,IC_BUILDER_FAILURE_TAG_NAME,
																										   nil]];
		}
	}
	
	return NO;
}

- (BOOL) addRelocators:(NSArray *) inLocatorsArray forBundle:(NSString *) inBundleIdentifier packageInfoElement:(NSXMLElement *) inPackageInfoElement
{
	NSDictionary * tMetadata=nil;
		
	if (inLocatorsArray!=nil && inBundleIdentifier!=nil && inPackageInfoElement!=nil)
	{
		NSString * tSearchID;
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		tSearchID=[[NSString stringWithString:@"relocate."] stringByAppendingString:inBundleIdentifier];
		
		if (tSearchID!=nil)
		{
			NSXMLElement * tRelocateElement;
			NSXMLElement * tLocatorElement;
			
			// relocate
			
			tRelocateElement=(NSXMLElement *) [NSXMLNode elementWithName:@"relocate"];
			
			if (tRelocateElement!=nil)
			{
				NSXMLElement * tBundleElement;
				id tAttribute;
												
				// search-id
				
				tAttribute=[NSXMLNode attributeWithName:@"search-id" stringValue:tSearchID];
				
				if (tAttribute!=nil)
				{
					[tRelocateElement addAttribute:tAttribute];
				}
				else
				{
					// Low Memory
					
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'search-id' attribute"];
				
					goto addRelocatorsBail;
				}
				
				tBundleElement=(NSXMLElement *) [NSXMLNode elementWithName:@"bundle"];
			
				if (tBundleElement!=nil)
				{
					id tAttribute;
					
					// search-id
					
					tAttribute=[NSXMLNode attributeWithName:@"id" stringValue:inBundleIdentifier];
					
					if (tAttribute!=nil)
					{
						[tBundleElement addAttribute:tAttribute];
					}
					else
					{
						// Low Memory
											
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'id' attribute"];
					
						goto addRelocatorsBail;
					}
					
					[tRelocateElement addChild:tBundleElement];
				}
				else
				{
					// Low Memory
			
					[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <bundle> element"];
					
					goto addRelocatorsBail;
				}
			}
			else
			{
				// Low Memory
			
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <relocate> element"];
				
				goto addRelocatorsBail;
			}
		
			// locators
		
			tLocatorElement=(NSXMLElement *) [NSXMLNode elementWithName:@"locator"];
		
			if (tLocatorElement!=nil)
			{
				NSEnumerator * tLocatorsEnumerator;
				
				tLocatorsEnumerator=[inLocatorsArray objectEnumerator];
				
				if (tLocatorsEnumerator!=nil)
				{
					NSDictionary * tDictionary;
					int tSearchIndex=1;
					NSMutableArray * tSearchIDsArray;
					NSXMLElement * tSearchElement=nil;
					
					if (sharedLocatorPluginsManager_==nil)
					{
						sharedLocatorPluginsManager_=[ICLocatorPluginsManager defaultManager];
					}
					
					tSearchIDsArray=[NSMutableArray array];
					
					if (tSearchIDsArray==nil)
					{
						goto addRelocatorsBail;
					}
					
					while (tDictionary=[tLocatorsEnumerator nextObject])
					{
						NSString * tLocatorIdentifier;
						
						tLocatorIdentifier=[tDictionary objectForKey:ICDOCUMENT_LOCATOR_IDENTIFIER];
						
						if (tLocatorIdentifier!=nil)
						{
							id tLocatorConverter;
												
							// Find the appropriate Locator Converter plugin
							
							tLocatorConverter=[sharedLocatorPluginsManager_ createPluginConverterControllerForIdentifier:tLocatorIdentifier documentDictionary:documentDictionary_];
							
							if (tLocatorConverter!=nil)
							{
								NSDictionary * tLocatorSettingsDictionary;
								
								tLocatorSettingsDictionary=[tDictionary objectForKey:ICDOCUMENT_LOCATOR_DICTIONARY];
								
								if (tLocatorSettingsDictionary!=nil)
								{
									NSString * tUniqueSearchID;
									NSDictionary * tErrorDictionary=nil;
									NSArray * tElementsArray;
									
									tUniqueSearchID=[NSString stringWithFormat:@"search.%d.%@",tSearchIndex,inBundleIdentifier];
									
									tElementsArray=[tLocatorConverter elementsWithSettings:tLocatorSettingsDictionary withUniqueSearchID:tUniqueSearchID error:&tErrorDictionary];
														
									if (tElementsArray!=nil)
									{
										NSUInteger i,tCount;
										
										tCount=[tElementsArray count];
										
										for(i=0;i<tCount;i++)
										{
											tSearchElement=[tElementsArray objectAtIndex:i];
											
											[tLocatorElement addChild:tSearchElement];
										}
										
										[tSearchIDsArray addObject:[NSString stringWithFormat:@"\'%@\'",tUniqueSearchID]];
									}
									else
									{
										if (tErrorDictionary!=nil)
										{
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_LOCATOR_CONVERSION_ERROR],IC_BUILDER_FAILURE_REASON,
																								 tLocatorIdentifier,IC_BUILDER_FAILURE_TAG_NAME,
																								 tErrorDictionary,IC_BUILDER_FAILURE_SECONDARY_REASON,
																								 nil];
																								 
											goto addRelocatorsBail;
										}
									}
									
									tSearchIndex++;
								}
								else
								{
									// Missing Information
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_LOCATOR_DICTIONARY,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									goto addRelocatorsBail;
								}
							}
							else
							{
								// Converter not found
													
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_LOCATOR_MISSING_CONVERTER],IC_BUILDER_FAILURE_REASON,
																					 tLocatorIdentifier,IC_BUILDER_FAILURE_TAG_NAME,
																					 nil];
								
								goto addRelocatorsBail;
							}							
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_LOCATOR_IDENTIFIER,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
									
							goto addRelocatorsBail;
						}
					}
					
					if ([tLocatorElement childCount]>0)
					{
						if ([tSearchIDsArray count]==1)
						{
							id tAttribute;
							
							// A TESTER
							
							// search-id
							
							tAttribute=[NSXMLNode attributeWithName:@"id" stringValue:tSearchID];
							
							if (tAttribute!=nil)
							{
								[tSearchElement removeAttributeForName:@"id"];
								
								[tSearchElement addAttribute:tAttribute];
							}
							else
							{
								// Low Memory
													
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'id' attribute"];
							
								goto addRelocatorsBail;
							}
						
							[inPackageInfoElement addChild:tRelocateElement];
				
							[inPackageInfoElement addChild:tLocatorElement];
						}
						else
						{
							NSXMLElement * tSearchScriptElement;
							
							// Merge all the results
							
							tSearchScriptElement=(NSXMLElement *) [NSXMLNode elementWithName:@"search"];
				
							if (tSearchScriptElement!=nil)
							{
								id tAttribute;
								NSXMLElement * tScriptElement;
								
								// type
								
								tAttribute=[NSXMLNode attributeWithName:@"type" stringValue:@"script"];
								
								if (tAttribute!=nil)
								{
									[tSearchScriptElement addAttribute:tAttribute];
								}
								else
								{
									// Low Memory
														
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'type' attribute"];
								
									goto addRelocatorsBail;
								}
								
								// id
								
								tAttribute=[NSXMLNode attributeWithName:@"id" stringValue:tSearchID];
								
								if (tAttribute!=nil)
								{
									[tSearchScriptElement addAttribute:tAttribute];
								}
								else
								{
									// Low Memory
														
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'id' attribute"];
								
									goto addRelocatorsBail;
								}
						
						
								// script
								
								tAttribute=[NSXMLNode attributeWithName:@"script" stringValue:@"processSearchResults()"];
								
								if (tAttribute!=nil)
								{
									[tSearchScriptElement addAttribute:tAttribute];
								}
								else
								{
									// Low Memory
														
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'script' attribute"];
								
									goto addRelocatorsBail;
								}
								
								tScriptElement=(NSXMLElement *) [NSXMLNode elementWithName:@"script"];
								
								if (tScriptElement!=nil)
								{
									NSString * tCodePart=nil;
																		
									tCodePart=[NSString stringWithFormat:@"\n\n\t\tfunction processSearchResults()\n\
\t\t{\n\
\t\t\tvar tIDsArray;\n\
\t\t\tvar tProcessedResults;\n\
\t\t\tvar i,tCount;\n\
\n\
\t\t\ttIDsArray = new Array(%@);\n\
\n\
\t\t\ttProcessedResults = new Array;\n\
\n\
\t\t\ttCount=tIDsArray.length;\n\
\n\
\t\t\tfor(i=0;i<tCount;i++)\n\
\t\t\t{\n\
\t\t\t\tvar tResults;\n\
\n\
\t\t\t\ttResults=my.search.results[tIDsArray[i]];\n\
\n\
\t\t\t\tif (typeof(tResults) == \'object\')\n\
\t\t\t\t{\n\
\t\t\t\t\tvar j,tLength;\n\
\n\
\t\t\t\t\ttLength=tResults.length;\n\
\n\
\t\t\t\t\tfor(j=0;j<tLength;j++)\n\
\t\t\t\t\t{\n\
\t\t\t\t\t\tvar tValue;\n\
\n\
\t\t\t\t\t\ttValue=tResults[j];\n\
\n\
\t\t\t\t\t\tif (tValue)\n\
\t\t\t\t\t\t{\n\
\t\t\t\t\t\t\ttProcessedResults.push(tValue);\n\
\t\t\t\t\t\t}\n\
\t\t\t\t\t}\n\
\t\t\t\t}\n\
\t\t\t\telse if (tResults)\n\
\t\t\t\t{\n\
\t\t\t\t\ttProcessedResults.push(tResults);\n\
\t\t\t\t}\n\
\t\t\t}\n\
\n\
\t\t\treturn tProcessedResults;\n\
\t\t}\n\n        ",[tSearchIDsArray componentsJoinedByString:@",\n\t\t\t                      "]];
								
								/*else
								{
									tCodePart=[NSString stringWithFormat:@"\n\n\t\tfunction processSearchResults()\n\
\t\t{\n\
\t\t\treturn my.search.results[%@];\n\
\t\t}\n\n        ",[tSearchIDsArray objectAtIndex:0]];
								}*/
								
									if (tCodePart!=nil)
									{
										NSXMLNode * tNode;
									
										tNode=[NSXMLNode textWithStringValue:tCodePart];
						
										if (tNode!=nil)
										{
											[tScriptElement addChild:tNode];
											
											[tSearchScriptElement addChild:tScriptElement]; 
											
											[tLocatorElement addChild:tSearchScriptElement];
											
											
											[inPackageInfoElement addChild:tRelocateElement];
					
											[inPackageInfoElement addChild:tLocatorElement];
										}
										else
										{
											// Low Memory
				
											[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create text node"];
										
											goto addRelocatorsBail;
										}
									}
									else
									{
										goto addRelocatorsBail;
									}
								}
								else
								{
									[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <script> element"];
							
									goto addRelocatorsBail;
								}
							}
							else
							{
								// Low Memory
				
								[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <search> element"];
							
								goto addRelocatorsBail;
							}
						}
					}
					
					return YES;
				}
			}
			else
			{
				// Low Memory
			
				[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <locator> element"];
			}
		}
	}

addRelocatorsBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];

	return NO;
}

- (BOOL) addBundleFromArray:(NSArray *) inArray toElement:(NSXMLElement *) inElement withPath:(NSString *) inPath packageInfoElement:(NSXMLElement *) inPackageInfoElement downgradableBundles:(NSMutableArray *) inDowngradableArray
{
	NSDictionary * tMetadata=nil;
	
	if (inArray!=nil && inElement!=nil && inPath!=nil)
	{
		NSEnumerator * tEnumerator;
								
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
		
		tEnumerator=[inArray objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tBundleInformationDictionary;
			NSUInteger tLength;
			
			tLength=[inPath length];
							
			while (tBundleInformationDictionary=[tEnumerator nextObject])
			{
				NSArray * tChildren;
				
				tChildren=[tBundleInformationDictionary objectForKey:@"Children"];
				
				if ([tBundleInformationDictionary objectForKey:@"id"]==nil)
				{
					if (tChildren!=nil)
					{
						[self addBundleFromArray:tChildren toElement:inElement withPath:inPath packageInfoElement:inPackageInfoElement downgradableBundles:inDowngradableArray];
					}
				}
				else
				{
					NSXMLElement * tBundleElement;
				
					tBundleElement=(NSXMLElement *) [NSXMLNode elementWithName:@"bundle"];
				
					if (tBundleElement!=nil)
					{
						NSEnumerator * tKeyEnumerator;
						
						tKeyEnumerator=[tBundleInformationDictionary keyEnumerator];
						
						if (tKeyEnumerator!=nil)
						{
							NSString * tKey;
							NSString * tPath=nil;
							BOOL tAddToPackageInfoElement=NO;
							
							while (tKey=[tKeyEnumerator nextObject])
							{
								id tAttribute;
								
								if ([tKey isEqualToString:@"Children"]==YES)
								{
								}
								else if ([tKey isEqualToString:@"path"]==YES)
								{
									NSUInteger tIndex;
									
									tPath=[tBundleInformationDictionary objectForKey:tKey];
									
									tIndex=[inDowngradableArray indexOfObject:tPath];
									
									if (tIndex!=NSNotFound)
									{
										[inDowngradableArray removeObjectAtIndex:tIndex];
									
										tAddToPackageInfoElement=YES;
										
										tAttribute=[NSXMLNode attributeWithName:tKey stringValue:[NSString stringWithFormat:@".%@",tPath]];
									}
									else
									{
										tAttribute=[NSXMLNode attributeWithName:tKey stringValue:[NSString stringWithFormat:@".%@",[tPath substringFromIndex:tLength]]];
									}
									
									if (tAttribute!=nil)
									{
										[tBundleElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create 'path' attribute"];
									
										goto addBundleFromArrayBail;
									}
								}
								else
								{
									tAttribute=[NSXMLNode attributeWithName:tKey stringValue:[tBundleInformationDictionary objectForKey:tKey]];
			
									if (tAttribute!=nil)
									{
										[tBundleElement addAttribute:tAttribute];
									}
									else
									{
										// Low Memory
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create '%@' attribute",tKey];
										
										goto addBundleFromArrayBail;
									}
								}
							}
							
							if (tAddToPackageInfoElement==YES)
							{
								[inPackageInfoElement addChild:tBundleElement];
							}
							else
							{
								[inElement addChild:tBundleElement];
							}
							
							if (tChildren!=nil && [tChildren count]>0)
							{
								[self addBundleFromArray:tChildren toElement:tBundleElement withPath:tPath packageInfoElement:inPackageInfoElement downgradableBundles:inDowngradableArray];
							}
						}
						else
						{
							// Low Memory
							
							goto addBundleFromArrayBail;
						}
					}
					else
					{
						// Low Memory
						
						[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Low Memory: Unable to create <bundle> element"];
					
						goto addBundleFromArrayBail;
					}
				}
			}
			
			return YES;
		}
	}

addBundleFromArrayBail:
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return NO;
}

- (NSString *) copyScriptAtPath:(NSString *) inSourcePath toPath:(NSString *) inDestinationPath
{
	NSDictionary * tMetadata=nil;
	
	if (inSourcePath!=nil && inDestinationPath!=nil)
	{
		NSString * tDestinationFolder;
		NSString * tBaseName;
		
		tDestinationFolder=[inDestinationPath stringByDeletingLastPathComponent];
		
		tBaseName=[[inDestinationPath lastPathComponent] stringByDeletingPathExtension];
		
		if (tDestinationFolder!=nil && tBaseName!=nil)
		{
			if ([fileManager_ fileExistsAtPath:inSourcePath]==YES)
			{
				NSString * tDestinationPath;
				int tIndex=1;
				NSString * tFileScriptName;
				
				tFileScriptName=tBaseName;
				
				do
				{
					tDestinationPath=[tDestinationFolder stringByAppendingPathComponent:tFileScriptName];
				
					if ([fileManager_ fileExistsAtPath:tDestinationPath]==NO)
					{
						if ([self copyObjectAtPath:inSourcePath toPath:tDestinationPath]==NO)
						{
							return nil;
						}
						
						return tFileScriptName;
					}
					
					tFileScriptName=[NSString stringWithFormat:@"%@_%d",tBaseName,tIndex];
					
					tIndex++;
				}
				while (tIndex<32765);
				
				if (tIndex==32765)
				{
					// A COMPLETER
				}
			}
			else
			{
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																	 inSourcePath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 nil];
			}
		}
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
	return nil;
}

- (BOOL) buildDistributionScriptsAtPath:(NSString *) inPath
{
	NSString * tTemporaryDirectoryPath;
	NSDictionary * tMetadata=nil;
	
	tTemporaryDirectoryPath=[self getDistributionScriptsPath];
	
	if (tTemporaryDirectoryPath!=nil)
	{
		// Check if we need to create the Scripts archive
		
		if ([ICFilesHierarchyUtilities isEmptyDirectoryAtPath:(char *) [tTemporaryDirectoryPath fileSystemRepresentation]]==NO)
		{
			NSString * tDestinationPath=nil;
			BOOL tContinue;
			NSTask * tTask;
			NSMutableArray * tArguments;
			int returnValue;
		
			[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
														forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_SCRIPTS
													   metadata:nil];
	
			if (rootIsYouDaddy_==NO)
			{
				seteuid(0);
				setegid(0);
			}
	
			// Set the privileges (root:wheel 0755 ) for the tTemporaryDirectoryPath path and its children
	
			tContinue=NO;
	
			if (
#ifdef __SET_CORRECT_PERMISSIONS__
				[ICFilesHierarchyUtilities changeOwner:0 group:0 atPath:tTemporaryDirectoryPath]==1 ||
#endif
				[ICFilesHierarchyUtilities changeMode:0755 atPath:tTemporaryDirectoryPath]==1)
			{
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																	 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																	 nil];
			}
			else
			{
				
				NSString * tSplitForksPath=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_SPLITFORKS_PATH;
				
				if ([[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_SPLITFORK_TOOL] isEqualToString:ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN]==YES)
				{
					tSplitForksPath=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN_PATH;
				}
				
				// Split Forks if needed
										
				if ([fileManager_ fileExistsAtPath:tSplitForksPath]==YES)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_SPLIT
															   metadata:nil];
					
					tTask=[NSTask new];
				
					if (tTask!=nil)
					{
						tArguments=[NSMutableArray arrayWithCapacity:1];
					
						[tArguments addObject:tTemporaryDirectoryPath];
						
						[tTask setArguments:tArguments];
						
						[tTask setLaunchPath:tSplitForksPath];
						
						[tTask launch];
						
						[tTask waitUntilExit];
						
						returnValue = [tTask terminationStatus];
						
						[tTask release];
						
						if (0!=returnValue)
						{
							switch(returnValue)
							{
								case -2:
									// Not a HFS or Extended HFS volume
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_SPLIT_NON_HFS_VOLUME],IC_BUILDER_FAILURE_REASON,
																						 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																						 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																						 nil];
									
									break;
									
								default:
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OPERATION_FAILED],IC_BUILDER_FAILURE_REASON,
																						 tSplitForksPath,IC_BUILDER_INFORMATION_FILE_PATH,
																						 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																						 [NSNumber numberWithInt:returnValue],IC_BUILDER_FAILURE_TOOL_RETURN_CODE,
																						 nil];
									
									
									// A COMPLETER (Recuperer sdterr si il y a)
									
									
									break;
							}
						}
						else
						{
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																		forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_SPLIT
																	   metadata:nil];
						
							tContinue=YES;
						}
					}
					else
					{
						// Low Memory
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					}
				}
				else
				{
					// Splitfork tool not found
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																		 tSplitForksPath,IC_BUILDER_INFORMATION_FILE_PATH,
																		 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																		 nil];
				}
			}
	
			if (tContinue==YES)
			{
				
				// Create the pax archive
				
				if ([fileManager_ fileExistsAtPath:ICPROJECTBUILDER_DITTO_PATH]==YES)
				{
					NSString * tLocalDestinationPath=nil;
					
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_PAX
															   metadata:nil];
					
					tTask=[NSTask new];
				
					if (tTask!=nil)
					{
						tDestinationPath=[inPath stringByAppendingPathComponent:@"Scripts"];
						
						tArguments=[NSMutableArray arrayWithCapacity:5];
						
						// Ditto
						
#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5
						
						[tArguments addObject:@"--noextattr"];
						
						[tArguments addObject:@"--noqtn"];
						
						[tArguments addObject:@"--noacl"];
						
						[tArguments addObject:@"--norsrc"];
						
#endif
						
						[tArguments addObject:@"-c"];
						
						[tArguments addObject:@"-z"];
					
						[tArguments addObject:@"."];
					
						if (rootIsYouDaddy_==NO)
						{
							tLocalDestinationPath=[[tTemporaryDirectoryPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:[tDestinationPath lastPathComponent]];
							
							[tArguments addObject:tLocalDestinationPath];
						}
						else
						{
							[tArguments addObject:tDestinationPath];
						}
					
						[tTask setArguments:tArguments];
					
						[tTask setLaunchPath:ICPROJECTBUILDER_DITTO_PATH];

						
						[tTask setCurrentDirectoryPath:tTemporaryDirectoryPath];
					
					
						/*NSPipe *pipe1,*pipe2;
						pipe1 = [NSPipe pipe];
						pipe2 = [NSPipe pipe];
						[tTask setStandardOutput: pipe1];
						[tTask setStandardError: pipe2];

						NSFileHandle *file1;
						file1 = [pipe1 fileHandleForReading];
						NSFileHandle *file2;
						file2 = [pipe2 fileHandleForReading];*/
						
						[tTask launch];
						
						/*NSData *data;
						data = [file1 readDataToEndOfFile];
						[data writeToFile:@"/tmp/stdout.txt" atomically:YES];
						data = [file2 readDataToEndOfFile];
						[data writeToFile:@"/tmp/stderr.txt" atomically:YES];*/
						
						[tTask waitUntilExit];
						
						
						
						returnValue = [tTask terminationStatus];
						
						[tTask release];
					
						if (0==returnValue)
						{
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																		forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_PAX
																	   metadata:nil];
																						   
							if (rootIsYouDaddy_==NO)
							{
								seteuid(userID_);
								setegid(groupID_);
								
								// Copy the file to the remote volume
								
								if ([self copyObjectAtPath:tLocalDestinationPath toPath:tDestinationPath]==NO)
								{
									return NO;
								}
								
								seteuid(0);
								setegid(0);
								
								// Remove Local file
								
								[fileManager_ removeFileAtPath:tLocalDestinationPath handler:self];
																
								seteuid(userID_);
								setegid(groupID_);
							}
							
							// Clean the file from disk
						
							if (rootIsYouDaddy_==NO)
							{
								seteuid(0);
								setegid(0);
							}
							
							if ([fileManager_ removeFileAtPath:tTemporaryDirectoryPath handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==YES)
							{
								if (rootIsYouDaddy_==NO)
								{
									seteuid(userID_);
									setegid(groupID_);
								}
								
								[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																			forStep:IC_BUILDER_INFORMATION_STEP_DISTRIBUTION_SCRIPTS
																		   metadata:nil];
								
								return YES;
							}
							else
							{
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																					 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																					 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																					 nil];
							}
						}
						else
						{
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OPERATION_FAILED],IC_BUILDER_FAILURE_REASON,
																				 ICPROJECTBUILDER_DITTO_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																				 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																				 [NSNumber numberWithInt:returnValue],IC_BUILDER_FAILURE_TOOL_RETURN_CODE,
																				 nil];
						}
					}
					else
					{
						// Low Memory
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					}
				}
				else
				{
					// Could not find ditto
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																		 ICPROJECTBUILDER_DITTO_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																		 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																		 nil];
				}
			}
		}
		else
		{
			return YES;
		}
		
	}
	
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
	
	return NO;
}

- (BOOL) buildScriptsAtPath:(NSString *) inPath forScriptsDictionary:(NSDictionary *) inScriptsDictionary ofPackageUUID:(NSString *) inPackageUUID
{
	NSDictionary * tMetadata=nil;
	NSString * tDestinationPath=nil;
	NSString * tTemporaryDirectoryPath;
	BOOL tContinue=YES;
		
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
												forStep:IC_BUILDER_INFORMATION_STEP_SCRIPTS_PAYLOAD
											   metadata:[NSDictionary dictionaryWithObjectsAndKeys:inPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,nil]];
	
	if (rootIsYouDaddy_==NO)
	{
		seteuid(0);
		setegid(0);
	}
	
	// Create the Temporary directory
	
	tTemporaryDirectoryPath=[scratchLocation_ stringByAppendingPathComponent:[NSString stringWithFormat:@"%d/%@",userID_,[[projectPath_ lastPathComponent] stringByDeletingPathExtension]]];
	
	if ([fileManager_ fileExistsAtPath:tTemporaryDirectoryPath]==YES)
	{
		// Delete the directory
		
		if ([fileManager_ removeFileAtPath:tTemporaryDirectoryPath handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==NO)
		{
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
															     tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
															     [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
															     nil];
			
			tContinue=NO;
		}
	}
	
	if (tContinue==YES && [self buildPath:tTemporaryDirectoryPath fixedPermissionAfter:NO]==YES)
	{
		NSMutableDictionary * tPackageBuildInformation;
		NSDictionary * tBundleScripts;
		NSString * tFileScriptName;
		
		tPackageBuildInformation=[[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES] objectForKey:inPackageUUID];
		
		[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_ASSEMBLE
												   metadata:nil];
		
		if (inScriptsDictionary!=nil)
		{
			NSMutableArray * tHierarchyArray;
			NSNumber * tNumber;
			
			tHierarchyArray=[inScriptsDictionary objectForKey:ICDOCUMENT_PACKAGE_SCRIPTS_FILE_HIERARCHY];
			
			if (tHierarchyArray!=nil)
			{
				NSUInteger i,tCount;
				NSDictionary * tChildDictionary;
				
				// Copy the file hierarchy
			
				tCount=[tHierarchyArray count];
				
				for(i=0;i<tCount;i++)
				{
					tChildDictionary=[tHierarchyArray objectAtIndex:i];
					
					if ([self buildFileHierarchyComponent:tChildDictionary atPath:tTemporaryDirectoryPath rootComponent:NO packagebuildInformation:nil]==NO)
					{
						return NO;
					}
				}
				
				// Remove the files defined as exceptions
				
				tNumber=[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS_PAYLOAD_ONLY];
				
				if (tNumber==nil || [tNumber boolValue]==NO)
				{
					if ([ICFilesHierarchyUtilities removeExceptions:[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS] fromFileHierarchyAtPath:tTemporaryDirectoryPath]==NO)
					{
						tContinue=NO;
					}
				}
			}
			
			if (tContinue==YES)
			{
				NSDictionary * tScriptDictionary;
				
				// Copy the scripts if needed
				
				tScriptDictionary=[inScriptsDictionary objectForKey:ICDOCUMENT_PACKAGE_SCRIPTS_PREINSTALL_PATH];
				
				if (tScriptDictionary!=nil && [tScriptDictionary count]>0)
				{
					NSString * tPath;
					 
					tPath=[self pathFromPathDictionary:tScriptDictionary];
					
					if (tPath!=nil)
					{
						tFileScriptName=[self copyScriptAtPath:tPath toPath:[tTemporaryDirectoryPath stringByAppendingPathComponent:@"preinstall"]];
						
						if (tFileScriptName!=nil)
						{
							[tPackageBuildInformation setObject:[NSString stringWithFormat:@"./%@",tFileScriptName] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_PREINSTALL_SCRIPT_PATH];
								
							[tPackageBuildInformation setObject:[NSNumber numberWithBool:YES] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_HAS_SCRIPTS];
						}
						else
						{
							return NO;
						}
					}
					else
					{
						return NO;
					}
				}
			
				tScriptDictionary=[inScriptsDictionary objectForKey:ICDOCUMENT_PACKAGE_SCRIPTS_POSTINSTALL_PATH];
			
				if (tScriptDictionary!=nil && [tScriptDictionary count]>0)
				{
					NSString * tPath;
					 
					tPath=[self pathFromPathDictionary:tScriptDictionary];
					
					if (tPath!=nil)
					{
						tNumber=[tScriptDictionary objectForKey:ICDOCUMENT_PATH_TYPE];
						
						if (tNumber!=nil)
						{
							int tPathType;
							
							tPathType=[tNumber intValue];
							
							if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
							{
								tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
							}
							else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
							{
								tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
							}
						}
						
						tFileScriptName=[self copyScriptAtPath:tPath toPath:[tTemporaryDirectoryPath stringByAppendingPathComponent:@"postinstall"]];
						
						if (tFileScriptName!=nil)
						{
							[tPackageBuildInformation setObject:[NSString stringWithFormat:@"./%@",tFileScriptName] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_POSTINSTALL_SCRIPT_PATH];
								
							[tPackageBuildInformation setObject:[NSNumber numberWithBool:YES] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_HAS_SCRIPTS];
						}
						else
						{
							return NO;
						}
					}
					else
					{
						return NO;
					}
				}
			}
		}
		
		// Copy the bundle pre-installation and post-installation scripts
					
		tBundleScripts=[tPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS];
					
		if (tBundleScripts!=nil)
		{
			NSMutableDictionary * tFinalDictionary;
				
			tFinalDictionary=[NSMutableDictionary dictionary];
				
			if (tFinalDictionary!=nil)
			{
				NSEnumerator * tEnumerator;
				
				tEnumerator=[tBundleScripts keyEnumerator];
				
				if (tEnumerator!=nil)
				{
					NSString * tBundleIdentifier;
					
					while (tBundleIdentifier=[tEnumerator nextObject])
					{
						NSDictionary * tPathDictionary;
						NSString * tPath;
						
						tPathDictionary=[tBundleScripts objectForKey:tBundleIdentifier];
						
						// Pre-installation
						
						tPath=[tPathDictionary objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_PREINSTALL];
						
						if (tPath!=nil)
						{
							tFileScriptName=[tFinalDictionary objectForKey:tPath];
							
							if (tFileScriptName==nil)
							{
								tFileScriptName=[self copyScriptAtPath:tPath toPath:[tTemporaryDirectoryPath stringByAppendingPathComponent:[tPath lastPathComponent]]];
								
								if (tFileScriptName!=nil)
								{
									[tFinalDictionary setObject:[NSString stringWithFormat:@"./%@",tFileScriptName] forKey:tPath];
								}
								else
								{
									return NO;
								}
							}
						}
						
						// Post-installation
						
						tPath=[tPathDictionary objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_POSTINSTALL];
						
						if (tPath!=nil)
						{
							tFileScriptName=[tFinalDictionary objectForKey:tPath];
							
							if (tFileScriptName==nil)
							{
								tFileScriptName=[self copyScriptAtPath:tPath toPath:[tTemporaryDirectoryPath stringByAppendingPathComponent:[tPath lastPathComponent]]];
								
								if (tFileScriptName!=nil)
								{
									[tFinalDictionary setObject:[NSString stringWithFormat:@"./%@",tFileScriptName] forKey:tPath];
								}
								else
								{
									return NO;
								}
							}
						}
					}
					
					if ([tFinalDictionary count]>0)
					{
						[tPackageBuildInformation setObject:tFinalDictionary forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_FINAL];
					}
				}
				else
				{
					// Low Memory
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					
					tContinue=NO;
				}
			}
			else
			{
				// Low Memory
					
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
			
				tContinue=NO;
			}
		}
					
		if (tContinue==YES)
		{
			NSTask * tTask;
			NSMutableArray * tArguments;
			int returnValue;
			
			// Set the privileges (root:wheel 0755 ) for the tTemporaryDirectoryPath path and its children
			
			[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
														forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_ASSEMBLE
													   metadata:nil];
			
			tContinue=NO;
			
			if (
#ifdef __SET_CORRECT_PERMISSIONS__
				[ICFilesHierarchyUtilities changeOwner:0 group:0 atPath:tTemporaryDirectoryPath]==1 ||
#endif
				[ICFilesHierarchyUtilities changeMode:0755 atPath:tTemporaryDirectoryPath]==1)
			{
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_INSUFFICIENT_PERMISSIONS_SET],IC_BUILDER_FAILURE_REASON,
																	 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																	 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																	 nil];
			}
			else
			{
				
				NSString * tSplitForksPath=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_SPLITFORKS_PATH;
				
				if ([[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_SPLITFORK_TOOL] isEqualToString:ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN]==YES)
				{
					tSplitForksPath=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN_PATH;
				}
				
				// Split Forks if needed
										
				if ([fileManager_ fileExistsAtPath:tSplitForksPath]==YES)
				{
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_SPLIT
															   metadata:nil];
					
					tTask=[NSTask new];
				
					if (tTask!=nil)
					{
						tArguments=[NSMutableArray arrayWithCapacity:1];
					
						[tArguments addObject:tTemporaryDirectoryPath];
						
						[tTask setArguments:tArguments];
						
						[tTask setLaunchPath:tSplitForksPath];
						
						[tTask launch];
						
						[tTask waitUntilExit];
						
						returnValue = [tTask terminationStatus];
						
						[tTask release];
						
						if (0!=returnValue)
						{
							switch(returnValue)
							{
								case -2:
									// Not a HFS or Extended HFS volume
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_SPLIT_NON_HFS_VOLUME],IC_BUILDER_FAILURE_REASON,
																						 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																						 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																						 nil];
									
									break;
									
								default:
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OPERATION_FAILED],IC_BUILDER_FAILURE_REASON,
																						 tSplitForksPath,IC_BUILDER_INFORMATION_FILE_PATH,
																						 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																						 [NSNumber numberWithInt:returnValue],IC_BUILDER_FAILURE_TOOL_RETURN_CODE,
																						 nil];
									
									
									// A COMPLETER (Recuperer sdterr si il y a)
									
									
									break;
							}
						}
						else
						{
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																		forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_SPLIT
																	   metadata:nil];
						
							tContinue=YES;
						}
					}
					else
					{
						// Low Memory
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					}
				}
				else
				{
					// Splitfork tool not found
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																		 tSplitForksPath,IC_BUILDER_INFORMATION_FILE_PATH,
																		 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																		 nil];
				}
			}
			
			if (tContinue==YES)
			{
				
				// Create the pax archive
				
				if ([fileManager_ fileExistsAtPath:ICPROJECTBUILDER_DITTO_PATH]==YES)
				{
					NSString * tLocalDestinationPath=nil;
					
					[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_PAX
															   metadata:nil];
					
					tTask=[NSTask new];
				
					if (tTask!=nil)
					{
						tDestinationPath=[inPath stringByAppendingPathComponent:@"Scripts"];
						
						tArguments=[NSMutableArray arrayWithCapacity:5];
						
						{
							// Ditto
							
#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5
							
							[tArguments addObject:@"--noextattr"];
							
							[tArguments addObject:@"--noqtn"];
							
							[tArguments addObject:@"--noacl"];
							
							[tArguments addObject:@"--norsrc"];
							
#endif
							
							[tArguments addObject:@"-c"];
							
							[tArguments addObject:@"-z"];
						
							[tArguments addObject:@"."];
						
							if (rootIsYouDaddy_==NO)
							{
								tLocalDestinationPath=[[tTemporaryDirectoryPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:[tDestinationPath lastPathComponent]];
								
								[tArguments addObject:tLocalDestinationPath];
							}
							else
							{
								[tArguments addObject:tDestinationPath];
							}
						
							[tTask setArguments:tArguments];
						
							[tTask setLaunchPath:ICPROJECTBUILDER_DITTO_PATH];
						}
						
						[tTask setCurrentDirectoryPath:tTemporaryDirectoryPath];
					
					
						/*NSPipe *pipe1,*pipe2;
						pipe1 = [NSPipe pipe];
						pipe2 = [NSPipe pipe];
						[tTask setStandardOutput: pipe1];
						[tTask setStandardError: pipe2];

						NSFileHandle *file1;
						file1 = [pipe1 fileHandleForReading];
						NSFileHandle *file2;
						file2 = [pipe2 fileHandleForReading];*/
						
						[tTask launch];
						
						/*NSData *data;
						data = [file1 readDataToEndOfFile];
						[data writeToFile:@"/tmp/stdout.txt" atomically:YES];
						data = [file2 readDataToEndOfFile];
						[data writeToFile:@"/tmp/stderr.txt" atomically:YES];*/
						
						[tTask waitUntilExit];
						
						
						
						returnValue = [tTask terminationStatus];
						
						[tTask release];
					
						if (0==returnValue)
						{
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																		forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_PAX
																	   metadata:nil];
																						   
							if (rootIsYouDaddy_==NO)
							{
								seteuid(userID_);
								setegid(groupID_);
								
								// Copy the file to the remote volume
								
								if ([self copyObjectAtPath:tLocalDestinationPath toPath:tDestinationPath]==NO)
								{
									return NO;
								}
								
								seteuid(0);
								setegid(0);
								
								// Remove Local file
								
								[fileManager_ removeFileAtPath:tLocalDestinationPath handler:self];
																
								seteuid(userID_);
								setegid(groupID_);
							}
							
							// Clean the file from disk
						
							if (rootIsYouDaddy_==NO)
							{
								seteuid(0);
								setegid(0);
							}
							
							if ([fileManager_ removeFileAtPath:tTemporaryDirectoryPath handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==YES)
							{
								if (rootIsYouDaddy_==NO)
								{
									seteuid(userID_);
									setegid(groupID_);
								}
								
								[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																			forStep:IC_BUILDER_INFORMATION_STEP_SCRIPTS_PAYLOAD
																		   metadata:nil];
								
								return YES;
							}
							else
							{
								tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																					 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																					 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																					 nil];
							}
						}
						else
						{
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OPERATION_FAILED],IC_BUILDER_FAILURE_REASON,
																				 ICPROJECTBUILDER_DITTO_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																				 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																				 [NSNumber numberWithInt:returnValue],IC_BUILDER_FAILURE_TOOL_RETURN_CODE,
																				 nil];
						}
					}
					else
					{
						// Low Memory
						
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
					}
				}
				else
				{
					// Could not find ditto
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																		 ICPROJECTBUILDER_DITTO_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																		 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																		 nil];
				}
			}
		}
	}

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
													forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
												   metadata:tMetadata];
	
	return NO;
}

- (BOOL) buildPayloadAndBomAtPath:(NSString *) inPath forFilesDictionary:(NSDictionary *) inFilesDictionary ofPackageUUID:(NSString *) inPackageUUID
{
	NSDictionary * tMetadata=nil;
	
	if (inFilesDictionary!=nil)
    {
		NSString * tDefaultLocation;
        
        [buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
													forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_PAYLOAD
												   metadata:[NSDictionary dictionaryWithObjectsAndKeys:inPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,nil]];
		
		tDefaultLocation=[inFilesDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION];
		
		if (tDefaultLocation!=nil)
		{
			NSNumber * tNumber;
			BOOL tSplitForks=YES;
			NSMutableDictionary * tHierarchyDictionary;
			 
			tNumber=[inFilesDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_SPLIT_FORKS];
			
			if (tNumber!=nil)
			{
				tSplitForks=[tNumber boolValue];
			}
			
			if (rootIsYouDaddy_==NO)
			{
				seteuid(0);
				setegid(0);
			}
			
			[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
														forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_ASSEMBLE
													   metadata:nil];
			
			tHierarchyDictionary=[inFilesDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY];
			
			if (tHierarchyDictionary!=nil)
			{
				NSString * tTemporaryDirectoryPath;
				NSString * tDestinationPath=nil;
				BOOL tContinue=YES;
				
				// Create the Temporary directory
				
				tTemporaryDirectoryPath=[scratchLocation_ stringByAppendingPathComponent:[NSString stringWithFormat:@"%d/%@",userID_,[[projectPath_ lastPathComponent] stringByDeletingPathExtension]]];
				
				if ([fileManager_ fileExistsAtPath:tTemporaryDirectoryPath]==YES)
				{
					// Delete the directory
					
					if ([fileManager_ removeFileAtPath:tTemporaryDirectoryPath handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==NO)
					{
						tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																			 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																			 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																			 nil];
			
						tContinue=NO;
					}
				}
				
				if (tContinue==YES && [self buildPath:tTemporaryDirectoryPath fixedPermissionAfter:NO]==YES)
				{
					NSTask * tTask;
					NSMutableArray * tArguments;
					NSString * tSplitForksPath=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_SPLITFORKS_PATH;
					int returnValue;
					uid_t tUID;
					gid_t tGID;
					int tPrivileges;
					NSMutableDictionary * tPackageBuildInformation;
					int64_t tHierarchySize=0;
					int64_t tNumberOfFiles=0;
					
					tPackageBuildInformation=[[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES] objectForKey:inPackageUUID];
					
					if ([[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_SPLITFORK_TOOL] isEqualToString:ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN]==YES)
					{
						tSplitForksPath=ICPROJECTBUILDER_RESOURCE_FORK_TOOL_GOLDIN_PATH;
					}
					
					// Find the beginning of the hierarchy according to the Default Location value
					
					tHierarchyDictionary=[ICPayloadUtilities firstItemForHierarchy:tHierarchyDictionary andDefaultLocation:tDefaultLocation];
					
					if (tHierarchyDictionary!=nil)
					{
						tContinue=NO;
						
						// Optimize the Hierarchy if needed (Remove the empty branch)
						
						if ([ICPayloadUtilities optimizeHierarchy:tHierarchyDictionary]!=-1)
						{
							[tPackageBuildInformation setObject:tTemporaryDirectoryPath forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_TEMPORARY_PAYLOAD_FOLDER];
						
							// Copy the file hierarchy
					
							if ([self buildFileHierarchyComponent:tHierarchyDictionary atPath:tTemporaryDirectoryPath rootComponent:YES packagebuildInformation:tPackageBuildInformation]==YES)
							{
								// Remove the files defined as exceptions
						
								if ([ICFilesHierarchyUtilities removeExceptions:[options_ objectForKey:ICPROJECTBUILDER_BUILD_OPTIONS_ARCHIVE_EXCEPTIONS] fromFileHierarchyAtPath:tTemporaryDirectoryPath]==YES)
								{
									// Look for bundles to populate the <bundle-version> section
						
									if ([self buildBundleVersionsDictionaryWithFileHierarchyAtPath:tTemporaryDirectoryPath ofPackageUUID:inPackageUUID]==YES)
									{
										// Set the privileges of the Default Location to the tTemporaryDirectoryPath path
						
										tNumber=[tHierarchyDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_UID];

										if (tNumber!=nil)
										{
											tUID=[tNumber intValue];
						
											tNumber=[tHierarchyDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_GID];
											
											if (tNumber!=nil)
											{
												tGID=[tNumber intValue];
						
												tNumber=[tHierarchyDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_PERMISSIONS];
												
												if (tNumber!=nil)
												{
													tPrivileges=[tNumber intValue];
#ifdef __SET_CORRECT_PERMISSIONS__						
													if ([ICFileSystemUtilities setFileOwner:tUID group:tGID atPath:tTemporaryDirectoryPath]==YES)
#endif
													{
														tContinue=YES;
													}
												}
												else
												{
													// Missing Information
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																										 ICDOCUMENT_PACKAGE_FILES_FILE_PERMISSIONS,IC_BUILDER_FAILURE_TAG_NAME,
																										 nil];
												}
											}
											else
											{
												// Missing Information
												
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																									 ICDOCUMENT_PACKAGE_FILES_FILE_GID,IC_BUILDER_FAILURE_TAG_NAME,
																									 nil];
											}
										}
										else
										{
											// Missing Information
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																								 ICDOCUMENT_PACKAGE_FILES_FILE_UID,IC_BUILDER_FAILURE_TAG_NAME,
																								 nil];
										}
									}
								}
							}
						}
						
						if (tContinue==YES)
						{
							[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																		forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_ASSEMBLE
																	   metadata:nil];
							
							// Split Forks if needed
					
							if (tSplitForks==YES)
							{
								tContinue=NO;
								
								if ([fileManager_ fileExistsAtPath:tSplitForksPath]==YES)
								{
									[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																				forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_SPLIT
																			   metadata:nil];
																		   
									tTask=[NSTask new];
								
									if (tTask!=nil)
									{
										tArguments=[NSMutableArray arrayWithCapacity:1];
										
										[tArguments addObject:tTemporaryDirectoryPath];
										
										[tTask setArguments:tArguments];
										
										[tTask setLaunchPath:tSplitForksPath];
										
										[tTask launch];
										
										[tTask waitUntilExit];
										
										returnValue = [tTask terminationStatus];
										
										[tTask release];
										
										if (0!=returnValue)
										{
											switch(returnValue)
											{
												case -2:
													// Not a HFS or Extended HFS volume
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_SPLIT_NON_HFS_VOLUME],IC_BUILDER_FAILURE_REASON,
																									 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																									 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																									 nil];
																		 
													break;
													
												default:
													
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OPERATION_FAILED],IC_BUILDER_FAILURE_REASON,
																									 tSplitForksPath,IC_BUILDER_INFORMATION_FILE_PATH,
																									 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																									 [NSNumber numberWithInt:returnValue],IC_BUILDER_FAILURE_TOOL_RETURN_CODE,
																									 nil];
													
													// A COMPLETER (Recuperer sdterr si il y a)
													
													
													break;
											}
										}
										else
										{
											tContinue=YES;
											
											[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																						forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_SPLIT
																					   metadata:nil];
										}
									}
									else
									{
										// Low Memory
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
									}
								}
								else
								{
									// Missing Splitfork tools
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																						 tSplitForksPath,IC_BUILDER_INFORMATION_FILE_PATH,
																						 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																						 nil];
								}
							}
							
							if (tContinue==YES)
							{
								// Compute the installed size
							
								if ([ICFilesHierarchyUtilities getSize:&tHierarchySize andNumberOfFiles:&tNumberOfFiles atPath:tTemporaryDirectoryPath]==YES)
								{
									[tPackageBuildInformation setObject:[NSNumber numberWithLongLong:(tHierarchySize/1024)] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_FILES_SIZE];
									
									[tPackageBuildInformation setObject:[NSNumber numberWithLongLong:tNumberOfFiles] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_NUMBER_OF_FILES];
								}
								else
								{
									// A COMPLETER
								}
								
								
								if ([ICFileSystemUtilities setFilePrivileges:tPrivileges atPath:tTemporaryDirectoryPath]==NO)
								{
									return NO;
								}
							
								tContinue=NO;
								
								// Create Archive.bom
						
								if ([fileManager_ fileExistsAtPath:ICPROJECTBUILDER_MKBOM_PATH]==YES)
								{
									NSString * tLocalDestinationPath=nil;
									
									[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																				forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_BOM
																			   metadata:nil];
									
									tTask=[NSTask new];
								
									if (tTask!=nil)
									{
										tDestinationPath=[inPath stringByAppendingPathComponent:@"Bom"];
										
										tArguments=[NSMutableArray arrayWithCapacity:2];
										
										[tArguments addObject:tTemporaryDirectoryPath];
										
										if (rootIsYouDaddy_==NO)
										{
											tLocalDestinationPath=[[tTemporaryDirectoryPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:@"Archive.bom"];
											
											[tArguments addObject:tLocalDestinationPath];
										}
										else
										{
											[tArguments addObject:tDestinationPath];
										}
										
										[tTask setArguments:tArguments];
										
										[tTask setLaunchPath:ICPROJECTBUILDER_MKBOM_PATH];
										
										[tTask launch];
										
										[tTask waitUntilExit];
										
										returnValue = [tTask terminationStatus];
										
										[tTask release];
										
										if (0==returnValue)
										{
											if (rootIsYouDaddy_==NO)
											{
												seteuid(userID_);
												setegid(groupID_);
												
												// Copy the file to the remote volume
												
												if ([self copyObjectAtPath:tLocalDestinationPath toPath:tDestinationPath]==YES)
												{
													seteuid(0);
													setegid(0);
													
													// Remove Local file
													
													[fileManager_ removeFileAtPath:tLocalDestinationPath handler:nil];
													
													seteuid(userID_);
													setegid(groupID_);
													
													tContinue=YES;
												}
											}
											else
											{
												tContinue=YES;
											}
										}
									}
									else
									{
										// Low Memory
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
									}
								}
								else
								{
									// Could not find mkbom
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																						 ICPROJECTBUILDER_MKBOM_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																						 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																						 nil];
								}
								
								if (tContinue==YES)
								{
									[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																				forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_BOM
																			   metadata:nil];
																			   
									// Create the pax archive
							
									if ([fileManager_ fileExistsAtPath:ICPROJECTBUILDER_DITTO_PATH]==YES)
									{
										[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_BEGIN
																					forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_PAX
																				   metadata:nil];
										
										tTask=[NSTask new];
									
										if (tTask!=nil)
										{
											NSString * tLocalDestinationPath=nil;
										
											tDestinationPath=[inPath stringByAppendingPathComponent:@"Payload"];
											
											tArguments=[NSMutableArray arrayWithCapacity:5];
											
											{
												// Ditto
#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5

												[tArguments addObject:@"--noextattr"];
												
												[tArguments addObject:@"--noqtn"];
												
												[tArguments addObject:@"--noacl"];
												
												[tArguments addObject:@"--norsrc"];
												
#endif
												
												[tArguments addObject:@"-c"];
												
												[tArguments addObject:@"-z"];
											
												[tArguments addObject:@"."];
											
												if (rootIsYouDaddy_==NO)
												{
													tLocalDestinationPath=[[tTemporaryDirectoryPath stringByDeletingLastPathComponent] stringByAppendingPathComponent:[tDestinationPath lastPathComponent]];
													
													[tArguments addObject:tLocalDestinationPath];
												}
												else
												{
													[tArguments addObject:tDestinationPath];
												}
											
												[tTask setArguments:tArguments];
											
												[tTask setLaunchPath:ICPROJECTBUILDER_DITTO_PATH];
											}
											
											[tTask setCurrentDirectoryPath:tTemporaryDirectoryPath];
											
											
											/*NSPipe *pipe1,*pipe2;
											pipe1 = [NSPipe pipe];
											pipe2 = [NSPipe pipe];
											[tTask setStandardOutput: pipe1];
											[tTask setStandardError: pipe2];

											NSFileHandle *file1;
											file1 = [pipe1 fileHandleForReading];
											NSFileHandle *file2;
											file2 = [pipe2 fileHandleForReading];*/
											
											[tTask launch];
											
											/*NSData *data;
											data = [file1 readDataToEndOfFile];
											[data writeToFile:@"/tmp/stdout.txt" atomically:YES];
											data = [file2 readDataToEndOfFile];
											[data writeToFile:@"/tmp/stderr.txt" atomically:YES];*/
											
											[tTask waitUntilExit];
											
											
											
											returnValue = [tTask terminationStatus];
											
											[tTask release];
										
											if (0==returnValue)
											{
												[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																							forStep:IC_BUILDER_INFORMATION_STEP_PAYLOAD_PAX
																						   metadata:nil];
												
												if (rootIsYouDaddy_==NO)
												{
													seteuid(userID_);
													setegid(groupID_);
													
													// Copy the file to the remote volume
													
													if ([self copyObjectAtPath:tLocalDestinationPath toPath:tDestinationPath]==NO)
													{
														return NO;
													}
													
													seteuid(0);
													setegid(0);
													
													// Remove Local file
													
													[fileManager_ removeFileAtPath:tLocalDestinationPath handler:self];
																					
													seteuid(userID_);
													setegid(groupID_);
												}
											
												if (tHierarchySize!=-1)
												{
													struct stat tFileStat;
													int64_t tArchiveSize;
													
													tArchiveSize=0;
													
													if (stat([tDestinationPath fileSystemRepresentation],&tFileStat)==0)
													{
														tArchiveSize=tFileStat.st_size/1024;
													}
													
													[tPackageBuildInformation setObject:[NSNumber numberWithLongLong:tArchiveSize] forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_ARCHIVE_SIZE];
												}
												
												// Clean the file from disk
										
												if (rootIsYouDaddy_==NO)
												{
													seteuid(0);
													setegid(0);
												}
												
												if ([fileManager_ removeFileAtPath:tTemporaryDirectoryPath handler:[ICProjectRemoverErrorHandler sharedRemoverErrorHandler]]==YES)
												{
													if (rootIsYouDaddy_==NO)
													{
														seteuid(userID_);
														setegid(groupID_);
													}
												
													[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_SUCCESS
																								forStep:IC_BUILDER_INFORMATION_STEP_PACKAGE_PAYLOAD
																							   metadata:[NSDictionary dictionaryWithObjectsAndKeys:inPackageUUID,IC_BUILDER_INFORMATION_PACKAGE_UUID,nil]];
													
													return YES;
												}
												else
												{
													tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_DELETED],IC_BUILDER_FAILURE_REASON,
																										 tTemporaryDirectoryPath,IC_BUILDER_INFORMATION_FILE_PATH,
																										 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																										 nil];
												}
											}
											else
											{
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OPERATION_FAILED],IC_BUILDER_FAILURE_REASON,
																									 ICPROJECTBUILDER_DITTO_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																									 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																									 [NSNumber numberWithInt:returnValue],IC_BUILDER_FAILURE_TOOL_RETURN_CODE,
																									 nil];
											}
										}
										else
										{
											// Low Memory
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
										}
									}
									else
									{
										// Could not find ditto
										
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_FOUND],IC_BUILDER_FAILURE_REASON,
																							 ICPROJECTBUILDER_DITTO_PATH,IC_BUILDER_INFORMATION_FILE_PATH,
																							 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_TOOL,
																							 nil];
									}
								}
							}
						}
					}
				}
				else
				{
					tMetadata=nil;
					
					// A COMPLETER
				}
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PACKAGE_FILES_FILE_HIERARCHY,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
		else
		{
			// Missing Information
			
			tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																 ICDOCUMENT_PACKAGE_FILES_DEFAULT_INSTALL_LOCATION,IC_BUILDER_FAILURE_TAG_NAME,
																 nil];
		}
	}
            
	// Size Does Matter
	
	/*if ([self buildArchiveInfoAtPath:[inPath stringByAppendingPathComponent:@"Resources"]]!=0)
	{
		// A COMPLETER
	}*/
    
	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];
	
    return NO;
}

- (BOOL) buildFileHierarchyComponent:(NSDictionary *) inDictionary atPath:(NSString *) inPath rootComponent:(BOOL) inRootComponent packagebuildInformation:(NSDictionary *) inPackageBuildInformation
{
    NSString * tPath;
    NSDictionary * tMetadata=nil;
	
    tPath=[inDictionary objectForKey:ICDOCUMENT_PATH];
    
    if (tPath!=nil)
    {
		NSArray * tChildren;
		NSUInteger i,tCount=0;
		NSString * tDestinationPath=nil;
		BOOL tShouldContinue=YES;
		
		tChildren=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_CHILDREN];
    
		if (tChildren!=nil)
		{
			 tCount=[tChildren count];
		}
    
		if (inRootComponent==NO)
		{
			NSNumber * tNumber;
			
			tShouldContinue=NO;
			
			tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_UID];
			
			if (tNumber!=nil)
			{
				uid_t tUID;
				
				tUID=[tNumber intValue];
			
				tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_GID];
				
				if (tNumber!=nil)
				{
					gid_t tGID;
					
					tGID=[tNumber intValue];
			
					tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_PERMISSIONS];
					
					if (tNumber!=nil)
					{
						int tPrivileges;
						
						tPrivileges=[tNumber intValue];
			
						tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_TYPE];
						
						if (tNumber!=nil)
						{
							int tType;
							int tPathType;
							BOOL tExceptionFolder=NO;
							NSDictionary * tPathDictionary;
							
							tType=[tNumber intValue];
							
							switch(tType)
							{
								case ICDOCUMENT_FILE_BASE_NODE_INVISIBLE:
								case ICDOCUMENT_FILE_BASE_NODE:	// Base Node
								case ICDOCUMENT_FILE_NEWFOLDER_NODE:	// New Folder
									tDestinationPath=[inPath stringByAppendingPathComponent:tPath];
									
									if ([fileManager_ createDirectoryAtPath:tDestinationPath attributes:nil]==NO)
									{
										BOOL isDirectory;
										
										if ([fileManager_ fileExistsAtPath:tDestinationPath isDirectory:&isDirectory]==NO || isDirectory==NO)
										{
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																								 tDestinationPath,IC_BUILDER_INFORMATION_FILE_PATH,
																								 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																								 nil];
											
											goto buildFileHierarchyComponentBail;
										}
									}
									
									if ([ICFileSystemUtilities setFilePrivileges:tPrivileges atPath:tDestinationPath]==NO)
									{
										return NO;
									}
									
#ifdef __SET_CORRECT_PERMISSIONS__

									if ([ICFileSystemUtilities setFileOwner:tUID group:tGID atPath:tDestinationPath]==NO)
									{
										return NO;
									}

#endif
									tShouldContinue=YES;
									
									break;
									
								case ICDOCUMENT_FILE_REALFILE_NODE:	// Real Node Item
								
									tNumber=[inDictionary objectForKey:ICDOCUMENT_PATH_TYPE];	// Support for relative path 09/12/04
							
									if (tNumber!=nil)
									{
										tPathType=[tNumber intValue];
									}
									else
									{
										tPathType=ICDOCUMENT_PATH_TYPE_GLOBAL;
									}
									
									if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_PROJECT)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceProjectPath_];
									}
									else if (tPathType==ICDOCUMENT_PATH_TYPE_RELATIVE_TO_REFERENCE_FOLDER)
									{
										tPath=[tPath stringByAbsolutingWithPath:referenceFolderPath_];
									}
									
									tDestinationPath=[inPath stringByAppendingPathComponent:[tPath lastPathComponent]];
									
									// We need to check the item is not expanded and empty in fact
									
									tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_EXPANDED];
							
									if (tNumber!=nil)				// Bug Fix 22/03/04
									{
										tExceptionFolder=[tNumber boolValue];
									}
									
									if (tCount==0 && tExceptionFolder==NO)
									{
										// Copy the path
										
										NSDictionary * tDictionary;
										
										if ([ICFileSystemUtilities copyPath:tPath toPath:inPath]==NO)
										{
											return NO;
										}
										
										tDictionary=[fileManager_ fileAttributesAtPath:tDestinationPath traverseLink:NO];
										
										if (tDictionary!=nil)
										{
											if ([[tDictionary objectForKey:NSFileType] isEqualToString:NSFileTypeSymbolicLink]==NO)
											{
												if ([ICFileSystemUtilities setFilePrivileges:tPrivileges atPath:tDestinationPath]==NO)
												{
													return NO;
												}
											}
											
											// Set the owner and group recursively
											
#ifdef __SET_CORRECT_PERMISSIONS__
					 
											if ([ICFilesHierarchyUtilities changeOwner:tUID group:tGID atPath:tDestinationPath]==1)
											{
												return NO;
											}
#endif

											tShouldContinue=YES;
										}
									}
									else
									{
										CFURLRef tURLRef;
										
										tURLRef=CFURLCreateWithFileSystemPath(kCFAllocatorDefault,(CFStringRef) tPath,kCFURLPOSIXPathStyle,TRUE);
										
										if (tURLRef!=NULL)
										{
											FSRef tFileRef;
											
											if (CFURLGetFSRef(tURLRef,&tFileRef)==TRUE)
											{
												FSCatalogInfo tCatalogInfo;
												OSErr tErr;
												
												// Release Memory
											
												CFRelease(tURLRef);
												
												tErr=FSGetCatalogInfo(&tFileRef,kFSCatInfoFinderInfo+kFSCatInfoFinderXInfo+kFSCatInfoAllDates,&tCatalogInfo,NULL,NULL,NULL);
												
												if (tErr==noErr)
												{
													NSDictionary * tFileAttributes;
													CFURLRef tDestinationURLRef;
													
													tFileAttributes=[fileManager_ fileAttributesAtPath:tPath traverseLink:NO];
													
													// Create the folder
												
													if ([fileManager_ createDirectoryAtPath:tDestinationPath attributes:tFileAttributes]==NO)
													{
														BOOL isDirectory;
														
														if ([fileManager_ fileExistsAtPath:tDestinationPath isDirectory:&isDirectory]==NO || isDirectory==NO)
														{
															tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_FILE_CANNOT_BE_CREATED],IC_BUILDER_FAILURE_REASON,
																												 tDestinationPath,IC_BUILDER_INFORMATION_FILE_PATH,
																												 [NSNumber numberWithBool:YES],IC_BUILDER_FAILURE_FILE_IS_FOLDER,
																												 nil];
															
															goto buildFileHierarchyComponentBail;
														}
													}
													
													if ([ICFileSystemUtilities setFilePrivileges:tPrivileges atPath:tDestinationPath]==NO)
													{
														return NO;
													}
												
#ifdef __SET_CORRECT_PERMISSIONS__

													if ([ICFileSystemUtilities setFileOwner:tUID group:tGID atPath:tDestinationPath]==NO)
													{
														return NO;
													}
												
#endif
													tDestinationURLRef=CFURLCreateWithFileSystemPath(kCFAllocatorDefault,(CFStringRef) tDestinationPath,kCFURLPOSIXPathStyle,TRUE);
													
													if (tDestinationURLRef!=NULL)
													{
														FSRef tDestinationFolderRef;
														
														if (CFURLGetFSRef(tDestinationURLRef,&tDestinationFolderRef)==TRUE)
														{
															tErr=FSSetCatalogInfo(&tDestinationFolderRef,kFSCatInfoFinderInfo+kFSCatInfoFinderXInfo+kFSCatInfoAllDates,&tCatalogInfo);
															
															tShouldContinue=YES;
														}
														else
														{
															tMetadata=nil;
														}
														
														// Release Memory
											
														CFRelease(tDestinationURLRef);
													}
													else
													{
														// Low Memory
														
														tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
													}
												}
												else
												{
													tMetadata=nil;
												}
											}
											else
											{
												// Release Memory
												
												CFRelease(tURLRef);
											}
										}
										else
										{
											// Low Memory
											
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
										}
									}
									
									break;
									
								default:
								
									// Incorrect Value
									
									tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_INCORRECT_VALUE],IC_BUILDER_FAILURE_REASON,
																						 ICDOCUMENT_PACKAGE_FILES_FILE_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																						 nil];
									
									break;
							}
				
							if (tShouldContinue==YES && inPackageBuildInformation!=nil && tType==ICDOCUMENT_FILE_REALFILE_NODE)
							{
								NSMutableDictionary * tBundleScriptsDictionary=nil;
								NSString * tBundleIdentifier=nil;
								NSBundle * tBundle;
								NSString * tScriptSourcePath;
								
								// Can Downgrade
								
								tNumber=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_CAN_DOWNGRADE];
								
								if (tNumber!=nil)
								{
									NSString * tPayloadRootFolder;
									
									tPayloadRootFolder=[inPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_TEMPORARY_PAYLOAD_FOLDER];
									
									if (tPayloadRootFolder!=nil)
									{
										NSMutableArray * tMutableArray;
										NSString * tRelativePath;
										
										tMutableArray=[inPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_DOWNGRADABLE];
										
										tRelativePath=[tDestinationPath substringFromIndex:[tPayloadRootFolder length]];
											
										if (tRelativePath!=nil)
										{
											[tMutableArray addObject:tRelativePath];
										}
										else
										{
											// Low Memory
										
											tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
										
											tShouldContinue=NO;
										}
									}
									else
									{
										// Missing Build Data
										
										[[ICBuilderLogManager sharedLogManager] logMessageWithLevel:ASL_LEVEL_CRIT format:@"Missing Build Data: ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_TEMPORARY_PAYLOAD_FOLDER"];
									
										tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_BUILD_DATA],IC_BUILDER_FAILURE_REASON,
																							 ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_TEMPORARY_PAYLOAD_FOLDER,IC_BUILDER_FAILURE_TAG_NAME,
																							 nil];
										
										tShouldContinue=NO;
									}
								}
								
								// Pre-installation and Post-installation script
								
								if (tShouldContinue==YES)
								{
									tPathDictionary=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_PREINSTALL_PATH];
									
									if (tPathDictionary!=nil)
									{
										tScriptSourcePath=[self pathFromPathDictionary:tPathDictionary];
										
										if (tScriptSourcePath==nil)
										{
											tMetadata=nil;
											
											tShouldContinue=NO;
										}
										else
										{
											// Find the bundle identifier
											
											tBundle=[NSBundle bundleWithPath:tPath];
											
											if (tBundle!=nil)
											{
												tBundleIdentifier=[tBundle bundleIdentifier];
											}
											
											if (tBundleIdentifier!=nil)
											{
												tBundleScriptsDictionary=[inPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS];
												
												if (tBundleScriptsDictionary!=nil)
												{
													[tBundleScriptsDictionary setObject:[NSMutableDictionary dictionaryWithObjectsAndKeys:tScriptSourcePath,ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_PREINSTALL,nil] forKey:tBundleIdentifier];
												}
											}
											else
											{
												tMetadata=nil;
											
												tShouldContinue=NO;
											}
										}
									}
								}
								
								if (tShouldContinue==YES)
								{
									tPathDictionary=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_BUNDLE_POSTINSTALL_PATH];
								
									if (tPathDictionary!=nil)
									{
										tScriptSourcePath=[self pathFromPathDictionary:tPathDictionary];
										
										if (tScriptSourcePath==nil)
										{
											tMetadata=nil;
											
											tShouldContinue=NO;
										}
										
										if (tBundleIdentifier==nil)
										{
											// Find the bundle identifier
										
											tBundle=[NSBundle bundleWithPath:tPath];
										
											if (tBundle!=nil)
											{
												tBundleIdentifier=[tBundle bundleIdentifier];
											}
											
											if (tBundleIdentifier!=nil)
											{
												tBundleScriptsDictionary=[inPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS];
											
												if (tBundleScriptsDictionary!=nil)
												{
													[tBundleScriptsDictionary setObject:[NSMutableDictionary dictionaryWithObjectsAndKeys:tScriptSourcePath,ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_POSTINSTALL,nil] forKey:tBundleIdentifier];
												}
											}
										}
										else
										{
											tBundleScriptsDictionary=[inPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS];
										
											if (tBundleScriptsDictionary!=nil)
											{
												NSMutableDictionary * tMutableDictionary;
											
												tMutableDictionary=[tBundleScriptsDictionary objectForKey:tBundleIdentifier];
											
												if (tMutableDictionary!=nil)
												{
													[tMutableDictionary setObject:tScriptSourcePath forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_SCRIPTS_POSTINSTALL];
												}
											}
										}
									}
								}
								
								if (debug_==NO)
								{
									// Locators
									
									if (tShouldContinue==YES)
									{
										NSArray * tLocatorsArray;
										
										tLocatorsArray=[inDictionary objectForKey:ICDOCUMENT_PACKAGE_FILES_FILE_LOCATORS];
										
										if (tLocatorsArray!=nil)
										{
											NSEnumerator * tLocatorEnumerator;
											
											tLocatorEnumerator=[tLocatorsArray objectEnumerator];
											
											if (tLocatorEnumerator!=nil)
											{
												NSDictionary * tLocatorDictionary;
												
												while (tLocatorDictionary=[tLocatorEnumerator nextObject])
												{
													tNumber=[tLocatorDictionary objectForKey:ICDOCUMENT_LOCATOR_STATE];
													
													if (tNumber!=nil && [tNumber boolValue]==YES)
													{
														if (tBundleIdentifier==nil)
														{
															// Find the bundle identifier
														
															tBundle=[NSBundle bundleWithPath:tPath];
														
															if (tBundle!=nil)
															{
																tBundleIdentifier=[tBundle bundleIdentifier];
															}
														}
												
														if (tBundleIdentifier!=nil)
														{
															NSMutableDictionary * tBuildLocatorsDictionary;
															NSMutableArray * tBuildLocatorsForBundle;
															
															tBuildLocatorsDictionary=[inPackageBuildInformation objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_LOCATORS];
															
															tBuildLocatorsForBundle=[tBuildLocatorsDictionary objectForKey:tBundleIdentifier];
															
															if (tBuildLocatorsForBundle==nil)
															{
																tBuildLocatorsForBundle=[NSMutableArray array];
																
																if (tBuildLocatorsForBundle!=nil)
																{
																	[tBuildLocatorsDictionary setObject:tBuildLocatorsForBundle forKey:tBundleIdentifier];
																}
															}
															
															if (tBuildLocatorsForBundle!=nil)
															{
																[tBuildLocatorsForBundle addObject:tLocatorDictionary];
															}
															else
															{
																tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
																
																tShouldContinue=NO;
																
																break;
															}
														}
													}
												}
											}
											else
											{
												tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_OUT_OF_MEMORY],IC_BUILDER_FAILURE_REASON,nil];
												
												tShouldContinue=NO;
											}
										}
									}
								}
							}
						}
						else
						{
							// Missing Information
							
							tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																				 ICDOCUMENT_PACKAGE_FILES_FILE_TYPE,IC_BUILDER_FAILURE_TAG_NAME,
																				 nil];
						}
					}
					else
					{
					   // Missing Information
					   
					   tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																			ICDOCUMENT_PACKAGE_FILES_FILE_PERMISSIONS,IC_BUILDER_FAILURE_TAG_NAME,
																			nil];
					}
				}
				else
				{
					// Missing Information
					
					tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																		 ICDOCUMENT_PACKAGE_FILES_FILE_GID,IC_BUILDER_FAILURE_TAG_NAME,
																		 nil];
				}
			}
			else
			{
				// Missing Information
				
				tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
																	 ICDOCUMENT_PACKAGE_FILES_FILE_UID,IC_BUILDER_FAILURE_TAG_NAME,
																	 nil];
			}
		}
		else
		{
			tDestinationPath=[inPath stringByAppendingPathComponent:@"/"];
		}
		
		if (tShouldContinue==YES)
		{
			if (tCount>0)
			{
				NSDictionary * tChildDictionary;
				
				for(i=0;i<tCount;i++)
				{
					tChildDictionary=[tChildren objectAtIndex:i];
					
					if ([self buildFileHierarchyComponent:tChildDictionary atPath:tDestinationPath rootComponent:NO packagebuildInformation:inPackageBuildInformation]==NO)
					{
						return NO;
					}
				}
			}
		
			return YES;
		}
	}
	else
	{
		// Missing Information
		
		tMetadata=[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedLong:IC_BUILDER_FAILURE_REASON_MISSING_INFORMATION],IC_BUILDER_FAILURE_REASON,
															 ICDOCUMENT_PATH,IC_BUILDER_FAILURE_TAG_NAME,
															 nil];
    }
	
buildFileHierarchyComponentBail:

	[buildNotificationCenter_ postNotificationWithState:IC_BUILDER_BUILD_STATE_END_FAILURE
												forStep:IC_BUILDER_INFORMATION_STEP_CURRENT
											   metadata:tMetadata];

	return NO;
	
}

#pragma mark -

- (BOOL) buildBundleVersionsDictionaryWithFileHierarchyAtPath:(NSString *) inFileHierarchyPath ofPackageUUID:(NSString *) inPackageUUID
{
    FTS * ftsp;
    FTSENT * tFile;
    char * tPath[2]={(char *) [inFileHierarchyPath fileSystemRepresentation],NULL};
    //NSMutableDictionary * tMutableDictionary;
    NSMutableArray * tMutableArray=nil;
	NSUInteger tLength;
	short tCurrentBundleLevel=-100;
	NSMutableArray * tLevelsArray;
	NSMutableArray * tParentsChildrenArray;
	
	tLength=[inFileHierarchyPath length];
	
    if ((ftsp = fts_open(tPath, FTS_PHYSICAL, 0)) == NULL)
    {
        return YES;
    }
    
	tLevelsArray=[NSMutableArray array];
	
	tParentsChildrenArray=[NSMutableArray array];
	
    //tMutableDictionary=[NSMutableDictionary dictionary];
    
    while ((tFile = fts_read(ftsp)) != NULL)
    {
        switch (tFile->fts_info)
        {
            case FTS_DC:
			case FTS_DNR:
            case FTS_ERR:
            case FTS_NS:
                    fts_close(ftsp);
                    
                    return 1;
            case FTS_D:
				        if (!strcmp(tFile->fts_name,"Contents"))
						{
							NSString * tAbsolutePath;
							
							tAbsolutePath=[NSString stringWithUTF8String:tFile->fts_path];
							
							if (tAbsolutePath!=nil)
							{
								NSString * tBundlePath;
								NSBundle * tBundle;
								
								tBundlePath=[tAbsolutePath stringByDeletingLastPathComponent];
								
								tBundle=[NSBundle bundleWithPath:tBundlePath];
								
								if (tBundle!=nil)
								{
									NSDictionary * tInfoDictionary;
									
									tInfoDictionary=[tBundle infoDictionary];
									
									if (tInfoDictionary!=nil)
									{
										NSMutableDictionary * tMutableBundleVersionDictionary=nil;
										
										tMutableBundleVersionDictionary=[NSMutableDictionary dictionary];
										
										if (tMutableBundleVersionDictionary!=nil)
										{
											id tObject;
											NSString * tVersionFilePath;
											NSDictionary * tVersionDictionary;
											NSString * tStringPath;
											NSMutableArray * tChildrenArray;
											
											tCurrentBundleLevel=tFile->fts_level;
											
											tStringPath=[tBundlePath substringFromIndex:tLength];
											
											[tMutableBundleVersionDictionary setObject:tStringPath forKey:@"path"];
											
											// Get Information from the Info.plist file
											
											// CFBundleShortVersionString
											
											tObject=[tInfoDictionary objectForKey:@"CFBundleShortVersionString"];
											
											if (tObject!=nil)
											{
												[tMutableBundleVersionDictionary setObject:tObject forKey:@"CFBundleShortVersionString"];
											}
											
											// CFBundleVersion
											
											tObject=[tInfoDictionary objectForKey:@"CFBundleVersion"];
											
											if (tObject!=nil)
											{
												[tMutableBundleVersionDictionary setObject:tObject forKey:@"CFBundleVersion"];
											}
											
											// CFBundleIdentifier
											
											tObject=[tInfoDictionary objectForKey:@"CFBundleIdentifier"];
											
											if (tObject!=nil)
											{
												[tMutableBundleVersionDictionary setObject:tObject forKey:@"CFBundleIdentifier"];
												
												[tMutableBundleVersionDictionary setObject:tObject forKey:@"id"];
											}
											
											
											// Look for a version.plist file
											
											
											tVersionFilePath=[tAbsolutePath stringByAppendingPathComponent:@"version.plist"];
											
											tVersionDictionary=[NSDictionary dictionaryWithContentsOfFile:tVersionFilePath];
											
											if (tVersionDictionary!=nil)
											{
												// BuildVersion
											
												tObject=[tVersionDictionary objectForKey:@"BuildVersion"];
												
												if (tObject!=nil)
												{
													[tMutableBundleVersionDictionary setObject:tObject forKey:@"BuildVersion"];
												}
												
												// ProjectName
											
												tObject=[tVersionDictionary objectForKey:@"ProjectName"];
												
												if (tObject!=nil)
												{
													[tMutableBundleVersionDictionary setObject:tObject forKey:@"ProjectName"];
												}
												
												// SourceVersion
												
												tObject=[tVersionDictionary objectForKey:@"SourceVersion"];
												
												if (tObject!=nil)
												{
													[tMutableBundleVersionDictionary setObject:tObject forKey:@"SourceVersion"];
												}

											
												// CFBundleShortVersionString
											
												tObject=[tVersionDictionary objectForKey:@"CFBundleShortVersionString"];
												
												if (tObject!=nil && [tMutableBundleVersionDictionary objectForKey:@"CFBundleShortVersionString"]==nil)
												{
													[tMutableBundleVersionDictionary setObject:tObject forKey:@"CFBundleShortVersionString"];
												}
												
												// CFBundleVersion
												
												tObject=[tVersionDictionary objectForKey:@"CFBundleVersion"];
												
												if (tObject!=nil && [tMutableBundleVersionDictionary objectForKey:@"CFBundleShortVersionString"]==nil)
												{
													[tMutableBundleVersionDictionary setObject:tObject forKey:@"CFBundleVersion"];
												}
											}
											
											if (tMutableArray==nil)
											{
												tMutableArray=[NSMutableArray array];
											}
											
											if (tMutableArray!=nil)
											{
												tChildrenArray=[tParentsChildrenArray lastObject];
											
												if (tChildrenArray!=nil)
												{
													[tChildrenArray addObject:tMutableBundleVersionDictionary];
												}
												else
												{
													[tMutableArray addObject:tMutableBundleVersionDictionary];
												}
												
												[tLevelsArray addObject:[NSNumber numberWithShort:tFile->fts_level]];
												
												tChildrenArray=[NSMutableArray array];
												
												[tMutableBundleVersionDictionary setObject:tChildrenArray forKey:@"Children"];
												
												[tParentsChildrenArray addObject:tChildrenArray];
											}
										}
									}
								}
							}
						}
						
						break;
						
			case FTS_DP:
			
				if (tCurrentBundleLevel==tFile->fts_level)
				{
					NSNumber * tNumber;
					
					[tParentsChildrenArray removeLastObject];
					
					[tLevelsArray removeLastObject];
					
					tNumber=[tLevelsArray lastObject];
					
					if (tNumber!=nil)
					{
						tCurrentBundleLevel=[tNumber shortValue];
					}
					else
					{
						tCurrentBundleLevel=-100;
					}
				}
				
				break;
            case FTS_SL:
            case FTS_SLNONE:
			case FTS_F:
                    continue;
            default:
                    break;
        }
    }
	
	if (tMutableArray!=nil)
	{
		NSMutableDictionary * tPackageBuildInformation;
		
		tPackageBuildInformation=[[buildInformation_ objectForKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGES] objectForKey:inPackageUUID];
		
		if (tPackageBuildInformation!=nil)
		{
			[tPackageBuildInformation setObject:tMutableArray forKey:ICPROJECTBUILDER_BUILD_INFORMATION_PACKAGE_BUNDLE_VERSIONS];
		}
	}
    
    fts_close(ftsp);
    
    return YES;
}

@end
